#include "ifinterface.h"

#include "cgcalc.h"
#include "cgmemory.h"
#include "cggraphs.h"

#include "ccnodes.h"
#include "ccmacros.h"

/* all routines implicitly need the following global variable. It is set by the
   exported functions explicitly */

static BLOCK_GRAPH *current_block_graph;

/********************************************************************************
 *																				*
 *		calculation of monadic/dyadic graphs									*
 *																				*
 ********************************************************************************/

static VOID calc_graph (register INSTRUCTION_GRAPH graph);
static INT16 sort_and_calc_graphs(INSTRUCTION_GRAPH graphs[], INT16 from, INT16 to,INT16 *sorted_arguments);

static VOID calculate_operator (INSTRUCTION_GRAPH graph, INT16 reg_use)
{	
	graph->save_result = graph->count > 1;
	
	graph->i_locals		= graph->save_result;
	graph->u_locals 	= max (0, graph->i_locals);
	graph->uregs		= reg_use;
	graph->ufloat_regs	= 0;
}

static VOID calculate_float_operator (INSTRUCTION_GRAPH graph, INT16 reg_use, INT16 float_reg_use)
{	
	graph->save_result = graph->count > 1;
	
	if (graph->save_result)
		graph->i_locals	= 2;
	else
		graph->i_locals = 0;
	
	graph->u_locals 	= max (0, graph->i_locals);
	graph->uregs		= reg_use;
	graph->ufloat_regs	= float_reg_use;
}

static VOID calculate_monadic_operator (INSTRUCTION_GRAPH graph,
										INSTRUCTION_GRAPH arg_graph,
										INT16 reg_use)
{	
	calc_graph (arg_graph);
	
	graph->save_result= graph->count > 1;
	
	graph->i_locals 	= arg_graph->i_locals + graph->save_result;
	graph->u_locals 	= max(arg_graph->u_locals,graph->i_locals);
	graph->uregs		= max(arg_graph->uregs, reg_use);
	graph->ufloat_regs	= arg_graph->ufloat_regs;
}

static VOID calculate_monadic_float_operator (	INSTRUCTION_GRAPH graph,
												INSTRUCTION_GRAPH arg_graph,
												INT16 reg_use,
												INT16 float_reg_use)
{
	calc_graph (arg_graph);
	
	graph->save_result= graph->count > 1;
	
	if (graph->save_result)
		graph->i_locals = arg_graph->i_locals + 2;
	else
		graph->i_locals = arg_graph->i_locals;

	graph->u_locals		= max(arg_graph->u_locals,graph->i_locals);
	graph->uregs		= max(arg_graph->uregs, reg_use);
	graph->ufloat_regs	= max(arg_graph->ufloat_regs, float_reg_use);
}

static VOID calculate_dyadic_operator (	INSTRUCTION_GRAPH graph,
										register INSTRUCTION_GRAPH arg1,
										register INSTRUCTION_GRAPH arg2,
										INT16 reg_use)
{
	INT16 i_locals;
	INT16 l_locals,r_locals;
	INT16 uregs;
	BOOL order_left;	
	
	calc_graph (arg1);
	calc_graph (arg2);
	
	graph->save_result	= graph->count > 1;
	
	i_locals =	arg1->i_locals + arg2->i_locals + graph->save_result;
	
	if (arg2->uregs >= N_REGS && !arg1->save_result)
		l_locals=max (arg1->u_locals, arg1->i_locals + 1 + arg2->u_locals);
	else
		l_locals=max (arg1->u_locals, arg1->i_locals + arg2->u_locals);
	
	if (arg1->uregs >= N_REGS && !arg2->save_result)
		r_locals=max (arg2->u_locals, arg2->i_locals + 1 + arg1->u_locals);
	else
		r_locals=max (arg2->u_locals, arg2->i_locals + arg1->u_locals);

	l_locals = max (l_locals,i_locals);
	r_locals = max (r_locals,i_locals);

	/*	if amount of locals needed for both orders are the same, determine order
		so that register use is maximised (floating point register use is 
		considered most important, as saving floating point stuff is more
		expensive than saving 'normal' values), else determine order on local
		use. Do this in such a way that when both are equally good, the left
		branch is taken first, as dyadic operator are built in such a way that
		it is slightly more efficient to evaluate the leftmost argument first.
	*/
		
	if (l_locals == r_locals)
	{	if (arg1->ufloat_regs == arg2->ufloat_regs)
			order_left=arg1->uregs >= arg2->uregs;		/* equal reg use: left first */
		else
			order_left=arg1->ufloat_regs > arg2->ufloat_regs;
	} else
		order_left=l_locals < r_locals;

	if (order_left)
		uregs = max (arg1->uregs, 1 + arg2->uregs);
	else
		uregs = max (arg2->uregs, 1 + arg1->uregs);
	uregs = min(uregs,N_REGS);

	graph->i_locals		= i_locals;
	if (order_left)
		graph->u_locals = l_locals;
	else
		graph->u_locals = r_locals;	
	graph->uregs		= max (uregs, reg_use);
	graph->ufloat_regs	= max (arg1->ufloat_regs, arg2->ufloat_regs);
	
	graph->order_left 	= order_left;
}

static VOID calculate_dyadic_float_operator (	INSTRUCTION_GRAPH graph,
												register INSTRUCTION_GRAPH arg1,
												register INSTRUCTION_GRAPH arg2,
												INT16 reg_use,
												INT16 float_reg_use)
{
	INT16 i_locals;
	INT16 l_locals,r_locals;
	INT16 ufloat_regs;
	BOOL order_left;	
	
	calc_graph (arg1);
	calc_graph (arg2);
	
	graph->save_result	= graph->count > 1;
	i_locals = arg1->i_locals + arg2->i_locals;
	if (graph->save_result)
		i_locals += 2;

	if (arg2->ufloat_regs >= N_FLOAT_REGS && !arg1->save_result)
		l_locals=max (arg1->u_locals, arg1->i_locals + 2 + arg2->u_locals);
	else
		l_locals=max (arg1->u_locals, arg1->i_locals + arg2->u_locals);
	
	if (arg1->ufloat_regs >= N_FLOAT_REGS && !arg2->save_result)
		r_locals=max (arg2->u_locals, arg2->i_locals + 2 + arg1->u_locals);
	else
		r_locals=max (arg2->u_locals, arg2->i_locals + arg1->u_locals);

	l_locals = max (l_locals,i_locals);
	r_locals = max (r_locals,i_locals);
	
	/* see calculate_dyadic_operator */
	
	if (l_locals == r_locals)
	{	if (arg1->ufloat_regs == arg2->ufloat_regs)
			order_left=arg1->uregs >= arg2->uregs;		/* equal reg use, left first */
		else
			order_left=arg1->ufloat_regs > arg2->ufloat_regs;
	} else
		order_left = l_locals < r_locals;
	
	if (order_left)
		ufloat_regs = max (arg1->ufloat_regs, 1 + arg2->ufloat_regs);
	else
		ufloat_regs = max (arg2->ufloat_regs, 1 + arg1->ufloat_regs);
	ufloat_regs = min(ufloat_regs,N_FLOAT_REGS);

	graph->i_locals		= i_locals;
	if (order_left)
		graph->u_locals = l_locals;
	else
		graph->u_locals = r_locals;
	
	graph->uregs		= max (arg1->uregs, arg2->uregs);
	graph->uregs		= max (graph->uregs, reg_use);
	graph->ufloat_regs	= max (ufloat_regs, float_reg_use);
	
	graph->order_left 	= order_left;
}

/********************************************************************************
 *																				*
 *		calculation of create and fill nodes									*
 *																				*
 ********************************************************************************/

static VOID calculate_create_r_operator (INSTRUCTION_GRAPH graph)
{
	INT16 u_locals, i_locals, uregs, ufloat_regs;
	INSTRUCTION_GRAPH desc_graph, float_graph;
	
	desc_graph = graph->parameters[0].pointer;
	float_graph = graph->parameters[1].pointer;
	
	calc_graph(desc_graph);
	calc_graph(float_graph);
	
	u_locals	= desc_graph->u_locals;
	i_locals	= desc_graph->i_locals;
	uregs		= max(desc_graph->uregs,2);			/* at least 2 regs needed */
	ufloat_regs	= desc_graph->ufloat_regs;
	
	u_locals	= max(u_locals, i_locals+float_graph->u_locals);
	i_locals	+= float_graph->i_locals;
	uregs		= max(uregs, float_graph->uregs);
	ufloat_regs	= max(ufloat_regs, float_graph->ufloat_regs);
	
	graph->save_result=graph->count > 1;
	
	graph->i_locals		= i_locals + graph->save_result;
	graph->u_locals		= max (u_locals, graph->i_locals);
	graph->uregs		= uregs;
	graph->ufloat_regs	= ufloat_regs;
}

static VOID calculate_create_operator (	INSTRUCTION_GRAPH graph)
{	
	INT16 *args;
	INT16 n_arguments, i_locals, u_locals, uregs, ufloat_regs;
	register INT16 i;
	
	n_arguments=graph->arity;
	args = alloc_block_mem(n_arguments*sizeof (INT16));
	n_arguments = sort_and_calc_graphs((INSTRUCTION_GRAPH *)&graph->parameters[0],0, n_arguments, args);

	i_locals	= 0;
	u_locals	= 0;
	
	if (graph->arity > 3 && graph->info->node_arity>1, graph->parameters[0].pointer!=NULL)
	{
		/*	This is a large create node. If the arity of the node is zero or
			one, the node does not have to be altered: its size will be inserted
			before the code and the garbage collector checks this. If the
			arity is larger, all empty fields must be initialised at
			runtime (with a pointer to the node itself), unless the node
			has become overwritten (indicated by the absence of code/descriptor).
		*/
		
		INT16 empty_count;
		
		empty_count=0;
		for (i=2; i<graph->arity; ++i)
			if (graph->parameters[i].pointer==NULL)
				++empty_count;
		
		if (empty_count > 1)
			uregs=3;				/* multiple empty fields have to be filled */
		else
			uregs=2;
	} else
		uregs=2;					/* normally at least 2 regs needed */
	
	ufloat_regs	= 0;
	
	for (i=0; i<n_arguments; ++i)
	{	INSTRUCTION_GRAPH graph_i;
	
		graph_i = graph->parameters[args[i]].pointer;
		
		u_locals = max(u_locals, graph_i->u_locals+i_locals);
		i_locals += graph_i->i_locals;
		uregs = max(uregs, graph_i->uregs);
		ufloat_regs = max(ufloat_regs, graph_i->ufloat_regs);
	}
		
	graph->save_result=graph->count > 1;
	
	graph->i_locals		= i_locals + graph->save_result;
	graph->u_locals		= max (u_locals, graph->i_locals);
	graph->uregs		= uregs;
	graph->ufloat_regs	= ufloat_regs;
}

static VOID calculate_fill_r_operator (register INSTRUCTION_GRAPH graph)
{
	INT16 i_locals,u_locals,uregs, ufloat_regs;
	INSTRUCTION_GRAPH address_graph, desc_graph, wait_graph, float_graph;
	
	address_graph	= graph->parameters[0].pointer;
	wait_graph		= graph->parameters[1].pointer;
	desc_graph		= graph->parameters[2].pointer;
	float_graph		= graph->parameters[3].pointer;
	
	if (address_graph->code==GCREATE
		&& address_graph->count==1)
		/* overwrite fill_r node with create_r node (forget wait graph: it's empty anyway) */
	{
		graph->code=GCREATE_R;
		graph->parameters[0].pointer=desc_graph;
		graph->parameters[1].pointer=float_graph;
		graph->arity=2;
		
		calculate_create_r_operator (graph);
		return;
	}
	
	calc_graph (address_graph);
	
	if (wait_graph!=NULL)
		calc_graph(wait_graph);
	
	calc_graph (desc_graph);
	calc_graph (float_graph);
	
	/* regs needed to compute address */

	u_locals		= address_graph->u_locals;
	i_locals		= address_graph->i_locals;
	uregs			= max (address_graph->uregs, 2);		/* at least 2 regs needed */
	ufloat_regs 	= address_graph->ufloat_regs;
	
	/* regs needed to compute descriptor */
	
	u_locals		= max(u_locals, i_locals+desc_graph->u_locals);
	i_locals		+= desc_graph->i_locals;
	uregs			= max(uregs, desc_graph->uregs);
	ufloat_regs		= max(ufloat_regs, desc_graph->ufloat_regs);
	
	/* regs needed to save waiting list (if any) */
	
	if (wait_graph!=NULL)
	{	u_locals	= max(u_locals, i_locals+wait_graph->u_locals);
		i_locals	+= wait_graph->i_locals;
		uregs		= max(uregs,wait_graph->uregs);
		ufloat_regs	= max(ufloat_regs,wait_graph->ufloat_regs);
	}
	
	/* regs needed to fill real */

	u_locals		= max(u_locals, i_locals+float_graph->u_locals);
	i_locals		+= float_graph->i_locals;
	uregs			= max(uregs, float_graph->uregs);
	ufloat_regs		= max(ufloat_regs, float_graph->ufloat_regs);
	
	if (address_graph->save_result)
		graph->save_result = true;
	else
	{	graph->save_result = graph->count > 1;
		i_locals += graph->save_result;
		u_locals++;
	}
	
	graph->i_locals		= i_locals;
	graph->u_locals		= u_locals;
	graph->uregs		= uregs;
	graph->ufloat_regs	= ufloat_regs;
}

static VOID calculate_fill_operator (	register INSTRUCTION_GRAPH graph)
{
	register INT16 i;
	INT16 n_arguments,i_locals,u_locals,uregs, ufloat_regs;
	INSTRUCTION_GRAPH addr_graph, wait_graph, create_graph;
	INT16 *args;
	
	create_graph=graph->parameters[0].pointer;
	
	if (create_graph->code==GCREATE && create_graph->count==1)
	{
		/* create node is unique: combine the create and fill node and build a
		   single create node. Forget the waitlist: it's empty */
				
		if (graph->arity>3)
		{
			/* fill node is big enough to become a create node (at least 3 args) */
	
			graph->code=GCREATE;
			for (i=2; i<graph->arity; ++i)
				graph->parameters[i-2]=graph->parameters[i];
				
			graph->arity-=2;
			
			calculate_create_operator (graph);
						
			return;
		}
		
		if (graph->arity>1)
		{	/* the fill is quite small (address plus 1 or 2 arguments) */
			
			INSTRUCTION_GRAPH desc_node;
			
			/* take a look at the descriptor, to see if it's a HNF node */
			
			desc_node=graph->parameters[2].pointer;
			
			if(desc_node->code==GLOAD_DESC)
			{	
				/* a hnf node is filled, this may safely be replaced by a
				   small create node (saves memory) */
				
				graph->code=GCREATE;
				for (i=2; i<graph->arity; ++i)
					graph->parameters[i-2]=graph->parameters[i];
			
				graph->arity-=2;
	
				calculate_create_operator (graph);
				
				return;
			}
			
			/* the fill node is too small to be replaced by a small create node.
			   Instead the create node will be filled by fill arguments and the
			   fill node will be replaced by an indirection node. */
			
			if(create_graph->arity<3)
				internal_error("calculate_fill_operator: create node too small %d",create_graph->arity);
			
			for(i=2; i<graph->arity; i++)
				create_graph->parameters[i-2]=graph->parameters[i];
			for(i-=2; i<create_graph->arity; i++)
				create_graph->parameters[i].pointer=NULL;
			
			create_graph->count=graph->count;
			
			graph->arity=1;
			graph->count=1;
			graph->code=GINDIRECT;
			graph->parameters[0].pointer=create_graph;
			
			calculate_create_operator (create_graph);
			return;
		}
		internal_error("calculate_fill: fill node with arity 0 or 1");
	}
	
	/* if ref count of create node (if any) > 1, it should have been created big enough
	   to hold the arguments of the fill node. */
	
	addr_graph=create_graph;
	wait_graph=graph->parameters[1].pointer;
	
	calc_graph(addr_graph);
	if (wait_graph!=NULL)
		calc_graph(wait_graph);
	
	n_arguments=graph->arity;
	args = alloc_block_mem(n_arguments*sizeof(INT16));
	n_arguments = sort_and_calc_graphs((INSTRUCTION_GRAPH *)&graph->parameters[0],
								2, n_arguments, args);

	/* regs needed to compute address */
	
	i_locals		= 0;
	u_locals		= 0;
	uregs			= max (addr_graph->uregs, 2);		/* at least 2 regs needed */
	ufloat_regs 	= addr_graph->ufloat_regs;
	
	for (i=2; i<n_arguments; ++i)
	{	INSTRUCTION_GRAPH graph_i;
	
		graph_i = graph->parameters[args[i]].pointer;
		
		if (args[i]==3 && wait_graph!=NULL)
		{
			/* compute regs to save waiting list */
			
			u_locals		=max(u_locals,wait_graph->u_locals+i_locals);
			i_locals		+=wait_graph->i_locals;
			uregs			=max(uregs,wait_graph->uregs);
			ufloat_regs		=max(ufloat_regs,wait_graph->ufloat_regs);
		}
		
		u_locals = max(u_locals, graph_i->u_locals+i_locals);
		i_locals += graph_i->i_locals;
		uregs = max (uregs, graph_i->uregs);
		ufloat_regs = max (ufloat_regs, graph_i->ufloat_regs);
	}
		   
	if (addr_graph->save_result)
	{	graph->save_result = true;
		i_locals = addr_graph->i_locals + i_locals;
		u_locals = max (addr_graph->u_locals,
						addr_graph->i_locals + u_locals);
	} else
	{	graph->save_result = graph->count > 1;
		i_locals = addr_graph->i_locals + i_locals + graph->save_result;
		u_locals = max (addr_graph->u_locals,
						addr_graph->i_locals + 1 + u_locals);
	}
	
	graph->i_locals		=i_locals;
	graph->u_locals		=u_locals;
	graph->uregs		=uregs;
	graph->ufloat_regs	=ufloat_regs;
}

static void calculate_keep_save_operator (register INSTRUCTION_GRAPH graph)
{
	int i_locals,u_locals,uregs, ufloat_regs;
	INSTRUCTION_GRAPH keep_graph, result_graph;
	
	keep_graph		= graph->parameters[0].pointer;
	result_graph	= graph->parameters[1].pointer;
		
	calc_graph (keep_graph);
	calc_graph (result_graph);
	
	/*	compute the amount of regs needed to linearise keep_graph first, note that
		the result of the keep/save graph will be popped */
	
	u_locals		= keep_graph->u_locals;
	i_locals		= keep_graph->i_locals;
	uregs			= keep_graph->uregs;
	ufloat_regs 	= keep_graph->ufloat_regs;
	
	u_locals		= max(u_locals, i_locals+result_graph->u_locals);
	i_locals		+= result_graph->i_locals;
	uregs			= max(uregs, result_graph->uregs);
	ufloat_regs		= max(ufloat_regs, result_graph->ufloat_regs);
	
	graph->save_result = graph->count > 1;
	i_locals += graph->save_result;
	u_locals = max(u_locals, i_locals);
	
	graph->i_locals		= i_locals;
	graph->u_locals		= u_locals;
	graph->uregs		= uregs;
	graph->ufloat_regs	= ufloat_regs;
}

/********************************************************************************
 *																				*
 *		graph calculation														*
 *																				*
 ********************************************************************************/

static VOID calc_graph (register INSTRUCTION_GRAPH graph)
{
	/*	take notice about so-called fpentry instructions !!!!
		these take the following form: ldc code, fpentry. This means these
		instructions not only use the floating point registers, but the
		normal registers as well!!!. They can be recognised by their 'FPU'
		prefix (instead of FP only).
	*/
	
	/*
	 	Note that there are 2 possibilties in general with some graph:
		- some computation must be performed with it (value needed)
		- it must be moved to some other place (value not needed).
		
		calc_graph calculates register uses for the first case: it calculates
		the reg use that would be needed to get the result in a register (be
		it Areg, or fAreg), so that something can be done with the result.
		The second case does not require loading the result in a register
		always: e.g. a floating point number that must be moved to a node,
		or to some local does not require any space on the floating point
		evaluation stack, so in these cases the register use might be too
		high. Note that this is of no concern, as  such moves are not part
		of some expression evaluation: they occur 'on top of the stack', or
		inside some fill node. In both cases the value will be 'popped'
		directly, so that no one can get into problems by the value that
		decreased the amount of available registers temporarily.
		
		exact computation of the register use is posible, but then dyadic
		(floating point!) operations have to adjust the register use of
		their parameters by the amount needed to load them. 'local' nodes
		should not increase the register use then. 
	*/
	
	switch (graph->code)
	{		
		case GLOCAL:
			graph->save_result = true;
	
			/* 1 reg needed to get it in Areg */
			
			graph->i_locals		= 0;
			graph->u_locals 	= 0;
			graph->uregs		= 1;
			graph->ufloat_regs	= 0;
			return;
			
		case GFLOAT_LOCAL:
			graph->save_result = true;
			
			/* 1 reg and 1 float reg needed to get it in FAreg */
			
			graph->i_locals		= 0;
			graph->u_locals		= 0;
			graph->uregs		= 1;	/* for loading address */
			graph->ufloat_regs	= 1;	/* for float value itself */
			return;
			
		case GLOAD_CONST:
		case GLOAD_ADDR:
		case GLOAD_DESC:
		case GLOAD_L:
		case GPROC_INFO:
		case GDESC_TABLE:
		case GID:
		case GMINT:
			calculate_operator (graph, 1);
			return;

		case GRANDOM:
		
			/* NOTE: floating point regs should be kept intact!! */
			
			calculate_operator (graph,N_REGS);
			return;
		
		case GFLOAD_CONST:
		{
			FLOAT f;
			
			/*	if the constant is zero, this node will result in a
				'floating load zero double' instruction that only needs
				one floating point register. Otherwise the floating point
				value will be stored in memory and loading it will require
				an address calculation (not shared with label: ldc, ldpi)
				and a load instruction (fpldnldb)
			*/
			
			f.join.high	=graph->parameters[0].int32;
			f.join.low	=graph->parameters[1].int32;
			
			if (f.f == 0.0)
				calculate_float_operator (graph,0,1);
			else
				calculate_float_operator (graph,1,1);
			return;
		}
		case GFLOAD_NL:
			calculate_monadic_float_operator (	graph,
												graph->parameters[0].pointer,
												1, 1);
			return;
		case GSTORE_L:
		{
			register INSTRUCTION_GRAPH store_graph,register_graph;
			INT16 local_offset;
			
			store_graph=graph->parameters[0].pointer;
			local_offset=graph->parameters[1].int16;
			
			/* instruction count should be 1 at most, as this instruction is
			   solely used to put results on the stack at the end of
			   a basic block */
			   
			if (graph->count > 1)
				internal_error ("instruction count of store_l node greater than 1");
				
			calc_graph (store_graph);
			
			if (store_graph->save_result)
				graph->i_locals = store_graph->i_locals;
			else
				graph->i_locals = store_graph->i_locals + 1;
			
			graph->uregs		= max(1,store_graph->uregs);
			graph->ufloat_regs	= store_graph->ufloat_regs;
			
			register_graph=NULL;
			if (local_offset < N_LOCALS)
				register_graph=current_block_graph->local_parameter_node[local_offset];
			
			/* if local used as parameter, storing something in it
			   could cause more locals to be used (if it hasn't been read yet) */
			
			if (register_graph!=NULL && register_graph->count>0)
			{	++graph->i_locals;
				graph->uregs	= max (graph->uregs, 2);
			}
			graph->u_locals = max (store_graph->u_locals,graph->i_locals);
			
			graph->save_result	= true;
			return;
		}
		case GFSTORE_L:
		{
			register INSTRUCTION_GRAPH store_graph,local_graph1,local_graph2;
			INT16 local_offset;
			
			store_graph		=graph->parameters[0].pointer;
			local_offset	=graph->parameters[1].int16;
			
			/* instruction count should be 1 at most, as this instruction is
			   solely used to put results on the stack at the end of
			   a basic block */
			   
			if (graph->count > 1)
				internal_error ("instruction count of store_l node greater than 1");
				
			calc_graph (store_graph);
			
			if (store_graph->save_result)
				graph->i_locals = store_graph->i_locals;
			else
				graph->i_locals = store_graph->i_locals + 2;
			
			graph->uregs		= max (1,store_graph->uregs);		/* for address of local */
			graph->ufloat_regs	= max (1,store_graph->ufloat_regs);
			
			local_graph1=NULL;
			local_graph2=NULL;
			if (local_offset < N_LOCALS-1)
			{	local_graph1=current_block_graph->local_parameter_node[local_offset];
				local_graph2=current_block_graph->local_parameter_node[local_offset+1];
			}
			
			/* if local used as parameter, storing something in it
			   could cause more locals to be used (if it hasn't been read yet) */
			
			if (local_graph1!=NULL && local_graph1->count>0)
			{	++graph->i_locals;
				graph->uregs	= max (graph->uregs, 2);
			}
			if (local_graph2!=NULL && local_graph2->count>0)
			{	++graph->i_locals;
				graph->uregs	= max (graph->uregs, 2);
			}
			graph->u_locals = max (store_graph->u_locals,graph->i_locals);
			
			graph->save_result	= true;
			return;
		}
		case GSTORE_NL:
		{
			register INSTRUCTION_GRAPH load_graph,store_graph,addr_graph;
			
			store_graph=graph->parameters[0].pointer;
			addr_graph=graph->parameters[1].pointer;
			load_graph=graph->parameters[3].pointer;
			
			/* instruction count should be 1 at most, as this instruction is
			   solely used to put results on the stack at the end of
			   a basic block (and it leaves no result behind) */
			   
			if (graph->count > 1)
				internal_error ("instruction count of store_nl node greater than 1");
				
			calculate_dyadic_operator(graph,store_graph,addr_graph,2);
			
			if (load_graph!=NULL && load_graph->count>0
				&& load_graph->code==GLOAD_NL)
			{	load_graph = load_graph->parameters[0].pointer;
				if (load_graph!=NULL
					&& load_graph->code==GLOAD_L)

				{	++graph->i_locals;
					graph->u_locals = max (graph->u_locals,graph->i_locals);
					graph->uregs = max (graph->uregs, 3);
				}
			}
			
			graph->save_result	= false;
			return;
		}
		case GFSTORE_NL:
		{
			register INSTRUCTION_GRAPH load_graph1,load_graph2,store_graph,addr_graph;
			INT16 i_locals, u_locals;
			
			store_graph=graph->parameters[0].pointer;
			addr_graph=graph->parameters[1].pointer;
			load_graph1=graph->parameters[3].pointer;
			load_graph2=graph->parameters[4].pointer;

			
			/* instruction count should be 1 at most, as this instruction is
			   solely used to put results on the stack at the end of
			   a basic block (and it leaves no result behind) */
			   
			if (graph->count > 1)
				internal_error ("instruction count of store_nl node greater than 1");
			
			calc_graph(store_graph);
			calc_graph(addr_graph);
	
			graph->save_result	= graph->count > 1;
				
			if (graph->save_result)
				i_locals =	store_graph->i_locals + addr_graph->i_locals + 2;
			else
				i_locals =	store_graph->i_locals + addr_graph->i_locals;

			/* the address will be computed last */

			if (addr_graph->ufloat_regs >= N_FLOAT_REGS && !store_graph->save_result)
				u_locals=max (store_graph->u_locals, store_graph->i_locals + 2 + addr_graph->u_locals);
			else
				u_locals=max (store_graph->u_locals, store_graph->i_locals + addr_graph->u_locals);

			u_locals = max (u_locals,i_locals);
	
			graph->i_locals		= i_locals;
			graph->u_locals 	= u_locals;
			graph->uregs		= max (store_graph->uregs, addr_graph->uregs);
			graph->ufloat_regs	= max (store_graph->ufloat_regs, 1 + addr_graph->ufloat_regs);
				
			graph->order_left 	= true;
			
			if (load_graph1!=NULL && load_graph1->count>0
				&& load_graph1->code==GLOAD_NL)
			{	load_graph1 = load_graph1->parameters[0].pointer;
				if (load_graph1!=NULL
					&& load_graph1->code==GLOAD_L)

				{	++graph->i_locals;
					graph->u_locals = max (graph->u_locals,graph->i_locals);
					graph->uregs = max (graph->uregs, 3);
				}
			}
			
			if (load_graph2!=NULL && load_graph2->count>0
				&& load_graph2->code==GLOAD_NL)
			{	load_graph2 = load_graph2->parameters[0].pointer;
				if (load_graph2!=NULL
					&& load_graph2->code==GLOAD_L)

				{	++graph->i_locals;
					graph->u_locals = max (graph->u_locals,graph->i_locals);
					graph->uregs = max (graph->uregs, 3);
				}
			}

			
			graph->save_result	= false;
			return;
		}

		case GNOT:
		case GADC:
		case GEQC:
		case GEQD:
		case G_BCOUNT:
		case GLOAD_NL:
		case GLOAD_NLP:
		case GLOAD_B:
			calculate_monadic_operator(	graph,
										graph->parameters[0].pointer,
										1);
			return;
				
		case GABS:
		
			/* computation of absolute value takes 3 regs */
			
			calculate_monadic_operator(	graph,
										graph->parameters[0].pointer,
										3);
			return;
		
		case GADD:
		case GSUB:
		case GDIFF:
		case GMUL:
		case GPROD:
		case GDIV:
		case GMOD:
		case GFJOIN:
		
		case GBSUB:
		case GWSUB:
		
		case GAND:
		case GOR:
		case GXOR:
		case GLSL:
		case GLSR:
		
		case GEQ:
		case GGT:
			calculate_dyadic_operator (graph,
									graph->parameters[0].pointer,
									graph->parameters[1].pointer,
									2);
			return;
		
		case GASL:
		case GASR:
			calculate_dyadic_operator (graph,
									graph->parameters[0].pointer,
									graph->parameters[1].pointer,
									N_REGS);
			return;

		case GFADD:
		case GFSUB:
		case GFMUL:
		case GFDIV:
			calculate_dyadic_float_operator (graph,
									graph->parameters[0].pointer,
									graph->parameters[1].pointer,
									0, 2);
		case GFEQ:
		case GFGT:
			calculate_dyadic_float_operator (graph,
									graph->parameters[0].pointer,
									graph->parameters[1].pointer,
									1, 2);
			return;
			
		case GFREM:
		
			/* NOTE! remainder sequence destroys values in (integer) evaluation stack */
			
			calculate_dyadic_float_operator (graph,
									graph->parameters[0].pointer,
									graph->parameters[0].pointer,
									N_REGS, N_FLOAT_REGS);
			return;
		case GFLOW:
		case GFHIGH:
		
			/* register use is the same as that of the float argument that has to be computed */
			
			calculate_monadic_float_operator (graph,
									graph->parameters[0].pointer,
									0, 0);
			return;
		case GFPUSQRT:
		
			/* NOTE! Areg needed */
			
			calculate_monadic_float_operator (graph,
									graph->parameters[0].pointer,
									1, N_FLOAT_REGS);
			return;
		
		case GFPUMULBY2:
		case GFPUDIVBY2:
		case GFPUEXPINC32:
		case GFPUEXPDEC32:
		
			/* NOTE! Areg needed as well */
			
			calculate_monadic_float_operator (graph,
									graph->parameters[0].pointer,
									1,1);
			return;
			
		
		case GFPUENTIER:
		
			/* NOTE! Areg and fAreg needed */
			
			calculate_monadic_float_operator (graph,
									graph->parameters[0].pointer,
									1,1);
			return;
		
		case GFITOR:
		
			/*	evaluate integer, store in (non) local, load address of local
				and load and convert integer in floating point reg */
			
			calculate_monadic_operator (graph,graph->parameters[0].pointer,1);
			graph->ufloat_regs = max(graph->ufloat_regs,1);
			return;

		case GFRTOI:
		
			/* evaluate float, and store in (non) local */
			
			calculate_monadic_float_operator(graph,graph->parameters[0].pointer,1,1);
			return;															

		case GCREATE:
			calculate_create_operator (graph);
			return;
		case GCREATE_S:
		
			/*	linerisation returns a string in the heap (NOT a string node):
				the length, plus the characters. The graph is monadic: its
				argument delivers the address of the string that must be copied
				in the heap. 3 regs are needed, as a 'move' instruction is used */
			
			calculate_monadic_operator (graph,
										graph->parameters[0].pointer,
										3);
			return;
			
		case GCREATE_R:
			
			calculate_fill_r_operator (graph);
			return;
			
		case GFILL:
			calculate_fill_operator (graph);
			return;
		case GFILL_R:
			calculate_fill_r_operator (graph);
			return;
		case GFILL_COPY:
		{
			INSTRUCTION_GRAPH wait_graph;
			
			/* calculate reg use to save waiting list */
			
			wait_graph=graph->parameters[2].pointer;
			if (wait_graph!=NULL)
				calc_graph(wait_graph);
			
			/* calculate reg use to copy node */
			
			calculate_dyadic_operator (graph,graph->parameters[0].pointer,graph->parameters[1].pointer,3);
			
			if (wait_graph!=NULL)
			{	graph->u_locals		=max(wait_graph->u_locals,graph->u_locals+wait_graph->i_locals);
				graph->i_locals		+=wait_graph->i_locals;
				graph->uregs		=max(graph->uregs,wait_graph->uregs);
				graph->ufloat_regs	=max(graph->ufloat_regs,wait_graph->ufloat_regs);
			}
			
			return;
		}	
	
		case GINDIRECT:
		{	INSTRUCTION_GRAPH arg;
		
			/* reg use of graph the same as it's argument */
			
			arg=graph->parameters[0].pointer;
			
			calc_graph(arg);
			graph->i_locals		=arg->i_locals;
			graph->u_locals 	=arg->u_locals;
			graph->uregs		=arg->uregs;
			graph->ufloat_regs	=arg->ufloat_regs;

			return;
		}

		case GKEEP:
		case GSAVE:
		
			calculate_keep_save_operator (graph);
			return;

		case GFSIN:
		case GFCOS:
		case GFTAN:
		case GFASIN:
		case GFACOS:
		case GFATAN:
		case GFLN:
		case GFLOG10:
		case GFEXP:
					
			/* NOTE! all regs needed: this will result in a call to a library function */
			
			calculate_monadic_float_operator (graph,
									graph->parameters[0].pointer,
									N_REGS, N_FLOAT_REGS);
			return;

		case GFPOWER:
		
			/* NOTE! all regs needed, see above */
			
			calculate_dyadic_float_operator (graph,
									graph->parameters[0].pointer,
									graph->parameters[1].pointer,
									N_REGS, N_FLOAT_REGS);
			return;

		case GEQ_SYMBOL:
		case GSTR_EQ:
		case GSTR_CMP:
		
			/*	NOTE! all normal regs needed, floating point regs should be kept
				intact */
			
			calculate_dyadic_operator (graph,graph->parameters[0].pointer,graph->parameters[1].pointer,N_REGS);
			return;
		
		case GEMPTY:
		
			/* result not needed */
			
			graph->save_result = false;
			graph->i_locals		= 0;
			graph->u_locals 	= 0;
			graph->uregs		= 0;
			graph->ufloat_regs	= 0;
			return;

		default:
			internal_error("calc_graph: unknown node: id = %d",graph->code);
	}
}

VOID calc_block_graph (BLOCK_GRAPH *block_graph,INSTRUCTION_GRAPH graph)
{
	current_block_graph=block_graph;
	calc_graph(graph);
}

/********************************************************************************
 *																				*
 *		sorting of graphs														*
 *																				*
 ********************************************************************************/

static INT16 sort_and_calc_graphs(INSTRUCTION_GRAPH graphs[], INT16 from, INT16 to,
						INT16 *sorted_arguments)
{
	register INT16 i;
	INT16 n_sorted;
	BOOL *argument_evaluated;
		
	argument_evaluated = alloc_block_mem(to * sizeof (BOOL));
	
	for (n_sorted=0; n_sorted<from; ++n_sorted)
		sorted_arguments[n_sorted]=n_sorted;

	for (i=from; i<to; ++i)
	{	INSTRUCTION_GRAPH a_graph;
		
		a_graph=graphs[i];
		
		if (a_graph!=NULL)
		{	calc_graph (a_graph);
			argument_evaluated[i]=false;
		} else
			argument_evaluated[i]=true;
	}
	
	for (;;)
	{	register INT16 first_argument_number;
		register INSTRUCTION_GRAPH a_graph_1;
		
		first_argument_number=from;
		while (first_argument_number<to && argument_evaluated[first_argument_number])
			++first_argument_number;
		
		if (first_argument_number>=to)
			break;
			
		a_graph_1=graphs[first_argument_number];
		
		for (i=first_argument_number+1; i<to; ++i){
			if (!argument_evaluated[i])
			{	register INSTRUCTION_GRAPH a_graph_2;
				register INT16 i1,i2;
				INT16 u1,u2;
				
				a_graph_2=graphs[i];
				
				i1=a_graph_1->i_locals;
				i2=a_graph_2->i_locals;
				
				u1=a_graph_1->u_locals;
				u2=a_graph_2->u_locals;
				
				if (i1<0)
				{	if (i2<0 && (u2<u1 || (u1==u2 && i2<i1)))
					{	first_argument_number=i;
						a_graph_1=a_graph_2;
					}
				} else if (i1==0)
				{	if (i2<0 || (i2==0 && u2<u1))
					{	first_argument_number=i;
						a_graph_1=a_graph_2;
					}
				} else
				{	if (i2<=0 || (u2-i2>u1-i1 || (u2-i2==u1-i1 && u2<u1)))
					{	first_argument_number=i;
						a_graph_1=a_graph_2;
					}
				}
			} /* endif not yet evaluated */
		} /* end of inner loop */
		
		argument_evaluated[first_argument_number]=true;
		sorted_arguments[n_sorted++]=first_argument_number;
		
	} /* end of outer loop */
	
	return n_sorted;
}

INT16 sort_and_calc_block_graphs(BLOCK_GRAPH *block_graph,INSTRUCTION_GRAPH graphs[], INT16 from, INT16 to,
						INT16 *sorted_arguments)
{	
	current_block_graph=block_graph;
	return sort_and_calc_graphs(graphs,from,to,sorted_arguments);
}

/********************************************************************************
 *																				*
 *		counting of graphs														*
 *																				*
 ********************************************************************************/
 
static VOID set_pref_local (register INSTRUCTION_GRAPH graph,register INT16 pref_local)
{
	switch (graph->code)
	{
		case GADD:
		case GSUB:
		case GDIFF:
		case GMUL:
		case GPROD:
		case GDIV:
		case GMOD:
		
		case GAND:
		case GOR:
		case GXOR:
		case GLSL:
		case GLSR:
		case GASL:
		case GASR:
		
		case GBSUB:
		case GWSUB:
		
		case GEQ:
		case GGT:
		case GEQ_SYMBOL:
		
		case GSTR_EQ:
		case GSTR_CMP:

		case GFADD:
		case GFSUB:
		case GFMUL:
		case GFDIV:
		case GFREM:
		
		case GFEQ:
		case GFGT:
		
		case GNOT:
		case GABS:
		case GADC:
		case GEQC:
		case GEQD:
		case GLOAD_B:

				
		case G_BCOUNT:
		
		case GFITOR:
		case GFRTOI:
		case GFPUENTIER:
		
		case GFSIN:
		case GFCOS:
		case GFTAN:
		case GFASIN:
		case GFACOS:
		case GFATAN:
		case GFLN:
		case GFLOG10:
		case GFEXP:
		case GFPUSQRT:
		case GFPUMULBY2:
		case GFPUDIVBY2:
		case GFPUEXPINC32:
		case GFPUEXPDEC32:

		case GFPOWER:
		case GCREATE:
		case GCREATE_S:

		case GFLOAD_NL:
		case GLOAD_NL:
		case GLOAD_NLP:

		case GLOCAL:
		case GFLOAT_LOCAL:
		case GLOAD_CONST:
		case GFLOAD_CONST:
		case GLOAD_ADDR:
		case GLOAD_DESC:
		case GRANDOM:
		case GMINT:
		case GPROC_INFO:
		case GID:
		case GDESC_TABLE:
		case GLOAD_L:

			if (graph->pref_local<0)
				graph->pref_local=pref_local;
			
			break;
			
		case GFJOIN:
		
			if (graph->pref_local<0)
			{	graph->pref_local=pref_local;
			
				set_pref_local(graph->parameters[0].pointer,pref_local);
				set_pref_local(graph->parameters[1].pointer,pref_local+1);
			}			
			break;
			
		case GFHIGH:
			
			if (graph->pref_local<0)
			{	graph->pref_local=pref_local;
				
				if (pref_local>0)
					set_pref_local(graph->parameters[0].pointer,pref_local-1);
			}
			break;
			
		case GFLOW:
		
			if (graph->pref_local<0)
			{	graph->pref_local=pref_local;
				set_pref_local(graph->parameters[0].pointer,pref_local);
			}
			break;
		
		case GFILL:
		case GFILL_R:
		
			if (graph->pref_local<0)
			{	graph->pref_local=pref_local;
				
				if (graph->parameters[0].pointer!=NULL)
					set_pref_local(graph->parameters[0].pointer,pref_local);
			}
			break;
		
		case GFILL_COPY:
		
			/* dest will be delivered, so set pref local of dest */
			
			if (graph->pref_local<0)
			{	graph->pref_local=pref_local;
				set_pref_local(graph->parameters[1].pointer,pref_local);
			}
			break;
		
		case GKEEP:
		case GSAVE:
		
			/*	set pref local of second argument (the result graph), if the
				keep/save graph is still reachable via another route, its pref
				local will be set somewhere else */
		
			if (graph->pref_local<0)
			{
				graph->pref_local=pref_local;
				set_pref_local(graph->parameters[1].pointer,pref_local);
			}
			break;
		
		case GINDIRECT:
		
			/* pref local of arg the same as the indirection node */
			
			if (graph->pref_local<0)
			{	graph->pref_local=pref_local;
				set_pref_local(graph->parameters[0].pointer,pref_local);
			}
			break;
			
		case GFSTORE_NL:
		case GSTORE_NL:
		
			internal_error ("set_pref_local: STORE NON LOCAL nodes don't deliver result");
		
		default:
			internal_error ("set_pref_local: unknown node: %d",graph->code);
	}
}

static VOID count_graph (register INSTRUCTION_GRAPH graph)
{
	switch (graph->code){
		case GADD:
		case GSUB:
		case GDIFF:
		case GMUL:
		case GPROD:
		case GDIV:
		case GMOD:
		
		case GAND:
		case GOR:
		case GXOR:
		case GLSL:
		case GLSR:
		case GASL:
		case GASR:
				
		case GBSUB:
		case GWSUB:
		
		case GEQ:
		case GGT:
		case GEQ_SYMBOL:
		
		case GSTR_EQ:
		case GSTR_CMP:

		case GFADD:
		case GFSUB:
		case GFMUL:
		case GFDIV:
		case GFREM:
		
		case GFEQ:
		case GFGT:
		case GFJOIN:
		
			if (++graph->count==1)
			{	count_graph (graph->parameters[0].pointer);
				count_graph (graph->parameters[1].pointer);
			}
			break;
	
		case GINDIRECT:
		
			/* the count of an indirection node should not be used */
			
			++graph->count;
			count_graph (graph->parameters[0].pointer);
			break;

		case GNOT:
		case GABS:
		case GADC:
		case GEQC:
		case GEQD:
				
		case G_BCOUNT:
		case GLOAD_B:
		
		case GFHIGH:
		case GFLOW:
		case GFITOR:
		case GFRTOI:
		case GFPUENTIER:
		
		case GFSIN:
		case GFCOS:
		case GFTAN:
		case GFASIN:
		case GFACOS:
		case GFATAN:
		case GFLN:
		case GFLOG10:
		case GFEXP:
		case GFPUSQRT:
		case GFPUMULBY2:
		case GFPUDIVBY2:
		case GFPUEXPINC32:
		case GFPUEXPDEC32:

			if (++graph->count==1)
				count_graph (graph->parameters[0].pointer);
			break;
		
		case GFPOWER:
			if (++graph->count==1)
			{	count_graph (graph->parameters[0].pointer);
				count_graph (graph->parameters[1].pointer);
			}
			break;
			
		case GCREATE:
		{
			INT16 n;
			
			++graph->count;
			if (graph->count == 1)
			{	++current_block_graph->heap_ptr->count;
				for (n=0; n<graph->arity; ++n)
					if (graph->parameters[n].pointer!=NULL)
						count_graph (graph->parameters[n].pointer);
			}
			break;
		}
		case GCREATE_S:
		{
			if (++graph->count == 1)
			{	++current_block_graph->heap_ptr->count;
				count_graph(graph->parameters[0].pointer);
			}
			break;
		}
		case GFILL:
		case GFILL_R:
		{
			INT16 n;
			
			if (++graph->count == 1)
			{	if (graph->parameters[0].pointer!=NULL)
					count_graph(graph->parameters[0].pointer);
				
				/* skip wait_graph */
				
				for (n=2; n<graph->arity; ++n)
					if (graph->parameters[n].pointer!=NULL)
						count_graph (graph->parameters[n].pointer);
			}
			break;
		}
		case GFILL_COPY:
			if (++graph->count==1)
			{	count_graph (graph->parameters[0].pointer);
				count_graph (graph->parameters[1].pointer);
				
				/* skip wait_graph */				
			}
			break;
		
		case GKEEP:
		
			/*	count the first argument (the keep graph) as well: this
				might be the only reference to it and it must be evaluated
			*/
		
			if (++graph->count==1)
			{
				count_graph (graph->parameters[0].pointer);
				count_graph (graph->parameters[1].pointer);
			}
			break;

		case GSAVE:
		
			/*	do NOT count the first argument (the save graph): it must only
				be saved if there are others that need it. This node is only
				needed to avoid overwriting of fields
			*/

			if (++graph->count==1)
				count_graph (graph->parameters[1].pointer);
			
			break;

		case GFLOAD_NL:
		case GLOAD_NL:
		case GLOAD_NLP:
			if (++graph->count==1)
				count_graph (graph->parameters[0].pointer);
			break;
		case GFSTORE_NL:
		case GSTORE_NL:
			if (++graph->count==1)
			{	count_graph (graph->parameters[0].pointer);
				count_graph (graph->parameters[1].pointer);
			}
			break;
		case GSTORE_L:
		case GFSTORE_L:
			if (++graph->count==1)
				count_graph (graph->parameters[0].pointer);
			break;
		case GLOCAL:
		case GFLOAT_LOCAL:
		case GLOAD_CONST:
		case GFLOAD_CONST:
		case GLOAD_ADDR:
		case GLOAD_DESC:
		case GRANDOM:
		case GMINT:
		case GPROC_INFO:
		case GDESC_TABLE:
		case GID:
		case GLOAD_L:
			++graph->count;
			break;
		default:
			internal_error ("count_graph: unknown node: %d",graph->code);
	}
}

VOID count_block_graph (BLOCK_GRAPH *block_graph, register INSTRUCTION_GRAPH graph)
{
	current_block_graph=block_graph;
	count_graph(graph);
	
	if (graph->code==GSTORE_L || graph->code==GFSTORE_L)
	{	
		register INT16 local;
		
		local=graph->parameters[1].int16;
		graph->pref_local=local;
		set_pref_local(graph->parameters[0].pointer,local);		
	}
}

/********************************************************************************
 *																				*
 *		marking of graphs														*
 *																				*
 ********************************************************************************/
 
static VOID mark_graph_2 (register INSTRUCTION_GRAPH graph)
{
	switch (graph->code)
	{
		case GADD:
		case GSUB:
		case GDIFF:
		case GMUL:
		case GPROD:
		case GDIV:
		case GMOD:
		
		case GAND:
		case GOR:
		case GXOR:
		case GLSL:
		case GLSR:
		case GASL:
		case GASR:
				
		case GBSUB:
		case GWSUB:
		
		case GEQ:
		case GGT:
		case GEQ_SYMBOL:
		
		case GSTR_EQ:
		case GSTR_CMP:

		case GFADD:
		case GFSUB:
		case GFMUL:
		case GFDIV:
		case GFREM:
		
		case GFEQ:
		case GFGT:
		case GFJOIN:
		
			if (graph->used_mark<2)
			{	graph->used_mark=2;
				mark_graph_2 (graph->parameters[0].pointer);
				mark_graph_2 (graph->parameters[1].pointer);
			}
			break;
				
		case GINDIRECT:
		
			if (graph->used_mark<2)
			{	graph->used_mark=2;
				mark_graph_2 (graph->parameters[0].pointer);
			}

		case GNOT:
		case GABS:
		case GADC:
		case GEQC:
		case GEQD:
				
		case G_BCOUNT:
		case GLOAD_B:
		
		case GFHIGH:
		case GFLOW:
		case GFITOR:
		case GFRTOI:
		case GFPUENTIER:
		
		case GFSIN:
		case GFCOS:
		case GFTAN:
		case GFASIN:
		case GFACOS:
		case GFATAN:
		case GFLN:
		case GFLOG10:
		case GFEXP:
		case GFPUSQRT:
		case GFPUMULBY2:
		case GFPUDIVBY2:
		case GFPUEXPINC32:
		case GFPUEXPDEC32:

		
			if (graph->used_mark<2)
			{	graph->used_mark=2;
				mark_graph_2 (graph->parameters[0].pointer);
			}
			break;
		
		case GFPOWER:
		
			if (graph->used_mark<2)
			{	graph->used_mark=2;
				mark_graph_2 (graph->parameters[0].pointer);
				mark_graph_2 (graph->parameters[1].pointer);
			}
			break;
			
		case GCREATE:
		{
			INT16 n;
						
			if (graph->used_mark<2)
			{	graph->used_mark=2;
				current_block_graph->heap_ptr->used_mark=2;
				for (n=0; n<graph->arity; ++n)
					if (graph->parameters[n].pointer!=NULL)
						mark_graph_2 (graph->parameters[n].pointer);
			}
			break;
		}
		case GCREATE_S:
		{
			if (graph->used_mark<2)
			{	graph->used_mark=2;
				current_block_graph->heap_ptr->used_mark=2;
				mark_graph_2 (graph->parameters[0].pointer);
			}
			break;
		}
		case GFILL:
		case GFILL_R:
		{
			INT16 n;
			
			if (graph->used_mark<2)
			{	graph->used_mark=2;
				for (n=0; n<graph->arity; ++n)
					if (graph->parameters[n].pointer!=NULL)
						mark_graph_2 (graph->parameters[n].pointer);
			}
			break;
		}
		case GFILL_COPY:
			if (graph->used_mark<2)
			{	graph->used_mark=2;
				mark_graph_2 (graph->parameters[0].pointer);
				mark_graph_2 (graph->parameters[1].pointer);
				
				if (graph->parameters[2].pointer!=NULL)
					mark_graph_2(graph->parameters[2].pointer);
			}
			break;
			
		case GKEEP:
		
			/* mark both graphs */
			
			if (graph->used_mark<2)
			{
				graph->used_mark=2;
				mark_graph_2 (graph->parameters[0].pointer);
				mark_graph_2 (graph->parameters[1].pointer);
			}
			break;

		case GSAVE:
			
			/*	only mark the second graph, the first one should only
				be marked if it is reachable by someone else */
			
			if (graph->used_mark<2)
			{
				graph->used_mark=2;
				mark_graph_2 (graph->parameters[1].pointer);
			}
			break;

		case GFLOAD_NL:
		case GLOAD_NL:
		case GLOAD_NLP:
			if (graph->used_mark<2)
			{	graph->used_mark=2;
				mark_graph_2 (graph->parameters[0].pointer);
			}
			break;
		case GFSTORE_NL:
		case GSTORE_NL:
			if (graph->used_mark<2)
			{	graph->used_mark=2;
				mark_graph_2 (graph->parameters[0].pointer);
				mark_graph_2 (graph->parameters[1].pointer);
			}
			break;
		case GSTORE_L:
		case GFSTORE_L:
			if (graph->used_mark<2)
			{	graph->used_mark=2;
				mark_graph_2 (graph->parameters[0].pointer);
			}
			break;
		case GLOCAL:
		case GFLOAT_LOCAL:
		case GLOAD_CONST:
		case GFLOAD_CONST:
		case GLOAD_ADDR:
		case GLOAD_DESC:
		case GLOAD_L:
		case GRANDOM:
		case GMINT:
		case GPROC_INFO:
		case GID:
		case GDESC_TABLE:
			graph->used_mark=2;
			break;
		default:
			internal_error ("mark graph 2: unknown node: %d",graph->code);
	}
}

VOID mark_block_graph_2 (BLOCK_GRAPH *block_graph, register INSTRUCTION_GRAPH graph)
{	
	current_block_graph=block_graph;
	mark_graph_2(graph);
}

static VOID mark_graph_1 (register INSTRUCTION_GRAPH graph)
{
	switch (graph->code)
	{
		case GADD:
		case GSUB:
		case GDIFF:
		case GMUL:
		case GPROD:
		case GDIV:
		case GMOD:
		
		case GAND:
		case GOR:
		case GXOR:
		case GLSL:
		case GLSR:
		case GASL:
		case GASR:
				
		case GBSUB:
		case GWSUB:
		
		case GEQ:
		case GGT:
		case GEQ_SYMBOL:
		
		case GSTR_EQ:
		case GSTR_CMP:

		case GFADD:
		case GFSUB:
		case GFMUL:
		case GFDIV:
		case GFREM:
		
		case GFEQ:
		case GFGT:
		case GFJOIN:
		
			if (!graph->used_mark)
			{	graph->used_mark=1;
				mark_graph_2 (graph->parameters[0].pointer);
				mark_graph_2 (graph->parameters[1].pointer);
			}
			break;
				
		case GINDIRECT:
		
			if (!graph->used_mark)
			{	graph->used_mark=1;
				mark_graph_1 (graph->parameters[0].pointer);
			}

		case GNOT:
		case GABS:
		case GADC:
		case GEQC:
		case GEQD:
				
		case G_BCOUNT:
		case GLOAD_B:
		
		case GFITOR:
		case GFRTOI:
		case GFPUENTIER:
		
		case GFSIN:
		case GFCOS:
		case GFTAN:
		case GFASIN:
		case GFACOS:
		case GFATAN:
		case GFLN:
		case GFLOG10:
		case GFEXP:
		case GFPUSQRT:
		case GFPUMULBY2:
		case GFPUDIVBY2:
		case GFPUEXPINC32:
		case GFPUEXPDEC32:
		
			if (!graph->used_mark)
			{	graph->used_mark=1;
				mark_graph_2 (graph->parameters[0].pointer);
			}
			break;
		
		case GFPOWER:
		
			if (!graph->used_mark)
			{	graph->used_mark=1;
				mark_graph_2 (graph->parameters[0].pointer);
				mark_graph_2 (graph->parameters[1].pointer);
			}
			break;
			
		case GCREATE:
		{
			INT16 n;
						
			if (!graph->used_mark)
			{	graph->used_mark=1;
				current_block_graph->heap_ptr->used_mark=2;
				for (n=0; n<graph->arity; ++n)
					if (graph->parameters[n].pointer!=NULL)
						mark_graph_2 (graph->parameters[n].pointer);
			}
			break;
		}
		case GCREATE_S:
		{
			if (!graph->used_mark)
			{	graph->used_mark=1;
				current_block_graph->heap_ptr->used_mark=2;
				mark_graph_2 (graph->parameters[0].pointer);
			}
			break;
		}
		case GFILL:
		case GFILL_R:
		{
			INT16 n;
			
			if (!graph->used_mark)
			{	graph->used_mark=1;
				for (n=0; n<graph->arity; ++n)
					if (graph->parameters[n].pointer!=NULL)
						mark_graph_2 (graph->parameters[n].pointer);
			}
			break;
		}
		
		case GKEEP:
		
			if (!graph->used_mark)
			{
				graph->used_mark=1;
				mark_graph_2 (graph->parameters[0].pointer);
				mark_graph_2 (graph->parameters[1].pointer);
			}
			break;

		case GSAVE:
			
			if (!graph->used_mark)
			{
				graph->used_mark=1;
				mark_graph_2 (graph->parameters[1].pointer);
			}
			break;

		case GFILL_COPY:
			if (!graph->used_mark)
			{	graph->used_mark=1;
				mark_graph_2 (graph->parameters[0].pointer);
				mark_graph_2 (graph->parameters[1].pointer);
				
				if (graph->parameters[2].pointer!=NULL)
					mark_graph_2(graph->parameters[2].pointer);
			}
			break;
		case GFLOAD_NL:
		case GLOAD_NL:
		case GLOAD_NLP:
			if (!graph->used_mark)
			{	graph->used_mark=1;
				mark_graph_2 (graph->parameters[0].pointer);
			}
			break;
		case GFSTORE_NL:
		case GSTORE_NL:
			if (!graph->used_mark)
			{	graph->used_mark=1;
				mark_graph_2 (graph->parameters[0].pointer);
				mark_graph_2 (graph->parameters[1].pointer);
			}
			break;
		case GSTORE_L:
		case GFSTORE_L:
			if (!graph->used_mark)
			{	graph->used_mark=1;
				mark_graph_2 (graph->parameters[0].pointer);
			}
			break;
		case GLOCAL:
		case GFLOAT_LOCAL:
		case GLOAD_CONST:
		case GFLOAD_CONST:
		case GLOAD_ADDR:
		case GLOAD_DESC:
		case GLOAD_L:
		case GRANDOM:
		case GMINT:
		case GPROC_INFO:
		case GDESC_TABLE:
		case GID:
			graph->used_mark=1;
			break;
			
		case GFHIGH:
		case GFLOW:		
			if (!graph->used_mark)
			{	graph->used_mark=1;
				mark_graph_1 (graph->parameters[0].pointer);
			}
			break;

		default:
			internal_error ("mark graph 1: unknown node: %d",graph->code);
	}
}

VOID mark_block_graph_1 (BLOCK_GRAPH *block_graph,register INSTRUCTION_GRAPH graph)
{	
	current_block_graph=block_graph;
	mark_graph_1(graph);
}

VOID mark_and_count_block_graph (BLOCK_GRAPH *block_graph,INSTRUCTION_GRAPH graph)
{	
	current_block_graph=block_graph;
	mark_graph_2 (graph);
	count_graph (graph);
}

/********************************************************************************
 *																				*
 *		initialisation															*
 *																				*
 ********************************************************************************/

VOID initialise_calculation(VOID)
{
	/* nothing */
}