#include <string.h>

#include "ifinterface.h"

#include "cgcode.h"
#include "cglabels.h"
#include "cggraphs.h"
#include "cgstacks.h"
#include "cglin.h"
#include "cgmemory.h"
#include "cginputs.h"

#include "ccnodes.h"
#include "ccintermediate.h"
#include "ccmacros.h"
#include "ccsignals.h"

/*	warnings:
	-	if sequential code is generated
	-	an elimination of a space leak has not been implemented (jmp_eval_update)
	-	polling only works for code without external parallelism, so instructions
		for exernal parallelism are removed
*/

static BOOL jmp_eval_update_warned;
static BOOL ignore_defer_warned;
static BOOL sequential_warned;
static BOOL poll_warned;

static VOID  warning_jmp_eval_update(VOID)
{
	if (!jmp_eval_update_warned)
		error("warning: jmp_eval_update optimisation not supported yet\n");
	jmp_eval_update_warned=true;
}	

static VOID  warning_ignore_defer(VOID)
{
	if (!ignore_defer_warned)
		error("warning: defer instructions are ignored\n");
	ignore_defer_warned=true;
}	

static VOID  warning_sequential(VOID)
{
	if (!sequential_warned)
		error("warning: parallel instructions ignored\n");
	sequential_warned=true;
}	

static VOID  poll_warning(VOID)
{
	/* a warning when polling is used for parallel instructions */
	
	if (!poll_warned)
		error("warning: using polling for suspending processes (internal parallelism only!)\n");
	
	poll_warned=true;
}	

/* global variable that indicates whether this piece of code is reacahble at
   all (used to avoid generating unnecessay code). It is implicitly used by the
   end_block routines */

BOOL code_reachable;

/* global variable that points to the current block graph structure */

static BLOCK_GRAPH *curr_block_graph;

/* the result of coding: a pointer to the first block: set at initialisation */

static INSTRUCTION_BLOCK *first_instr_block;

/* .o label that has been encountered most recently. Such label is considered to be
   used as 'jsr' label. This will be used to determine whether the ctop should be
   stored in a local or not. If a .o is placed before a label that isn't used for
   jsr's this cannot result in an error, but it would foul up the 'analysis' */

static INSTRUCTION_BLOCK *last_jsr_entry;

/* the current waitlist. If it is NULL no waitlist is available. If one is available
   (because a getWL instruction has been executed) the only valid instruction
   sequences are:
   
   	getWL					or:				getWL
   	fill...									fill...
   	release									putWL
   	
   	the first one updates a node with a HNF node and releases the waitlist of the old
   	node. The second one overwrites an empty node with a closure and keeps the old
   	waitlist of the empty node (not supported anymore).
*/

static INSTRUCTION_GRAPH wait_graph;

/****************************************************************************
 *																			*
 *		allocation of c-top on entry of label								*
 *																			*
 ****************************************************************************/

static VOID fix_ctop_local()
{	
	/* last jsr entry points to block before first label where the ctop should
	   be stored in local. For the following blocks the ctop should be stored
	   there as well. last_jsr_entry does not point to the block where the
	   label is defined, as this block is not available at the moment of
	   definition yet.
	*/
	
	if (last_jsr_entry!=NULL)
		last_jsr_entry=last_jsr_entry->next;
	
	while (last_jsr_entry!=NULL)
	{
		register LABEL *label;

		label=last_jsr_entry->start_label;
		
		if(label!=NULL && (!(label->flags & FIXED_CTOP)))
		{	
			label->flags|=FIXED_CTOP;
			label->flags&=~CTOP_STACK;
		}
		last_jsr_entry=last_jsr_entry->next;
	}
}

static VOID fix_ctop_stack()
{
	/* see routine above */
	
	if (last_jsr_entry!=NULL)
		last_jsr_entry=last_jsr_entry->next;

	while (last_jsr_entry!=NULL)
	{
		register LABEL *label;

		label=last_jsr_entry->start_label;
		
		if(label!=NULL && (!(label->flags & FIXED_CTOP)))
			label->flags|=(FIXED_CTOP | CTOP_STACK);
		
		last_jsr_entry=last_jsr_entry->next;
	}
}

/********************************************************************************
 *																				*
 *		auxiliary functions														*
 *																				*
 ********************************************************************************/

static VOID instruction_call (	CHAR *subroutine,
								INT16 demanded_a_stack_size, INT16 demanded_b_stack_size, U_INT32 demanded_vector[],
								INT16 offered_a_stack_size, INT16 offered_b_stack_size, U_INT32 offered_vector[])
{
	LABEL *label;
	
	label = enter_label2(subroutine, USE_LABEL | GCALL_LABEL | JSR_LABEL | IMPORT_LABEL);

	/* note no room is reserved on the stack for a return address. It is assumed this
	   routine will only be used for functions that do not consume stack space (!)
	   This is the case for most library routines (else they should reserve it
	   themselves). Note that library routines will make use of the return address
	   local, so the return address must be stored on the stack before calling the
	   library routine */
	
	set_label_layout (label,demanded_a_stack_size,demanded_b_stack_size,demanded_vector);

	insert_label_layout_block (curr_block_graph,GCALL,label);
	
	curr_block_graph=begin_layout_block(offered_a_stack_size, offered_b_stack_size, offered_vector);
	
	fix_ctop_stack();
}

static VOID instruction_signal (INT16 signal,
								INT16 demanded_a_stack_size, INT16 demanded_b_stack_size, U_INT32 demanded_vector[],
								INT16 offered_a_stack_size, INT16 offered_b_stack_size, U_INT32 offered_vector[])
{
	insert_layout_block (curr_block_graph,SIGNAL,signal,demanded_a_stack_size,demanded_b_stack_size,demanded_vector);
	curr_block_graph=begin_layout_block(offered_a_stack_size, offered_b_stack_size, offered_vector);
}

/****************************************************************************************
 *                                                                                      *
 *								code generation											*
 *																						*
 ****************************************************************************************/

VOID code_no_op (VOID)
{
}

/****************************************************************************************
 *                                                                                      *
 *       simple arithmetic and logic operations    										*
 *																						*
 ****************************************************************************************/

VOID code_pushI (INT32 i)
{
	INSTRUCTION_GRAPH graph_1;
	
	graph_1=g_load_const (i);
	s_push_b (curr_block_graph,graph_1);
}

VOID code_pushI_a (INT16 a_offset)
{
	INSTRUCTION_GRAPH graph_1,graph_2;
	
	graph_1=s_get_a (curr_block_graph,a_offset);
	graph_2=g_load_nl (graph_1,ARG1_OFFSET);
	
	s_push_b (curr_block_graph,graph_2);
}

VOID code_addI (VOID)
{
	INSTRUCTION_GRAPH graph_1,graph_2,graph_3;

	graph_1=s_get_pop_b(curr_block_graph);
	graph_2=s_get_b (curr_block_graph,0);
	
	graph_3=g_addI(graph_1,graph_2);
	
	s_put_b (curr_block_graph,0,graph_3);
}

VOID code_subI (VOID)
{
	INSTRUCTION_GRAPH graph_1,graph_2,graph_3;

	graph_1=s_get_pop_b(curr_block_graph);
	graph_2=s_get_b (curr_block_graph,0);

	graph_3=g_subI(graph_1,graph_2);	
	
	s_put_b (curr_block_graph,0,graph_3);
}

VOID code_incI (VOID)
{
	code_pushI((INT32)1);
	code_addI();
}

VOID code_decI (VOID)
{
	code_pushI((INT32)-1);
	code_addI();
}

VOID code_mulI (VOID)
{
	INSTRUCTION_GRAPH graph_1,graph_2,graph_3;

	graph_1=s_get_pop_b(curr_block_graph);
	graph_2=s_get_b (curr_block_graph,0);

	graph_3=g_mulI(graph_1,graph_2);
	
	s_put_b (curr_block_graph,0,graph_3);
}

VOID code_divI (VOID)
{
	INSTRUCTION_GRAPH graph_1,graph_2,graph_3;
	
	graph_1=s_get_pop_b(curr_block_graph);
	graph_2=s_get_b (curr_block_graph,0);
	
	graph_3=g_divI(graph_1,graph_2);
	
	s_put_b (curr_block_graph,0,graph_3);
}

VOID code_modI (VOID)
{
	INSTRUCTION_GRAPH graph_1,graph_2,graph_3;

	graph_1=s_get_pop_b(curr_block_graph);
	graph_2=s_get_b (curr_block_graph,0);

	graph_3=g_modI(graph_1,graph_2);
	
	s_put_b (curr_block_graph,0,graph_3);
}

VOID code_andI (VOID)
{
	INSTRUCTION_GRAPH graph_1,graph_2,graph_3;

	graph_1=s_get_pop_b(curr_block_graph);
	graph_2=s_get_b (curr_block_graph,0);

	graph_3=g_andI(graph_1,graph_2);
	
	s_put_b (curr_block_graph,0,graph_3);
}

VOID code_andB (VOID)
{
	INSTRUCTION_GRAPH graph_1,graph_2,graph_3;

	graph_1=s_get_pop_b(curr_block_graph);
	graph_2=s_get_b (curr_block_graph,0);

	graph_3=g_andB(graph_1,graph_2);
	
	s_put_b (curr_block_graph,0,graph_3);
}

VOID code_orI (VOID)
{
	INSTRUCTION_GRAPH graph_1,graph_2,graph_3;
	
	graph_1=s_get_pop_b(curr_block_graph);
	graph_2=s_get_b (curr_block_graph,0);

	graph_3=g_orI(graph_1,graph_2);

	s_put_b (curr_block_graph,0,graph_3);
}

VOID code_orB (VOID)
{
	INSTRUCTION_GRAPH graph_1,graph_2,graph_3;
	
	graph_1=s_get_pop_b(curr_block_graph);
	graph_2=s_get_b (curr_block_graph,0);

	graph_3=g_orB(graph_1,graph_2);	
	
	s_put_b (curr_block_graph,0,graph_3);
}

VOID code_xorI (VOID)
{
	INSTRUCTION_GRAPH graph_1,graph_2,graph_3;
	
	graph_1=s_get_pop_b(curr_block_graph);
	graph_2=s_get_b (curr_block_graph,0);
	
	graph_3=g_xorI(graph_1,graph_2);
	
	s_put_b (curr_block_graph,0,graph_3);
}

VOID code_xorB (VOID)
{
	INSTRUCTION_GRAPH graph_1,graph_2,graph_3;
	
	graph_1=s_get_pop_b(curr_block_graph);
	graph_2=s_get_b (curr_block_graph,0);

	graph_3=g_xorB(graph_1,graph_2);	
	
	s_put_b (curr_block_graph,0,graph_3);
}

VOID code_notI (VOID)
{
	INSTRUCTION_GRAPH graph_1, graph_2;
	
	graph_1=s_get_b(curr_block_graph,0);

	graph_2=g_notI(graph_1);
	
	s_put_b(curr_block_graph,0,graph_2);
}

VOID code_notB (VOID)
{
	INSTRUCTION_GRAPH graph_1,graph_2;
	
	graph_1=s_get_b (curr_block_graph,0);

	graph_2=g_notB(graph_1);
	
	s_put_b (curr_block_graph,0,graph_2);
}

VOID code_L_shiftl (VOID)
{
	INSTRUCTION_GRAPH graph_1,graph_2,graph_3;

	graph_1=s_get_pop_b(curr_block_graph);
	graph_2=s_get_b (curr_block_graph,0);

	graph_3=g_L_shiftl(graph_1,graph_2);
	
	s_put_b (curr_block_graph,0,graph_3);
}

VOID code_L_shiftr (VOID)
{
	INSTRUCTION_GRAPH graph_1,graph_2,graph_3;

	graph_1=s_get_pop_b(curr_block_graph);
	graph_2=s_get_b (curr_block_graph,0);

	graph_3=g_L_shiftr(graph_1,graph_2);
	
	s_put_b (curr_block_graph,0,graph_3);
}

VOID code_A_shiftl (VOID)
{
	INSTRUCTION_GRAPH graph_1,graph_2,graph_3;

	graph_1=s_get_pop_b(curr_block_graph);
	graph_2=s_get_b (curr_block_graph,0);

	graph_3=g_A_shiftl(graph_1,graph_2);
	
	s_put_b (curr_block_graph,0,graph_3);
}

VOID code_A_shiftr (VOID)
{
	INSTRUCTION_GRAPH graph_1,graph_2,graph_3;

	graph_1=s_get_pop_b(curr_block_graph);
	graph_2=s_get_b (curr_block_graph,0);

	graph_3=g_A_shiftr(graph_1,graph_2);
	
	s_put_b (curr_block_graph,0,graph_3);
}

VOID code_eqI (VOID)
{
	INSTRUCTION_GRAPH graph_1,graph_2,graph_3;
	
	graph_1=s_get_pop_b(curr_block_graph);
	graph_2=s_get_b (curr_block_graph,0);

	graph_3=g_eqI(graph_1,graph_2);
	
	s_put_b (curr_block_graph,0,graph_3);
}

VOID code_gtI (VOID)
{
	INSTRUCTION_GRAPH graph_1,graph_2,graph_3;
	
	graph_1=s_get_pop_b(curr_block_graph);
	graph_2=s_get_b (curr_block_graph,0);

	graph_3=g_gtI(graph_1,graph_2);
	
	s_put_b (curr_block_graph,0,graph_3);
}

VOID code_ltI (VOID)
{	
	INSTRUCTION_GRAPH graph_1,graph_2,graph_3;
	
	graph_1=s_get_pop_b(curr_block_graph);
	graph_2=s_get_b (curr_block_graph,0);

	graph_3=g_gtI(graph_2,graph_1);
	
	s_put_b (curr_block_graph,0,graph_3);
}

VOID code_eqI_a (INT32 value,INT16 a_offset)
{
	code_pushI_a(a_offset);
	code_pushI(value);
	code_eqI();
}

VOID code_eqI_b (INT32 value,INT16 b_offset)
{
	code_push_b(b_offset);
	code_pushI(value);
	code_eqI();
}

VOID code_ItoC (VOID)
{
	code_pushI((INT32)255);
	code_andI();
}

VOID code_CtoI (VOID)
{
	/* dummy operation: type of character on top of stack becomes integer */
}

VOID code_randomI (VOID)
{
	INSTRUCTION_GRAPH graph_1;
	
	if (random_label==NULL)
	{	random_label=enter_label2 ("new_random",USE_LABEL | JUMP_LABEL | JSR_LABEL | IMPORT_LABEL);
		set_label_layout(random_label,0,0,e_vector);
	}
	
	graph_1 = g_random();
	s_push_b (curr_block_graph,graph_1);
}

/**************************************************************************************
 *                                                                                    *
 *          arithmic operations on reals                                              *
 *                                                                                    *
 **************************************************************************************/

VOID code_pushR (FLOAT *r)
{
	s_push_float(curr_block_graph,g_fload_const (r));
}

VOID code_pushR_a (INT16 a_offset)
{
	INSTRUCTION_GRAPH graph_1,graph_2;
	
	graph_1=s_get_a (curr_block_graph,a_offset);
	graph_2=g_fload_nl (graph_1,ARG1_OFFSET);
	
	s_push_float(curr_block_graph,graph_2);
}

static VOID code_monadic_real_operator (INT16 g_code)
{
	INSTRUCTION_GRAPH arg,result;
	
	arg=s_get_float(curr_block_graph,0);

	result=g_instruction1 (g_code,arg);
	
	s_put_float(curr_block_graph,0,result);
}

VOID code_sinR (VOID)
{
	if (sin_label==NULL)
	{	sin_label=enter_label2 ("SIN",USE_LABEL | JUMP_LABEL | JSR_LABEL | IMPORT_LABEL);
		set_label_layout(sin_label,0,0,e_vector);
	}

	code_monadic_real_operator (GFSIN);
}

VOID code_cosR (VOID)
{	
	if (cos_label==NULL)
	{	cos_label=enter_label2 ("COS",USE_LABEL | JUMP_LABEL | JSR_LABEL | IMPORT_LABEL);
		set_label_layout(cos_label,0,0,e_vector);
	}

	code_monadic_real_operator (GFCOS);
}

VOID code_tanR (VOID)
{	
	if (tan_label==NULL)
	{	tan_label=enter_label2 ("TAN",USE_LABEL | JUMP_LABEL | JSR_LABEL | IMPORT_LABEL);
		set_label_layout(tan_label,0,0,e_vector);
	}

	code_monadic_real_operator (GFTAN);
}

VOID code_asinR (VOID)
{	
	if (asin_label==NULL)
	{	asin_label=enter_label2 ("ASIN",USE_LABEL | JUMP_LABEL | JSR_LABEL | IMPORT_LABEL);
		set_label_layout(asin_label,0,0,e_vector);
	}

	code_monadic_real_operator (GFASIN);
}

VOID code_acosR (VOID)
{	
	if (acos_label==NULL)
	{	acos_label=enter_label2 ("ACOS",USE_LABEL | JUMP_LABEL | JSR_LABEL | IMPORT_LABEL);
		set_label_layout(acos_label,0,0,e_vector);
	}

	code_monadic_real_operator (GFACOS);
}

VOID code_atanR (VOID)
{	
	if (atan_label==NULL)
	{	atan_label=enter_label2 ("ATAN",USE_LABEL | JUMP_LABEL | JSR_LABEL | IMPORT_LABEL);
		set_label_layout(atan_label,0,0,e_vector);
	}

	code_monadic_real_operator (GFATAN);
}

VOID code_sqrtR (VOID)
{
	code_monadic_real_operator (GFPUSQRT);
}

VOID code_expR (VOID)
{	
	if (exp_label==NULL)
	{	exp_label=enter_label2 ("EXP",USE_LABEL | JUMP_LABEL |JSR_LABEL | IMPORT_LABEL);
		set_label_layout(exp_label,0,0,e_vector);
	}

	code_monadic_real_operator (GFEXP);
}

VOID code_lnR (VOID)
{	
	if (ln_label==NULL)
	{	ln_label=enter_label2 ("LN",USE_LABEL | JUMP_LABEL | JSR_LABEL | IMPORT_LABEL);
		set_label_layout(ln_label,0,0,e_vector);
	}

	code_monadic_real_operator (GFLN);
}

VOID code_log10R (VOID)
{	
	if (log10_label==NULL)
	{	log10_label=enter_label2 ("LOG10",USE_LABEL | JUMP_LABEL | JSR_LABEL | IMPORT_LABEL);
		set_label_layout(log10_label,0,0,e_vector);
	}

	code_monadic_real_operator (GFLOG10);
}

VOID code_powR (VOID)
{
	INSTRUCTION_GRAPH arg1,arg2,result;
		
	if (power_label==NULL)
	{	power_label=enter_label2 ("POWER",USE_LABEL | JUMP_LABEL | JSR_LABEL | IMPORT_LABEL);
		set_label_layout(power_label,0,0,e_vector);
	}

	arg1=s_get_pop_float	(curr_block_graph);
	arg2=s_get_float		(curr_block_graph,0);
	
	result=g_instruction2 (GFPOWER,arg1,arg2);

	s_put_float(curr_block_graph,0,result);
}

VOID code_entierR (VOID)
{
	INSTRUCTION_GRAPH arg,result;
	
	arg=s_get_pop_float(curr_block_graph);
	
	result=g_instruction1 (GFPUENTIER,arg);
	
	s_push_b (curr_block_graph,result); /* entier delivers an integer */
}

VOID code_addR (VOID)
{
	INSTRUCTION_GRAPH arg1,arg2,result;
	
	arg1=s_get_pop_float	(curr_block_graph);
	arg2=s_get_float		(curr_block_graph,0);
	
	result=g_addR(arg1,arg2);

	s_put_float(curr_block_graph,0,result);
}

VOID code_subR (VOID)
{
	INSTRUCTION_GRAPH arg1,arg2,result;
	
	arg1=s_get_pop_float	(curr_block_graph);
	arg2=s_get_float		(curr_block_graph,0);
	
	result=g_subR(arg1,arg2);
	
	s_put_float(curr_block_graph,0,result);
}

VOID code_incR (VOID)
{
	INSTRUCTION_GRAPH arg1,arg2,result;
	FLOAT one;

	arg1=s_get_float(curr_block_graph,0);
		
	one.f = 1.0;
	arg2=g_fload_const (&one);
	
	result=g_addR(arg1,arg2);
	
	s_put_float(curr_block_graph,0,result);
}

VOID code_decR (VOID)
{
	INSTRUCTION_GRAPH arg1,arg2,result;
	FLOAT one;

	arg1=s_get_float(curr_block_graph,0);
		
	one.f = 1.0;
	arg2=g_fload_const (&one);
	
	result=g_subR(arg1,arg2);
	
	s_put_float(curr_block_graph,0,result);
}

VOID code_mulR (VOID)
{
	INSTRUCTION_GRAPH arg1,arg2,result;
	
	arg1=s_get_pop_float	(curr_block_graph);
	arg2=s_get_float		(curr_block_graph,0);
	
	result=g_mulR(arg1,arg2);

	s_put_float(curr_block_graph,0,result);
}

VOID code_divR (VOID)
{
	INSTRUCTION_GRAPH arg1,arg2,result;
	
	arg1=s_get_pop_float	(curr_block_graph);
	arg2=s_get_float		(curr_block_graph,0);
	
	result=g_divR(arg1,arg2);

	s_put_float(curr_block_graph,0,result);
}

VOID code_remR (VOID)
{
	INSTRUCTION_GRAPH arg1,arg2,result;
	
	arg1=s_get_pop_float	(curr_block_graph);
	arg2=s_get_float		(curr_block_graph,0);
	
	result=g_remR(arg1,arg2);

	s_put_float(curr_block_graph,0,result);
}

VOID code_eqR (VOID)
{
	INSTRUCTION_GRAPH arg1,arg2,result;
	
	arg1=s_get_pop_float	(curr_block_graph);
	arg2=s_get_pop_float	(curr_block_graph);
	
	result=g_eqR(arg1,arg2);
	
	s_push_b (curr_block_graph,result);
}

VOID code_gtR (VOID)
{
	INSTRUCTION_GRAPH arg1,arg2,result;
	
	arg1=s_get_pop_float	(curr_block_graph);
	arg2=s_get_pop_float	(curr_block_graph);
	
	result=g_gtR(arg1,arg2);
	
	s_push_b (curr_block_graph,result);
}

VOID code_ltR (VOID)
{
	INSTRUCTION_GRAPH arg1,arg2,result;
	
	arg1=s_get_pop_float	(curr_block_graph);
	arg2=s_get_pop_float	(curr_block_graph);
	
	result=g_gtR(arg2,arg1);
	
	s_push_b (curr_block_graph,result);
}

VOID code_eqR_a (FLOAT *value,INT16 a_offset)
{
	INSTRUCTION_GRAPH graph_1,graph_2,graph_3,graph_4;
	
	graph_1=s_get_a (curr_block_graph,a_offset);
	graph_2=g_fload_nl (graph_1,ARG1_OFFSET);	
	graph_3=g_fload_const (value);
	
	graph_4=g_eqR(graph_2,graph_3);
	
	s_push_b (curr_block_graph,graph_4);
}

VOID code_eqR_b (FLOAT *value,INT16 b_offset)
{
	INSTRUCTION_GRAPH arg1,arg2,result;
	
	arg1=s_get_float (curr_block_graph,b_offset);
	arg2=g_fload_const (value);

	result=g_eqR(arg1,arg2);
	
	s_push_b (curr_block_graph,result);
}

VOID code_ItoR (VOID)
{
	INSTRUCTION_GRAPH graph_1,graph_2;
	
	graph_1=s_get_pop_b(curr_block_graph);
	graph_2=g_fitor (graph_1);
	s_push_float(curr_block_graph,graph_2);
}

VOID code_RtoI (VOID)
{
	INSTRUCTION_GRAPH graph_1,graph_2;
	
	graph_1=s_get_pop_float(curr_block_graph);
	graph_2=g_frtoi (graph_1);
	
	s_push_b (curr_block_graph,graph_2);
}

/****************************************************************************************
 *                                                                                      *
 *       stack manipulations        		     										*
 *																						*
 ****************************************************************************************/

VOID code_pop_a (INT16 n)
{
	while (n>0)
	{	s_pop_a(curr_block_graph);
		--n;
	}
}

VOID code_pop_b (INT16 n)
{
	while (n>0)
	{	s_pop_b(curr_block_graph);
		--n;
	}
}

VOID code_cleanup_a (INT16 n_elements,INT16 a_offset)
{
	INT16 element_n;
	
	for (element_n=n_elements-1; element_n>=0; --element_n)
	{	INSTRUCTION_GRAPH graph_1=s_get_a (curr_block_graph,element_n);
		s_put_a (curr_block_graph,element_n+a_offset,graph_1);
	}
	while (a_offset!=0)
	{	s_pop_a(curr_block_graph);
		--a_offset;
	}
}

VOID code_cleanup_b (INT16 n_elements,INT16 b_offset)
{
	INT16 element_n;
	
	for (element_n=n_elements-1; element_n>=0; --element_n)
	{	INSTRUCTION_GRAPH graph_1=s_get_b (curr_block_graph,element_n);
		s_put_b (curr_block_graph,element_n+b_offset,graph_1);
	}
	while (b_offset!=0)
	{	s_pop_b(curr_block_graph);
		--b_offset;
	}
}

VOID code_push_a (INT16 a_offset)
{
	INSTRUCTION_GRAPH graph_1;
	
	graph_1=s_get_a (curr_block_graph,a_offset);
	s_push_a (curr_block_graph,graph_1);
}

VOID code_push_b (INT16 b_offset)
{
	INSTRUCTION_GRAPH graph_1;
	
	graph_1=s_get_b (curr_block_graph,b_offset);
	s_push_b (curr_block_graph,graph_1);
}

VOID code_update_a (INT16 a_offset_1,INT16 a_offset_2)
{
	if (a_offset_1!=a_offset_2)
	{	INSTRUCTION_GRAPH graph_1;
		
		graph_1=s_get_a (curr_block_graph,a_offset_1);
		s_put_a (curr_block_graph,a_offset_2,graph_1);
	}
}

VOID code_update_b (INT16 b_offset_1,INT16 b_offset_2)
{
	if (b_offset_1!=b_offset_2)
	{	INSTRUCTION_GRAPH graph_1;
		
		graph_1=s_get_b (curr_block_graph,b_offset_1);
		s_put_b (curr_block_graph,b_offset_2,graph_1);
	}
}

VOID code_push_arg (INT16 a_offset,INT16 arity,INT16 argument_number)
{
	INSTRUCTION_GRAPH graph_1,graph_2,graph_3;
	
	graph_1=s_get_a (curr_block_graph,a_offset);
	if (argument_number==1)
		graph_3=g_load_nl(graph_1,ARG1_OFFSET);
	else if (argument_number==2 && arity==2)
		graph_3=g_load_nl(graph_1,ARG2_OFFSET);
	else
	{	graph_2=g_load_nl(graph_1,HNF_ARG_PTR_OFFSET);
		graph_3=g_load_nl(graph_2,argument_number-2);
	}
	
	s_push_a (curr_block_graph,graph_3);
}


VOID code_push_arg_b (INT16 a_offset)
{
	INSTRUCTION_GRAPH graph_1, arg_n_graph;
	
	arg_n_graph=s_get_b(curr_block_graph,0);

	if (arg_n_graph->code==GLOAD_CONST && arg_n_graph->parameters[0].int32!=2)
	{
		/* it is clear where the argument can be found, so generate inline
		   code */
		
		INT32 arg_n;
		
		arg_n=arg_n_graph->parameters[0].int32;
		
		s_pop_b(curr_block_graph);
		s_pop_b(curr_block_graph);
		
		code_push_arg (a_offset,arg_n,arg_n);
	} else
	{
		graph_1=s_get_a(curr_block_graph,a_offset);
		s_push_a(curr_block_graph,graph_1);
	
		instruction_call ("push_arg_b", 1,2,i_i_vector, 1,0,e_vector);
	}
}

VOID code_push_args (INT16 a_offset,INT16 arity,INT16 n_arguments)
{
	INSTRUCTION_GRAPH graph_1,graph_2,graph_3;
	
	if (n_arguments==0)
		return;
		
	graph_1=s_get_a (curr_block_graph,a_offset);
	graph_2=g_load_nl (graph_1,ARG1_OFFSET);
	
	if (n_arguments!=1)
	{	if (n_arguments==2 && arity==2)
		{	graph_3=g_load_nl (graph_1,ARG2_OFFSET);
			s_push_a (curr_block_graph,graph_3);
		} else		
		{	graph_3=g_load_nl (graph_1,HNF_ARG_PTR_OFFSET);
			--n_arguments;
			
			while (n_arguments!=0)
			{	INSTRUCTION_GRAPH graph_4;
			
				--n_arguments;
				graph_4=g_load_nl (graph_3, n_arguments);
				s_push_a (curr_block_graph,graph_4);
			}
		}
	}

	s_push_a (curr_block_graph,graph_2);
}

VOID code_push_node (STRING *label_name,INT16 n_arguments)
{
	INSTRUCTION_GRAPH graph_1,graph_2,graph_3,first_argument;
	LABEL *label;
	
	if (!options.parallel_code && strcmp (label_name->string,reserve_symbol)==0)
		label=enter_reserve_label(3);
	else
		label=enter_label (label_name,USE_LABEL | GCALL_LABEL | JSR_LABEL);
	
	set_label_layout(label,1,0,e_vector);
		
	graph_1=s_get_a (curr_block_graph,0);
		
	if (n_arguments>0)
	{
		register INT16 argument_n;
		
		argument_n=n_arguments;
		
		/* push all arguments, the first one should be on top */
		
		while (argument_n>1)
		{
			INSTRUCTION_GRAPH argument;

			--argument_n;
			argument=g_load_nl (graph_1,ARG1_OFFSET+argument_n);
			s_push_a (curr_block_graph,argument);
		}
		
		/* push first argument */
		
		first_argument=g_load_nl(graph_1,ARG1_OFFSET);
		s_push_a (curr_block_graph,first_argument);
		
	}
	
	graph_2=g_load_addr (label);
	
	/*	use 'large' fill to enable the node to be overwritten by the compiler.
		The arity of the actual node should be 0 and its size 3. If the node
		becomes overwritten by a reserved node, the waiting list field must be
		initialised as well. One should take care not to overwrite the
		original argument that was stored there. For this we use a save graph
	*/
	
	if (label->entry_type & RESERVED_NODE)
	{
		INSTRUCTION_GRAPH zero_waitlist;
		
		/* initialise waiting list to zero */
		
		zero_waitlist=g_load_const((INT32)0);
		
		if (n_arguments==0)
			graph_3=g_fill_4(graph_1,wait_graph,graph_2,zero_waitlist,NULL,0,3,nil_label);
		else
			graph_3=g_fill_4(graph_1,wait_graph,graph_2,g_save(first_argument,zero_waitlist),NULL,0,3,nil_label);
			
	} else
		graph_3=g_fill_4 (graph_1,wait_graph,graph_2,NULL,NULL,0,3,nil_label);
	
	s_put_a (curr_block_graph,n_arguments,graph_3);
}

VOID code_repl_args (INT16 arity,INT16 n_arguments)
{
	INSTRUCTION_GRAPH graph_1,graph_2,graph_3;
	
	if (n_arguments==0)
		return;
		
	graph_1=s_get_pop_a(curr_block_graph);
	graph_2=g_load_nl (graph_1,ARG1_OFFSET);
	
	if (n_arguments!=1)
	{	if (n_arguments==2 && arity==2)
		{	graph_3=g_load_nl (graph_1,ARG2_OFFSET);
			s_push_a (curr_block_graph,graph_3);
		} else
		{	
			graph_3=g_load_nl (graph_1, HNF_ARG_PTR_OFFSET);
			--n_arguments;
			
			while (n_arguments!=0)
			{	INSTRUCTION_GRAPH graph_4;
				
				--n_arguments;
				graph_4=g_load_nl (graph_3, n_arguments);
				s_push_a (curr_block_graph,graph_4);
			}
		}
	}
	s_push_a (curr_block_graph,graph_2);
}

VOID code_repl_arg (INT16 arity,INT16 argument_n)
{
	INSTRUCTION_GRAPH graph_1,graph_2;
	
	graph_1=s_get_pop_a(curr_block_graph);
	
	if (argument_n < 2)
		graph_2 = g_load_nl (graph_1, ARG1_OFFSET);
	else if (argument_n == 2 && arity == 2)
		graph_2 = g_load_nl (graph_1, ARG2_OFFSET);
	else
	{	INSTRUCTION_GRAPH graph_3;
		
		graph_3=g_load_nl (graph_1, HNF_ARG_PTR_OFFSET);
		graph_2=g_load_nl (graph_3, argument_n-2);
	}
	
	s_push_a (curr_block_graph,graph_2);
}

/**************************************************************************************
 *                                                                                    *
 *           repl_args_b                				                              *
 *                                                                                    *
 **************************************************************************************/
 
VOID code_repl_args_b (VOID)
{	
	/* This library routine may not be called in the normal way (by means of
	   "instruction_call"), as the correspondence between this block and the
	   following is completely unknown. Also, for this block stack checks can't
	   be combined */

	if(repl_args_label==NULL)
	{	repl_args_label = enter_label2("repl_args_b", USE_LABEL | GCALL_LABEL | JSR_LABEL | IMPORT_LABEL);
		set_label_layout (repl_args_label,1,2,i_i_vector);
	}

	end_label_layout_block (curr_block_graph,repl_args_label);
	
	if (code_reachable)
	{
		/*	if the code is not reachable, the instruction block returned is the
			last useful one, and this one is ready */
		
		append_instr_label_int16(curr_block_graph->instruction_block,GCALL,repl_args_label,0); /* no increase of b-stack */
		block_ready(curr_block_graph,true);
	}
	curr_block_graph=begin_layout_block(1,0,e_vector);
	
	fix_ctop_stack();
}

/**************************************************************************************
 *                                                                                    *
 *          create and fill instructions                                              *
 *                                                                                    *
 **************************************************************************************/
		
VOID code_create (INT16 n_arguments)
{
	INSTRUCTION_GRAPH graph_1,graph_3;
	
	/* the reserve label and cycle in spine label have a defer descriptor by default
	   (defined in _system.abc). Create will deliver a reserved node with the
	   waiting list set to zero. The arity of the node will be zero. The size will be
	   set to 3 at least */
	
	if (n_arguments<=2)
	{
		/* this node has the minimum size: 3 words */
		
		graph_1=g_load_addr (enter_reserve_label(3));
		graph_3=g_create_3 (graph_1,g_load_const((INT32)0),NULL,0,3,nil_label);
	}
	else
	{	INT16 n;
		
		/*	special reserve label to ensure the garbage collector keeps the
			total node. Its arity is 0 and its size n_arguments+1 */
		
		graph_1=g_load_addr (enter_reserve_label(n_arguments+1));
		
		graph_3=g_create_m (n_arguments+1,0,n_arguments+1,nil_label);	/* code+args				*/
		graph_3->parameters[0].pointer=graph_1;							/* special reserve label	*/
		graph_3->parameters[1].pointer=g_load_const((INT32)0);
		
		/* all arguments except waiting list are NULL */
		
		for (n=1; n<n_arguments; ++n)
			graph_3->parameters[n+1].pointer=NULL;
	}
	
	s_push_a (curr_block_graph,graph_3);
}

VOID code_fill (STRING *descriptor_name,INT16 arity,STRING *code_name,INT16 a_offset)
{
	INSTRUCTION_GRAPH graph_1,graph_2,graph_3,graph_4,graph_5,graph_6;
	LABEL *descriptor_label, *code_label;
	
	/*	If a waiting list has been extracted from this node earlier, the corresponding
		global variable will not be NULL. It will be stored in the node to
		indicate the waiting list must be saved before the corresponding field
		gets overwritten. We do not use a SAVE graph for this, for optimisation
		reasons: it is now easy to check if a waiting list should be saved.
		This is mainly usefull if one want to avoid retrieving empty waiting
		lists
	*/
	
	graph_1 = s_get_a (curr_block_graph,a_offset);
	
	if (strcmp (code_name->string,hnf_symbol)==0)
	{	
		/* a rnf node is constructed */
		
		descriptor_label=enter_label (descriptor_name,USE_LABEL | DATA_LABEL);

		graph_2 = g_load_desc_id(descriptor_label,arity);
		
		switch (arity)
		{	case 0:
				graph_4=g_fill_3 (graph_1,wait_graph,graph_2,NULL,arity,arity+1,descriptor_label);
				break;
			case 1:
				graph_5=s_get_pop_a(curr_block_graph);
				graph_4=g_fill_3 (graph_1,wait_graph,graph_2,graph_5,arity,arity+1,descriptor_label);
				break;
			case 2:
				graph_5=s_get_pop_a(curr_block_graph);
				graph_6=s_get_pop_a(curr_block_graph);
				graph_4=g_fill_4 (graph_1,wait_graph,graph_2,graph_5,graph_6,arity,arity+1,descriptor_label);
				break;
			default:
			{
				INT16 n_arguments;
				PARAMETER *parameter;
				
				graph_5=s_get_pop_a(curr_block_graph);
				
				graph_3=g_create_m (arity-1,0,arity-1,nil_label);
				parameter=graph_3->parameters;
				for (n_arguments=arity-1; n_arguments>0; --n_arguments)
				{	graph_6=s_get_pop_a(curr_block_graph);
					parameter->pointer=graph_6;
					++parameter;
				}

				graph_4=g_fill_4 (graph_1,wait_graph,graph_2,graph_5,graph_3,arity,3,descriptor_label);
			}
		}
		
		s_put_a (curr_block_graph,a_offset-arity,graph_4);

	} else
	{
		/* a closure is constructed */
		
		INT16 n_arguments;
		PARAMETER *parameter;

		/* if "_" for descriptor, it is non-existent, use nil for descriptor, */
		
		if (descriptor_name->string[0]=='_' && descriptor_name->length==1)
			descriptor_label=nil_label;
		else
			descriptor_label=enter_label (descriptor_name,USE_LABEL | DATA_LABEL);

		/* if arity < 0, a node is filled with a selector function (arity 1) */
		
		if (arity<0)
			arity=1;

		if (!options.parallel_code && strcmp (code_name->string,reserve_symbol)==0)
			code_label=enter_reserve_label(3);
		else
			code_label=enter_label (code_name,USE_LABEL | GCALL_LABEL | JSR_LABEL);
		
		set_label_layout(code_label,1,0,e_vector);
		
		if (arity<2)
		{
			/* fill with 'larger' node to enable it to get overwritten */
			
			graph_2=g_fill_m (graph_1,wait_graph,3,arity,3,descriptor_label);
			graph_2->parameters[3].pointer=NULL;
			graph_2->parameters[4].pointer=NULL;
			
		} else
			graph_2=g_fill_m (graph_1,wait_graph,arity+1,arity,arity+1,descriptor_label);
		
		/*	address and 'waitlist save indication' initialised, overwrite node
			by new args */
		
		parameter=&graph_2->parameters[2];

		parameter->pointer=g_load_addr (code_label);
		++parameter;
				
		for (n_arguments=arity; n_arguments>0; --n_arguments)
		{
			parameter->pointer=s_get_pop_a(curr_block_graph);
			++parameter;
		}
		
		optimise_fill_m(graph_2,arity+1);

		s_put_a (curr_block_graph,a_offset-arity,graph_2);
	}
}

VOID code_fillI (INT32 value,INT16 a_offset)
{
	INSTRUCTION_GRAPH graph_1,graph_2,graph_3,graph_4;
	
	INT_label->flags|=USE_LABEL;
		
	graph_1=s_get_a (curr_block_graph,a_offset);
	
	graph_2=g_load_desc_id(INT_label,0);
	graph_3=g_load_const (value);
	graph_4=g_fill_3 (graph_1,wait_graph,graph_2,graph_3,0,2,INT_label);
	
	s_put_a (curr_block_graph,a_offset,graph_4);
}

VOID code_fillB (INT32 value,INT16 a_offset)
{
	INSTRUCTION_GRAPH graph_1,graph_2,graph_3,graph_4;
	
	BOOL_label->flags|=USE_LABEL;
		
	graph_1=s_get_a (curr_block_graph,a_offset);

	graph_2=g_load_desc_id (BOOL_label,0);
	graph_3=g_load_const (value);
	graph_4=g_fill_3 (graph_1,wait_graph,graph_2,graph_3,0,2,BOOL_label);
	
	s_put_a (curr_block_graph,a_offset,graph_4);
}

VOID code_fillC (INT32 value,INT16 a_offset)
{
	INSTRUCTION_GRAPH graph_1,graph_2,graph_3,graph_4;
	
	CHAR_label->flags|=USE_LABEL;
		
	graph_1=s_get_a (curr_block_graph,a_offset);
	
	graph_2=g_load_desc_id (CHAR_label,0);
	graph_3=g_load_const (value);
	graph_4=g_fill_3 (graph_1,wait_graph,graph_2,graph_3,0,2,CHAR_label);
	
	s_put_a (curr_block_graph,a_offset,graph_4);
}

VOID code_fillR (FLOAT *value,INT16 a_offset)
{
	INSTRUCTION_GRAPH graph_1,graph_2,graph_3,graph_4;

	REAL_label->flags|=USE_LABEL;
		
	graph_1=s_get_a (curr_block_graph,a_offset);
	
	graph_2=g_load_desc_id (REAL_label,0);
	graph_3=g_fload_const (value);
	
	graph_4=g_fill_r(graph_1,wait_graph,graph_2,graph_3);
	
	s_put_a (curr_block_graph,a_offset,graph_4);
}

VOID code_fillS (STRING *string,INT16 a_offset)
{
	INSTRUCTION_GRAPH graph_1,graph_2,graph_3,graph_4;
	
	STRING2_label->flags|=USE_LABEL;
	
	graph_1=s_get_a (curr_block_graph,a_offset);
	
	graph_2=g_load_desc_id (STRING2_label,0);
	graph_3=g_create_string (string);
	graph_4=g_fill_3 (graph_1,wait_graph,graph_2,graph_3,0,2,STRING2_label);
	
	s_put_a (curr_block_graph,a_offset,graph_4);
}

VOID code_fillI_b (INT16 b_offset,INT16 a_offset)
{
	INSTRUCTION_GRAPH graph_1,graph_2,graph_3,graph_4;
	
	INT_label->flags|=USE_LABEL;
	
	graph_1=s_get_a (curr_block_graph,a_offset);
	
	graph_2=g_load_desc_id (INT_label,0);
	graph_3=s_get_b (curr_block_graph,b_offset);
	graph_4=g_fill_3 (graph_1,wait_graph,graph_2,graph_3,0,2,INT_label);

	s_put_a (curr_block_graph,a_offset,graph_4);
}

VOID code_fillB_b (INT16 b_offset,INT16 a_offset)
{
	INSTRUCTION_GRAPH graph_1,graph_2,graph_3,graph_4;
	
	BOOL_label->flags|=USE_LABEL;
		
	graph_1=s_get_a (curr_block_graph,a_offset);
	
	graph_2=g_load_desc_id (BOOL_label,0);
	graph_3=s_get_b (curr_block_graph,b_offset);
	graph_4=g_fill_3 (graph_1,wait_graph,graph_2,graph_3,0,2,BOOL_label);

	s_put_a (curr_block_graph,a_offset,graph_4);
}

VOID code_fillC_b (INT16 b_offset,INT16 a_offset)
{
	INSTRUCTION_GRAPH graph_1,graph_2,graph_3,graph_4;
	
	CHAR_label->flags|=USE_LABEL;
		
	graph_1=s_get_a (curr_block_graph,a_offset);
	
	graph_2=g_load_desc_id (CHAR_label,0);
	graph_3=s_get_b (curr_block_graph,b_offset);
	graph_4=g_fill_3 (graph_1,wait_graph,graph_2,graph_3,0,2,CHAR_label);

	s_put_a (curr_block_graph,a_offset,graph_4);
}

VOID code_fillF_b (INT16 b_offset,INT16 a_offset)
{
	INSTRUCTION_GRAPH graph_1,graph_2,graph_3,graph_4,graph_5;
		
	FILE_label->flags|=USE_LABEL;
		
	graph_1=s_get_a (curr_block_graph,a_offset);
	graph_2=g_load_desc_id (FILE_label,0);
	
	graph_3=s_get_b (curr_block_graph,b_offset);
	graph_4=s_get_b (curr_block_graph,b_offset+1);
	
	graph_5=g_fill_4 (graph_1,wait_graph,graph_2,graph_3,graph_4,0,3,FILE_label);

	s_put_a (curr_block_graph,a_offset,graph_5);
}

VOID code_fillR_b (INT16 b_offset,INT16 a_offset)
{
	INSTRUCTION_GRAPH graph_1,graph_2,graph_3,graph_4,graph_5,graph_6;
	
	REAL_label->flags|=USE_LABEL;
		
	graph_1=s_get_b (curr_block_graph,b_offset);
	graph_2=s_get_b (curr_block_graph,b_offset+1);
	
	graph_3=s_get_a (curr_block_graph,a_offset);
	
	graph_4=g_load_desc_id (REAL_label,0);
	
	graph_5=g_fjoin (graph_1,graph_2);
	
	graph_6=g_fill_r (graph_3,wait_graph,graph_4,graph_5);
	
	s_put_a (curr_block_graph,a_offset,graph_6);
}

VOID code_fillS_symbol (INT16 a_offset_1,INT16 a_offset_2)
{
	INSTRUCTION_GRAPH graph_1,graph_2;
	
	graph_1=s_get_a (curr_block_graph,a_offset_1);
	graph_2=s_get_a (curr_block_graph,a_offset_2);
	
	s_push_a (curr_block_graph,graph_2);
	s_push_a (curr_block_graph,graph_1);
	
	/*	local 5 contains the destination node
		local 6 contains the source node */
	
	/*	this instruction may be called when a waiting list is available. The
		routine below leaves the waiting list on top of the a-stack */
	
	if (wait_graph==NULL)
		instruction_call ("DESC_TO_STRING",2,0,e_vector,0,0,e_vector);
	else
	{
		/*	there's a waiting list that should be preserved. It should be the waiting
			list of the node that is going to be overwritten */
		
		if(wait_graph->code==GLOAD_CONST)
		{
			/*	the waitlist has been defined locally, so it doesn't have to be
				taken out of the node itself */
			
			instruction_call ("DESC_TO_STRING",2,0,e_vector,0,0,e_vector);
			
			/*	just keep the same graph in the waitlist register */
		
		} else
		{
			/*	waitlist will be taken out of this node before it is filled.
				It gets pushed on the a-stack */
			
			instruction_call ("DESC_TO_STRING",2,0,e_vector,1,0,e_vector);
			wait_graph=s_get_pop_a(curr_block_graph);
		}
	}
}

VOID code_fill_a (INT16 from_offset,INT16 to_offset)
{
	INSTRUCTION_GRAPH graph_1,graph_2,graph_3;
	
	if (from_offset==to_offset)
		return;
	
	graph_1=s_get_a (curr_block_graph,from_offset);
	graph_2=s_get_a (curr_block_graph,to_offset);
	graph_3=g_fill_copy(graph_1,graph_2,wait_graph,3*WORD_SIZE);
	
	s_put_a (curr_block_graph,to_offset,graph_3);
}

void code_set_defer (INT16 a_offset)
{
	INSTRUCTION_GRAPH closure;
	
	/*
		Note that one may set the code generator to ignore defer instructions
		
		Create a new entry for the original node entry of closure at the
		given offset. The decriptor of this entry should be set to _Defer,
		its arity should be the same as the arity of the original node.
	*/
	
	if (options.ignore_defer)
	{
		warning_ignore_defer();
		return;
	}
	
	closure=s_get_a(curr_block_graph,a_offset);
	
	switch (closure->code)
	{
		case GFILL:
		{
			INSTRUCTION_GRAPH code_address;

			code_address=closure->parameters[2].pointer;
			if (code_address->code==GLOAD_ADDR)
			{
				LABEL *entry_label;
				LABEL *new_entry;

				/* code address, it is not a descriptor: code is not GLOAD_DESC_ID */
				
				entry_label=code_address->parameters[0].label;
				new_entry=enter_defer_label(closure->info,entry_label);
				
				closure->parameters[2].pointer=g_load_addr(new_entry);
				return;
			}
			break;
		}
		case GCREATE:
		{
			INSTRUCTION_GRAPH code_address;
			
			code_address=closure->parameters[0].pointer;
			if (code_address->code==GLOAD_ADDR)
			{
				LABEL *entry_label;
				LABEL *new_entry;

				/* code address, it is not a descriptor: code is not GLOAD_DESC_ID */
				
				entry_label=code_address->parameters[0].label;
				new_entry=enter_defer_label(closure->info,entry_label);
				
				closure->parameters[0].pointer=g_load_addr(new_entry);
				return;
			}
		}
	}
	error ("argument of set_defer is not a closure\n");
}

VOID code_set_entry (STRING *label_name,INT16 a_offset)
{
	INSTRUCTION_GRAPH graph_1,graph_2,graph_3;
	LABEL *label;
	
	/* set entry does not initialise the waitlist field */
	
	if (!options.parallel_code && strcmp (label_name->string,reserve_symbol)==0)
		label=enter_reserve_label(3);
	else
		label=enter_label (label_name, USE_LABEL | GCALL_LABEL | JSR_LABEL);
	set_label_layout(label,1,0,e_vector);
	
	graph_1=s_get_a (curr_block_graph,a_offset);
	graph_2=g_load_addr (label);
	
	/* fill with 'large' node to enable it to be overwritten */
	
	graph_3=g_fill_4 (graph_1,wait_graph,graph_2,NULL,NULL,0,3,nil_label);
	
	s_put_a (curr_block_graph,a_offset,graph_3);
}

/**************************************************************************************
 *                                                                                    *
 *          operations on nodes                                                       *
 *                                                                                    *
 **************************************************************************************/
 
 VOID code_eq_desc (STRING *descriptor_name,INT16 arity,INT16 a_offset)
{
	INSTRUCTION_GRAPH graph_1,graph_2,graph_3;
	LABEL *descriptor;
	
	descriptor=enter_label (descriptor_name,USE_LABEL | DATA_LABEL);
	
	graph_1=s_get_a (curr_block_graph,a_offset);
	graph_2=g_load_nl (graph_1,DESCRIPTOR_OFFSET);
	graph_3=g_equal_desc_id (graph_2, descriptor, arity);
	
	s_push_b (curr_block_graph,graph_3);
}

VOID code_eq_nulldesc (STRING *descriptor_name,INT16 a_offset)
{
	INSTRUCTION_GRAPH graph_1,graph_2,graph_3,graph_4,graph_5,graph_6,graph_7,
	graph_8,graph_9;
	LABEL *descriptor;
	
	descriptor=enter_label (descriptor_name,USE_LABEL | DATA_LABEL);
	
	graph_1=s_get_a(curr_block_graph,a_offset);
	graph_2=g_load_nl(graph_1,DESCRIPTOR_OFFSET);
	graph_3=g_desc_table ();
	graph_4=g_addI(graph_2,graph_3);		/* get real descriptor address */
	
	graph_5=g_load_nl(graph_4,0);			/* get arity */
	graph_6=g_load_const((INT32)3);
	graph_7=g_L_shiftl(graph_5, graph_6);	/* get offset to start of descriptor */
	graph_8=g_subI(graph_2,graph_7);		/* get offset of descriptor [0] */
	
	graph_9=g_equal_desc_id (graph_8,descriptor,0);
	
	s_push_b (curr_block_graph,graph_9);
}

VOID code_eq_symbol (INT16 a_offset_1,INT16 a_offset_2)
{
	INSTRUCTION_GRAPH graph_1,graph_2, graph_3;
	
	if (eq_symbol_label==NULL)
	{	eq_symbol_label=enter_label2 ("EQ_SYMBOL",USE_LABEL | JUMP_LABEL | JSR_LABEL | IMPORT_LABEL);
		set_label_layout(eq_symbol_label,2,0,e_vector);
	}
	
	graph_1=s_get_a (curr_block_graph,a_offset_1);
	graph_2=s_get_a (curr_block_graph,a_offset_2);
			
	graph_3 = g_instruction2(GEQ_SYMBOL, graph_1, graph_2);
	s_push_b (curr_block_graph,graph_3);
}

VOID code_get_desc_arity (INT16 a_offset)
{
	INSTRUCTION_GRAPH graph_1,graph_2,graph_3,graph_4,graph_5,graph_6,graph_7,
	graph_8,graph_9;
	
	graph_1=s_get_a(curr_block_graph,a_offset);
	graph_2=g_load_nl(graph_1,DESCRIPTOR_OFFSET);
	graph_3=g_desc_table();
	graph_4=g_addI(graph_2,graph_3);		/* get real desc address */
	
	graph_5=g_load_nl(graph_4,0);			/* get arity */
	graph_6=g_load_const ((INT32)3);
	graph_7=g_L_shiftl(graph_5, graph_6);	/* offset to start of descriptor */
	
	graph_8=g_subI(graph_4,graph_7);
	graph_9=g_load_nl(graph_8,ARITY_OFFSET);

	s_push_b (curr_block_graph,graph_9);
}

VOID code_get_node_arity (INT16 a_offset)
{
	INSTRUCTION_GRAPH graph_1,graph_2,graph_3,graph_4,graph_5;
	
	graph_1=s_get_a (curr_block_graph,a_offset);
	graph_2=g_load_nl (graph_1,DESCRIPTOR_OFFSET);
	graph_3=g_desc_table ();
	graph_4=g_addI(graph_2,graph_3);
	graph_5=g_load_nl (graph_4,0);		/* get arity */
	
	s_push_b (curr_block_graph,graph_5);
}

/**************************************************************************************
 *                                                                                    *
 *          operations on strings                                                     *
 *                                                                                    *
 **************************************************************************************/
 
 VOID code_cmpS (INT16 a_offset_1,INT16 a_offset_2)
{
	INSTRUCTION_GRAPH graph_1,graph_2,graph_3,graph_4,graph_5;
	
	if (cmp_string_label==NULL)
	{	cmp_string_label=enter_label2 ("STR_CMP",USE_LABEL | JUMP_LABEL | JSR_LABEL | IMPORT_LABEL);
		set_label_layout(cmp_string_label,0,0,e_vector);
	}
		
	graph_1=s_get_a (curr_block_graph,a_offset_1);
	graph_2=s_get_a (curr_block_graph,a_offset_2);
	
	graph_3=g_load_nl(graph_1, ARG1_OFFSET);
	graph_4=g_load_nl(graph_2, ARG1_OFFSET);
	
	graph_5=g_instruction2(GSTR_CMP, graph_3, graph_4);
	
	s_push_b(curr_block_graph,graph_5);
}

VOID code_eqS_a (STRING *string,INT16 a_offset)
{
	INSTRUCTION_GRAPH graph_1,graph_2,graph_3,graph_4;
	LABEL *string_label;	
	
	if (equal_string_label==NULL)
	{	equal_string_label=enter_label2 ("STR_EQ",USE_LABEL | JUMP_LABEL | JSR_LABEL | IMPORT_LABEL);
		set_label_layout(equal_string_label,0,0,e_vector);
	}
	
	string_label=define_string (string);
	
	graph_1=s_get_a (curr_block_graph,a_offset);
	graph_2=g_load_nl (graph_1,ARG1_OFFSET);
	graph_3=g_load_addr (string_label);
	
	graph_4=g_instruction2(GSTR_EQ, graph_2, graph_3);
	
	s_push_b(curr_block_graph,graph_4);
}

VOID code_catS (INT16 source_offset_1,INT16 source_offset_2,INT16 destination_offset)
{
	INSTRUCTION_GRAPH graph_1,graph_2,graph_3;
				
	graph_1=s_get_a (curr_block_graph,source_offset_1);
	graph_2=s_get_a (curr_block_graph,source_offset_2);
	graph_3=s_get_a (curr_block_graph,destination_offset);
	
	s_push_a (curr_block_graph,graph_3);
	s_push_a (curr_block_graph,graph_2);
	s_push_a (curr_block_graph,graph_1);
	
	/*	local5 contains the destination
		local6 contains source2
		local7 contains source1	*/

	instruction_call ("STR_CAT",3,0,e_vector,0,0,e_vector);
}

VOID code_indexS (INT16 string_offset)
{
	INSTRUCTION_GRAPH graph_1,graph_2,graph_3,graph_4,graph_5,graph_6;
	
	graph_1=s_get_b (curr_block_graph,0);
	graph_2=g_load_nlp (graph_1, STRING_FIRST_CHAR_OFFSET);
	
	graph_3=s_get_a (curr_block_graph,string_offset);
	graph_4=g_load_nl (graph_3,ARG1_OFFSET);
	
	graph_5=g_byte_sub (graph_2, graph_4);
	graph_6=g_load_b (graph_5);
	
	s_put_b (curr_block_graph,0,graph_6);
}

VOID code_updateS (INT16 source_offset,INT16 destination_offset)
{
	INSTRUCTION_GRAPH graph_1,graph_2;
		
	graph_1=s_get_a (curr_block_graph,source_offset);
	graph_2=s_get_a (curr_block_graph,destination_offset);
	
	s_push_a (curr_block_graph,graph_2);
	s_push_a (curr_block_graph,graph_1);
	
	/*	local1 contains an index
		local2 contains a character
		local5 contains the destination
		local6 contains the source */
	
	instruction_call ("STR_UPDATE",2,2,i_i_vector,0,0,e_vector);
}

VOID code_sliceS (INT16 source_offset,INT16 destination_offset)
{
	INSTRUCTION_GRAPH graph_1,graph_2;
		
	graph_1=s_get_a (curr_block_graph,source_offset);
	graph_2=s_get_a (curr_block_graph,destination_offset);
	
	s_push_a (curr_block_graph,graph_2);
	s_push_a (curr_block_graph,graph_1);
	
	/*	local1 contains index 'to'
		local2 contains index 'from'
		local5 contains the destination
		local6 contains the source */
	
	instruction_call ("STR_SLICE",2,2,i_i_vector,0,0,e_vector);
}

VOID code_lenS (INT16 string_offset)
{
	INSTRUCTION_GRAPH graph_1,graph_2,graph_3;
	
	graph_1=s_get_a (curr_block_graph,string_offset);
	graph_2=g_load_nl (graph_1,ARG1_OFFSET);
	graph_3=g_load_nl (graph_2,STRING_LENGTH_OFFSET);
	
	s_push_b (curr_block_graph,graph_3);
}

/************************************************************************************
 *																					*
 *		print operations															*
 *																					*
 ************************************************************************************/

VOID code_print (STRING *string)
{
	INSTRUCTION_GRAPH graph_1;
	LABEL *string_label;
	
	string_label=define_string (string);
	
	graph_1=g_load_addr(string_label);
	s_push_b (curr_block_graph,graph_1);
	
	instruction_call("print_string",0,1,i_vector,0,0,e_vector);
}

VOID code_print_sc (STRING *string)
{
	INSTRUCTION_GRAPH graph_1;
	LABEL *string_label;
		
	string_label=define_string (string);
	
	graph_1=g_load_addr(string_label);
	s_push_b (curr_block_graph,graph_1);
	
	instruction_call("print_sc_string",0,1,i_vector,0,0,e_vector);
}

VOID code_print_symbol (INT16 a_offset)
{
	INSTRUCTION_GRAPH graph_1;
		
	graph_1=s_get_a (curr_block_graph,a_offset);
	s_push_a (curr_block_graph,graph_1);
	
	instruction_call("print_symbol",1,0,e_vector,0,0,e_vector);
}

VOID code_print_symbol_sc (INT16 a_offset)
{
	INSTRUCTION_GRAPH graph_1;
		
	graph_1=s_get_a (curr_block_graph,a_offset);
	s_push_a (curr_block_graph,graph_1);

	instruction_call("print_symbol_sc",1,0,e_vector,0,0,e_vector);
}

VOID code_printD (VOID)
{	
	instruction_call("printD",0,1,i_vector,0,0,e_vector);
}

VOID code_pushD (STRING *descriptor)
{
	INSTRUCTION_GRAPH graph_1;
	LABEL *label;
	
	label=enter_label (descriptor, USE_LABEL);
	
	if (label->flags & DATA_LABEL)	
	{
		/*	descriptors must have been defined earlier, otherwise the label
			is assumed to point to a string. */
				
		switch (label->def_type)
		{
			case D_NODATA:
				break;							/* just use this label */
		
			case D_STRING:
				if (label!=label->def.string->label)
					internal_error("code_pushD: data labels do not match\n");
				break;
				
			case D_DESCRIPTOR:
			
				/* get string label of descriptor and set its use flag */
				
				label=label->def.descriptor->string_label;
				label->flags|=USE_LABEL;
				break;
			
			default:
				error("pushD: label \"%s\" has incorrect definition type %d\n",descriptor->string,label->def_type);
				return;
		}
				
	} else
	{
		/* assume this is a string label, just indicate it is a data label */
		
		label->flags|=DATA_LABEL;
	}
	
	graph_1=g_load_addr(label);
	s_push_b (curr_block_graph,graph_1);
}

/************************************************************************************
 *																					*
 *		file operations																*
 *																					*
 ************************************************************************************/

VOID code_pushF_a (INT16 a_offset)
{
	INSTRUCTION_GRAPH graph_1,graph_2,graph_3;
	
	graph_1=s_get_a (curr_block_graph,a_offset);
	graph_2=g_load_nl (graph_1,ARG1_OFFSET);
	graph_3=g_load_nl (graph_1,ARG2_OFFSET);
	
	s_push_b (curr_block_graph,graph_3);
	s_push_b (curr_block_graph,graph_2);
}

VOID code_openF (INT16 file_name_offset,INT16 file_type_offset)
{
	INSTRUCTION_GRAPH graph_1,graph_2;
			
	graph_1=s_get_a (curr_block_graph,file_type_offset);
	graph_2=s_get_a (curr_block_graph,file_name_offset);
	
	s_push_a (curr_block_graph,graph_1);
	s_push_a (curr_block_graph,graph_2);
	
	instruction_call("open_file",2,0,e_vector,0,2,i_i_vector);
}

VOID code_reopenF (VOID)
{
	instruction_call("reopen_file",0,2,i_i_vector,0,2,i_i_vector);	
}

VOID code_endF (VOID)
{
	instruction_call("end_file",0,2,i_i_vector,0,1,i_vector);
}

VOID code_getFC (VOID)
{	
	instruction_call("file_getC",0,2,i_i_vector,0,3,i_i_i_vector);
}

VOID code_getFS (INT16 string_offset)
{
	INSTRUCTION_GRAPH graph_1;
		
	graph_1=s_get_a (curr_block_graph,string_offset);
	s_push_a (curr_block_graph,graph_1);
	
	instruction_call("file_getS",1,3,i_i_i_vector,0,2,i_i_vector);
}

VOID code_putFC (VOID)
{
	instruction_call("file_putC",0,3,i_i_i_vector,0,2,i_i_vector);
}

VOID code_putFS (INT16 a_offset)
{
	INSTRUCTION_GRAPH graph_1;
		
	graph_1=s_get_a (curr_block_graph,a_offset);
	s_push_a (curr_block_graph,graph_1);
	
	instruction_call("file_putS",1,2,i_i_vector,0,2,i_i_vector);
}

/**************************************************************************************
 *                                                                                    *
 *          labels etc.                                                               *
 *                                                                                    *
 **************************************************************************************/

VOID code_desc (STRING *label_name,STRING *node_entry_name,STRING *lazy_entry_name,INT16 arity,STRING *descriptor_name)
{
	LABEL *label,*node_entry_label,*lazy_entry_label;
	
	register INT16 n;
	
	if (code_reachable)
		error("descriptor placed in code area\n");
	
	for (n=0; n<arity-1; ++n)
	{	if (n>MAX_YET_ARGS_NEEDED_ARITY)
		{	if (yet_args_needed_label==NULL)
			{
				yet_args_needed_label=enter_label2 ("yet_args_needed", USE_LABEL | GCALL_LABEL | JSR_LABEL | IMPORT_LABEL);
				set_label_layout(yet_args_needed_label,2,0,e_vector);

				switch (options.assembler)
				{	case PARC_ASSEMBLER:

						/* Par.C linker cannot resolve words with ext references */
						
						init_yet_args_needed_label=enter_label2 ("init_yet_args_needed", USE_LABEL | GCALL_LABEL | JSR_LABEL);
						set_label_layout(init_yet_args_needed_label,2,0,e_vector);
						define_label(init_yet_args_needed_label);
						break;
					
					case PACT_ASSEMBLER:
						init_yet_args_needed_label=yet_args_needed_label;
						break;
					
					default:
						internal_error("code_desc: unknown assembler type (1)");
				}
				
				break;
			}
		} else
		{	LABEL **yet_args_needed_label_p;
			LABEL **init_yet_args_needed_label_p;
		
			yet_args_needed_label_p=&yet_args_needed_labels[n];
			init_yet_args_needed_label_p=&init_yet_args_needed_labels[n];
			
			if (*yet_args_needed_label_p==NULL)
			{	CHAR slabel_name[64];
			
				sprintf (slabel_name, "yet_args_needed_%d", n);
				*yet_args_needed_label_p=enter_label2 (slabel_name,USE_LABEL | GCALL_LABEL | JSR_LABEL | IMPORT_LABEL);
				set_label_layout(*yet_args_needed_label_p,2,0,e_vector);

				switch (options.assembler)
				{	case PARC_ASSEMBLER:
						sprintf (slabel_name, "init_yet_args_needed_%d", n);
						*init_yet_args_needed_label_p=enter_label2 (slabel_name, USE_LABEL | GCALL_LABEL | JSR_LABEL);
						set_label_layout(*init_yet_args_needed_label_p,2,0,e_vector);
						define_label(*init_yet_args_needed_label_p);
						break;
						
					case PACT_ASSEMBLER:
						*init_yet_args_needed_label_p=*yet_args_needed_label_p;
						break;
					
					default:
						internal_error("code_desc: unknown assembler type (2)");
				}
			}
		}
	}
	
	if (arity>0)
	{
		/* determine code of lazy entry */
		
		if (strcmp(lazy_entry_name->string,add_arg_symbol)==0)
		{
			/* call standard routine to add last argument */
			
			if (arity-1>MAX_YET_ARGS_NEEDED_ARITY)
			{
				if (yet_args_needed_label==NULL)
				{
					yet_args_needed_label=enter_label2 ("yet_args_needed", USE_LABEL | GCALL_LABEL | JSR_LABEL | IMPORT_LABEL);
					set_label_layout(yet_args_needed_label,2,0,e_vector);

					switch (options.assembler)
					{	case PARC_ASSEMBLER:
					
							init_yet_args_needed_label=enter_label2 ("init_yet_args_needed", USE_LABEL | GCALL_LABEL | JSR_LABEL);
							set_label_layout(init_yet_args_needed_label,2,0,e_vector);
							define_label(init_yet_args_needed_label);
							break;
						
						case PACT_ASSEMBLER:
						
							init_yet_args_needed_label=yet_args_needed_label;
							break;
						
						default:
							internal_error("code_desc: unknown assembler type (3)");
					}				
				}
				lazy_entry_label=init_yet_args_needed_label;

			} else
			{	LABEL **yet_args_needed_label_p;
				LABEL **init_yet_args_needed_label_p;
			
				yet_args_needed_label_p=&yet_args_needed_labels[arity-1];
				init_yet_args_needed_label_p=&init_yet_args_needed_labels[arity-1];
				
				if (*yet_args_needed_label_p==NULL)
				{	CHAR slabel_name[64];
				
					sprintf (slabel_name, "yet_args_needed_%d", arity-1);
					*yet_args_needed_label_p=enter_label2 (slabel_name,USE_LABEL | GCALL_LABEL | JSR_LABEL | IMPORT_LABEL);
					set_label_layout(*yet_args_needed_label_p,2,0,e_vector);

					switch (options.assembler)
					{	case PARC_ASSEMBLER:
							sprintf (slabel_name, "init_yet_args_needed_%d", arity-1);
							*init_yet_args_needed_label_p=enter_label2 (slabel_name, USE_LABEL | GCALL_LABEL | JSR_LABEL);
							set_label_layout(*init_yet_args_needed_label_p,2,0,e_vector);
							define_label(*init_yet_args_needed_label_p);
							break;
						
						case PACT_ASSEMBLER:
							*init_yet_args_needed_label_p=*yet_args_needed_label_p;
							break;
						
						default:
							internal_error("code_desc: unknown assembler type (4)");
					}		
				}
				lazy_entry_label=*init_yet_args_needed_label_p;
			}
		
		} else
		{	if (strcmp(lazy_entry_name->string,hnf_symbol)==0)
			{
				if (init_hnf_label==NULL)
				{	hnf_label=enter_label2 (hnf_symbol, USE_LABEL | GCALL_LABEL | JSR_LABEL | IMPORT_LABEL);
				
					switch (options.assembler)
					{	case PARC_ASSEMBLER:
							init_hnf_label=enter_label2 (init_hnf_symbol, USE_LABEL | GCALL_LABEL | JSR_LABEL);
							define_label(init_hnf_label);
							break;
						
						case PACT_ASSEMBLER:
							init_hnf_label=hnf_label;
							break;
							
						default:
							internal_error("code_desc: unknown assembler type (5)");
					}
				}
				lazy_entry_label=init_hnf_label;
			} else
				lazy_entry_label=enter_label (lazy_entry_name,USE_LABEL | GCALL_LABEL | JSR_LABEL);
		}
	} else
	{
		/* arity 0: no lazy entry */
		
		lazy_entry_label=NULL;
	}

	switch (options.assembler)
	{	case PARC_ASSEMBLER:
	
			/* the node entry is either a local label (in which case the linker can
			   fill in it's address), or the HNF label, in which case an initialisation
			   routine should be used */
			
			if (strcmp(node_entry_name->string,hnf_symbol)==0)
			{
				if (init_hnf_label==NULL)
				{	hnf_label=enter_label2 (hnf_symbol, USE_LABEL | GCALL_LABEL | JSR_LABEL | IMPORT_LABEL);
					init_hnf_label=enter_label2 (init_hnf_symbol, USE_LABEL | GCALL_LABEL | JSR_LABEL);
				
					define_label(init_hnf_label);
				}		
				node_entry_label=init_hnf_label;
		
			} else
				node_entry_label=enter_label (node_entry_name,USE_LABEL | GCALL_LABEL | JSR_LABEL);
			
			break;
			
		case PACT_ASSEMBLER:
			node_entry_label=enter_label (node_entry_name,USE_LABEL | GCALL_LABEL | JSR_LABEL);
			break;
		
		default:
			node_entry_label=NULL;
			internal_error("code_desc: unknown assembler type (6)");
	}

	label=enter_label (label_name,DATA_LABEL);
	define_label(label);
	
	if (label == nil_label)
	{
		/* this is the system module that defines nil and should be linked first.
		   a routine will be added that passes information to the runtime system.
		   for this reason the start label must be imported, as this address has
		   to be passed as well.
		*/
		
		start_label=enter_label2(start_symbol, IMPORT_LABEL | JSR_LABEL | USE_LABEL | JUMP_LABEL);
	}	

	define_label_descriptor (label, node_entry_label, lazy_entry_label, arity, descriptor_name);
	
	lin_cdata(get_current_block());
}

VOID code_export (STRING *label_name)
{	LABEL *label;

	label=enter_label (label_name,EXPORT_LABEL | JSR_LABEL | USE_LABEL);
}

VOID code_import (STRING *label_name)
{	LABEL *label;

	label=enter_label (label_name,IMPORT_LABEL | JSR_LABEL);
}

VOID code_impdesc (STRING *desc_name, STRING *ealabel_name)
{
	LABEL *descriptor_label;
	
	descriptor_label=enter_label (desc_name, IMPORT_LABEL | DATA_LABEL);

	if (ealabel_name != NULL)
	{
		/* associate ea_label with this descriptor */
		
		descriptor_label->def_type=IMP_DESCRIPTOR;
		
		if (ealabel_name->string[0]=='_' && ealabel_name->length==1)
		{
			/* use default eval_fill label */
			
			if (eval_fill_label==NULL)
				eval_fill_label=enter_label2("eval_fill",IMPORT_LABEL);
			
			descriptor_label->def.ea_label=eval_fill_label;
			
		} else
			descriptor_label->def.ea_label=enter_label (ealabel_name,IMPORT_LABEL);
	}
}

VOID code_start (STRING *label_name)
{
	STRING s;
	
	if (strcmp (nostart_symbol, label_name->string)==0)
		return;
	
	s.string=start_symbol;
	s.length=strlen(start_symbol);
	
	code_o_label (&s, 0, 0, e_vector);
	code_export (&s);
	
	code_d_jmp (label_name,0,0,e_vector);
}

VOID code_n_o_label (STRING *label_name,INT16 arity,STRING *descriptor,STRING *ealabel, INT16 offered_a_stack_size,INT16 offered_b_stack_size, U_INT32 offered_vector[])
{
	LABEL *label;
	LABEL *descriptor_label;
	
	if (ealabel != NULL)
		warning_jmp_eval_update();
	
	/* if descriptor is "_", this entry has no descriptor */
	
	if (descriptor->string[0]=='_' && descriptor->length==1)
		descriptor_label=nil_label;	
	else
		descriptor_label=enter_label (descriptor,USE_LABEL | DATA_LABEL);
	
	if (code_reachable)
		label=enter_label (label_name,JSR_LABEL | REACHABLE);
	else
		label=enter_label (label_name,JSR_LABEL);
	
	set_label_layout (label,offered_a_stack_size,offered_b_stack_size,offered_vector);
	define_label(label);

	/* this label is certainly a jsr entry */

	if (code_reachable)
	{
		end_layout_block (curr_block_graph,offered_a_stack_size,offered_b_stack_size,offered_vector);
		
		if (last_jsr_entry == NULL)
			last_jsr_entry=curr_block_graph->instruction_block;

		block_ready(curr_block_graph,true);
			
	} else
	{
		/*	previous block should have been marked ready already and this block
			is not needed */
		
		end_dummy_block(curr_block_graph);
		last_jsr_entry=get_current_block();
	}
	
	curr_block_graph=begin_layout_block(offered_a_stack_size, offered_b_stack_size, offered_vector);
	
	code_reachable = true;
	
	curr_block_graph->start_label		=label;
	curr_block_graph->start_descriptor	=descriptor_label;
	curr_block_graph->start_arity		=arity;	

	define_symbol(label);
}

VOID code_o_label (STRING *label_name,INT16 offered_a_stack_size,INT16 offered_b_stack_size, U_INT32 offered_vector[])
{
	LABEL *label;

	if (code_reachable)
		label=enter_label (label_name,JSR_LABEL | REACHABLE);
	else
		label=enter_label (label_name,JSR_LABEL);
	
	set_label_layout (label,offered_a_stack_size,offered_b_stack_size,offered_vector);
	define_label(label);

	/* consider this label to be a jsr entry */

	if (code_reachable)
	{	
		end_layout_block (curr_block_graph,offered_a_stack_size,offered_b_stack_size,offered_vector);
		
		if(last_jsr_entry==NULL)
			last_jsr_entry=curr_block_graph->instruction_block;

		block_ready(curr_block_graph,true);
				
	} else
	{
		end_dummy_block (curr_block_graph);
		last_jsr_entry=get_current_block();
	}

	curr_block_graph=begin_layout_block(offered_a_stack_size, offered_b_stack_size, offered_vector);
	
	code_reachable = true;
			
	curr_block_graph->start_label=label;
	
	define_symbol(label);
}

VOID code_label (STRING *label_name)
{
	LABEL *label;
	
	if (code_reachable)
		label=enter_label (label_name,REACHABLE);
	else
		label=enter_label (label_name,0);
	
	define_label(label);
	
	if(code_reachable)
	{
		end_label_layout_block (curr_block_graph,label);
		block_ready(curr_block_graph,true);

	} else
		end_dummy_label_block (curr_block_graph,label);

	curr_block_graph=begin_label_layout_block(label);
	
	code_reachable = true;
	
	curr_block_graph->start_label = label;
		
	define_symbol(label);
}

VOID code_string (STRING *label_name, STRING *string)
{
	LABEL *label;
	
	if (code_reachable)
		error("string placed in code area");
	
	label=enter_label (label_name,DATA_LABEL);
	define_label(label);
	
	define_label_string (label, string);
	
	lin_cdata(get_current_block());
}

/************************************************************************************
 *																					*
 *		administration																*
 *																					*
 ************************************************************************************/

VOID code_keep (INT16 keep_offset,INT16 result_offset)
{
	INSTRUCTION_GRAPH keep_graph,result_graph,graph;

	keep_graph=s_get_a (curr_block_graph, keep_offset);
	result_graph=s_get_a (curr_block_graph, result_offset);
	
	graph=g_keep (keep_graph,result_graph);

	s_put_a (curr_block_graph, result_offset,graph);
}

/**************************************************************************************
 *                                                                                    *
 *          instructions for flow of control                                          *
 *                                                                                    *
 **************************************************************************************/

VOID code_d_rtn (INT16 demanded_a_stack_size,INT16 demanded_b_stack_size, U_INT32 demanded_vector[])
{
	INT16 ctop_offset;
	
	ctop_offset=get_ctop_offset(demanded_b_stack_size,demanded_vector);
	end_layout_block (curr_block_graph,demanded_a_stack_size,demanded_b_stack_size,demanded_vector);
	
	if (code_reachable)
	{
		append_instr_short (curr_block_graph->instruction_block,RTN,ctop_offset);
		block_ready(curr_block_graph,true);
		code_reachable = false;
	}
	
	curr_block_graph=begin_layout_block(0, 0, e_vector);
	
	fix_ctop_local();
}

static VOID clean_up_return(BLOCK_GRAPH *block,INT16 c_top_offset)
{
	/*	jsr's leave the old return address on top of the stack after they return,
		so this return address should be popped and all values on top of it
		should be shifted down */
	
	INSTRUCTION_GRAPH return_address;
	register INT16 offset;
	
	return_address=s_get_b(block,c_top_offset);			
	for (offset=c_top_offset-1; offset>=0; --offset)
		s_put_b (block,offset+1,s_get_b(block,offset));
	s_pop_b(block);

	/* due to optimisations the return address will be considered unused in the previous
	   block. To avoid removal of the return address: mark return load graph */
	
	return_address->used_mark=2;
}

VOID code_d_jsr_o (STRING *label_name,INT16 demanded_a_stack_size,
					INT16 demanded_b_stack_size, U_INT32 demanded_vector[],
					INT16 offered_a_stack_size,
					INT16 offered_b_stack_size, U_INT32 offered_vector[])
{
	LABEL *label;

	if (options.parallel_code)
		label=enter_label (label_name,USE_LABEL | JUMP_LABEL | JSR_LABEL);
	else
		label=enter_label (label_name,USE_LABEL | GCALL_LABEL | JSR_LABEL);

	if (!strcmp (label_name->string,e_system_sAP_symbol))
	{		
		set_label_layout(label,2,0,e_vector);
		
		insert_graph_in_b_stack(curr_block_graph,NULL,0,e_vector);
			
		insert_label_layout_block (curr_block_graph,JSR_AP,label);
		curr_block_graph=begin_layout_block(1,0,e_vector);
		
		clean_up_return(curr_block_graph,offered_b_stack_size);

	} else
	{				
		set_label_layout(label,demanded_a_stack_size,demanded_b_stack_size,demanded_vector);
		
		/* reserve an empty word on b-stack behind parameters to store return address */
		
		insert_graph_in_b_stack (curr_block_graph,NULL,demanded_b_stack_size,demanded_vector);

		insert_label_layout_block (curr_block_graph,JSR,label);
		curr_block_graph=begin_layout_block(offered_a_stack_size, offered_b_stack_size, offered_vector);
		
		clean_up_return(curr_block_graph,offered_b_stack_size);

	}
	fix_ctop_stack();
}

VOID code_d_jmp (	STRING *label_name, INT16 demanded_a_stack_size,
					INT16 demanded_b_stack_size, U_INT32 demanded_vector[])
{
	LABEL *label;
	
	label=enter_label (label_name,USE_LABEL | JUMP_LABEL);
	
	if (!strcmp (label_name->string,e_system_sAP_symbol))
	{	
		set_label_layout(label,2,0,e_vector);
		
		end_label_layout_block (curr_block_graph,label);
		
		if (code_reachable)
		{
			lin_cdata(curr_block_graph->instruction_block);
			append_instr_void (curr_block_graph->instruction_block,JMP_AP);
			block_ready(curr_block_graph,true);
			code_reachable = false;			
		} else
			lin_cdata(get_current_block());

		curr_block_graph=begin_layout_block(0,0,e_vector);
		
	} else
	{
		set_label_layout(label,demanded_a_stack_size,demanded_b_stack_size,demanded_vector);
	
		end_label_layout_block (curr_block_graph,label);
		
		if (code_reachable)
		{
			append_instr_label (curr_block_graph->instruction_block,JUMP,label);
			block_ready(curr_block_graph,true);		
			code_reachable = false;
		}
		
		curr_block_graph=begin_layout_block(0,0,e_vector);
	}
	last_jsr_entry=NULL;
}

VOID code_jmp (STRING *label_name)
{
	LABEL *label;
	
	label=enter_label (label_name, USE_LABEL | JUMP_LABEL);
	
	if (!strcmp (label_name->string,e_system_sAP_symbol))
	{
		set_label_layout(label,2,0,e_vector);

		end_label_layout_block (curr_block_graph,label);
		
		if (code_reachable)
		{
			lin_cdata(curr_block_graph->instruction_block);
			append_instr_void (curr_block_graph->instruction_block,JMP_AP);
			block_ready(curr_block_graph,true);
			code_reachable = false;
		} else
			lin_cdata(get_current_block());
		
		curr_block_graph=begin_layout_block(0,0,e_vector);
		
	} else
	{
		end_label_layout_block (curr_block_graph,label);
	
		if (code_reachable)
		{
			append_instr_label (curr_block_graph->instruction_block,JUMP,label);
			block_ready(curr_block_graph,true);		
			code_reachable = false;
		}
		
		curr_block_graph=begin_layout_block(0,0,e_vector);
	}
}

VOID code_jmp_false (STRING *label_name)
{
	INSTRUCTION_GRAPH bool_graph;
	LABEL *label;
	
	label=enter_label (label_name,USE_LABEL | JUMP_LABEL);
	
	bool_graph=s_get_pop_b(curr_block_graph);
	
	if (options.optimise_arith && bool_graph->code==GLOAD_CONST)
	{
		/*	optimise for constant conditional: generate jump if boolean false,
			else do not even break block */
		
		if (bool_graph->parameters[0].int32 == (INT32)false)
		{
			end_label_layout_block (curr_block_graph,label);

			if (code_reachable)
			{
				lin_cdata(curr_block_graph->instruction_block);
				append_instr_label (curr_block_graph->instruction_block,JUMP,label);
				block_ready(curr_block_graph,true);
				code_reachable = false;
			
			} else
				lin_cdata(get_current_block());
			
			curr_block_graph=begin_label_layout_block(label);
		}
	} else
	{
		end_graph_label_layout_block (curr_block_graph,bool_graph,label);
	
		if (code_reachable)
		{
			append_instr_label(curr_block_graph->instruction_block,CJUMP,label);
			block_ready(curr_block_graph,true);
		}
		curr_block_graph=begin_label_layout_block(label);
	}
}

VOID code_jmp_true (STRING *label_name)
{
	INSTRUCTION_GRAPH bool_graph;
	LABEL *label;
	
	label=enter_label (label_name,USE_LABEL | JUMP_LABEL);
	
	bool_graph=s_get_pop_b(curr_block_graph);
	
	if (options.optimise_arith && bool_graph->code==GLOAD_CONST)
	{
		/*	optimise for constant conditional: generate jump if boolean true,
			else do not even break block */
		
		if (bool_graph->parameters[0].int32 == (INT32)true)
		{
			end_label_layout_block (curr_block_graph,label);

			if (code_reachable)
			{
				lin_cdata(curr_block_graph->instruction_block);
				append_instr_label (curr_block_graph->instruction_block,JUMP,label);
				block_ready(curr_block_graph,true);		
				code_reachable = false;
			
			} else
				lin_cdata(get_current_block());
			
			curr_block_graph=begin_label_layout_block(label);
		}
	} else
	{
		end_graph_label_layout_block (curr_block_graph,bool_graph,label);
		
		if (code_reachable)
		{
			INSTRUCTION_BLOCK *instruction_block;
			
			instruction_block=curr_block_graph->instruction_block;
			
			append_instr_integer(instruction_block,EQC,0);
			append_instr_label(instruction_block,CJUMP,label);
			block_ready(curr_block_graph,true);
		}
		curr_block_graph=begin_label_layout_block(label);
	}
}
 
VOID code_halt (VOID)
{		
	instruction_call ("halt",0,0,e_vector,0,0,e_vector);
	end_layout_block (curr_block_graph,0,0,e_vector);		/* force eval of previous blocks */
	
	if (code_reachable)
	{
		block_ready(curr_block_graph,true);
		code_reachable=false;
	}
	
	fix_ctop_stack();
}

INSTRUCTION_BLOCK *code_end (BOOL inline_end)
{
	if (code_reachable)
	{
		if (!inline_end)
			error("error: code improperly ended\n");
		
		end_layout_block (curr_block_graph,0,0,e_vector);	/* force eval of previous blocks */
		lin_cdata(curr_block_graph->instruction_block);
		block_ready(curr_block_graph,true);
	
	}else
	{
		end_dummy_block(curr_block_graph);
		lin_cdata(get_current_block());
	}
	
	return first_instr_block;
}

VOID code_jmp_eval (VOID)
{
	end_layout_block (curr_block_graph,1,0,e_vector);
	
	if (code_reachable)
	{
		/* if the node is already in HNF this routine just performs a return */

		lin_cdata(curr_block_graph->instruction_block);
		append_instr_void (curr_block_graph->instruction_block,JMP_EVAL);
		block_ready(curr_block_graph,true);
		code_reachable = false;
	} else
		lin_cdata(get_current_block());
		
	last_jsr_entry=NULL;
	curr_block_graph=begin_layout_block(0,0,e_vector);
}

VOID code_jmp_eval_upd (VOID)
{
	/*	Note this instruction is not implemented correctly at this moment.
	
		The basic netto behaviour of this instruction is as follows:
		- evaluate top elem
		- overwrite second elem with result
		- pop top elem
		- return
		NOTE: for parallel code a release on the second elem must be
		performed as well...
	
		This scheme introduces a space leak: the second elem is kept on the
		stack, while the top elem is evaluated. The stack grows needlessly,
		which is a problem with recursive functions.
	
		This instruction avoids this by performing the following actions:
		- if top elem in HNF
			- overwrite second elem with top node
			- pop top node
			- return
		- else
			get special code that is stored before node and jump to it.
			this code will reserve the second node, pop the top node,
	       	push its args and overwrite the node by an indirection to the
			node to be updated. Next the ea_entry of the popped node is
			jumped to: this will update the reserved node with te result.
	
		Note the transputer code generator must put the return address at
		the correct place for the ea_entry. Conditional jumps
		should be used: no context switching allowed here.
	
	*/

	warning_jmp_eval_update();
	
	code_jsr_eval();			/* evaluate top elem */
	
	if (options.parallel_code)
	{
		code_getWL(1);
		code_fill_a(0,1);			/* overwrite second elem */
		code_releaseWL();
	
	}else
		code_fill_a(0,1);
	
	code_pop_a(1);				/* pop top elem */
	code_d_rtn(1,0,e_vector);	/* and return */
}

VOID code_jsr_eval (VOID)
{
	INSTRUCTION_GRAPH eval_node;
	
	eval_node=s_get_a(curr_block_graph,0);			/* forces the node into a local */
	
	insert_delay_layout_a_param_block (curr_block_graph,JSR_EVAL);
	curr_block_graph=begin_layout_block(0,0,e_vector);
}

/********************************************************************
 *																	*
 *		instructions for parallelism								*
 *																	*
 ********************************************************************/

/********************************************************************************
 *																				*
 *		waitlist routines														*
 *																				*
 ********************************************************************************/

VOID code_setwait (INT16 a_offset)
{
	/* put the current reducer in the waiting list of the node indicated by the
	   offset. Note that a node with a waiting list has arity 1: the waiting
	   list is an ordinary graph. This routine assumes the node already has
	   a waiting list, so its code does not need to be changed.
	*/
	
	if (options.parallel_code)
	{
		if (options.poll_eval)
		{
			/* local waiting lists not needed, polling is used */
			
			poll_warning();
		
		} else
		{
			INSTRUCTION_GRAPH graph_1,graph_2,graph_3,graph_4,graph_5,graph_6;

			release_proc1_label->flags|=USE_LABEL;
		
			graph_1=s_get_a (curr_block_graph,a_offset);
			
			graph_2=g_load_addr (release_proc1_label);
			graph_3=g_load_nl (graph_1,WAITLIST_OFFSET);
			graph_4=g_id();
			
			/* create waitlist node with arity 1 release_proc code */
			
			graph_5=g_create_3(graph_2,graph_3,graph_4,1,3,nil_label);
			
			/* fill old node with it, but without destroying the other values in the
			   node by using a 'transparant' fill.
			*/
			
			graph_6=g_transparant_fill_4(graph_1,wait_graph,NULL,graph_5,NULL,0,3,nil_label);
			
			s_put_a(curr_block_graph,a_offset,graph_6);
		}
	} else
		warning_sequential();
}

VOID code_setfirstwait (STRING *label_name,INT16 a_offset)
{
	/* 	make the (reserved) node at the given offset a node with a waiting list,
		and put the current reducer in it.
	*/
	
	if (options.parallel_code)
	{
		if (options.poll_eval)
		{
			/*	local waiting list not needed. the node does not have to be
				reserved (it should already be reserved). Just return, the reducer
				will start polling.
			*/
			
			poll_warning();
			
		} else
		{
			INSTRUCTION_GRAPH graph_1,graph_2,graph_3,graph_4,graph_5,graph_6,graph_7;
			LABEL *label;

			label=enter_label (label_name,USE_LABEL | GCALL_LABEL | JSR_LABEL);
			set_label_layout(label,1,0,e_vector);

			release_proc_label->flags|=USE_LABEL;

			graph_1=s_get_a (curr_block_graph,a_offset);
			
			/*	create first waitlist node: use arity 0 release_proc code */
						
			graph_2=g_load_addr (release_proc_label);
			graph_3=g_load_const ((INT32)0);
			graph_4=g_id();

			graph_5=g_create_3(graph_2,graph_3,graph_4,0,4,nil_label);

			/*	fill the node with new code and waitlist, this does not need to
				be done with a transparant fill as no fields need to be preserved.
			*/
			
			graph_6=g_load_addr (label);
			graph_7=g_fill_4(graph_1,wait_graph,graph_6,graph_5,NULL,0,3,nil_label);
			
			s_put_a (curr_block_graph,a_offset,graph_7);
		}
	} else
	{
		/*	sequential code, do nothing: sequential programs cannot reach this
			code
		*/
		
		warning_sequential();
	}
}

VOID code_getWL (INT16 a_offset)
{
	/* preserve the waitlist of the indicated node. A fill instruction should
	   follow. The global variable wait_graph is set to the waiting list field
	   of the given node. The fill instructions use this variable to determine
	   whether the waiting list should be preserved or not
	*/
	
	INSTRUCTION_GRAPH node_address;

	if (options.parallel_code)
	{
		if (options.poll_eval)
		{
			/* no local waiting lists needed */
			
			poll_warning();
			
		} else
		{
			if (wait_graph!=NULL)
				error("getWL: waitlist register overwritten\n");
	
			node_address=s_get_a (curr_block_graph,a_offset);
			wait_graph=g_load_nl(node_address,WAITLIST_OFFSET);
		}
	} else
		warning_sequential();
}

VOID code_putWL (INT16 a_offset)
{
	/*	This instruction should be used after overwriting an empty node with a
		closure. If no waiting list has been saved earlier or if the saved
		waiting list is empty this instruction will do nothing. Else it will
		replace the filled node with a reserved one that contains the saved
		waiting list and it will start up a process on a copy of the closure.
	*/
	
	if (options.parallel_code)
	{
		if (options.poll_eval)
		{
			/* no local waiting lists needed */
			
			poll_warning();
			
		} else
		{
			INSTRUCTION_GRAPH update,reduce,code;
	
			if (wait_graph==NULL)
			{	error("WARNING putWL: no preserved waitlist, putWL ignored\n");
				return;
			}

			if(wait_graph->code==GLOAD_CONST)
			{
				/*	The saved waiting list is empty. One does not have to
					keep it, just keep the node filled with the closure */
				
				wait_graph=NULL;
				return;
			}

			/*	It is unknown whether the waitlist is empty or not.			
				it will have been taken out of a (reserved) node.
				If this is the same node as the one we are putting
				the waitlist in now, nothing has to be done, except
				ignoring the fill (which will be taken care of by
				'g_reserve_closure'). If the waiting list has been taken
				from another node, we will have to fill the node
				explicitly with this waiting list. The latter has not
				yet been implemented: the code of the node would
				have to depend on the runtime state of the list */

			update=s_get_a(curr_block_graph,a_offset);			
			reduce=g_duplicate_closure(update);
			
			if (update->code==GFILL && (update->parameters[1].pointer==wait_graph || update->parameters[1].pointer==NULL))
				update=g_reserve_closure(update);
			else
				internal_error("putWL: cannot put WL in node");
			
			wait_graph=NULL;
			
			/*	Evaluate the closure if some reducer is waiting for the
				result. This will be determined at runtime */

			ea_process_label->flags|=USE_LABEL;
			code=g_load_addr(ea_process_label);

			s_put_a(curr_block_graph,a_offset,update);
			s_push_a(curr_block_graph,update);			/* node to update	*/
			s_push_a(curr_block_graph,reduce);			/* node to reduce 	*/
			s_push_b(curr_block_graph,code);			/* reduction code	*/
			
			instruction_call ("evaluate_if_requested",2,1,i_vector,0,0,e_vector);
		}
	} else
		warning_sequential();
}

VOID code_releaseWL (VOID)
{
	if (options.parallel_code)
	{
		if (options.poll_eval)
		{
			/* no waitlists, no releasing */
			
			poll_warning();
			
		} else
		{		
			if (wait_graph==NULL)
			{	error("WARNING releaseWL: no waitlist, releaseWL ignored\n");
				return;
			}
			
			if (wait_graph->code==GLOAD_CONST && wait_graph->parameters[0].int32==0)
			{
				/* no need to release node: waiting list is empty */
				
				wait_graph=NULL;
				return;
			}
			
			/* push the waiting list on the a-stack and insert release block */
			
			s_push_a (curr_block_graph,wait_graph);
			
			insert_delay_layout_a_param_block(curr_block_graph,RELEASE);
			curr_block_graph=begin_layout_block (0,0,e_vector);
			
			s_pop_a(curr_block_graph);
			
			wait_graph=NULL;
		}
	} else
		warning_sequential();
}

/********************************************************************
 *																	*
 *		stopping / starting reducers								*
 *																	*
 ********************************************************************/
 
VOID code_suspend (VOID)
{
	if (options.parallel_code)
	{
		if (options.poll_eval)
			code_suspend_uneval();
		else
		{
			insert_delay_layout_block(curr_block_graph,SUSPEND);
			curr_block_graph=begin_layout_block(0,0,e_vector);
		}
	} else
		warning_sequential();
}

VOID code_suspend_uneval(VOID)
{
	if (options.parallel_code)
	{
		if (options.poll_eval)
			poll_warning();
		
		/*	insert a suspend_uneval instruction as usual, a loop will be
			generated for it. This instruction has become obsolete */
		
		insert_delay_layout_a_param_block(curr_block_graph,SUSPEND_UNEVAL);
		curr_block_graph=begin_layout_block (0,0,e_vector);

	} else
		warning_sequential();
}

VOID code_new_int_reducer (STRING *label, INT16 a_offset)
{
	if (options.parallel_code)
	{
		if (options.any_proc_eval)
		{
			/* start up reducer on a node and don't care who actually reduces it */
		
			INSTRUCTION_GRAPH reduce, update, code;
			LABEL *code_label;
			
			code_label = enter_label(label, USE_LABEL | GCALL_LABEL | JSR_LABEL);
			set_label_layout(code_label,1,0,e_vector);
			
			code			= g_load_addr (code_label);
			update			= s_get_a (curr_block_graph,a_offset);
			reduce			= g_load_const((INT32) 0);
			
			s_push_a (curr_block_graph,update);			/* node to (reduce and) update	*/
			s_push_a (curr_block_graph,reduce);			/* no separate node to reduce 	*/
			
			s_push_b (curr_block_graph,code);			/* startup code */
		
			instruction_call ("start_new_reducer",2,1,i_vector,0,0,e_vector);

		} else
		{
		/*	Make sure the new reducer reduces the node node it is supposed to reduced.
			Note that just starting a new reducer on a node doesn't accomplish this
			as another reducer might reach the node earlier than this one.
			
			This wasn't any problem until strict continuations were introduced.
			Now it is possible a reducer hits a node that has the SC annotation,
			and evaluates it. The result that is left on the a-stack is not what
			one would expect in general. This is solved by restricting such a node
			to be evaluated by a specific reducer only (this one), that is started
			at a special code address that takes into account the different
			behaviour of the reducer. The passed code address is ignored at the
			moment!!
			
			Restricting the evaluation is done as follows: The reducer is started
			on a COPY of the original node (this copy is made here, at compile
			time). The original node will be transformed into a reserved one
			(other reducers will access the original node and will get blocked
			if they try to evaluate it). The reserved node is pushed on the
			stack first, next the copy is pushed. A new reducer will be started
			at a special code address that will reduce the top of the stack
			first and update the node below it afterwards, thus releasing any
			blocked process.
			
			Unfortunately the Clean compiler does not yet generate code that
			will update the reserved node early...
		*/
		
			INSTRUCTION_GRAPH reduce, update, code;
		
			error("warning: new_int_reducer will ignore code address \'%s\'\n",label->string);
			
			ea_process_label->flags|=USE_LABEL;
			
			code		=g_load_addr(ea_process_label);
			update		=s_get_a(curr_block_graph,a_offset);
			reduce		=g_duplicate_closure(update);
			update		=g_reserve_closure(update);
			
			s_push_a (curr_block_graph,update);			/* node to update	*/
			s_push_a (curr_block_graph,reduce);			/* node to reduce 	*/
			s_push_b (curr_block_graph,code);			/* startup code */
			
			instruction_call ("start_new_reducer",2,1,i_vector,0,0,e_vector);
		}
	} else
		warning_sequential();
}

VOID code_stop_reducer (VOID)
{
	if (options.parallel_code)
	{
		instruction_call ("stop_reducer",0,0,e_vector,0,0,e_vector);
		end_layout_block (curr_block_graph,0,0,e_vector);				/* force eval of block */
		
		if (code_reachable)
		{
			block_ready(curr_block_graph,true);
			code_reachable = false;
		}
	} else
		warning_sequential();
}

VOID code_force_cs (VOID)
{
	if (options.parallel_code)
	{
		insert_delay_layout_block(curr_block_graph,CS);
		curr_block_graph=begin_layout_block(0,0,e_vector);
	} else
		warning_sequential();
}

VOID code_enable_cs (VOID)
{
	if (options.parallel_code)
	{
		insert_delay_layout_block(curr_block_graph,ENABLE_CS);
		curr_block_graph=begin_layout_block(0,0,e_vector);
	} else
		warning_sequential();
}

VOID code_disable_cs (VOID)
{
	if (options.parallel_code)
	{
		insert_delay_layout_block(curr_block_graph,DISABLE_CS);
		curr_block_graph=begin_layout_block(0,0,e_vector);
	} else
		warning_sequential();
}

/********************************************************************
 *																	*
 *		computation of new processor id's							*
 *																	*
 ********************************************************************/

VOID code_number_processors(VOID)
{
	INSTRUCTION_GRAPH graph_1,graph_2;
	
	graph_1 = g_processors();
	graph_2 = g_load_nl(graph_1,N_PROCS);
	s_push_b (curr_block_graph,graph_2);
}

VOID code_current_processor(VOID)
{
	INSTRUCTION_GRAPH graph_1,graph_2;

	graph_1 = g_processors();
	graph_2 = g_load_nl(graph_1,CURR_PROC);
	s_push_b (curr_block_graph,graph_2);
}

VOID code_channel_processor(INT16 a_offset)
{
	INSTRUCTION_GRAPH graph_1,graph_2,graph_3,graph_4;

	graph_1 = s_get_a (curr_block_graph,a_offset);
	graph_2 = g_load_nl (graph_1, CHAN_TID_OFFSET);
	graph_3 = g_load_const ((INT32)5);
	graph_4 = g_L_shiftr (graph_2,graph_3);
	
	s_push_b (curr_block_graph,graph_4);
}

VOID code_distanceP(VOID)
{
	/*	call runtime system, get processor at distance (on top of b-stack)
		from current processors */

	instruction_signal (S_PROC_AT_DIST,0,1,i_vector,0,1,i_vector);
}

VOID code_neighbour_processor(VOID)
{
	INSTRUCTION_GRAPH distance, id;
	
	/*	at this moment only distances relative to the current processor
		are supported. This is optimised to 'distanceP'. */
	
	distance	=s_get_pop_b(curr_block_graph);
	id			=s_get_pop_b(curr_block_graph);
	
	if (id->code == GLOAD_NL)
	{
		INSTRUCTION_GRAPH	address;
		INT16				offset;
		
		address	=id->parameters[0].pointer;
		offset	=id->parameters[1].int16;
		
		if (address->code == GPROC_INFO && offset==CURR_PROC)
		{
			/* neigbour relative to current processor */
			
			s_push_b(curr_block_graph,distance);
			code_distanceP();
			return;
		}
	}
	
	/*	no optimisation possible, generate real neighbourP (currently
		not implemented) */
	
	code_current_processor();
	
	error("WARNING: neighbour_processor is currently replaced by currentP\n");
}

VOID code_ItoP(VOID)
{
	INSTRUCTION_GRAPH graph_1,graph_2,graph_3,graph_4,graph_5,graph_6;
	
	graph_1 = g_processors();
	graph_2 = s_get_b (curr_block_graph,0);
	
	/* make integer positive... */
	
	graph_3 = g_abs (graph_2);
	graph_4 = g_load_nl (graph_1,N_PROCS);
		
	graph_5 = g_modI (graph_3,graph_4);
	graph_6 = g_addI (graph_5,g_load_const((INT32)1));
	
	s_put_b (curr_block_graph,0,graph_6);
}

VOID code_new_processor(VOID)
{
	INSTRUCTION_GRAPH graph_1,graph_2,graph_3,graph_4,graph_5;
	
	graph_1 = g_processors();
	graph_2 = g_load_nl(graph_1,CURR_PROC);
	graph_3 = g_load_nl(graph_1,N_PROCS);
	graph_4 = g_modI(graph_2,graph_3);
	graph_5 = g_addI(graph_4,g_load_const((INT32)1));
	
	s_push_b (curr_block_graph,graph_5);
}

VOID code_random_processor(VOID)
{
	INSTRUCTION_GRAPH graph_1,graph_2,graph_3,graph_4,graph_5;
	
	if (random_label==NULL)
	{	random_label=enter_label2 ("new_random",USE_LABEL | JUMP_LABEL | JSR_LABEL | IMPORT_LABEL);
		set_label_layout(random_label,0,0,e_vector);
	}
	
	graph_1 = g_processors();
	graph_2 = g_random();						/* always positive */
	graph_3 = g_load_nl(graph_1,N_PROCS);
	graph_4 = g_modI(graph_2,graph_3);
	graph_5 = g_addI(graph_4,g_load_const((INT32)1));

	s_push_b (curr_block_graph,graph_5);
}

/********************************************************************/

VOID code_create_channel(STRING *code_label)
{	
	if (options.parallel_code && !options.poll_eval)
	{
		LABEL *label;

		label=enter_label (code_label,USE_LABEL | GCALL_LABEL | JSR_LABEL);
		label->entry_type=(CHANNEL_NODE | DEFERED_NODE);

		if (label != channel_label && options.verbose_warnings)
			error("WARNING: label \'%s\' will be ignored (create_channel)\n",code_label->string);
				
		/* top of stack (b1) contains a pid */
		
		instruction_call ("create_channel",0,1,i_vector,1,0,e_vector);
		
		/* top of stack (a1) contains a ptr to a channel node */
		
	} else
	{
		/*	sequential code, or polling code, pop the pid and create an empty
			node on the stack */
		
		code_pop_b(1);
		code_create(0);

		if (options.poll_eval)
			poll_warning();
		else
			warning_sequential();		
	}
}

VOID code_send_graph(STRING *code, INT16 graph_offset, INT16 channel_offset)
{
	if (options.parallel_code && !options.poll_eval)
	{
		INSTRUCTION_GRAPH graph,channel_graph,code_graph;
		LABEL *code_label;

		code_label=enter_label(code, USE_LABEL | DATA_LABEL);
		set_label_layout(code_label,0,0,e_vector);

		graph			=s_get_a(curr_block_graph,graph_offset);
		channel_graph	=s_get_a(curr_block_graph,channel_offset);
		code_graph		=g_load_desc_id(code_label,0);				/* the descriptor offset is sent */
				
		s_push_a (curr_block_graph,channel_graph);
		s_push_a (curr_block_graph,graph);
		s_push_b (curr_block_graph,code_graph);
		
		/* local a1 contains a pointer to the channel				*/
		/* local a2 contains the graph								*/
		/* b1 contains the descriptor offset of the reduction code	*/
		
		instruction_call ("send_graph",2,1,i_vector,0,0,e_vector);
		
	} else
	{
		STRING symbol, code;
		
		/* the channel node is assumed empty, as this module is generated with
		   the sequential/poll flag. Fill it with a defer node to the graph */
		
		code_push_a(graph_offset);
		
		symbol.string=Defer_symbol;
		symbol.length=strlen(Defer_symbol);
		
		code.string=defer_code_symbol;
		code.length=strlen(defer_code_symbol);
		
		code_fill(&symbol, 1, &code, channel_offset+1);
	
		if (options.poll_eval)
		{
			STRING start;
			
			/* an internal reducer must be started up on the defer node */
			
			start.string=hnf_reducer_symbol;
			start.length=strlen(hnf_reducer_symbol);
			
			code_new_int_reducer(&start,channel_offset);
			poll_warning();
			
		} else
			warning_sequential();
	}
}

/************************************************************************************
 *																					*
 *			inline instructions														*
 *																					*
 ************************************************************************************/

BOOL code_d_begin_inline(INT16 demanded_a_stack_size,INT16 demanded_b_stack_size, U_INT32 demanded_vector[])
{
	BOOL result;
	
	end_layout_block (curr_block_graph,demanded_a_stack_size,demanded_b_stack_size,demanded_vector);
	
	if (code_reachable)
		block_ready(curr_block_graph,true);
	
	/*	insert a new block without any abc instructions, and default stack layout.
		Then append the inline instructions and indicate the block is ready.
	*/
	
	curr_block_graph=begin_layout_block(0,0,e_vector);
	code_reachable=true;	
	end_layout_block(curr_block_graph,0,0,e_vector);
	
	result=parse_inline_instructions (curr_block_graph->instruction_block);
	block_ready(curr_block_graph,true);
	
	curr_block_graph=begin_layout_block(0,0,e_vector);

	return result;
}

VOID code_o_end_inline(INT16 offered_a_stack_size,INT16 offered_b_stack_size, U_INT32 offered_vector[])
{
	/*	End default layout block. This function overrides the default stack layout
		that is available when ']' is encountered at the end of an inline block.
		It is only invoked when stack information follows the ']'. */
	
	end_layout_block(curr_block_graph,0,0,e_vector);
	
	if (code_reachable)
		block_ready(curr_block_graph,true);
	
	curr_block_graph=begin_layout_block(offered_a_stack_size, offered_b_stack_size, offered_vector);
}

VOID code_o(INT16 offered_a_stack_size,INT16 offered_b_stack_size, U_INT32 offered_vector[])
{
	end_layout_block(curr_block_graph,offered_a_stack_size, offered_b_stack_size, offered_vector);
	
	if (code_reachable)
		block_ready(curr_block_graph,true);
		
	curr_block_graph=begin_layout_block(offered_a_stack_size, offered_b_stack_size, offered_vector);
}

/************************************************************************************
 *																					*
 *		interface between inline instructions and PABC instructions					*
 *																					*
 ************************************************************************************/

VOID code_in (INT16 n_a_parameters, INT16 n_b_parameters, U_INT32 vector[], STRING *params)
{
	/*	check if parameter description is valid and if so, store them in the
		correct places. */

	if (!check_parameters(n_a_parameters,n_b_parameters,vector,params))
	{
		n_a_parameters=0;
		n_b_parameters=0;
		vector=e_vector;
	}
	
	end_layout_block (curr_block_graph,n_a_parameters,n_b_parameters,vector);
	
	if (code_reachable)
		block_ready(curr_block_graph,true);
		
	begin_layout_block(n_a_parameters,n_b_parameters,vector);
}


VOID code_out (INT16 n_a_parameters, INT16 n_b_parameters, U_INT32 vector[], STRING *params)
{
	/* just check the current stack layout */
	
	check_parameters(n_a_parameters,n_b_parameters,vector,params);	
}

/************************************************************************************
 *																					*
 *		entry attributes															*
 *																					*
 ************************************************************************************/

static VOID set_entry_attribute(STRING *entry,U_INT16 attribute)
{
	LABEL *label;
	
	label=enter_label(entry,GCALL_LABEL | JSR_LABEL);
	label->entry_type|=attribute;
}

VOID code_entry_defer(STRING *entry)
{
	if (options.ignore_defer)
	{
		warning_ignore_defer();
		return;
	}
	
	set_entry_attribute(entry,DEFERED_NODE);
}

VOID code_entry_reserve(STRING *entry)
{
	set_entry_attribute(entry,RESERVED_NODE);
}

VOID code_entry_channel(STRING *entry)
{
	set_entry_attribute(entry,CHANNEL_NODE);
}

/************************************************************************************
 *																					*
 *			directives for placement return address									*
 *																					*
 ************************************************************************************/

VOID code_return_local(STRING *label_name)
{
	LABEL *label;
	
	label=enter_label (label_name,JSR_LABEL);
	
	if ((label->flags & FIXED_CTOP) && label->flags & CTOP_STACK)
		error("code_return_local: return address already fixed on stack");
	else
	{	
		label->flags|=FIXED_CTOP;
		label->flags&=~CTOP_STACK;
	}
}

/************************************************************************************
 *																					*
 *			not yet implemented														*
 *																					*
 ************************************************************************************/

VOID code_unimplemented (VAR_ARGS)
{
	error ("unimplemented instruction\n");
}

/***********************************************************************
 *                                                                     *
 *          initialisation                                             *
 *                                                                     *
 ***********************************************************************/

VOID initialize_coding (VOID)
{	
	jmp_eval_update_warned	=false;
	ignore_defer_warned		=false;
	sequential_warned		=false;
	poll_warned				=false;
	
	/* initialise imported modules */
	
	first_instr_block=initialize_linearization();
	wait_graph=NULL;

	/* default start settings */

	last_jsr_entry		=NULL;
	code_reachable		=false;
	curr_block_graph	=begin_layout_block(0,0,e_vector);
}