/*******************************************************************************
 *                                                                             *
 *     Concurrent Clean parser.c                                               *
 *     =========================                                               *
 *                                                                             *
 *                                                                             *
 *                                                                             *
 *     Author: Eric Nocker, John van Groningen, Marco Kesseler                 *
 *     At: Department of Computer Science                                      *
 *         University of Nijmegen                                              *
 *                                                                             *
 *     Version: 0.6                                                            *
 *     Date:    July 19, 1990                                                  *
 *                                                                             *
 ******************************************************************************/

#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <limits.h>

#include "ifinterface.h"

#include "cgscanner.h"
#include "cginputabc.h"
#include "cgmemory.h"
#include "cgcode.h"
#include "cgstacks.h"

#include "ccmacros.h"

/********************************************************************************
 *																				*
 *		global variables														*
 *																				*
 ********************************************************************************/

#define TABLE_SIZE		500

typedef struct ABC_INSTRUCTION *ABC_INSTRUCTION_P;

typedef struct ABC_INSTRUCTION
{
	CHAR	*name;
	U_INT16	kind;
	VOID    (*parse) (ABC_INSTRUCTION_P instr);
	VOID	(*code) (VAR_ARGS);
	
} ABC_INSTRUCTION;

typedef struct TABLE_ENTRY
{
	struct TABLE_ENTRY		*next;
	ABC_INSTRUCTION			*instr;

} TABLE_ENTRY;

static TABLE_ENTRY 			*instr_table;			/* the instruction table */
static ABC_INSTRUCTION_P	nop_instr = NULL;		/* returned for empty strings */

static BOOL CompDirectiveParsed;	/* booleans to tell what has been parsed */
static BOOL CodeDirectiveParsed;
static BOOL StartDirectiveParsed;
static BOOL InfoDirectiveExpected;

/* info of latest demanded directive */

static U_INT32		demanded_vector [MAX_VECTOR_SIZE];
static INT16		demanded_a_stack_size=0;
static INT16		demanded_b_stack_size=0;
static BOOL			demanded_flag;

/* info of latest offered directive */

static U_INT32		offered_vector [MAX_VECTOR_SIZE];
static INT16		offered_a_stack_size=0;
static INT16		offered_b_stack_size=0;
static BOOL			offered_flag;

/* global variables to keep track of latest .n directive */

static INT16		entry_arity;
static SSTRING		entry_descriptor_chars;
static SSTRING		entry_ealabel_chars;
static STRING		entry_descriptor;
static STRING		entry_ealabel;
static BOOL			entry_flag;

/* global variables to keep track of latest jsr (delayed until .o) */

static STRING		jsr_label;
static SSTRING		jsr_label_chars;
static BOOL			jsr_flag;

/* global variable to indicate ']' has been parsed (delayed until .o) */

static BOOL			end_asm_flag;

/* info about .d jsr .o and .n .o label combination */

static BOOL			offer_expected;				/* true if .o should be parsed */
static BOOL			demand_instr_expected;		/* true of one of these expected */

/*	code info, used to indicate amount parsed already. when the instruction number
	becomes larger than the previous one + the step amount information is printed.
	The step amount is set to 1% of the number of instructions. To avoid printing
	info when the total amount is unknown these values are initialised as follows
	- instruction_number: 0
	- previous instruction number: 0
	- step: maxint
*/

INT32 n_instructions;
INT32 n_descriptors;
INT32 n_labels;

/*******************************************************************************
 *                                                                             *
 *    The parser, instruction hash table                                       *
 *                                                                             *
 *    The instruction hash table should be initialised at system startup       *
 *    and not be changed. Note that instructions which are entered first       *
 *    can be found fast.                                                       *
 *                                                                             *
 ******************************************************************************/

enum
{
	Directive,			Instruction,		No_op,
	
	DepDir,				ExpDir,
	O_Dir,				D_Dir,
	
	D_Instr				/* instruction that may follow a .d directive */
};

static U_INT16 instr_hash (CHAR *s)
{
	register U_INT16 h;
	
	/* add the character values of the string */
	
	h=0;
	while(*s != '\0')
		h+=*s++;
		
	return h;
}

static VOID put_instruction_name(ABC_INSTRUCTION *instruction)
{
	TABLE_ENTRY *entry;
	
	entry=&instr_table [instr_hash (instruction->name) % TABLE_SIZE];
	
	if (entry->instr!=NULL)
	{
		TABLE_ENTRY *link;
		
		link=entry;
		while (link->next!=NULL)
			link=link->next;
		
		for (entry=instr_table; entry<&instr_table [TABLE_SIZE]; ++entry)
			if (entry->instr==NULL)
				break;
		
		if (entry==&instr_table [TABLE_SIZE])
			fatal_error ("not enough entries in instruction hash tabel");
		
		link->next=entry;
	}
	
	entry->instr=instruction;

	if (instruction->kind==No_op)
		nop_instr=instruction;
}

static VOID Parse_empty			(ABC_INSTRUCTION_P instr);
static VOID Parse_empty_or_n	(ABC_INSTRUCTION_P instr);
static VOID Parse_n				(ABC_INSTRUCTION_P instr);
static VOID Parse_nn			(ABC_INSTRUCTION_P instr);
static VOID Parse_nnn			(ABC_INSTRUCTION_P instr);

static VOID Parse_LABEL			(ABC_INSTRUCTION_P instr);
static VOID Parse_INTEGER		(ABC_INSTRUCTION_P instr);
static VOID Parse_BOOL			(ABC_INSTRUCTION_P instr);
static VOID Parse_CHAR			(ABC_INSTRUCTION_P instr);
static VOID Parse_STRING		(ABC_INSTRUCTION_P instr);
static VOID Parse_REAL			(ABC_INSTRUCTION_P instr);

static VOID Parse_LABELn		(ABC_INSTRUCTION_P instr);
static VOID Parse_INTEGERn		(ABC_INSTRUCTION_P instr);
static VOID Parse_BOOLn			(ABC_INSTRUCTION_P instr);
static VOID Parse_CHARn			(ABC_INSTRUCTION_P instr);
static VOID Parse_STRINGn		(ABC_INSTRUCTION_P instr);
static VOID Parse_REALn			(ABC_INSTRUCTION_P instr);

static VOID Parse_LABELnn		(ABC_INSTRUCTION_P instr);
static VOID Parse_LABELnLABELn	(ABC_INSTRUCTION_P instr);

static VOID Parse_Rtn			(ABC_INSTRUCTION_P instr);
static VOID Parse_Jmp			(ABC_INSTRUCTION_P instr);
static VOID Parse_Jsr			(ABC_INSTRUCTION_P instr);

static VOID ParseD_Comp			(ABC_INSTRUCTION_P instr);
static VOID ParseD_Code			(ABC_INSTRUCTION_P instr);
static VOID ParseD_Start		(ABC_INSTRUCTION_P instr);
static VOID ParseD_Deps			(ABC_INSTRUCTION_P instr);
static VOID ParseD_Stop			(ABC_INSTRUCTION_P instr);
static VOID ParseD_Desc			(ABC_INSTRUCTION_P instr);
static VOID ParseD_Exps			(ABC_INSTRUCTION_P instr);
static VOID ParseD_Inline		(ABC_INSTRUCTION_P instr);
static VOID ParseD_End			(ABC_INSTRUCTION_P instr);
static VOID ParseD_O			(ABC_INSTRUCTION_P instr);
static VOID ParseD_D			(ABC_INSTRUCTION_P instr);
static VOID ParseD_N			(ABC_INSTRUCTION_P instr);
static VOID ParseD_ImpLab		(ABC_INSTRUCTION_P instr);
static VOID ParseD_ImpDesc		(ABC_INSTRUCTION_P instr);
static VOID ParseD_String		(ABC_INSTRUCTION_P instr);
static VOID ParseD_Keep			(ABC_INSTRUCTION_P instr);
static VOID Parse_InParameters	(ABC_INSTRUCTION_P instr);
static VOID Parse_OutParameters	(ABC_INSTRUCTION_P instr);

static ABC_INSTRUCTION instructions[] =
{
	{"pop_a",			Instruction,	Parse_n,			(VOID *)code_pop_a},
	{"push_a",			Instruction,	Parse_n,			(VOID *)code_push_a},
	{"update_a",		Instruction,	Parse_nn,			(VOID *)code_update_a},
	{"push_args",		Instruction,	Parse_nnn,			(VOID *)code_push_args},
	{"push_node",		Instruction,	Parse_LABELn,		(VOID *)code_push_node},
	{"pop_b",			Instruction,	Parse_n,			(VOID *)code_pop_b},
	{"push_b",			Instruction,	Parse_n,			(VOID *)code_push_b},
	{"update_b",		Instruction,	Parse_nn,			(VOID *)code_update_b},
	{"create",			Instruction,	Parse_empty_or_n,	(VOID *)code_create},
	{"no_op",			No_op,			Parse_empty,		(VOID *)code_no_op},
	{"jmp",				D_Instr,		Parse_Jmp,			(VOID *)code_jmp},
	{"jmp_false",		Instruction,	Parse_LABEL,		(VOID *)code_jmp_false},
	{"jmp_true",		Instruction,	Parse_LABEL,		(VOID *)code_jmp_true},
	{"jsr",				D_Instr,		Parse_Jsr,			(VOID *)code_d_jsr_o},
	{"jsr_eval",		Instruction,	Parse_empty,		(VOID *)code_jsr_eval},
	{"jmp_eval",		Instruction,	Parse_empty,		(VOID *)code_jmp_eval},
	{"jmp_eval_upd",	Instruction,	Parse_empty,		(VOID *)code_jmp_eval_upd},
	{"rtn",				D_Instr,		Parse_Rtn,			(VOID *)code_d_rtn},
	{"fill",			Instruction,	Parse_LABELnLABELn,	(VOID *)code_fill},
	{"fillI",			Instruction,	Parse_INTEGERn,		(VOID *)code_fillI},
	{"fillB",			Instruction,	Parse_BOOLn,		(VOID *)code_fillB},
	{"fillC",			Instruction,	Parse_CHARn,		(VOID *)code_fillC},
	{"fillR",			Instruction,	Parse_REALn,		(VOID *)code_fillR},
	{"fillS",			Instruction,	Parse_STRINGn,		(VOID *)code_fillS},
	{"fillI_b",			Instruction,	Parse_nn,			(VOID *)code_fillI_b},
	{"fillB_b",			Instruction,	Parse_nn,			(VOID *)code_fillB_b},
	{"fillC_b",			Instruction,	Parse_nn,			(VOID *)code_fillC_b},
	{"fillR_b",			Instruction,	Parse_nn,			(VOID *)code_fillR_b},
	{"fillF_b",			Instruction,	Parse_nn,			(VOID *)code_fillF_b},
	{"fill_a",			Instruction,	Parse_nn,			(VOID *)code_fill_a},
	{"set_entry",		Instruction,	Parse_LABELn,		(VOID *)code_set_entry},
	{"get_node_arity",	Instruction,	Parse_n,			(VOID *)code_get_node_arity},
	{"get_desc_arity",	Instruction,	Parse_n,			(VOID *)code_get_desc_arity},
	{"eq_desc",			Instruction,	Parse_LABELnn,		(VOID *)code_eq_desc},
	{"eq_nulldesc",		Instruction,	Parse_LABELn,		(VOID *)code_eq_nulldesc},
	{"eqI_a",			Instruction,	Parse_INTEGERn,		(VOID *)code_eqI_a},
	{"eqB_a",			Instruction,	Parse_BOOLn,		(VOID *)code_eqI_a},
	{"eqC_a",			Instruction,	Parse_CHARn,		(VOID *)code_eqI_a},
	{"eqR_a",			Instruction,	Parse_REALn,		(VOID *)code_eqR_a},
	{"eqS_a",			Instruction,	Parse_STRINGn,		(VOID *)code_eqS_a},
	{"pushB",			Instruction,	Parse_BOOL,			(VOID *)code_pushI},
	{"pushI",			Instruction,	Parse_INTEGER,		(VOID *)code_pushI},
	{"pushC",			Instruction,	Parse_CHAR,			(VOID *)code_pushI},
	{"pushR",			Instruction,	Parse_REAL,			(VOID *)code_pushR},
	{"fillS_symbol",	Instruction,	Parse_nn,			(VOID *)code_fillS_symbol},
	{"eq_symbol",		Instruction,	Parse_nn,			(VOID *)code_eq_symbol},
	{"eqI_b",			Instruction,	Parse_INTEGERn,		(VOID *)code_eqI_b},
	{"eqB_b",			Instruction,	Parse_BOOLn,		(VOID *)code_eqI_b},
	{"eqC_b",			Instruction,	Parse_CHARn,		(VOID *)code_eqI_b},
	{"eqR_b",			Instruction,	Parse_REALn,		(VOID *)code_eqR_b},
	{"pushI_a",			Instruction,	Parse_n,			(VOID *)code_pushI_a},
	{"pushB_a",			Instruction,	Parse_n,			(VOID *)code_pushI_a},
	{"pushC_a",			Instruction,	Parse_n,			(VOID *)code_pushI_a},
	{"pushR_a",			Instruction,	Parse_n,			(VOID *)code_pushR_a},
	{"pushF_a",			Instruction,	Parse_n,			(VOID *)code_pushF_a},
	{"repl_args",		Instruction,	Parse_nn,			(VOID *)code_repl_args},
	{"repl_arg",		Instruction,	Parse_nn,			(VOID *)code_repl_arg},
	{"repl_args_b",		Instruction,	Parse_empty,		(VOID *)code_repl_args_b},
	{"push_arg",		Instruction,	Parse_nnn,			(VOID *)code_push_arg},
	{"push_arg_b",		Instruction,	Parse_n,			(VOID *)code_push_arg_b},
	{"addI",			Instruction,	Parse_empty,		(VOID *)code_addI},
	{"subI",			Instruction,	Parse_empty,		(VOID *)code_subI},
	{"mulI",			Instruction,	Parse_empty,		(VOID *)code_mulI},
	{"divI",			Instruction,	Parse_empty,		(VOID *)code_divI},
	{"modI",			Instruction,	Parse_empty,		(VOID *)code_modI},
	{"incI",			Instruction,	Parse_empty,		(VOID *)code_incI},
	{"decI",			Instruction,	Parse_empty,		(VOID *)code_decI},
	{"eqI",				Instruction,	Parse_empty,		(VOID *)code_eqI},
	{"gtI",				Instruction,	Parse_empty,		(VOID *)code_gtI},
	{"ltI",				Instruction,	Parse_empty,		(VOID *)code_ltI},
	{"randomI",			Instruction,	Parse_empty,		(VOID *)code_randomI},
	{"ItoR",			Instruction,	Parse_empty,		(VOID *)code_ItoR},
	
	{"andB",			Instruction,	Parse_empty,		(VOID *)code_andB},
	{"orB",				Instruction,	Parse_empty,		(VOID *)code_orB},
	{"xorB",			Instruction,	Parse_empty,		(VOID *)code_xorB},
	{"eqB",				Instruction,	Parse_empty,		(VOID *)code_eqI},
	{"notB",			Instruction,	Parse_empty,		(VOID *)code_notB},
	{"shiftrB",			Instruction,	Parse_empty,		(VOID *)code_L_shiftr},
	{"shiftlB",			Instruction,	Parse_empty,		(VOID *)code_L_shiftl},
	
	{"and%",			Instruction,	Parse_empty,		(VOID *)code_andI},
	{"or%",				Instruction,	Parse_empty,		(VOID *)code_orI},
	{"xor%",			Instruction,	Parse_empty,		(VOID *)code_xorI},
	{"not%",			Instruction,	Parse_empty,		(VOID *)code_notI},
	{"shiftr%",			Instruction,	Parse_empty,		(VOID *)code_A_shiftr},
	{"shiftl%",			Instruction,	Parse_empty,		(VOID *)code_A_shiftl},

	{"eqC",				Instruction,	Parse_empty,		(VOID *)code_eqI},
	{"ltC",				Instruction,	Parse_empty,		(VOID *)code_ltI},
	{"gtC",				Instruction,	Parse_empty,		(VOID *)code_gtI},
	{"CtoI",			Instruction,	Parse_empty,		(VOID *)code_CtoI},
	{"ItoC",			Instruction,	Parse_empty,		(VOID *)code_ItoC},

	{"addR",			Instruction,	Parse_empty,		(VOID *)code_addR},
	{"subR",			Instruction,	Parse_empty,		(VOID *)code_subR},
	{"mulR",			Instruction,	Parse_empty,		(VOID *)code_mulR},
	{"divR",			Instruction,	Parse_empty,		(VOID *)code_divR},
	{"remR",			Instruction,	Parse_empty,		(VOID *)code_remR},
	{"incR",			Instruction,	Parse_empty,		(VOID *)code_incR},
	{"decR",			Instruction,	Parse_empty,		(VOID *)code_decR},
	{"eqR",				Instruction,	Parse_empty,		(VOID *)code_eqR},
	{"ltR",				Instruction,	Parse_empty,		(VOID *)code_ltR},
	{"gtR",				Instruction,	Parse_empty,		(VOID *)code_gtR},
	{"sinR",			Instruction,	Parse_empty,		(VOID *)code_sinR},
	{"cosR",			Instruction,	Parse_empty,		(VOID *)code_cosR},
	{"tanR",			Instruction,	Parse_empty,		(VOID *)code_tanR},
	{"asinR",			Instruction,	Parse_empty,		(VOID *)code_asinR},
	{"acosR",			Instruction,	Parse_empty,		(VOID *)code_acosR},
	{"atanR",			Instruction,	Parse_empty,		(VOID *)code_atanR},
	{"lnR",				Instruction,	Parse_empty,		(VOID *)code_lnR},
	{"log10R",			Instruction,	Parse_empty,		(VOID *)code_log10R},
	{"sqrtR",			Instruction,	Parse_empty,		(VOID *)code_sqrtR},
	{"powR",			Instruction,	Parse_empty,		(VOID *)code_powR},
	{"expR",			Instruction,	Parse_empty,		(VOID *)code_expR},
	{"entierR",			Instruction,	Parse_empty,		(VOID *)code_entierR},
	{"RtoI",			Instruction,	Parse_empty,		(VOID *)code_RtoI},

	{"lenS",			Instruction,	Parse_n,			(VOID *)code_lenS},
	{"cmpS",			Instruction,	Parse_nn,			(VOID *)code_cmpS},
	{"catS",			Instruction,	Parse_nnn,			(VOID *)code_catS},
	{"sliceS",			Instruction,	Parse_nn,			(VOID *)code_sliceS},
	{"indexS",			Instruction,	Parse_n,			(VOID *)code_indexS},
	{"updateS",			Instruction,	Parse_nn,			(VOID *)code_updateS},

	{"setwait",			Instruction,	Parse_n,			(VOID *)code_setwait},
	{"setfirstwait",	Instruction,	Parse_LABELn,		(VOID *)code_setfirstwait},
	{"getWL",			Instruction,	Parse_n,			(VOID *)code_getWL},
	{"putWL",			Instruction,	Parse_n,			(VOID *)code_putWL},
	{"release",			Instruction,	Parse_empty,		(VOID *)code_releaseWL},
	
	{"suspend",			Instruction,	Parse_empty,		(VOID *)code_suspend},
	{"suspend_uneval",	Instruction,	Parse_empty,		(VOID *)code_suspend_uneval},
	{"new_int_reducer",	Instruction,	Parse_LABELn,		(VOID *)code_new_int_reducer},
	{"stop_reducer",	Instruction,	Parse_empty,		(VOID *)code_stop_reducer},
	{"force_cswitch",	Instruction,	Parse_empty,		(VOID *)code_force_cs},
	{"enable_cs",		Instruction,	Parse_empty,		(VOID *)code_enable_cs},
	{"disable_cs",		Instruction,	Parse_empty,		(VOID *)code_disable_cs},

	{"set_defer",		Instruction,	Parse_n,			(VOID *)code_set_defer},
		
	{"create_channel",	Instruction,	Parse_LABEL,		(VOID *)code_create_channel},
	{"send_request",	Instruction,	Parse_n,			(VOID *)code_unimplemented},
	{"send_graph",		Instruction,	Parse_LABELnn,		(VOID *)code_send_graph},
	{"new_ext_reducer",	Instruction,	Parse_LABELn,		(VOID *)code_unimplemented},
		
	{"numberP",			Instruction,	Parse_empty,		(VOID *)code_number_processors},
	{"currentP",		Instruction,	Parse_empty,		(VOID *)code_current_processor},
	{"newP",			Instruction,	Parse_empty,		(VOID *)code_new_processor},
	{"randomP",			Instruction,	Parse_empty,		(VOID *)code_random_processor},
	{"channelP",		Instruction,	Parse_n,			(VOID *)code_channel_processor},
	{"neighbourP",		Instruction,	Parse_empty,		(VOID *)code_neighbour_processor},
	{"ItoP",			Instruction,	Parse_empty,		(VOID *)code_ItoP},
	{"distanceP",		Instruction,	Parse_empty,		(VOID *)code_distanceP},

	{"halt",			Instruction,	Parse_empty,		(VOID *)code_halt},
	{"pause",			Instruction,	Parse_empty,		(VOID *)code_unimplemented},
	{"print",			Instruction,	Parse_STRING,		(VOID *)code_print},
	{"print_sc",		Instruction,	Parse_STRING,		(VOID *)code_print_sc},
	{"print_symbol",	Instruction,	Parse_n,			(VOID *)code_print_symbol},
	{"print_symbol_sc",	Instruction,	Parse_n,			(VOID *)code_print_symbol_sc},
	{"printD",			Instruction,	Parse_empty,		(VOID *)code_printD},
	{"pushD",			Instruction,	Parse_LABEL,		(VOID *)code_pushD},
	{"openF",			Instruction,	Parse_nn,			(VOID *)code_openF},
	{"getFC",			Instruction,	Parse_empty,		(VOID *)code_getFC},
	{"getFS",			Instruction,	Parse_n,			(VOID *)code_getFS},
	{"putFC",			Instruction,	Parse_empty,		(VOID *)code_putFC},
	{"putFS",			Instruction,	Parse_n,			(VOID *)code_putFS},
	{"endF",			Instruction,	Parse_empty,		(VOID *)code_endF},
	{"reopenF",			Instruction,	Parse_empty,		(VOID *)code_reopenF},

	{".comp",			Directive,		ParseD_Comp,		(VOID *)code_no_op},
	{".code",			Directive,		ParseD_Code,		(VOID *)code_no_op},
	{".start",			Directive,		ParseD_Start,		(VOID *)code_start},
	{".depend",			DepDir,			ParseD_Deps,		(VOID *)code_no_op},
	{".endinfo",		Directive,		ParseD_Stop,		(VOID *)code_no_op},
	{".desc",			Directive,		ParseD_Desc,		(VOID *)code_desc},
	{".export",			ExpDir,			ParseD_Exps,		(VOID *)code_export},
	{".inline",			Directive,		ParseD_Inline,		(VOID *)code_no_op},
	{".end",			Directive,		ParseD_End,			(VOID *)code_no_op},
	{".o",				O_Dir,			ParseD_O,			(VOID *)code_no_op},
	{".d",				D_Dir,			ParseD_D,			(VOID *)code_no_op},
	{".n",				Directive,		ParseD_N,			(VOID *)code_no_op},
	{".implab",			Directive,		ParseD_ImpLab,		(VOID *)code_import},
	{".impdesc",		Directive,		ParseD_ImpDesc,		(VOID *)code_impdesc},
	{".string",			Directive,		ParseD_String,		(VOID *)code_string},
	{".keep",			Directive,		ParseD_Keep,		(VOID *)code_keep},
	
	{".entry_defer",	Directive,		Parse_LABEL,		(VOID *)code_entry_defer},
	{".entry_reserve",	Directive,		Parse_LABEL,		(VOID *)code_entry_reserve},
	{".entry_channel",	Directive,		Parse_LABEL,		(VOID *)code_entry_channel},
	
	{".return_local",	Directive,		Parse_LABEL,		(VOID *)code_return_local},
	
	{"in",				Instruction,	Parse_InParameters,	(VOID *)code_in},
	{"out",				Instruction,	Parse_OutParameters,(VOID *)code_out}
};

static VOID initialise_instruction_table(VOID)
{
	register INT16 i;
	
	instr_table = (TABLE_ENTRY *) alloc_heap_mem((U_INT32)TABLE_SIZE*sizeof (TABLE_ENTRY));
			
	for (i = 0; i < TABLE_SIZE; i++)
	{	instr_table[i].instr  	= NULL;
		instr_table[i].next 	= NULL;
	}

	for (i=0; i<sizeof (instructions)/sizeof (ABC_INSTRUCTION); ++i)
		put_instruction_name (&instructions[i]);
}

static ABC_INSTRUCTION_P get_instruction (STRING *s)
{
	TABLE_ENTRY *entry;
		
	if (s->length == 0)
		return nop_instr;
	
	entry=&instr_table [instr_hash(s->string) % TABLE_SIZE];
	
	while (entry!=NULL && entry->instr!=NULL)
	{
		if (strcmp (entry->instr->name,s->string)==0)
			return entry->instr;
		
		entry=entry->next;		
	}
	return NULL;
}

/*******************************************************************************
 *                                                                             *
 *    The parser, error messages                                               *
 *                                                                             *
 ******************************************************************************/

static VOID reset_flags(VOID)
{	demanded_flag			=false;
	offered_flag			=false;
	entry_flag				=false;
	jsr_flag				=false;
	end_asm_flag			=false;
	offer_expected			=false;
	demand_instr_expected	=false;
}
		
static VOID MissingNumber (CHAR *InstrName)
{	scan_error (InstrName,"wrong or missing argument: number expected");
	reset_flags();
	SkipToChar ('\n');
}

static VOID MissingBoolean (CHAR *InstrName)
{	scan_error (InstrName,"wrong or missing argument: boolean expected");
	reset_flags();
	SkipToChar ('\n');
}

static VOID MissingCharacter (CHAR *InstrName)
{	scan_error (InstrName,"wrong or missing argument: character expected");
	reset_flags();
	SkipToChar ('\n');
}

static VOID MissingIdentifier (CHAR *InstrName)
{	scan_error (InstrName,"wrong or missing argument: identifier expected");
	reset_flags();
	SkipToChar ('\n');
}

static VOID MissingReal (CHAR *InstrName)
{	scan_error (InstrName,"wrong or missing argument: real expected");
	reset_flags();
	SkipToChar ('\n');
}

static VOID MissingString (CHAR *InstrName)
{	scan_error (InstrName,"wrong or missing argument: string expected");
	reset_flags();
	SkipToChar ('\n');
}

static VOID MissingVector (CHAR *InstrName)
{	scan_error (InstrName,"wrong or missing argument: vector expected");
	reset_flags();
	SkipToChar ('\n');
}

static VOID MissingParameters (CHAR *InstrName)
{	scan_error (InstrName,"wrong or missing argument: parameter descriptor expected");
	reset_flags();
	SkipToChar ('\n');
}

static VOID NotExpected (CHAR *InstrName, CHAR *what)
{	SSTRING s;

	strcpy (s, what);
	strcat (s, " not expected");
	
	scan_error (InstrName,s);
	reset_flags();
	SkipToChar ('\n');
}

static VOID Expected (CHAR *InstrName, CHAR *what)
{	SSTRING s;

	strcpy (s, what);
	strcat (s, " expected");
	
	scan_error (InstrName,s);
	reset_flags();
}

static VOID Unknown (CHAR *InstrName, CHAR *what)
{	SSTRING s;

	strcpy (s, "unknown ");
	strcat (s, what);
	
	scan_error (InstrName,s);
	reset_flags();
	SkipToChar ('\n');
}

static VOID TooMany (CHAR *InstrName, CHAR *what)
{	SSTRING s;

	strcpy (s, what);
	strcat (s, " encountered more than once");
	
	scan_warning (InstrName,s);
	reset_flags();
	SkipToChar ('\n');
}

/*******************************************************************************
 *                                                                             *
 *    The parser, instructions                                                 *
 *                                                                             *
 ******************************************************************************/

static VOID Parse_empty (ABC_INSTRUCTION_P instr)
{	
	(*instr->code) ();
}

static VOID Parse_empty_or_n (ABC_INSTRUCTION_P instr)
{
	INT16 n;
	
	if (! scan_number16 (&n))
		(*instr->code) (-1);
	else
		(*instr->code) (n);
}

static VOID Parse_n (ABC_INSTRUCTION_P instr)
{	
	INT16 n; 

	if (! scan_number16 (&n))
	{	MissingNumber (instr->name);
		return;
	}
	(*instr->code) (n);
}


static VOID Parse_BOOL (ABC_INSTRUCTION_P instr)
{	INT32 b; 

	if (! scan_bool (&b))
	{	MissingBoolean (instr->name);
		return;
	}
	(*instr->code) (b);
}


static VOID Parse_CHAR (ABC_INSTRUCTION_P instr)
{	INT32 c; 

	if (! scan_char (&c))
	{	MissingCharacter (instr->name);
		return;
	}
	(*instr->code) (c);
}


static VOID Parse_INTEGER (ABC_INSTRUCTION_P instr)
{	INT32 b;

	if (! scan_number32 (&b))
	{	MissingNumber (instr->name);
		return;
	}
	(*instr->code) (b);
}


static VOID Parse_LABEL (ABC_INSTRUCTION_P instr)
{	
	STRING	s;
	SSTRING chars;
	
	init_string(s,chars);

	if (! scan_ident (&s))
	{	MissingIdentifier (instr->name);
		return;
	}
	(*instr->code) (&s);
}


static VOID Parse_REAL (ABC_INSTRUCTION_P instr)
{	FLOAT r; 

	if (! scan_real (&r))
	{	MissingReal (instr->name);
		return;
	}
	(*instr->code) (&r);
}


static VOID Parse_STRING (ABC_INSTRUCTION_P instr)
{	
	STRING	s;
	SSTRING chars;

	init_string(s,chars);
	
	if (! scan_string (&s))
	{	MissingString (instr->name);
		return;
	}
	(*instr->code) (&s);
}


static VOID Parse_BOOLn (ABC_INSTRUCTION_P instr)
{	INT32 b;
	INT16 n; 

	if (! scan_bool (&b))
	{	MissingBoolean (instr->name);
		return;
	}
	if (! scan_number16 (&n))
	{	MissingNumber (instr->name);
		return;
	}
	(*instr->code) (b,n);
}


static VOID Parse_CHARn (ABC_INSTRUCTION_P instr)
{	INT32 c;
	INT16 n; 

	if (! scan_char (&c))
	{	MissingCharacter (instr->name);
		return;
	}
	if (! scan_number16 (&n))
	{	MissingNumber (instr->name);
		return;
	}
	(*instr->code) (c,n);
}


static VOID Parse_nn (ABC_INSTRUCTION_P instr)
{	INT16 n1, n2; 

	if (! scan_number16 (&n1))
	{	MissingNumber (instr->name);
		return;
	}
	if (! scan_number16 (&n2))
	{	MissingNumber (instr->name);
		return;
	}
	(*instr->code) (n1,n2);
}


static VOID Parse_LABELn (ABC_INSTRUCTION_P instr)
{
	STRING	s;
	SSTRING chars;
	INT16 n; 

	init_string(s,chars);
	
	if (! scan_ident (&s))
	{	MissingIdentifier (instr->name);
		return;
	}
	if (! scan_number16 (&n))
	{	MissingNumber (instr->name);
		return;
	}
	(*instr->code) (&s,n);
}


static VOID Parse_INTEGERn (ABC_INSTRUCTION_P instr)
{	INT32 b;
	INT16 n; 

	if (! scan_number32 (&b))
	{	MissingNumber (instr->name);
		return;
	}
	if (! scan_number16 (&n))
	{	MissingNumber (instr->name);
		return;
	}
	(*instr->code) (b,n);
}

static VOID Parse_REALn (ABC_INSTRUCTION_P instr)
{	FLOAT r;
	INT16 n; 

	if (! scan_real (&r))
	{	MissingReal (instr->name);
		return;
	}
	if (! scan_number16 (&n))
	{	MissingNumber (instr->name);
		return;
	}
	(*instr->code) (&r,n);
}


static VOID Parse_STRINGn (ABC_INSTRUCTION_P instr)
{	
	STRING	s;
	SSTRING chars;
	INT16 n;

	init_string(s,chars);
	
	if (! scan_string (&s))
	{	MissingString (instr->name);
		return;
	}
	if (! scan_number16 (&n))
	{	MissingNumber (instr->name);
		return;
	}
	(*instr->code) (&s,n);
}


static VOID Parse_nnn (ABC_INSTRUCTION_P instr)
{
	INT16 n1, n2, n3; 

	if (! scan_number16 (&n1) || ! scan_number16 (&n2) || ! scan_number16 (&n3))
	{	MissingNumber (instr->name);
		return;
	}
	(*instr->code) (n1,n2,n3);	
}

static VOID Parse_LABELnn (ABC_INSTRUCTION_P instr)
{
	STRING	s;
	SSTRING chars;
	INT16 n1, n2; 

	init_string(s,chars);
	
	if (! scan_ident (&s))
	{	MissingIdentifier (instr->name);
		return;
	}
	if (! scan_number16 (&n1))
	{	MissingNumber (instr->name);
		return;
	}
	if (! scan_number16 (&n2))
	{	MissingNumber (instr->name);
		return;
	}
	(*instr->code) (&s,n1,n2);
}


static VOID Parse_LABELnLABELn (ABC_INSTRUCTION_P instr)
{	
	STRING s1, s2;
	SSTRING	chars1,chars2;
	INT16 n1, n2; 

	init_string(s1,chars1);
	init_string(s2,chars2);
	
	if (! scan_ident (&s1))
	{	MissingIdentifier (instr->name);
		return;
	}
	if (! scan_number16 (&n1))
	{	MissingNumber (instr->name);
		return;
	}
	if (! scan_ident (&s2))
	{	MissingIdentifier (instr->name);
		return;
	}
	if (! scan_number16 (&n2))
	{	MissingNumber (instr->name);
		return;
	}
	(*instr->code) (&s1,n1,&s2,n2);
}

static VOID Parse_Rtn (ABC_INSTRUCTION_P instr)
{
	demand_instr_expected=false;
	
	if (demanded_flag)
	{	demanded_flag=false;
		code_d_rtn(demanded_a_stack_size,demanded_b_stack_size, demanded_vector);
		return;
	}
	Expected (instr->name, ".d directive");
}

static VOID Parse_Jmp (ABC_INSTRUCTION_P instr)
{	
	STRING label;
	SSTRING	chars;
	
	init_string(label,chars);
	
	if (!scan_ident(&label))
	{	MissingIdentifier (instr->name);
		return;
	}
	demand_instr_expected=false;
	
	if (demanded_flag)
	{	demanded_flag=false;
		code_d_jmp(&label, demanded_a_stack_size,demanded_b_stack_size, demanded_vector);
		return;
	}
	code_jmp(&label);
}

static VOID Parse_Jsr (ABC_INSTRUCTION_P instr)
{
	if (!scan_ident(&jsr_label))
	{	MissingIdentifier (instr->name);
		return;
	}
	jsr_flag=true;
	demand_instr_expected=false;
	
	if (demanded_flag)
	{	offer_expected = true;
		return;
	}
	Expected (instr->name, ".d directive");
}

/************************************************************************************
 *																					*
 *		Clean <-> Transputer interface												*
 *																					*
 ************************************************************************************/

/*
	The Clean compiler generates 'in' and 'out' instructions. The transputer
	code generator uses these instruction to store parameters at an appropriate
	place and to provide some kind of compatibility check with inline
	transputer code. In the future it may also generate some code to arange
	parameters so that compatibilty is kept, to reserve memory on the stack,
	or push constants (etc...), but at the moment it doesn't.
	
	The parameter of this instruction is an array of statements about the
	expected layout of the stack. In Clean it looks like "var0:place0,
	var1:place1, ...". The Clean compiler transforms this into "a0:place0
	a1:place1 ...", in other words stack+offset:place. Statement for the
	a-stack, b-stack, or any other stack can be ordered arbitrarily.
	
	The "place" statements depend on the code generator used, just as
	inline transputer-code is compiler dependent. This is not serious as
	the placement directives have only meaning when inline assembly is
	used. For inline PABC instructions it should be omitted. Note that the
	"in" instruction alone is not sufficient to read transputer instructions:
	a square bracket "[...]" should be used to indicate the start of a block of
	inline transputer instructions...
	
	The place statement can be of the following form for the transputer:
	
	L+offset:	a particular local
	A+offset:	on the a-stack on the given offset
	B+offset:	on the b-stack on the given offset
	
	Note: offsets are decimal.
	
	The placement statements must reflect the state of both stack-tops.
	"Holes" are not permitted.
*/

static VOID Parse_InParameters (ABC_INSTRUCTION_P instr)
{
	STRING params;
	SSTRING	chars;
	
	if (!options.inline_instructions)
	{
		NotExpected(instr->name,"inline transputer instructions");
		return;
	}
	
	init_string(params,chars);

	if (!scan_parameters(&demanded_a_stack_size,&demanded_b_stack_size,demanded_vector,&params))
	{
		MissingParameters (instr->name);
		return;
	}

	/* test if the parameters are the same as the ones offered on entry... */

	if (demanded_b_stack_size == offered_b_stack_size)
	{
		/* on entry the same b-parameters have been passed */
		
		if (demanded_a_stack_size == offered_a_stack_size)
		{
			/* a-parameters exactly the same, store all in the right place... */
			
			(*instr->code) (demanded_a_stack_size,demanded_b_stack_size,demanded_vector,&params);
			return;
		}
		
		if (demanded_a_stack_size+1 == offered_a_stack_size)
		{
			/*	one more a-parameter passed than demanded: this is considered
				an implicit parameter result node. */
			
			(*instr->code) (offered_a_stack_size,demanded_b_stack_size,demanded_vector,&params);
			return;
		
		}
		
		/*	offered a-parameters are not the same as the ones in the
			'in' instruction. Give a warning and store parameters
			according to 'in' instruction anyway...
		*/

		scan_warning(instr->name,"demanded %d input a-stack parameters, offered (.o) %d a-stack parameters\n",demanded_a_stack_size,offered_a_stack_size);
		(*instr->code) (demanded_a_stack_size,demanded_b_stack_size,demanded_vector,&params);
		return;
	}

	if (demanded_a_stack_size == offered_a_stack_size || demanded_a_stack_size + 1 == offered_a_stack_size)
		scan_warning(instr->name,"demanded %d input b-stack parameters, offered (.o) %d b-stack parameters\n",demanded_b_stack_size,offered_b_stack_size);
	else
		scan_warning(instr->name,"demanded input parameters do not coincide with offered (.o) parameters\n");
	
	(*instr->code) (demanded_a_stack_size,demanded_b_stack_size,demanded_vector,&params);
}

static VOID Parse_OutParameters (ABC_INSTRUCTION_P instr)
{
	STRING params;
	SSTRING	chars;
	
	if (!options.inline_instructions)
	{
		NotExpected(instr->name,"inline transputer instructions");
		return;
	}
	
	init_string(params,chars);

	if (!scan_parameters(&demanded_a_stack_size,&demanded_b_stack_size,demanded_vector,&params))
	{
		MissingParameters (instr->name);
		return;
	}
	
	(*instr->code) (demanded_a_stack_size,demanded_b_stack_size,demanded_vector,&params);
}

/*******************************************************
 *                                                     *
 *     The parser, directives                          *
 *                                                     *
 *******************************************************/
 
static VOID ParseD_Comp (ABC_INSTRUCTION_P instr)
{
	SSTRING chars;
	STRING opts;
	INT32 version;

	init_string(opts,chars);
	
	if		(CompDirectiveParsed)			TooMany (instr->name, ".comp directive");
	else if	(! scan_number32 (& version))	MissingNumber (instr->name);
	else if (! scan_ident (&opts))			MissingIdentifier (instr->name);
	else									CompDirectiveParsed = true;
}


static VOID ParseD_Code (ABC_INSTRUCTION_P instr)
{
	if		(CodeDirectiveParsed)		TooMany (instr->name, ".code directive");
	else if (! scan_number32 (&n_instructions) || ! scan_number32 (&n_descriptors) || !scan_number32 (&n_labels))
		MissingNumber (instr->name);
	else
		CodeDirectiveParsed  = true;
}


static VOID ParseD_Start (ABC_INSTRUCTION_P instr)
{
	SSTRING chars;
	STRING startlabel;
	
	init_string(startlabel,chars);
			
	if		(StartDirectiveParsed)		TooMany (instr->name, ".start directive");
	else if	(! scan_ident (&startlabel))	MissingIdentifier (instr->name);
	else
	{	code_start (&startlabel);
		StartDirectiveParsed   = true;
	}
}

static VOID import_file(STRING *oldname);

static VOID ParseD_Deps (ABC_INSTRUCTION_P instr)
{
	SSTRING chars;
	STRING s;

	init_string(s,chars);
	
	if		(! InfoDirectiveExpected)		NotExpected (instr->name, ".depend directive");
	else if (! scan_string (&s))			MissingString (instr->name);
	else
	{	import_file (&s);
		while (scan_string (&s))
			import_file (&s);
	}
}

static VOID ParseD_Exps (ABC_INSTRUCTION_P instr)
{	
	SSTRING chars;
	STRING s;
	
	init_string(s,chars);
	
	if		(InfoDirectiveExpected)		NotExpected (instr->name, ".export directive");
	else if (! scan_ident (&s))			MissingIdentifier (instr->name);
	else
	{	code_export (&s);
		while (scan_ident (&s))
			code_export (&s);
	}
}

static VOID ParseD_Imps (ABC_INSTRUCTION_P instr)
{
	SSTRING chars;
	STRING s;
	
	init_string(s,chars);
	
	if (! scan_ident (&s))
		MissingIdentifier (instr->name);
	else
	{
		code_import (&s);
		while (scan_ident (&s))
			code_import (&s);
	}
}

static VOID ParseD_Stop (ABC_INSTRUCTION_P instr)
{
	if		(! InfoDirectiveExpected)		NotExpected (instr->name, ".endinfo directive");
	else
	{
		short main;
		char *name;

		InfoDirectiveExpected = false;

		name=get_curr_file_name(&main);
			
		if (main)
			set_file_stage("reading \'%s\'",name);
	}
}


static VOID ParseD_Inline (ABC_INSTRUCTION_P instr)
{
	SSTRING chars;
	STRING s;
	
	init_string(s,chars);
	
	if		(InfoDirectiveExpected)			NotExpected (instr->name, ".inline directive");
	else if (!scan_ident(&s))				MissingIdentifier (instr->name);
}


static VOID ParseD_End (ABC_INSTRUCTION_P instr)
{
	if		(InfoDirectiveExpected)			NotExpected (instr->name, ".end directive");
}


static VOID ParseD_ImpLab (ABC_INSTRUCTION_P instr)
{
	SSTRING chars;
	STRING s;
	
	init_string(s,chars);
	
	if		(InfoDirectiveExpected)			NotExpected (instr->name, ".implab directive");
	else if (! scan_ident (&s))				MissingIdentifier (instr->name);
	else
	{	code_import(&s);
		while (scan_ident (&s))
			code_import(&s);
	}
}


static VOID ParseD_ImpDesc (ABC_INSTRUCTION_P instr)
{	
	SSTRING chars1,chars2;
	STRING s1,s2;
	
	init_string(s1,chars1);
	init_string(s2,chars2);
	
	if		(InfoDirectiveExpected)			NotExpected (instr->name, ".impdesc");
	else if (! scan_ident (&s1))			MissingIdentifier (instr->name);
	else
	{
		if (scan_ident (&s2))
			code_impdesc (&s1,&s2);
		else
			code_impdesc (&s1,NULL);
	}
}

static VOID ParseD_D (ABC_INSTRUCTION_P instr)
{
	if		(InfoDirectiveExpected)						NotExpected (instr->name, ".d directive");
	else if (!scan_number16 (&demanded_a_stack_size) ||
			 !scan_number16 (&demanded_b_stack_size))		MissingNumber (instr->name);
	else if (!ScanBVector (&demanded_b_stack_size,
			 demanded_vector))							MissingVector (instr->name);
	else
	{	demanded_flag = true;
		demand_instr_expected=true;
	}
}

static VOID ParseD_O (ABC_INSTRUCTION_P instr)
{			
	if		(InfoDirectiveExpected)						NotExpected (instr->name, ".o directive");
	else if (!scan_number16 (&offered_a_stack_size) ||
			 !scan_number16 (&offered_b_stack_size))	MissingNumber (instr->name);
	else if (!ScanBVector (&offered_b_stack_size,
			 offered_vector))							MissingVector (instr->name);
	else
	{
		offer_expected=false;
	
		if (jsr_flag)
		{	if (demanded_flag)
			{	code_d_jsr_o(&jsr_label,
							(INT16)demanded_a_stack_size, (INT16)demanded_b_stack_size, demanded_vector,
							(INT16)offered_a_stack_size, (INT16)offered_b_stack_size, offered_vector);
				demanded_flag=false;
				jsr_flag=false;
			} else
				Expected(instr->name,".d directive missing");
		
		} else
		{
			/*	just administer the availablity of offered parameters */
						
			offered_flag = true;
		}
	}
}

static VOID ParseD_N (ABC_INSTRUCTION_P instr)
{	
	if		(! scan_number16 (&entry_arity))		MissingNumber (instr->name);
	else if	(! scan_ident (&entry_descriptor))		MissingIdentifier(instr->name);
	else
	{
		entry_flag=true;
		offer_expected=true;

		if (! scan_ident (&entry_ealabel))
			entry_ealabel.length=0;
	}
}

static VOID ParseD_Desc (ABC_INSTRUCTION_P instr)
{	
	SSTRING desc_chars, node_entry_chars, lazy_entry_chars, repr_chars;
	STRING desc, node_entry, lazy_entry, repr;
	INT16 arity;
	
	init_string(desc,desc_chars);
	init_string(node_entry, node_entry_chars);
	init_string(lazy_entry, lazy_entry_chars);
	init_string(repr, repr_chars);
	
	if		(InfoDirectiveExpected)			NotExpected (instr->name, ".desc directive");
	else if (! scan_ident (&desc))			MissingIdentifier (instr->name);
	else if (! scan_ident (&node_entry))	MissingIdentifier (instr->name);
	else if (! scan_ident (&lazy_entry))	MissingIdentifier (instr->name);
	else if (! scan_number16 (&arity))		MissingNumber (instr->name);
	else if (! scan_string (&repr))			MissingString (instr->name);
	else									code_desc (&desc,&node_entry,&lazy_entry,arity,&repr);
}

static VOID ParseD_String (ABC_INSTRUCTION_P instr)
{
	SSTRING label_chars, string_chars;
	STRING label, string;
	
	init_string(label,label_chars);
	init_string(string, string_chars);
	
	if		(InfoDirectiveExpected)				NotExpected(instr->name, ".string directive");
	else if (! scan_ident (&label))				MissingIdentifier (instr->name);
	else if (! scan_string (&string))			MissingString(instr->name);
	else										code_string(&label,&string);
}

static VOID ParseD_Keep (ABC_INSTRUCTION_P instr)
{	INT16 keep_offset, result_offset;

	if		(InfoDirectiveExpected)			NotExpected (instr->name, ".keep directive");
	else if	(! scan_number16 (&keep_offset))	MissingNumber (instr->name);
	else if (! scan_number16 (&result_offset))	MissingNumber (instr->name);
	else										code_keep (keep_offset,result_offset);
}

/*******************************************************************************
 *                                                                             *
 *    The parser, main                                                         *
 *                                                                             *
 ******************************************************************************/

static VOID parse_inline(VOID)
{
	if (options.inline_instructions)
	{
		/*	one may insert a .d directive before the inline instruction,
			but this will be ignored */
		
		demand_instr_expected=false;
		demanded_flag=false;	
		
		if (!scan_number16 (&demanded_a_stack_size))
		{
			/* no stack layout information */
			
			end_asm_flag=code_d_begin_inline(0,0,e_vector);
			
		} else
		{
			/* parse stack information */

			if (!scan_number16 (&demanded_b_stack_size))
			{
				MissingNumber (NULL);
				return;	
			}
			
			if (!ScanBVector (&demanded_b_stack_size,demanded_vector))
			{
				MissingVector (NULL);
				return;	
			}	
			end_asm_flag=code_d_begin_inline(demanded_a_stack_size,demanded_b_stack_size, demanded_vector);
		}
		
		if (!scan_number16 (&offered_a_stack_size))
			return;										/* default already set */
		
		if (!scan_number16 (&offered_b_stack_size))
		{
			MissingNumber (NULL);
			return;
		}
		
		if (!ScanBVector (&offered_b_stack_size,offered_vector))
		{
			MissingVector (NULL);
			return;
		}
		code_o_end_inline(offered_a_stack_size,offered_b_stack_size,offered_vector);

		/* Now indicate the inline block has been parsed. The offer-expected
		   flag is set, which allows for efficient clearing of the end_asm flag
		   (without having to do this after every instruction...) */
		
		offer_expected=true;
		return;
	}
	
	NotExpected ("[", "inline transputer instructions");
}

static VOID ParseDirectiveLine (VOID)
{	
	SSTRING chars;
	STRING s;
	ABC_INSTRUCTION_P instr;
	
	init_string(s,chars);
	
	if (!scan_ident (&s))
	{
		Expected(".","directive");
		return;
	}
	
	if (!(instr = get_instruction (&s)))
	{
		Unknown (s.string, "directive");
		return;
	}

	if (offer_expected && instr->kind != O_Dir)
	{
		/* this is not a '.o' directive, but perhaps a ']' has been
		   parsed... */
		
		if (end_asm_flag)
		{
			offer_expected=false;
			end_asm_flag=false;
		
		} else
		{	
			Expected(instr->name, ".o directive");
			return;
		}
	}
	if (demand_instr_expected && instr->kind != D_Instr)
	{
		Expected(instr->name, "jsr, jmp, rtn or {");
		return;
	}
	
	(*instr->parse) (instr);
}

static VOID ParseImportDirectiveLine (VOID)
{	
	SSTRING chars;
	STRING s;
	ABC_INSTRUCTION_P instr;
	
	init_string(s,chars);

	if (!scan_ident(&s))
	{
		/* a directive should follow */
		
		Expected (".", "directive");
		return;
	}

	instr = get_instruction (&s);
	if 	(instr)
	{	switch (instr->kind)
		{	case DepDir:	ParseD_Deps (instr); break;	/* recursive depend */
			case ExpDir:	ParseD_Imps (instr); break; /* export: import in this module */
			default:		;
		}
	}
	else
		Unknown (s.string, "directive");
}

static VOID parse_label (STRING *s)
{
	if (offered_flag)
	{	if (entry_flag)
		{
			if (entry_ealabel.length==0)
				code_n_o_label(s,entry_arity,&entry_descriptor,NULL,offered_a_stack_size,offered_b_stack_size,offered_vector);
			else
				code_n_o_label(s,entry_arity,&entry_descriptor,&entry_ealabel,offered_a_stack_size,offered_b_stack_size,offered_vector);
				
			entry_flag=false;
			offered_flag=false;
			
			return;
		}

		code_o_label(s, offered_a_stack_size,offered_b_stack_size, offered_vector);
		offered_flag=false;
		return;
	}
	
	code_label (s);
}

static VOID ParseInstructionLine (VOID)
{
	SSTRING chars;
	STRING s;
	ABC_INSTRUCTION_P instr;
	CHAR colon;

	init_string(s,chars);
	
	if (!scan_ident(&s))
	{
		/* perhaps this is the start of a block of inline transputer instructions */
		
		if (scan_inline_start())
			parse_inline();
		
		return;
	}
	
	if (scan_colon(&colon))
	{	parse_label(&s);
		if (!scan_ident(&s))
			return;
	}
	
	if (! (instr = get_instruction (&s)))
	{	Unknown (s.string, "instruction");
		return;
	}

	if (offered_flag)
	{
		/* start new block according to description */
		
		code_o((INT16)offered_a_stack_size, (INT16)offered_b_stack_size, offered_vector);
		offered_flag=false;
	}

	if (offer_expected)
	{
		/* this is not a '.o' directive, but perhaps a ']' has been parsed */
		
		if (end_asm_flag)
		{
			offer_expected=false;
			end_asm_flag=false; /* not necessary to end block */
			
		} else
		{
			Expected(instr->name, ".o directive");
			return;
		}
	}
	
	if (demand_instr_expected && instr->kind != D_Instr)
	{	Expected(instr->name, "jsr, jmp, rtn or [");
		return;
	}
		
	(*instr->parse) (instr);
}

static INT16 Parse_Info_Line (VOID)
{	
	skip_whites ();
	if (is_directive())
		ParseDirectiveLine ();
	else if (scan_inline_start())
		parse_inline();

	return SkipComment ();
}

static INT16 Parse_Line (VOID)
{	
	skip_whites ();
	if (is_directive())
		ParseDirectiveLine ();
	else
		ParseInstructionLine ();

	return SkipComment ();
}

static INT16 Parse_ImportLine (VOID)
{
	skip_whites ();
	
	if (is_directive())
		ParseImportDirectiveLine ();
	else if (scan_inline_start())
		skip_assembler();

	return SkipRest ();
}

/********************************************************************
 *                                                                  *
 *        external definitions                                      *
 *                                                                  *
 ********************************************************************/
 
INSTRUCTION_BLOCK *generate_intermediate_code(CHAR *filename)
{
	INT16 result;
	INSTRUCTION_BLOCK *code;

	initialise_instruction_table();
	
	reset_flags();
	init_string(jsr_label,			jsr_label_chars);
	init_string(entry_descriptor,	entry_descriptor_chars);
	init_string(entry_ealabel,		entry_ealabel_chars);
		
	CompDirectiveParsed = false;
	CodeDirectiveParsed = false;
	StartDirectiveParsed = false;
	InfoDirectiveExpected = true;

	n_instructions	=-1;
	n_descriptors	=-1;
	n_labels		=-1;

	if (open_source_file(filename))
	{			
		handle_compute_event();
		result  = 0;
	
		set_file_stage("reading \'%s\'",filename);

		while (InfoDirectiveExpected && result == PROCEED && nr_errors < options.max_errors)
			result = Parse_Info_Line ();

		while (result == PROCEED && nr_errors < options.max_errors)
		{	result = Parse_Line ();
			handle_generate_intermediate();
		}

		code=code_end(end_asm_flag);
		
		close_input_file();
						
	}else
	{	fatal_error("input file \'%s\' couldn't be opened\n",filename);
		code=NULL;
	}
	
	return code;
}

static VOID import_file(STRING *filename)
{
	INT16 result;
	
	/* ignoring depends speeds up parsing: mostly needed if no implicit imports */
	
	if (options.ignore_depend)
		return;
	
	if (open_depend_file(filename))
	{
		char *previous_name;
		short previous_main;

		set_file_stage("importing \'%s\'",filename->string);

		/* a .endinfo should follow, this will reset the code stage string */
		
		result  = 0;
	
		while (result == PROCEED && nr_errors < options.max_errors)
			result = Parse_ImportLine ();

		close_input_file();

		previous_name=get_curr_file_name(&previous_main);

		if (!previous_main)
			set_file_stage("importing \'%s\'",previous_name);
			
		/* if main, the endinfo directive will restore the file stage */

    }else
    	error("file \'%s\' couldn't be imported\n",filename->string);    
}