#include <string.h>

#include "ifinterface.h"

#include "cginputs.h"
#include "cgscanner.h"
#include "cglabels.h"
#include "cgmemory.h"
#include "cglin.h"
#include "cgstacks.h"
#include "cgwrites.h"

#include "ccintermediate.h"
#include "ccmacros.h"

#define zerochar 48

/********************************************************************************
 *																				*
 *		instructions to generate code											*
 *																				*
 ********************************************************************************/

static TINSTRUCTION_DEF *label_tinstruction;	/* used for definition of labels */

static TINSTRUCTION *first_macro_instruction;
static TINSTRUCTION *last_macro_instruction;

static TINSTRUCTION **first_instruction;
static TINSTRUCTION *last_instruction;

INT32 number_of_inline_instructions;

/********************************************************************************
 *																				*
 *		construction of instruction args										*
 *																				*
 ********************************************************************************/
 
static TINSTRUCTION_NODE *new_integer(INT16 kind,INT32 number)
{
	TINSTRUCTION_NODE *new_node;
	
	new_node=alloc_heap_mem(sizeof(TINSTRUCTION_NODE)+sizeof(TINSTRUCTION_PARAMETER));
	new_node->kind=kind;
	new_node->parameter[0].number=number;
	
	return new_node;
}

static TINSTRUCTION_NODE *new_float(FLOAT number)
{
	TINSTRUCTION_NODE *new_node;
	
	new_node=alloc_heap_mem(sizeof(TINSTRUCTION_NODE)+2*sizeof(TINSTRUCTION_PARAMETER));
	new_node->kind=AFLOAT;
	*(FLOAT *)&new_node->parameter[0].number=number;
	
	return new_node;
}

static TINSTRUCTION_NODE *new_label(INT16 kind,LABEL *label)
{
	TINSTRUCTION_NODE *new_node;
	
	new_node=alloc_heap_mem(sizeof(TINSTRUCTION_NODE)+sizeof(TINSTRUCTION_PARAMETER));
	new_node->kind=kind;
	new_node->parameter[0].label=label;
	
	return new_node;
}

static TINSTRUCTION_NODE *new_string(STRING *string)
{
	TINSTRUCTION_NODE *new_node;
	
	new_node=alloc_heap_mem(sizeof(TINSTRUCTION_NODE)+sizeof(TINSTRUCTION_PARAMETER));
	new_node->kind=ASTRING;
	new_node->parameter[0].string=new_string_from_string(string);
	
	return new_node;
}

static TINSTRUCTION_NODE *new_literal(STRING *string)
{
	TINSTRUCTION_NODE *new_node;
	
	new_node=alloc_heap_mem(sizeof(TINSTRUCTION_NODE)+sizeof(TINSTRUCTION_PARAMETER));
	new_node->kind=ALITERAL;
	new_node->parameter[0].literal=new_chars_from_string(string);
	
	return new_node;
}

static TINSTRUCTION_NODE *new_monadic(INT16 kind, TINSTRUCTION_NODE *arg)
{
	TINSTRUCTION_NODE *new_node;
	
	new_node=alloc_heap_mem(sizeof(TINSTRUCTION_NODE)+sizeof(TINSTRUCTION_PARAMETER));
	new_node->kind=kind;
	new_node->parameter[0].pointer=arg;
	
	return new_node;
}

static TINSTRUCTION_NODE *new_dyadic(INT16 kind, TINSTRUCTION_NODE *arg1, TINSTRUCTION_NODE *arg2)
{
	TINSTRUCTION_NODE *new_node;
	
	new_node=alloc_heap_mem(sizeof(TINSTRUCTION_NODE)+2*sizeof(TINSTRUCTION_PARAMETER));
	new_node->kind=kind;
	new_node->parameter[0].pointer=arg1;
	new_node->parameter[1].pointer=arg2;
	
	return new_node;
}

/************************************************************************************
 *																					*
 *		construction of instructions												*
 *																					*
 ************************************************************************************/
 
static VOID define_macro_instr(TINSTRUCTION_DEF *instr)
{
	TINSTRUCTION *new_instruction;

	new_instruction=alloc_heap_mem (sizeof(TINSTRUCTION));
	
	new_instruction->operator=instr;
	new_instruction->next=NULL;
	
	if(last_macro_instruction)
		last_macro_instruction->next=new_instruction;
	else
		first_macro_instruction=new_instruction;
	last_macro_instruction=new_instruction;
}

static VOID define_macro_instr1(TINSTRUCTION_DEF *instr, TINSTRUCTION_NODE *arg)
{
	TINSTRUCTION *new_instruction;

	new_instruction=alloc_heap_mem (sizeof(TINSTRUCTION)+sizeof(TINSTRUCTION_NODE *));
	
	new_instruction->operator=instr;
	new_instruction->operand[0]=arg;
	new_instruction->next=NULL;
	
	if(last_macro_instruction)
		last_macro_instruction->next=new_instruction;
	else
		first_macro_instruction=new_instruction;
	last_macro_instruction=new_instruction;
}

static VOID inline_instr(TINSTRUCTION_DEF *instr)
{
	TINSTRUCTION *new_instruction;

	++number_of_inline_instructions;
	
	new_instruction=alloc_heap_mem (sizeof(TINSTRUCTION));
	
	new_instruction->operator=instr;	
	new_instruction->next=NULL;
	
	if(last_instruction)
		last_instruction->next=new_instruction;
	else
		*first_instruction=new_instruction;
	last_instruction=new_instruction;
}

static VOID inline_instr1(TINSTRUCTION_DEF *instr, TINSTRUCTION_NODE *arg)
{
	TINSTRUCTION *new_instruction;

	++number_of_inline_instructions;
	
	new_instruction=alloc_heap_mem (sizeof(TINSTRUCTION)+sizeof(TINSTRUCTION_NODE *));
	
	new_instruction->operator=instr;
	new_instruction->operand[0]=arg;
	new_instruction->next=NULL;
	
	if(last_instruction)
		last_instruction->next=new_instruction;
	else
		*first_instruction=new_instruction;
	last_instruction=new_instruction;
}

/************************************************************************************
 *																					*
 *		conversion of labels														*
 *																					*
 ************************************************************************************/

/* if an inline label is entered, this label may start with '.D', '.J', or '.G'.
   The code generator uses these prefices as well. Therefore these prefices
   are stripped and the flags of the label are altered accordingly. This is
   useful to detect doubly defined labels and other errors (e.g. a data label
   may not be used as code and vice versa). It should be noted that a jump
   label may also be defined as gcall label, so one should not issue an error
   if this happens
*/

static INT16 convert_label (register STRING *name, register U_INT16 *flags)
{
	/* the type of the label is returned, i.e. ALABEL, AJLABEL, AGLABEL, or
	   ADLABEL, so that this can be used in an inline instruction to determine
	   the particular use of this label */
	
	if (*flags & INLINE_LABEL)
	{
		if (name->length >= 2 && name->string[0]=='.')
		{	switch(name->string[1])
			{	case 'D':
					*flags &=~INLINE_LABEL;
					*flags |= DATA_LABEL;
					name->string+=2;
					return ADLABEL;
					
				case 'G':
					*flags &=~INLINE_LABEL;
					*flags |= GCALL_LABEL;
					name->string+=2;
					return AGLABEL;

				case 'J':
					*flags &=~INLINE_LABEL;
					*flags |= JUMP_LABEL;
					name->string+=2;
					return AJLABEL;

				case 'L':
					error("warning: .L labels used internally\n");
					return ALABEL;
					
				default:
					return ALABEL;
			}
		}
		return ALABEL;
		
	} else
		internal_error("convert_label: only conversion possible for inline labels");
	
	return ALABEL;
}

/********************************************************************************
 *																				*
 *		The inline parser, instruction hash table								*
 *																				*
 *		The instruction hash table should be initialised at system startup		*
 *		and not be changed. Note that instructions which are entered first		*
 *		can be found fast.														*
 *																				*
 ********************************************************************************/

static VOID parse_e					(TINSTRUCTION_DEF* instr);
static VOID parse_op				(TINSTRUCTION_DEF* instr);
static VOID parse_words				(TINSTRUCTION_DEF* instr);
static VOID parse_bytes				(TINSTRUCTION_DEF* instr);
static VOID parse_global			(TINSTRUCTION_DEF* instr);
static VOID parse_attrib			(TINSTRUCTION_DEF* instr);
static VOID parse_extern			(TINSTRUCTION_DEF* instr);
static VOID parse_macro				(TINSTRUCTION_DEF* instr);
static VOID parse_include			(TINSTRUCTION_DEF* instr);
static VOID parse_macro_application	(TINSTRUCTION_DEF* macro);
static VOID parse_inline_string		(TINSTRUCTION_DEF* instr);
static VOID parse_inline_float		(TINSTRUCTION_DEF* instr);

static VOID define_e				(TINSTRUCTION_DEF* instr);
static VOID define_op				(TINSTRUCTION_DEF* instr);
static VOID define_words			(TINSTRUCTION_DEF* instr);
static VOID define_bytes			(TINSTRUCTION_DEF* instr);
static VOID define_error			(TINSTRUCTION_DEF* instr);
static VOID define_inline_string	(TINSTRUCTION_DEF* instr);
static VOID define_inline_float		(TINSTRUCTION_DEF* instr);

#define TABLE_SIZE		500			/* 8*500 fits in memory block... */

typedef struct TMACRO_NODE
{
	struct TMACRO_NODE	*left;
	struct TMACRO_NODE	*right;
	TINSTRUCTION_DEF	macro;
	
} TMACRO_NODE;

typedef struct TABLE_ENTRY
{
	struct TABLE_ENTRY		*next;
	TINSTRUCTION_DEF		*instr;

} TABLE_ENTRY;

static TABLE_ENTRY 			*instr_table;			/* the instruction table */
static TMACRO_NODE			*macros;				/* the macro tree */

static U_INT16 instr_hash (CHAR *s)
{
	register U_INT16 h;
	
	/* add the character values of the string */
	
	h=0;
	while(*s != '\0')
		h+=*s++;
		
	return h;
}

static VOID put_instruction_name(TINSTRUCTION_DEF *instruction)
{
	TABLE_ENTRY *entry;
	
	entry=&instr_table [instr_hash (instruction->name) % TABLE_SIZE];
	
	if (entry->instr!=NULL)
	{
		TABLE_ENTRY *link;
		
		link=entry;
		while (link->next!=NULL)
			link=link->next;
		
		for (entry=instr_table; entry<&instr_table [TABLE_SIZE]; ++entry)
			if (entry->instr==NULL)
				break;
		
		if (entry==&instr_table [TABLE_SIZE])
			fatal_error ("not enough entries in instruction hash tabel");
		
		link->next=entry;
	}
	
	entry->instr=instruction;

	if (instruction->kind==TLABEL_DEF)
		label_tinstruction=instruction;
}

static TINSTRUCTION_DEF *put_macro_name (STRING *s)
{
	register TMACRO_NODE **macro_p,*new_macro;

	macro_p=&macros;
	while (*macro_p!=NULL)
	{	TMACRO_NODE *macro=*macro_p;
		INT16 r=strcmp (s->string,macro->macro.name);
		if (r==0)
		{
			error("macro %s defined twice\n",s->string);
			return NULL;
		}
		if (r<0)
			macro_p=&macro->left;
		else
			macro_p=&macro->right;
	}
	
	new_macro=alloc_heap_mem(sizeof (TMACRO_NODE));
	
	new_macro->left		=NULL;
	new_macro->right	=NULL;

	new_macro->macro.name			=new_chars_from_string(s);
	new_macro->macro.kind			=TMACRO;
	new_macro->macro.arity			=0;
	new_macro->macro.n_labels		=0;
	new_macro->macro.instructions	=NULL;
	new_macro->macro.parse			=parse_macro_application;
	new_macro->macro.define			=define_error;
	
	*macro_p=new_macro;
	return &new_macro->macro;
}

static TINSTRUCTION_DEF instructions[] =
{
	{"ldc",			TINSTR,	parse_op,	define_op,	1,0,NULL},
	{"ldl",			TINSTR,	parse_op,	define_op,	1,0,NULL},
	{"stl",			TINSTR,	parse_op,	define_op,	1,0,NULL},
	{"ldlp",		TINSTR,	parse_op,	define_op,	1,0,NULL},
	{"adc",			TINSTR,	parse_op,	define_op,	1,0,NULL},
	{"eqc",			TINSTR,	parse_op,	define_op,	1,0,NULL},
	{"j",			TINSTR,	parse_op,	define_op,	1,0,NULL},
	{"cj",			TINSTR,	parse_op,	define_op,	1,0,NULL},
	{"ldnl",		TINSTR,	parse_op,	define_op,	1,0,NULL},
	{"stnl",		TINSTR,	parse_op,	define_op,	1,0,NULL},
	{"ldnlp",		TINSTR,	parse_op,	define_op,	1,0,NULL},
	{"call",		TINSTR,	parse_op,	define_op,	1,0,NULL},
	{"ajw",			TINSTR,	parse_op,	define_op,	1,0,NULL},
	
	{"rev",			TINSTR,	parse_e,	define_e,	0,0,NULL},
	{"add",			TINSTR,	parse_e,	define_e,	0,0,NULL},
	{"sub",			TINSTR,	parse_e,	define_e,	0,0,NULL},
	{"mul",			TINSTR,	parse_e,	define_e,	0,0,NULL},
	{"div",			TINSTR,	parse_e,	define_e,	0,0,NULL},
	{"rem",			TINSTR,	parse_e,	define_e,	0,0,NULL},
	{"sum",			TINSTR,	parse_e,	define_e,	0,0,NULL},
	{"diff",		TINSTR,	parse_e,	define_e,	0,0,NULL},
	{"prod",		TINSTR,	parse_e,	define_e,	0,0,NULL},
	{"and",			TINSTR,	parse_e,	define_e,	0,0,NULL},
	{"or",			TINSTR,	parse_e,	define_e,	0,0,NULL},
	{"xor",			TINSTR,	parse_e,	define_e,	0,0,NULL},
	{"not",			TINSTR,	parse_e,	define_e,	0,0,NULL},
	{"shl",			TINSTR,	parse_e,	define_e,	0,0,NULL},
	{"shr",			TINSTR,	parse_e,	define_e,	0,0,NULL},
	{"gt",			TINSTR,	parse_e,	define_e,	0,0,NULL},
	
	{"lend",		TINSTR,	parse_e,	define_e,	0,0,NULL},
	{"bcnt",		TINSTR,	parse_e,	define_e,	0,0,NULL},
	{"wcnt",		TINSTR,	parse_e,	define_e,	0,0,NULL},
	{"ldpi",		TINSTR,	parse_e,	define_e,	0,0,NULL},
	{"mint",		TINSTR,	parse_e,	define_e,	0,0,NULL},
	{"bsub",		TINSTR,	parse_e,	define_e,	0,0,NULL},
	{"wsub",		TINSTR,	parse_e,	define_e,	0,0,NULL},
	{"move",		TINSTR,	parse_e,	define_e,	0,0,NULL},
	{"in",			TINSTR,	parse_e,	define_e,	0,0,NULL},
	{"out",			TINSTR,	parse_e,	define_e,	0,0,NULL},
	{"lb",			TINSTR,	parse_e,	define_e,	0,0,NULL},
	{"sb",			TINSTR,	parse_e,	define_e,	0,0,NULL},
	{"outbyte",		TINSTR,	parse_e,	define_e,	0,0,NULL},
	{"outword",		TINSTR,	parse_e,	define_e,	0,0,NULL},
	{"gcall",		TINSTR,	parse_e,	define_e,	0,0,NULL},
	{"gajw",		TINSTR,	parse_e,	define_e,	0,0,NULL},
	{"ret",			TINSTR,	parse_e,	define_e,	0,0,NULL},
	{"startp",		TINSTR,	parse_e,	define_e,	0,0,NULL},
	{"endp",		TINSTR,	parse_e,	define_e,	0,0,NULL},
	{"runp",		TINSTR,	parse_e,	define_e,	0,0,NULL},
	{"stopp",		TINSTR,	parse_e,	define_e,	0,0,NULL},
	{"ldpri",		TINSTR,	parse_e,	define_e,	0,0,NULL},
	{"ldtimer",		TINSTR,	parse_e,	define_e,	0,0,NULL},
	{"tin",			TINSTR,	parse_e,	define_e,	0,0,NULL},
	
	{"alt",			TINSTR,	parse_e,	define_e,	0,0,NULL},
	{"altwt",		TINSTR,	parse_e,	define_e,	0,0,NULL},
	{"altend",		TINSTR,	parse_e,	define_e,	0,0,NULL},
	{"talt",		TINSTR,	parse_e,	define_e,	0,0,NULL},
	{"taltwt",		TINSTR,	parse_e,	define_e,	0,0,NULL},
	{"enbs",		TINSTR,	parse_e,	define_e,	0,0,NULL},
	{"diss",		TINSTR,	parse_e,	define_e,	0,0,NULL},
	{"enbc",		TINSTR,	parse_e,	define_e,	0,0,NULL},
	{"disc",		TINSTR,	parse_e,	define_e,	0,0,NULL},
	{"enbt",		TINSTR,	parse_e,	define_e,	0,0,NULL},
	{"dist",		TINSTR,	parse_e,	define_e,	0,0,NULL},
	
	{"csub0",		TINSTR,	parse_e,	define_e,	0,0,NULL},
	{"ccnt1",		TINSTR,	parse_e,	define_e,	0,0,NULL},
	{"testerr",		TINSTR,	parse_e,	define_e,	0,0,NULL},
	{"stoperr",		TINSTR,	parse_e,	define_e,	0,0,NULL},
	{"seterr",		TINSTR,	parse_e,	define_e,	0,0,NULL},
	{"xword",		TINSTR,	parse_e,	define_e,	0,0,NULL},
	{"cword",		TINSTR,	parse_e,	define_e,	0,0,NULL},
	{"xdble",		TINSTR,	parse_e,	define_e,	0,0,NULL},
	{"csngl",		TINSTR,	parse_e,	define_e,	0,0,NULL},
	{"ladd",		TINSTR,	parse_e,	define_e,	0,0,NULL},
	{"lsub",		TINSTR,	parse_e,	define_e,	0,0,NULL},
	{"lsum",		TINSTR,	parse_e,	define_e,	0,0,NULL},
	{"ldiff",		TINSTR,	parse_e,	define_e,	0,0,NULL},
	{"lmul",		TINSTR,	parse_e,	define_e,	0,0,NULL},
	{"ldiv",		TINSTR,	parse_e,	define_e,	0,0,NULL},
	{"lshl",		TINSTR,	parse_e,	define_e,	0,0,NULL},
	{"lshr",		TINSTR,	parse_e,	define_e,	0,0,NULL},
	
	{"norm",		TINSTR,	parse_e,	define_e,	0,0,NULL},
	{"resetch",		TINSTR,	parse_e,	define_e,	0,0,NULL},
	{"testpranal",	TINSTR,	parse_e,	define_e,	0,0,NULL},
	{"sthf",		TINSTR,	parse_e,	define_e,	0,0,NULL},
	{"stlf",		TINSTR,	parse_e,	define_e,	0,0,NULL},
	{"sttimer",		TINSTR,	parse_e,	define_e,	0,0,NULL},
	{"sthb",		TINSTR,	parse_e,	define_e,	0,0,NULL},
	{"stlb",		TINSTR,	parse_e,	define_e,	0,0,NULL},
	{"saveh",		TINSTR,	parse_e,	define_e,	0,0,NULL},
	{"savel",		TINSTR,	parse_e,	define_e,	0,0,NULL},
	{"clrhalterr",	TINSTR,	parse_e,	define_e,	0,0,NULL},
	{"sethalterr",	TINSTR,	parse_e,	define_e,	0,0,NULL},
	{"testhalterr",	TINSTR,	parse_e,	define_e,	0,0,NULL},
	
	{"dup",			TINSTR,	parse_e,	define_e,	0,0,NULL},
	{"move2dinit",	TINSTR,	parse_e,	define_e,	0,0,NULL},
	{"move2dall",	TINSTR,	parse_e,	define_e,	0,0,NULL},
	{"move2dnonzero",TINSTR,parse_e,	define_e,	0,0,NULL},
	{"move2dzero",	TINSTR,	parse_e,	define_e,	0,0,NULL},
	{"crcword",		TINSTR,	parse_e,	define_e,	0,0,NULL},
	{"crcbyte",		TINSTR,	parse_e,	define_e,	0,0,NULL},
	{"bitcnt",		TINSTR,	parse_e,	define_e,	0,0,NULL},
	{"bitrevword",	TINSTR,	parse_e,	define_e,	0,0,NULL},
	{"bitrevnbits",	TINSTR,	parse_e,	define_e,	0,0,NULL},
	{"wsubdb",		TINSTR,	parse_e,	define_e,	0,0,NULL},
	
	{"fpdup",		TINSTR,	parse_e,	define_e,	0,0,NULL},
	{"fprev",		TINSTR,	parse_e,	define_e,	0,0,NULL},
	{"fpldnlsn",	TINSTR,	parse_e,	define_e,	0,0,NULL},
	{"fpldnldb",	TINSTR,	parse_e,	define_e,	0,0,NULL},
	{"fpldnlsni",	TINSTR,	parse_e,	define_e,	0,0,NULL},
	{"fpldnldbi",	TINSTR,	parse_e,	define_e,	0,0,NULL},
	{"fpstnlsn",	TINSTR,	parse_e,	define_e,	0,0,NULL},
	{"fpstnldb",	TINSTR,	parse_e,	define_e,	0,0,NULL},
	
	{"fpurn",		TINSTR,	parse_e,	define_e,	0,0,NULL},
	{"fpurz",		TINSTR,	parse_e,	define_e,	0,0,NULL},
	{"fpurp",		TINSTR,	parse_e,	define_e,	0,0,NULL},
	{"fpurm",		TINSTR,	parse_e,	define_e,	0,0,NULL},
	
	{"fpadd",		TINSTR,	parse_e,	define_e,	0,0,NULL},
	{"fpsub",		TINSTR,	parse_e,	define_e,	0,0,NULL},
	{"fpmul",		TINSTR,	parse_e,	define_e,	0,0,NULL},
	{"fpdiv",		TINSTR,	parse_e,	define_e,	0,0,NULL},
	{"fpusqrtfirst",TINSTR,	parse_e,	define_e,	0,0,NULL},
	{"fpusqrtstep",	TINSTR,	parse_e,	define_e,	0,0,NULL},
	{"fpusqrtlast",	TINSTR,	parse_e,	define_e,	0,0,NULL},
	{"fpremfirst",	TINSTR,	parse_e,	define_e,	0,0,NULL},
	{"fpremstep",	TINSTR,	parse_e,	define_e,	0,0,NULL},
	
	{"fpldzerosn",	TINSTR,	parse_e,	define_e,	0,0,NULL},
	{"fpldzerodb",	TINSTR,	parse_e,	define_e,	0,0,NULL},
	{"fpumulby2",	TINSTR,	parse_e,	define_e,	0,0,NULL},
	{"fpudivby2",	TINSTR,	parse_e,	define_e,	0,0,NULL},
	{"fpuexpinc32",	TINSTR,	parse_e,	define_e,	0,0,NULL},
	{"fpuexpdec32",	TINSTR,	parse_e,	define_e,	0,0,NULL},
	{"fpuabs",		TINSTR,	parse_e,	define_e,	0,0,NULL},
	
	{"fpldnladdsn",	TINSTR,	parse_e,	define_e,	0,0,NULL},
	{"fpldnladddb",	TINSTR,	parse_e,	define_e,	0,0,NULL},
	{"fpldnlmulsn",	TINSTR,	parse_e,	define_e,	0,0,NULL},
	{"fpldnlmuldb",	TINSTR,	parse_e,	define_e,	0,0,NULL},
	
	{"fpchkerr",	TINSTR,	parse_e,	define_e,	0,0,NULL},
	{"fptesterr",	TINSTR,	parse_e,	define_e,	0,0,NULL},
	{"fpuseterr",	TINSTR,	parse_e,	define_e,	0,0,NULL},
	{"fpuclrerr",	TINSTR,	parse_e,	define_e,	0,0,NULL},
	
	{"fpgt",		TINSTR,	parse_e,	define_e,	0,0,NULL},
	{"fpeq",		TINSTR,	parse_e,	define_e,	0,0,NULL},
	{"fpordered",	TINSTR,	parse_e,	define_e,	0,0,NULL},
	{"fpnan",		TINSTR,	parse_e,	define_e,	0,0,NULL},
	{"fpnotfinite",	TINSTR,	parse_e,	define_e,	0,0,NULL},
	{"fpur32tor64",	TINSTR,	parse_e,	define_e,	0,0,NULL},
	{"fpur64tor32",	TINSTR,	parse_e,	define_e,	0,0,NULL},
	{"fpint",		TINSTR,	parse_e,	define_e,	0,0,NULL},
	{"fpstnli32",	TINSTR,	parse_e,	define_e,	0,0,NULL},
	{"fpuchki32",	TINSTR,	parse_e,	define_e,	0,0,NULL},
	{"fpuchki64",	TINSTR,	parse_e,	define_e,	0,0,NULL},
	
	{"fprtoi32",	TINSTR,	parse_e,	define_e,	0,0,NULL},
	{"fpi32tor32",	TINSTR,	parse_e,	define_e,	0,0,NULL},
	{"fpi32tor64",	TINSTR,	parse_e,	define_e,	0,0,NULL},
	{"fpb32tor64",	TINSTR,	parse_e,	define_e,	0,0,NULL},
	{"fpunoround",	TINSTR,	parse_e,	define_e,	0,0,NULL},

	{".word",		TWORD_DEF,		parse_words,			define_words,			1,0,NULL},
	{".byte",		TBYTE_DEF,		parse_bytes,			define_bytes,			1,0,NULL},
	{".string",		TSTRING_DEF,	parse_inline_string,	define_inline_string,	1,0,NULL},
	{".float",		TFLOAT_DEF,		parse_inline_float,		define_inline_float,	1,0,NULL},

	{":",			TLABEL_DEF,		parse_e,				define_error,			1,0,NULL},
	
	{".align",		TALIGN,			parse_e,				define_e,				0,0,NULL},
	{".global",		TGLOBAL,		parse_global,			define_error,			1,0,NULL},
	{".extern",		TEXTERN,		parse_extern,			define_error,			1,0,NULL},
	{".attrib",		TATTRIB,		parse_attrib,			define_error,			1,0,NULL},

	{".macro",		TMACRO,			parse_macro,			define_error,			0,0,NULL},
	{".include",	TINCLUDE,		parse_include,			define_error,			0,0,NULL},
};

static VOID initialise_instruction_table(VOID)
{
	register INT16 i;
	
	instr_table = (TABLE_ENTRY *) alloc_heap_mem((U_INT32)TABLE_SIZE*sizeof (TABLE_ENTRY));
	
	for (i = 0; i < TABLE_SIZE; i++)
	{	instr_table[i].instr  	= NULL;
		instr_table[i].next 	= NULL;
	}

	for (i=0; i<sizeof (instructions)/sizeof (TINSTRUCTION_DEF); ++i)
		put_instruction_name (&instructions[i]);
}

static TINSTRUCTION_DEF *get_instruction (STRING *s)
{
	register TMACRO_NODE *macro_p;
	TABLE_ENTRY *entry;
		
	entry=&instr_table [instr_hash(s->string) % TABLE_SIZE];
	
	while (entry!=NULL && entry->instr!=NULL)
	{
		if (strcmp (entry->instr->name,s->string)==0)
			return entry->instr;
		
		entry=entry->next;		
	}
	
	/* instruction not found, try macro */
		
	macro_p=macros;
	while (macro_p!=NULL)
	{
		INT16 r=strcmp (s->string,macro_p->macro.name);
		if (r==0)
			return &macro_p->macro;
		if (r<0)
			macro_p=macro_p->left;
		else
			macro_p=macro_p->right;
	}
	return NULL;
}

/*******************************************************************************
 *                                                                             *
 *    The parser, error messages                                               *
 *                                                                             *
 ******************************************************************************/

static void MissingExpression (char *s)
{	scan_error (s,"wrong or missing argument: expression expected");
	SkipToChar ('\n');
}

static void MissingEqual (char *s)
{	scan_error (s,"missing =");
	SkipToChar ('\n');
}

static void MissingIdentifier (char *s)
{	scan_error (s,"wrong or missing argument: identifier expected");
	SkipToChar ('\n');
}

static void MissingMacroName (char *s)
{	scan_error (s,"macro name expected");
	SkipToChar ('\n');
}

static VOID MissingOpen (char *s)
{	scan_error (s,"\'(\' exprected");
	SkipToChar ('\n');
}

static VOID MissingClose (char *s)
{	scan_error (s,"\')\' exprected");
	SkipToChar ('\n');
}

static void MissingFileName (char *s)
{	scan_error (s,"filename expected");
	SkipToChar ('\n');
}

static void MissingString (char *s)
{	scan_error (s,"string expected");
	SkipToChar ('\n');
}

static void MissingFloat (char *s)
{	scan_error (s,"float expected");
	SkipToChar ('\n');
}

static void UndefinedMacroLabel(char *s)
{	scan_error (s,"macro label not in define list");
	SkipToChar ('\n');
}

static void Unknown (char *s, char *what)
{	char line[STRINGSIZE+1];

	strcpy (line, "unknown ");
	strcat (line, what);
	
	scan_error (s,line);
	SkipToChar ('\n');
}

/****************************************************************************
 *																			*
 *		the parser, parsing of defined instructions							*
 *																			*
 ****************************************************************************/

static BOOL parse_attribute_identifier(register STRING *s)
{
	SSTRING	chars;
	STRING 	i;

	init_string(i,chars);
	
	if (!scan_assembler_ident(&i))
		return false;
	
	/* append identifier (without label conversion) */
	
	if (s->length+i.length > STRINGSIZE)
		return false;
	
	memcpy (&s->string[s->length],i.string,i.length+1);
	s->length+=i.length;
	
	enter_label(&i, INLINE_LABEL | USE_LABEL | DEF_LABEL);

	return true;
}

static BOOL parse_attribute_value(register STRING *s)
{
	if (is_number())
	{
		INT32 n;
		
		if (!scan_number32(&n))
			return false;
		
		/* append decimal number */
		
		if (s->length > STRINGSIZE -12)
			return false;					/* no room for integer */
		
		s->length+=(int)sprintf(&s->string[s->length],"%ld",(long)n);
		
		return true;
	}

	return parse_attribute_identifier(s);
}

/************************************************************************************
 *																					*
 *		expressions																	*
 *																					*
 ************************************************************************************/

static TINSTRUCTION_NODE *parse_term(VOID)
{
	SSTRING			chars;
	STRING			i;
	LABEL			*label;
	U_INT16			label_flags;
	INT16			label_type;

	skip_whites();

	if (is_number())
	{
		INT32 n;
		
		if (!scan_number32(&n))
			return NULL;
		
		return new_integer(AINTEGER,n);
	}
	
	if (is_char())
	{
		INT32 c;
		
		if (!scan_char(&c))
			return NULL;

		return new_integer(AINTEGER,c);
	}				
	
	init_string (i,chars);
	
	if (!scan_assembler_ident(&i))
		return NULL;
			
	if((label=find_macro_label(&i))!=NULL)
	{
		/* this is a macro argument or locally defined label */
		
		if(label->flags & MACRO_ARGUMENT)
			return new_integer(APARAMETER,label->number);
		
		return new_integer(ALOCAL,label->number);
	}
	
	label_flags=(INLINE_LABEL | USE_LABEL);
	label_type=convert_label(&i, &label_flags);
	label=enter_label(&i, label_flags);
	
	if (label->def_type == D_EXPRESSION)
	{
		/* this labels defines an expression, just return this expression */
		
		return label->def.expression;
	}
	
	/* return label */
	
	return new_label(label_type,label);
}

/* lowest number is highest prio */

static TINSTRUCTION_NODE	*parse_prio4_expression(VOID);
#define parse_expression()	parse_prio4_expression()

static TINSTRUCTION_NODE *parse_prio0_expression(VOID)
{
	/* terms and expressions in brackets (highest prio) */
	
	CHAR	c;
	
	if (scan_open (&c))
	{	
		TINSTRUCTION_NODE *e;
		
		/* any expression may be placed between brackets */
		
		e=parse_expression();
		
		if (e==NULL)
			return NULL;
		
		if (!scan_close(&c))			
		{	scan_error (NULL,"closing bracket missing");
			return NULL;
		}
		
		switch (e->kind)
		{
			case AINTEGER:
			case ALABEL:
			case ALOCAL:
				return e;
				
			default:
				return new_monadic(ABRACKET,e);
		}
	}
		
	return parse_term();
}

static TINSTRUCTION_NODE *parse_prio1_expression(VOID)
{
	/* prio 0 expressions with prio 1 operators (monadic) */

	INT16 operator;
	
	if (scan_monadic_assembler_operator(&operator))
	{
		TINSTRUCTION_NODE *e;
		
		/* only prio 2 expressions allowed for monadic operators
		   associativity is from right to left */
		
		e=parse_prio1_expression();
		
		if (e->kind == AINTEGER)
		{
			switch(operator)
			{
				case AMONADD:
					return e;
					
				case AMONSUB:
					return new_integer(AINTEGER,-e->parameter[0].number);
					
				default:
					return new_monadic(operator,e);
			}
		} else
			return new_monadic(operator,e);
	}
	
	return parse_prio0_expression();
}

static TINSTRUCTION_NODE *parse_prio2_expression(VOID)
{	
	INT16 operator;
	TINSTRUCTION_NODE *e1, *e2;

	/* expressions of prio 1 with operators of prio 2 */
	
	e1=parse_prio1_expression();
	
	if (e1==NULL)
		return NULL;
	
	while (scan_prio2_assembler_operator(&operator))
	{		
		/* associativity is from left to right */
		
		e2=parse_prio1_expression();
		
		if (e2==NULL)
			return NULL;
		
		if (e1->kind == AINTEGER && e2->kind == AINTEGER)
		{
			switch(operator)
			{
				case AMUL:
					e1=new_integer(AINTEGER,e1->parameter[0].number*e2->parameter[0].number);
					break;

				case ADIV:
					e1=new_integer(AINTEGER,e1->parameter[0].number/e2->parameter[0].number);
					break;

				case AMOD:
					e1=new_integer(AINTEGER,e1->parameter[0].number%e2->parameter[0].number);
					break;
					
				default:
					e1=new_dyadic(operator,e1,e2);
					break;
			}
		} else
			e1=new_dyadic(operator,e1,e2);
	}
	
	return e1;
}

static TINSTRUCTION_NODE *parse_prio3_expression(VOID)
{	
	/* expressions of prio 2 with operators of prio 3 */

	INT16 operator;
	TINSTRUCTION_NODE *e1, *e2;
	
	e1=parse_prio2_expression();
	
	if (e1==NULL)
		return false;
	
	while (scan_prio3_assembler_operator(&operator))
	{		
		/* associativity is from left to right */
		
		e2=parse_prio2_expression();
		
		if (e2==NULL)
			return NULL;

		if (e1->kind == AINTEGER && e2->kind == AINTEGER)
		{
			switch(operator)
			{
				case AADD:
					e1=new_integer(AINTEGER,e1->parameter[0].number+e2->parameter[0].number);
					break;

				case ASUB:
					e1=new_integer(AINTEGER,e1->parameter[0].number-e2->parameter[0].number);
					break;
					
				default:
					e1=new_dyadic(operator,e1,e2);
					break;
			}
		} else
			e1=new_dyadic(operator,e1,e2);
	}
	
	return e1;
}

static TINSTRUCTION_NODE *parse_prio4_expression(VOID)
{	
	/* expressions of prio 3 with operators of prio 4 */

	INT16 operator;
	TINSTRUCTION_NODE *e1, *e2;
	
	e1=parse_prio3_expression();
	
	if (e1==NULL)
		return false;
	
	while (scan_prio4_assembler_operator(&operator))
	{	
		/* associativity is from left to right */
		
		e2=parse_prio3_expression();
		
		if (e2==NULL)
			return NULL;

		if (e1->kind == AINTEGER && e2->kind == AINTEGER)
		{
			switch(operator)
			{
				case AOR:
					e1=new_integer(AINTEGER,e1->parameter[0].number | e2->parameter[0].number);
					break;

				case AAND:
					e1=new_integer(AINTEGER,e1->parameter[0].number & e2->parameter[0].number);
					break;
				
				default:
					e1=new_dyadic(operator,e1,e2);
					break;
			}
		} else
			e1=new_dyadic(operator,e1,e2);
	}
	
	return e1;
}

static TINSTRUCTION_NODE *parse_string (VOID)
{
	SSTRING	chars;
	STRING	string;
	
	init_string(string,chars);
	
	if (!scan_string(&string))
		return NULL;
	
	return new_string(&string);
}

static TINSTRUCTION_NODE *parse_float (VOID)
{
	FLOAT f; 

	if (! scan_real (&f))
		return NULL;
	
	return new_float(f);
}

/************************************************************************************
 *																					*
 *		parse functions stored in instruction table									*
 *																					*
 ************************************************************************************/

static VOID parse_e (TINSTRUCTION_DEF *instr)
{
	inline_instr(instr);
}

static VOID parse_op (TINSTRUCTION_DEF *instr)
{
	TINSTRUCTION_NODE *e;
	
	e=parse_expression();
	
	if (e != NULL)
	{	
		inline_instr1(instr,e);
		return;
	}

	MissingExpression(instr->name);
}

static VOID parse_words (TINSTRUCTION_DEF *instr)
{
	TINSTRUCTION_NODE *words;
	
	words=parse_expression();
	
	if (words != NULL)
	{	
		CHAR c;
		
		while (scan_assembler_separator(&c))
		{
			TINSTRUCTION_NODE *next_word;
			
			next_word=parse_expression();
			
			if (next_word==NULL)
			{
				MissingExpression(instr->name);
				return;
			}
			
			words=new_dyadic(ACOMMA,words,next_word);
		}
		inline_instr1(instr,words);
		
		return;
	}

	MissingExpression(instr->name);
}

static VOID parse_bytes (TINSTRUCTION_DEF *instr)
{
	CHAR c;
	TINSTRUCTION_NODE *bytes;

	if ((bytes=parse_string()) == NULL &&
		(bytes=parse_expression()) == NULL)	
	{
		MissingExpression(instr->name);
		return;
	}
	
	while (scan_assembler_separator(&c))
	{	
		TINSTRUCTION_NODE *next_bytes;
		
		if ((next_bytes=parse_string()) == NULL &&
			(next_bytes=parse_expression()) == NULL)
		{
			MissingExpression(instr->name);
			return;
		}
		
		bytes=new_dyadic(ACOMMA,bytes,next_bytes);
	}
	
	inline_instr1(instr,bytes);
}

static VOID parse_inline_string (TINSTRUCTION_DEF *instr)
{
	TINSTRUCTION_NODE *string;
	
	string=parse_string();
	
	if (string != NULL)
	{
		inline_instr1(instr,string);
		return;
	}

	MissingString(instr->name);
}

static VOID parse_inline_float (TINSTRUCTION_DEF *instr)
{
	TINSTRUCTION_NODE *f;
	
	f=parse_float();
	
	if (f != NULL)
	{
		inline_instr1(instr,f);
		return;
	}

	MissingFloat(instr->name);
}

static VOID parse_extern (TINSTRUCTION_DEF *instr)
{
	SSTRING	chars;
	STRING	string;
	U_INT16	flags;

	init_string(string,chars);

	if (!scan_assembler_ident(&string))
	{
		MissingIdentifier(instr->name);
		return;
	}

	flags=(INLINE_LABEL | IMPORT_LABEL);
	convert_label(&string, &flags);
	enter_label(&string,flags);
}

static VOID parse_global (TINSTRUCTION_DEF *instr)
{
	SSTRING	chars;
	STRING	string;
	U_INT16	flags;

	init_string(string,chars);

	if (!scan_assembler_ident(&string))
	{
		MissingIdentifier(instr->name);
		return;
	}
	
	flags=(INLINE_LABEL | EXPORT_LABEL);
	convert_label(&string, &flags);
	enter_label(&string,flags);
}

static VOID parse_attrib (TINSTRUCTION_DEF *instr)
{
	SSTRING		chars;
	STRING		s;
	CHAR		c;
	
	options.output_default_attributes=false;
	
	init_string(s,chars);
	
	if (!parse_attribute_identifier(&s))
	{	
		MissingIdentifier(instr->name);
		return;
	}
	
	if (!scan_equal(&c))
	{	MissingEqual(instr->name);
		return;
	}
	
	s.string[s.length++]=c;
	
	if (!parse_attribute_value(&s))
	{
		MissingExpression(instr->name);
		return;
	}
	
	inline_instr1(instr,new_literal(&s));
}

static VOID parse_macro_application(TINSTRUCTION_DEF *macro)
{	
	U_INT16		i;
	TINSTRUCTION *new_instruction;
	
	++number_of_inline_instructions;
	
	new_instruction=alloc_heap_mem(sizeof(TINSTRUCTION)+macro->arity*sizeof(TINSTRUCTION_NODE *));
	new_instruction->operator=macro;
	new_instruction->next=NULL;
		
	if(last_instruction)
		last_instruction->next=new_instruction;
	else
		*first_instruction=new_instruction;
	last_instruction=new_instruction;
	
	for(i=0; i<macro->arity; i++)
	{
		TINSTRUCTION_NODE *argument;
				
		if (is_string())
		{
			argument=parse_string();
			
			if (argument != NULL)
			{	
				new_instruction->operand[i]=argument;
				continue;
			}
			
			scan_error (macro->name,"string expected as macro argument");
			SkipToChar ('\n');
			return;
		}
		
		argument=parse_expression();
		
		if (argument != NULL)
		{
			new_instruction->operand[i]=argument;
			continue;
		}
		
		scan_error (macro->name,"expression expected as macro argument");
		SkipToChar ('\n');
		return;
	}
}

/****************************************************************************
 *																			*
 *		the parser, parsing of macro definition								*
 *																			*
 ****************************************************************************/

static BOOL parse_macro_argument (INT16 i)
{
	/* definition of formal parameter i */
	
	SSTRING chars;
	STRING	s;
	
	init_string(s,chars);
	
	if (scan_assembler_ident(&s))
	{
		LABEL *label;

		/* define local label */
		
		label=enter_macro_label(&s,MACRO_ARGUMENT,i);
		define_label(label);
		
		return true;
	}
	return false;
}

static BOOL parse_macro_label_declaration (INT16 i)
{
	/* declaration of local label i */
	
	SSTRING chars;
	STRING s;

	init_string(s,chars);
	
	if (scan_assembler_ident(&s))
	{	
		/* define local label */
		
		enter_macro_label(&s,MACRO_LOCAL,i);
		return true;
	}
	return false;
}

static TINSTRUCTION_DEF *parse_macro_name (VOID)
{
	SSTRING chars;
	STRING s;

	init_string(s,chars);
	
	if (scan_assembler_ident(&s))
		return put_macro_name (&s);
	return NULL;
}

/********************************************************************************
 *																				*
 *		macro definition functions stored in instruction table					*
 *																				*
 ********************************************************************************/

static VOID define_e(TINSTRUCTION_DEF *instr)
{
	define_macro_instr(instr);
}

static VOID define_op(TINSTRUCTION_DEF *instr)
{
	TINSTRUCTION_NODE *e;
	
	e=parse_expression();
	
	if (e != NULL)
	{	
		define_macro_instr1(instr,e);
		return;
	}

	MissingExpression(instr->name);	
}

static VOID define_words (TINSTRUCTION_DEF *instr)
{
	TINSTRUCTION_NODE *words;
	
	words=parse_expression();
	
	if (words != NULL)
	{	
		CHAR c;
		
		while (scan_assembler_separator(&c))
		{
			TINSTRUCTION_NODE *next_word;
			
			next_word=parse_expression();
			
			if (next_word==NULL)
			{
				MissingExpression(instr->name);
				return;
			}
			
			words=new_dyadic(ACOMMA,words,next_word);
		}
		define_macro_instr1(instr,words);
		
		return;
	}

	MissingExpression(instr->name);
}

static VOID define_bytes (TINSTRUCTION_DEF *instr)
{
	CHAR c;
	TINSTRUCTION_NODE *bytes;

	if ((bytes=parse_string()) == NULL &&
		(bytes=parse_expression()) == NULL)	
	{
		MissingExpression(instr->name);
		return;
	}
	
	while (scan_assembler_separator(&c))
	{	
		TINSTRUCTION_NODE *next_bytes;
		
		if ((next_bytes=parse_string()) == NULL &&
			(next_bytes=parse_expression()) == NULL)
		{
			MissingExpression(instr->name);
			return;
		}
		
		bytes=new_dyadic(ACOMMA,bytes,next_bytes);
	}
	
	define_macro_instr1(instr,bytes);
}

static TINSTRUCTION_NODE *parse_parameter(VOID)
{
	SSTRING chars;
	STRING s;
	LABEL *argument;
		
	init_string(s,chars);
		
	if(!scan_assembler_ident(&s))
		return NULL;

	argument=find_macro_label(&s);

	if(!argument)
		return NULL;

	return new_integer(APARAMETER,argument->number);
}

static VOID define_inline_string (TINSTRUCTION_DEF *instr)
{
	/* the string may be an argument or a string. If it is a string the string
	   is used as argument, otherwise the argument number is used
	*/

	if (is_string())
	{	TINSTRUCTION_NODE *string;
	
		string=parse_string();
	
		if (string != NULL)
		{
			define_macro_instr1(instr,string);
			return;
		}

		MissingString(instr->name);
		
	} else
	{
		TINSTRUCTION_NODE *p;
		
		p=parse_parameter();
		
		if (p != NULL)
		{
			define_macro_instr1(instr,p);
			return;
		}
			
		scan_error (instr->name,"a macro argument expected");
		SkipToChar ('\n');
		return;
	}
}

static VOID define_inline_float (TINSTRUCTION_DEF *instr)
{
	/* the float may be an argument or a float. If it is a float the float itself
	   is used as argument, otherwise the argument number is used (@A...)
	*/

	if (is_number())
	{
		TINSTRUCTION_NODE *f;
		
		f=parse_float();
		
		if (f != NULL)
		{
			define_macro_instr1(instr,f);
			return;
		}
	
		MissingFloat(instr->name);
		
	} else
	{
		TINSTRUCTION_NODE *p;
		
		p=parse_parameter();
		
		if (p != NULL)
		{
			define_macro_instr1(instr,p);
			return;
		}
		
		scan_error (instr->name,"a macro argument expected");
		SkipToChar ('\n');
		return;
	}
}

static VOID define_error(TINSTRUCTION_DEF *instr)
{
	scan_error (instr->name,"illegal instruction within macro definition");
	SkipToChar ('\n');
}

/********************************************************************
 *																	*
 *		macro parser												*
 *																	*
 ********************************************************************/

static VOID define_macro_args(TINSTRUCTION_DEF *macro)
{
	CHAR c;
	int i=0;

	if (!scan_open(&c))
	{
		MissingOpen(macro->name);
		return;
	}
	
	if (parse_macro_argument(i))
	{
		++i;
		while (scan_assembler_separator(&c))
		{
			if (!parse_macro_argument(i))
			{
				MissingIdentifier(macro->name);
				return;
			}
			
			++i;
		}
	}

	if (!scan_close(&c))
	{	MissingClose(macro->name);
		return;
	}

	macro->arity=i;
}

static VOID declare_macro_local_labels(TINSTRUCTION_DEF *macro)
{
	CHAR	c;
	INT16	i;
	
	i=0;
	
	while (!scan_begin_macro(&c))
	{
		while(parse_macro_label_declaration(i))
			i++;
		
		skip_assembler_comment();
	}
	macro->n_labels=i;
}

static VOID parse_macro_instruction(VOID)
{
	SSTRING	chars;
	STRING	s;
	CHAR	c;
	TINSTRUCTION_DEF *instr;

	init_string(s,chars);
	
	if (!scan_assembler_ident(&s))
		return;
	
	if (scan_colon(&c))					/* skip optional colon */
	{
		LABEL *label;
		
		if(!(label=find_macro_label(&s)))
		{	UndefinedMacroLabel(s.string);
			return;
		}
		define_label(label);
	
		/* define label instruction with number of label */
	
		define_macro_instr1(label_tinstruction,new_integer(ALOCAL,label->number));

		init_string(s,chars);
		
		if (!scan_assembler_ident(&s))
			return;
	}
	
	if (!(instr=get_instruction(&s)))
	{	Unknown (s.string, "instruction");
		return;
	}
	
	(*instr->define)(instr);
}

static INT16 parse_macro_line(VOID)
{
	CHAR c;
	
	parse_macro_instruction();
	
	if (scan_end_macro(&c))
		return END_OF_MACRO;
	
	return skip_assembler_comment ();
}

static VOID parse_macro(TINSTRUCTION_DEF *instr)
{
	INT16 result;
	TINSTRUCTION_DEF *macro;
	
	if (!(macro=parse_macro_name()))
	{	
		skip_assembler_comment();
		
		if (!(macro=parse_macro_name()))
		{
			MissingMacroName(instr->name);
			return;
		}
	}
	
	define_macro_args(macro);
	skip_assembler_comment();

	declare_macro_local_labels(macro);
	
	first_macro_instruction=NULL;
	last_macro_instruction=NULL;
	
	result = 0;
	while (result == PROCEED && nr_errors < options.max_errors)
		result = parse_macro_line();
	
	macro->instructions=first_macro_instruction;
	
	free_macro_labels();
}

/****************************************************************
 *																*
 *		include files											*
 *																*
 ****************************************************************/

static VOID IncludeFile(STRING *filename);

static VOID parse_include(TINSTRUCTION_DEF *instr)
{
	SSTRING chars;
	STRING s;

	init_string(s,chars);
	
	if (!scan_assembler_ident(&s))
	{	MissingFileName(instr->name);
		return;
	}
	IncludeFile(&s);
}

/*******************************************************************************
 *                                                                             *
 *    The parser, main                                                         *
 *                                                                             *
 ******************************************************************************/

static INT16 parse_line(VOID)
{
	SSTRING	chars;
	STRING	s;
	CHAR	c;
	TINSTRUCTION_DEF *instr;
	
	init_string(s,chars);
	
	if (!scan_assembler_ident(&s))
		return skip_assembler_comment ();
	
	if (scan_colon(&c))					/* skip optional colon */
	{
		LABEL	*label;
		U_INT16	flags;
		INT16	type;
		
		flags=INLINE_LABEL;
		type=convert_label(&s, &flags);
		label=enter_label(&s,flags);
		define_inline_label(label,flags);

		inline_instr1(label_tinstruction,new_label(type,label));

		init_string(s,chars);
		
		if (!scan_assembler_ident(&s))
			return skip_assembler_comment ();
	}
	
	if (scan_equal(&c))
	{
		LABEL				*ident;
		U_INT16				flags;
		TINSTRUCTION_NODE	*e;

		e=parse_expression();
		
		if (e == NULL)
		{
			MissingExpression(s.string);
			return skip_assembler_comment();
		}
		
		flags=INLINE_LABEL;
		convert_label(&s, &flags);
		ident=enter_label(&s, flags);

		ident->def_type=D_EXPRESSION;
		ident->def.expression=e;
		
		define_inline_label(ident,flags);
		
		return skip_assembler_comment();
	}
	
	if (!(instr=get_instruction(&s)))
	{
		Unknown (s.string, "instruction");
		return skip_assembler_comment ();
	}
	
	(*instr->parse)(instr);
	
	return skip_assembler_comment ();
}

VOID init_inline (VOID)
{
	number_of_inline_instructions=0;
	macros=NULL;
	
	if (options.inline_instructions)
		initialise_instruction_table();
}

static VOID IncludeFile (STRING *filename)
{
	INT16 result;
	
	if (open_depend_file(filename))
	{
		char *previous_name;
		short previous_main;

		set_file_stage("including \'%s\'",filename->string);
				
		result  = 0;
		
		while (result == PROCEED && nr_errors < options.max_errors)
		{
			result = parse_line ();
			handle_compute_event(); 
			
			/*	do not call 'handle_generate_intermediate', this should only
				be done for the main file */
		}

		close_input_file();
		
		previous_name=get_curr_file_name(&previous_main);

		if (previous_main)
			set_file_stage("reading \'%s\'",previous_name);
		else
			set_file_stage("including \'%s\'",previous_name);
			
    }else
    	error("file \'%s\' couldn't be included\n",filename->string);    
}

BOOL parse_inline_instructions (INSTRUCTION_BLOCK *block)
{
	INT16 result;
	
	first_instruction=append_inline_block(block);
	last_instruction=*first_instruction;
	
	result=0;
	
	while (result == PROCEED  && nr_errors < options.max_errors)
	{	result = parse_line ();
		handle_generate_intermediate();
	}
	
	return (result == END_OF_INLINE);
}
