#include "ifinterface.h"

#include "ccmacros.h"
#include "ccintermediate.h"

#include "cglocal.h"
#include "cgmemory.h"
#include "cglin.h"

/*	This module inserts instructions that move locals between external
	memory and internal memory. At this moment it uses the same scheme
	as the code generator for the M680x0 processor. Note the slight
	adaption below.
	
	-	normally the real local will be allocated that contains a value which
		will not be used for the longest time when all locals are in use. The
		only exception is the following.
	-	if a parameter is stored in some virtual register, this parameter
		must be stored in the same real register *eventually* and we will, by
		default, store it there already when the store instruction is
		encountered to avoid having to move the value at the end of the
		basic block (if it was placed in the wrong real registers). Thus,
		at such a point we will allocate this real register regardless of
		the next use of the value that is stored in it (The same happens when
		such a parameter is used hereafter . This means it is not
		a good idea to allocate this local shortly before (if the next use of
		the 'previous' value is large), as it may be needed shortly hereafter.
		The point is that the next use of the 'previous' value should be small:
		the distance to the point where it will be thrown out regardless.
		Note that we do not know in advance which virtual local will be placed
		in which real local, so we cannot connect this 'fixed' local point
		to some virtual local a priori (see also the M680x0 implementation for
		this: it just checks the next use of the parameter local some virtual
		local resides and adjusts the real next use if this is earlier than the
		next use of the value in the real local). Note that for our adjustment we
		need to know from what point the next use of the parameter local should be
		considered. We could keep track of it at every instruction, but it would
		be better to store it at the real local and use the next field of the
		parameter register after this point. The next use of some real local
		is at any point: the minimum of the next use of its virtual local and the
		'fixed' points of the real local (or only the fixed point when
		it is empty). When the fixed point of some local is nearby it will
		automatically be passed by for allocation (except when all other locals
		are used before this point).

		Note that a parameter will (a) not be used anymore in the same block,
		so that the next use of it's value is set to the end of the block, or
		(b) can be accessed for read access, so that the next use is this
		read access.	
*/

/************************************************************************************
 *																					*
 *		local typedefs																*
 *																					*
 ************************************************************************************/

#define NOT_STORED			-1
#define	NO_LOCAL			-1

#define UNUSED				0		/* not used anymore							*/
#define	NEXT_BLOCK			1		/* used in next block						*/
#define LAST_INSTRUCTION	2		/* used by last instruction of this block	*/

/* flags for group use */

#define USE					1		/* bit true if group uses local				*/
#define DEF					2		/* bit true if group defines local			*/
#define	FIXED				4		/* bit true if local is fixed				*/
#define	LOW					8		/* bit true if local is low word of float	*/
#define	HIGH				16		/* bit true if local is high word of float	*/

typedef struct VIRTUAL_LOCAL
{
	INT16	next_use;				/* next group the virtual local is used			*/
	BOOL	used;					/* true if local used lateron					*/
	INT16	number;					/* the number of this virtual local				*/
	INT16	real_local;				/* real local the virtual local is placed		*/
	
	INT16	offset;					/* offset in external memory					*/
	U_INT16	group_flags;			/* indicates for each group the local use		*/
	
	
} VIRTUAL_LOCAL;

typedef struct REAL_LOCAL
{
	INT16	next_use;				/* next group that uses this local				*/
	BOOL	used;					/* true if local used lateron					*/
	INT16	number;					/* the number of this real local				*/
	INT16	virtual_local;			/* the virtual local associated with this local	*/
	
	BOOL	altered;				/* true if contents have been altered			*/
	
	struct REAL_LOCAL	*next;		/* next unlocked local */
	struct REAL_LOCAL	*prev;		/* previous unlocked local */
	BOOL				locked;		/* true if local locked */
		
} REAL_LOCAL;

/************************************************************************************
 *																					*
 *		global variables															*
 *																					*
 ************************************************************************************/

/*	end_locals indicates which (virtual) locals should be kept after the basic block.
	This includes the locals for which a store node has been evaluated
	(the parameters) and the locals that are dedicated (stack pointers,
	heap pointers, the return address).
	
	Note that the amount of real locals cannot exceed 32 at this moment and
	that the amount of virtual locals allocated is always more than the
	amount of real locals (otherwise this module would not be necessary at
	all).
*/
	
static U_INT32	end_locals;				/* end_a_registers... */

/*	virtual_locals is an array of virtual locals, that is allocated for each
	block. deallocation is done implicitly
*/

static VIRTUAL_LOCAL 	*virtual_memory;
static REAL_LOCAL		real_memory[N_LOCALS];
static REAL_LOCAL		*unlocked_list;

/*	the virtual_memory_use indicates how many words are stored in virtual memory.
	usually this is less than the total amount of virtual memory. The virtual
	memory offset is the offset from the a-stack pointer to the virtual memory
	area. If the a-stack shrinks, this is equal to -stack_increase, as the
	stack pointer  will be decreased at the start of the basic block. On the
	other hand, if the a-stack grows, the offset is equal to stack_increase,
	as the a-stack pointer will be increased at the end of the basic block in this
	case...
*/

static INT16			virtual_memory_use;		/* number of virtual memory words used */
static INT16			virtual_memory_offset;	/* offset from a-stack to virtual memory */

/************************************************************************************
 *																					*
 *		(un)locking	of real locals													*
 *																					*
 ************************************************************************************/

static VOID lock_locals(VOID)
{
	register INT16 n;
	register REAL_LOCAL *local;

	/* lock all locals */
	
	unlocked_list=NULL;
	
	for (n=0,local=&real_memory[0]; n<N_LOCALS; ++n,++local)
		local->locked=true;
}

#if _COMPILE_UNUSED_

static VOID unlock_locals (VOID)
{
	register INT16 n;
	register REAL_LOCAL *local;
	
	/* unlock all locals */

	local=&real_memory[0];
	
	unlocked_list=local;
	
	local->prev=NULL;
	local->next=local+1;
	local->locked=false;
	
	for (n=1,++local; n<N_LOCALS-1; ++n,++local)
	{
		local->prev=local-1;
		local->next=local+1;
		local->locked=false;
	}
	local->prev=local-1;
	local->next=NULL;
	local->locked=false;
}

#endif

static VOID lock_local (REAL_LOCAL *local)
{	
	REAL_LOCAL *next, *prev;
	
	if (local->locked)
		return;
	
	local->locked=true;

	/* get local out of unlocked list */
	
	next=local->next;
	prev=local->prev;
	
	if (next != NULL)
		next->prev=prev;
	
	if (prev != NULL)
		prev->next=next;
	else
		unlocked_list=next;	
}

static VOID unlock_local (REAL_LOCAL *local)
{	
	if (!local->locked)
		return;

	/* put (i.e. push) local in unlocked list */
	
	local->locked=false;
	local->next=unlocked_list;
	local->prev=NULL;
	
	if (unlocked_list != NULL)
		unlocked_list->prev=local;
	
	unlocked_list=local;
}

/*	if a local number is out of range the local is considered locked, so it
	can not be allocated
*/

#define is_unlocked_local(n)	((n)>=0 && (n)<N_LOCALS && !real_memory[n].locked)
#define is_locked_local(n)		((n)<0 || (n)>=N_LOCALS || real_memory[n].locked)

#define first_unlocked_local()	unlocked_list
#define next_unlocked_local(l)	l->next;

/************************************************************************************
 *																					*
 *		insertion of load/store/move instructions									*
 *																					*
 ************************************************************************************/

static INSTRUCTION *insert_instruction (INT16 code, INT16 arg_size, INSTRUCTION *insertion_point)
{
	/*	This routine inserts an instruction before the insertion point. It
		assumes the insertion point is not NULL and it is not the first
		instruction.  This can be garantueed, as at least a group instruction
		should precede.
	*/
	
	INSTRUCTION *instruction, *prev_instruction;

	instruction=alloc_heap_mem(sizeof (INSTRUCTION)+arg_size);
	instruction->code=code;

	prev_instruction=insertion_point->prev;
	
	prev_instruction->next	=instruction;
	insertion_point->prev	=instruction;
	
	instruction->next=insertion_point;
	instruction->prev=prev_instruction;
	
	return instruction;
}

static VOID append_load (INT16 local, INT16 offset, INSTRUCTION_BLOCK *block)
{
	append_instr_short(block,LDL,A_STACK_POINTER);
	append_instr_short(block,LDNL,offset+virtual_memory_offset);
	append_instr_short(block,STL,local);
}

static VOID insert_load (INT16 local, INT16 offset, INSTRUCTION *insertion_point)
{
	INSTRUCTION *instruction;
	
	instruction=insert_instruction(LDL,sizeof(PARAMETER),insertion_point);
	instruction->parameters[0].int16=A_STACK_POINTER;

	instruction=insert_instruction(LDNL,sizeof(PARAMETER),insertion_point);
	instruction->parameters[0].int16=offset+virtual_memory_offset;
	
	instruction=insert_instruction(STL,sizeof(PARAMETER),insertion_point);
	instruction->parameters[0].int16=local;
}

static VOID insert_store (INT16 local,INT16 offset, INSTRUCTION *insertion_point)
{
	INSTRUCTION *instruction;
	
	instruction=insert_instruction(LDL,sizeof(PARAMETER),insertion_point);
	instruction->parameters[0].int16=local;

	instruction=insert_instruction(LDL,sizeof(PARAMETER),insertion_point);
	instruction->parameters[0].int16=A_STACK_POINTER;
	
	instruction=insert_instruction(STNL,sizeof(PARAMETER),insertion_point);
	instruction->parameters[0].int16=offset+virtual_memory_offset;
}

static VOID insert_move (INT16 destination, INT16 source, INSTRUCTION *insertion_point)
{
	INSTRUCTION *instruction;
	
	instruction=insert_instruction(LDL,sizeof(PARAMETER),insertion_point);
	instruction->parameters[0].int16=source;
	
	instruction=insert_instruction(STL,sizeof(PARAMETER),insertion_point);
	instruction->parameters[0].int16=destination;
}

/************************************************************************************
 *																					*
 *		adjustment of instructions, i.e. replace virtual locals by real ones		*
 *																					*
 ************************************************************************************/

static VOID adjust_group_instructions (INSTRUCTION *first_instruction)
{
	INT16	real_number, virtual_number;
	register INSTRUCTION *instruction;
	
	/* replace each virtual local by its current real local */

	for (instruction=first_instruction; instruction!=NULL && instruction->code != GROUP; instruction=instruction->next)
	{
		switch (instruction->code)
		{
			case LDL:
			case STL:
			{	
				virtual_number=instruction->parameters[0].virtual->local;
				real_number=virtual_memory[virtual_number].real_local;
				
				if (real_number == NO_LOCAL)
					internal_error("adjust_instructions (1)");
					
				instruction->parameters[0].int16=real_number;
				
				break;
			}
			case LDLP:
			{
				register INSTRUCTION *next_instruction;
				
				next_instruction=instruction->next;
				
				switch (next_instruction->code)
				{
					case FPLDNLDB:
					case FPSTNLDB:
					{
						INT16 high_virtual_number, high_real_number;
						
						virtual_number		=instruction->parameters[0].fvirtual->low.local;
						high_virtual_number	=instruction->parameters[0].fvirtual->high.local;
						real_number			=virtual_memory[virtual_number].real_local;
						high_real_number	=virtual_memory[high_virtual_number].real_local;
						
						if (real_number==NO_LOCAL || high_real_number==NO_LOCAL)
							internal_error("adjust_instructions (2)");

						if (real_number+1 != high_real_number)
							internal_error("adjust_instructions: float not allocated in subsequent locals");

						instruction->parameters[0].int16=real_number;
						
						break;
					}
					case FPI32TOR64:
					case FPSTNLI32:
					{
						virtual_number=instruction->parameters[0].virtual->local;
						real_number=virtual_memory[virtual_number].real_local;
				
						if (real_number == NO_LOCAL)
							internal_error("adjust_instructions (3)");
					
						instruction->parameters[0].int16=real_number;

						break;
					}
					default:
						internal_error("adjust_instructions (4)");
				}
			}
			default:
				break;
		}
	}
}

/************************************************************************************
 *																					*
 *		computation of next uses													*
 *																					*
 ************************************************************************************/

static VOID use_local (PARAMETER *parameter, INT16 instruction_number)
{
	INT16			virtual_local_number;
	VIRTUAL_LOCAL	*virtual_local;
		
	/* remember number of local */
	
	virtual_local_number=parameter->int16;
	virtual_local=&virtual_memory[virtual_local_number];
	
	/* replace local by virtual local structure */
	
	parameter->virtual=alloc_block_mem(sizeof(VIRTUAL_PARAMETER));
	
	/* copy virtual local field into parameter */
	
	parameter->virtual->local			=virtual_local_number;
	parameter->virtual->next_use		=virtual_local->next_use;
	parameter->virtual->used			=virtual_local->used;
	
	/* adjust virtual local fields */
	
	virtual_local->next_use			=instruction_number;
	virtual_local->used				=true;
}

static VOID use_flocal (PARAMETER *parameter, INT16 instruction_number)
{
	INT16			virtual_local_number;
	VIRTUAL_LOCAL	*virtual_local_low, *virtual_local_high;
	
	/* remember number of local */
	
	virtual_local_number	=parameter->int16;
	virtual_local_low		=&virtual_memory[virtual_local_number];
	virtual_local_high		=virtual_local_low+1;
	
	/* replace local by virtual local structure */
	
	parameter->fvirtual=alloc_block_mem(sizeof(VIRTUAL_FPARAMETER));
	
	/* copy both virtual local fields into parameter */
		
	parameter->fvirtual->low.local		=virtual_local_number;
	parameter->fvirtual->low.next_use	=virtual_local_low->next_use;
	parameter->fvirtual->low.used		=virtual_local_low->used;

	parameter->fvirtual->high.local		=virtual_local_number+1;
	parameter->fvirtual->high.next_use	=virtual_local_high->next_use;
	parameter->fvirtual->high.used		=virtual_local_high->used;
	
	/* adjust virtual local fields. */
	
	virtual_local_low->next_use			=instruction_number;
	virtual_local_low->used				=true;
	virtual_local_high->next_use		=instruction_number;
	virtual_local_high->used			=true;
}

static VOID define_local (PARAMETER *parameter, INT16 instruction_number)
{
	INT16			virtual_local_number;
	VIRTUAL_LOCAL	*virtual_local;

	/* remember number of local */
	
	virtual_local_number=parameter->int16;
	virtual_local=&virtual_memory[virtual_local_number];
	
	/* replace local by virtual local structure */
	
	parameter->virtual=alloc_block_mem(sizeof(VIRTUAL_PARAMETER));
	
	/* copy virtual local field into parameter */
	
	parameter->virtual->local			=virtual_local_number;
	parameter->virtual->next_use		=virtual_local->next_use;
	parameter->virtual->used			=virtual_local->used;

	/*	if a parameter local is stored, the next use of the virtual local becomes
		this point: this avoids allocating this local easily just before
		this instruction. If a non-parameter local is stored, the previous value
		becomes unused. In either case the used flag becomes false;
	*/
	
	if (virtual_local_number < N_LOCALS)
		virtual_local->next_use=instruction_number;
	else
		virtual_local->next_use=UNUSED;
	
	virtual_local->used=false;
}

static VOID define_flocal (PARAMETER *parameter, INT16 instruction_number)
{
	INT16			virtual_local_number;
	VIRTUAL_LOCAL	*virtual_local_low, *virtual_local_high;
	
	/* remember number of local */
	
	virtual_local_number	=parameter->int16;
	virtual_local_low		=&virtual_memory[virtual_local_number];
	virtual_local_high		=virtual_local_low+1;
	
	/* replace local by virtual local structure */
	
	parameter->fvirtual=alloc_block_mem(sizeof(VIRTUAL_FPARAMETER));
	
	/* copy both virtual local fields into parameter */
		
	parameter->fvirtual->low.local		=virtual_local_number;
	parameter->fvirtual->low.next_use	=virtual_local_low->next_use;
	parameter->fvirtual->low.used		=virtual_local_low->used;

	parameter->fvirtual->high.local		=virtual_local_number+1;
	parameter->fvirtual->high.next_use	=virtual_local_high->next_use;
	parameter->fvirtual->high.used		=virtual_local_high->used;
	
	/*	adjust virtual local fields, see also define_local. Floating point
		values are only stored in the exact parameter locals if both
		words fit in these parameters. */
	
	if (virtual_local_number < N_LOCALS-1)
	{
		/* both words fit in parameter locals */
		
		virtual_local_low->next_use		=instruction_number;
		virtual_local_high->next_use	=instruction_number;
	} else
	{
		/* one or both words do not fit */
		
		virtual_local_low->next_use		=UNUSED;
		virtual_local_high->next_use	=UNUSED;
	}
	virtual_local_low->used		=false;
	virtual_local_high->used	=false;
}

static VOID compute_next_uses (INSTRUCTION_BLOCK *block)
{
	INT16 n;
	INT16 instruction_number;
	INSTRUCTION *instruction;
	
	/* for all real locals: test if they should be kept at the end */
	
	for (n=0; n<N_LOCALS; ++n)
	{
		if (end_locals is_bit_set bit(n))
		{
			virtual_memory[n].next_use			=NEXT_BLOCK;
			virtual_memory[n].used				=true;
		}
	}
	
	/* now compute next uses from back to front */
	
	instruction			=block->last_instruction;
	instruction_number	=LAST_INSTRUCTION;
	
	while (instruction != NULL)
	{
		switch (instruction->code)
		{
			case LDL:
				use_local(&instruction->parameters[0],instruction_number);
				break;
			case STL:
				define_local(&instruction->parameters[0],instruction_number);
				break;
			case LDLP:
			{
				/* only a few things possible */
				
				INSTRUCTION *next_instruction;
				
				next_instruction=instruction->next;
				
				if (next_instruction == NULL)
					internal_error("compute_next_uses: ldlp... (no next instruction)");
				
				switch (next_instruction->code)
				{
					case FPLDNLDB:
						use_flocal(&instruction->parameters[0],instruction_number);
						break;	
					case FPSTNLDB:
						define_flocal(&instruction->parameters[0],instruction_number);
						break;
					case FPI32TOR64:
						use_local(&instruction->parameters[0],instruction_number);
						break;
					case FPSTNLI32:
						define_local(&instruction->parameters[0],instruction_number);
						break;
					default:
						internal_error("compute_next_uses: ldlp... (wrong instruction)");
				}
				break;
			}
			default:
			
				/* other instructions simply do not affect locals */
				
				break;
		}
		instruction=instruction->prev;
		++instruction_number;
	}
}

/************************************************************************************
 *																					*
 *		intialisation and allocation												*
 *																					*
 ************************************************************************************/

static VOID allocate_virtual_locals(INT16 n_virtual_locals)
{
	INT16 n;
	VIRTUAL_LOCAL *local;

	if (n_virtual_locals <= N_LOCALS)
		internal_error("allocate_virtual_locals");
	
	virtual_memory=alloc_block_mem(n_virtual_locals * sizeof(VIRTUAL_LOCAL));
	
	for (n=0,local=virtual_memory; n<n_virtual_locals; ++n,++local)
	{
		/* make virtual local empty */
		
		local->next_use			=UNUSED;
		local->used				=false;
		local->number			=n;
		local->real_local		=NO_LOCAL;
		local->offset			=NOT_STORED;
		local->group_flags		=0;
	}
}

static VOID initialise_locals(INSTRUCTION_BLOCK *block, INT16 n_virtual_locals)
{
	INT16 n;

	allocate_virtual_locals	(n_virtual_locals);	
	compute_next_uses		(block);

	/* the real local are assumed to hold the first N_LOCALS virtual locals */
	
	for (n=0; n<N_LOCALS; ++n)
	{
		real_memory[n].number=n;
		real_memory[n].altered=false;
		
		if (virtual_memory[n].used)
		{
			/* value will be used lateron */

			real_memory[n].virtual_local	=n;
			virtual_memory[n].real_local	=n;
			
			real_memory[n].used				=true;
			real_memory[n].next_use			=virtual_memory[n].next_use;
			
		} else
		{
			/* value in local can be destroyed */
			
			real_memory[n].virtual_local	=NO_LOCAL;
			
			real_memory[n].used				=false;
			real_memory[n].next_use			=UNUSED;
		}
	}
}

/************************************************************************************
 *																					*
 *		set group info																*
 *																					*
 ************************************************************************************/

static VOID compute_group_locked_locals (VOID)
{
	register INT16			n;
	register REAL_LOCAL		*real_local;

	/*
		determine which locals are 'free', i.e. those that may be allocated to
		place some non-fixed virtual local in. Each free local obeys to the
		following rules:
		-	not local of needed fixed one
		-	not the a-stack pointer
		-	not needed virtual local that is already in memory
		
		All this is done by first locking all locals and next unlocking those
		that are free. (this is slighty more efficient than the other way around:
		now the free list is built directly). In the current implementation it is
		possible that some locals get locked here, while it eventually
		will get reallocated,...(possible when a fixed local is not at its
		correct position) In that case these locals should get unlocked after the
		moment of reallocation, which is not optimal, unfortunately.
		
		Note that it is quite complex to prevent infinite recursion when locals
		that have to be reallocated are unlocked at beforehand: suppose at some
		point one starts to move such a value to its correct position: this
		position may hold some (non-)fixed value that could be placed in the local
		we are moving, resulting in moving this value again... NOTE THAT THIS
		PROBLEM ALSO ARISES WHEN 2 MISPLACED FIXED LOCALS ARE AT EACH OTHERS
		POSITION (so 2 moves just won't do), we are just lucky that in the
		current, suboptimal implementation it is not possible to have 2
		or more fixed locals at the wrong place: a local can only be misplaced
		if it changes from non-fixed to fixed (otherwise it would have been allocated
		correcty before). This can only happen for the low word of the floating
		point that is stored over the boundary of real and non-real locals...
		(and even that is not possible really...)
	*/

	lock_locals();

	/* determine which real locals are free */
	
	for (n=0, real_local=&real_memory[0]; n < N_LOCALS; ++n,++real_local)
	{
		if (!(virtual_memory[n].group_flags & FIXED) && n != A_STACK_POINTER)
		{
			INT16 virtual_number;
			
			/*	this local is not fixed, nor the a-stack, it us free if:
				-	it is empty
				-	contains a value that is not needed
			*/
			
			virtual_number=real_local->virtual_local;
			
			if (virtual_number == NO_LOCAL)
				unlock_local(real_local);
			else if (virtual_memory[virtual_number].group_flags == 0)
				unlock_local(real_local);
		}
	}
}

static VOID compute_group_virtual_uses (INSTRUCTION *last_instruction, INT16 n_virtual_locals)
{
	register INT16			n;
	register INSTRUCTION	*i;

	/*	determine for each virtual local in this group the following
		- if it is used (before any definition)
		- if it is defined
		- if it should be placed in a fixed real local
		- if it is the low word of a float
		- if it is the high word of a float
		
		-	the next_... fields of the virtual local are set according
			to the last instruction in this group that uses this local,
			which should indicate the next use of this local after this
			group. Note that these fields remain the same if this
			group does nothing with the locals
		
		first all bits are cleared for all virtual locals
		
		Note that if a local is used after definition by this block
		the use flag must not be set: the local does not have to be loaded for
		THIS use of it. For this reason all instructions are traversed backward:
		a definition clears the use flag (set by the instructions after it).
	*/
	
	for (n=0; n<n_virtual_locals; ++n)
		virtual_memory[n].group_flags=0;
		
	for (i=last_instruction; i!=NULL && i->code != GROUP; i=i->prev)
	{
		VIRTUAL_LOCAL			*virtual_local;
		VIRTUAL_PARAMETER		*parameter;

		switch (i->code)
		{
			case LDL:
			{
				parameter		=i->parameters[0].virtual;
				n				=parameter->local;
				virtual_local	=&virtual_memory[n];
			
				if (n < N_LOCALS)
					virtual_local->group_flags|=(USE | FIXED);	
				else
					virtual_local->group_flags|=USE;
				
				virtual_local->next_use		=parameter->next_use;
				virtual_local->used			=parameter->used;

				break;
			}
			case STL:
			{
				parameter		=i->parameters[0].virtual;
				n				=parameter->local;
				virtual_local	=&virtual_memory[n];

				/* set def flag and clear use flag */
				
				if (n < N_LOCALS)
					virtual_local->group_flags|=(DEF | FIXED);
				else
					virtual_local->group_flags|=DEF;
					
				virtual_local->group_flags	&=~USE;	
				virtual_local->next_use		=parameter->next_use;
				virtual_local->used			=parameter->used;

				break;
			}
			case LDLP:
			{
				VIRTUAL_FPARAMETER	*fparameter;
				VIRTUAL_LOCAL		*next_virtual_local;
				
				switch (i->next->code)
				{
					case FPLDNLDB:
					{
						fparameter			=i->parameters[0].fvirtual;
						n					=fparameter->low.local;
						virtual_local		=&virtual_memory[n];
						next_virtual_local	=virtual_local+1;
				
						if (n+1 != fparameter->high.local)
							internal_error("compute_group_use (1)");

						if (n < N_LOCALS-1)
						{	virtual_local->group_flags			|=(USE | LOW | FIXED);
							next_virtual_local->group_flags		|=(USE | HIGH | FIXED);
						} else
						{	virtual_local->group_flags			|=(USE | LOW);
							next_virtual_local->group_flags		|=(USE | HIGH);
						}
						virtual_local->next_use			=fparameter->low.next_use;
						virtual_local->used				=fparameter->low.used;
						next_virtual_local->next_use	=fparameter->high.next_use;
						next_virtual_local->used		=fparameter->high.used;
						
						break;
					}
					case FPSTNLDB:
					{
						fparameter			=i->parameters[0].fvirtual;
						n					=fparameter->low.local;
						virtual_local		=&virtual_memory[n];
						next_virtual_local	=virtual_local+1;
				
						if (n+1 != fparameter->high.local)
							internal_error("compute_group_use (2)");

						/* set def flag and clear use flag */
						
						if (n < N_LOCALS-1)
						{	virtual_local->group_flags			|=(DEF | LOW | FIXED);
							next_virtual_local->group_flags		|=(DEF | HIGH | FIXED);
						} else
						{	virtual_local->group_flags			|=(DEF | LOW);
							next_virtual_local->group_flags		|=(DEF | HIGH);
						}
						virtual_local->group_flags		&=~USE;
						virtual_local->next_use			=fparameter->low.next_use;
						virtual_local->used				=fparameter->low.used;
						next_virtual_local->group_flags	&=~USE;
						next_virtual_local->next_use	=fparameter->high.next_use;
						next_virtual_local->used		=fparameter->high.used;

						break;
					}
					case FPI32TOR64:
					{
						parameter		=i->parameters[0].virtual;
						n				=parameter->local;
						virtual_local	=&virtual_memory[n];
						
						if (n < N_LOCALS)
							virtual_local->group_flags|=(USE | FIXED);
						else
							virtual_local->group_flags|=USE;
						
						virtual_local->next_use		=parameter->next_use;
						virtual_local->used			=parameter->used;

						break;
					}
					case FPSTNLI32:
					{
						parameter		=i->parameters[0].virtual;
						n				=parameter->local;
						virtual_local	=&virtual_memory[n];
						
						/* set def flag and clear use flag */
						
						if (n < N_LOCALS)
							virtual_local->group_flags|=(DEF | FIXED);
						else
							virtual_local->group_flags|=DEF;
						
						virtual_local->group_flags	&=~USE;
						virtual_local->next_use		=parameter->next_use;
						virtual_local->used			=parameter->used;

						break;
					}
					default:
						internal_error("compute_group_use: ldlp... (wrong instruction)");
				}
				break;
			}
			default:
			
				/* other instructions simply do not affect locals */
				
				break;
		}
	}	
}

static VOID compute_group_real_next_uses (VOID)
{
	register INT16			n;
	register REAL_LOCAL		*real_local;
	
	for (n=0, real_local=&real_memory[0]; n<N_LOCALS; ++n,++real_local)
	{
		if (real_local->virtual_local != NO_LOCAL)
		{
			VIRTUAL_LOCAL *virtual_local;
				
			/*	real next use of some real local is the maximum of the next_use
				of its virtual local and the next use of the corresponding parameter
				register: if the parameter local is in the real local this is trivial
				(both are the same). otherwise this indicates what will happen most
				early: use of the parameter local or use of the virtual local */
			
			virtual_local=&virtual_memory[real_local->virtual_local];
			
			real_local->next_use		=max(virtual_memory[real_local->number].next_use,virtual_local->next_use);
			real_local->used			=virtual_local->used;
		
			/* see if this group will alter this real local */
			
			if (virtual_local->group_flags & DEF)
				real_local->altered=true;
		}
	}
}

/************************************************************************************
 *																					*
 *		local allocation, searching	for free locals									*
 *																					*
 ************************************************************************************/

static REAL_LOCAL *find_real_local (VOID)
{
	INT16 min_next_use;
	register REAL_LOCAL *match, *local;
	
	/*	find some suited real local for a non-fixed local. it should obey the following:
		- not the a-stack pointer: this will be the pointer to the virtual locals
		- local not needed by some fixed local in this group
		- value of local is not needed in this block
		- minimal next use
		
		Note that the first 3 requirements are met if the local is unlocked
	*/

	/* find first candidate */
	
	local=first_unlocked_local();
	
	if (local != NULL)
	{
		match=local;
		min_next_use=match->next_use;

		local=next_unlocked_local(local);
		
		while (local != NULL)
		{
			if (local->next_use < min_next_use)
			{
				match=local;
				min_next_use=match->next_use;
			}
			local=next_unlocked_local(local);
		}
		return match;
	}
	
	internal_error("find_real_local: could not find free local");
	return NULL;
}

static REAL_LOCAL *find_real_float_local (VOID)
{
	INT16 min_next_use;
	register REAL_LOCAL *match, *local;
	
	/*	see find_real_local() This routine searches for 2 subsequent real locals that
		obey the same rules as in find_real_local(). The next_use of the total float
		local is the maximum (the worst) of the next_use of both words.
	*/

	/* find first candidate */

	local=first_unlocked_local();

	while (local != NULL && is_locked_local(local->number+1))
		local=next_unlocked_local(local);
	
	if (local != NULL)
	{
		/* 2 subsequent free locals found */
		
		match=local;
		min_next_use=max(local->next_use,(local+1)->next_use);
		
		local=next_unlocked_local(local);
		
		while (local != NULL)
		{
			if (is_unlocked_local(local->number+1))
			{
				INT16 new_next_use;
				
				/* found another two candidates */
				
				new_next_use=max(local->next_use,(local+1)->next_use);
				
				if (new_next_use < min_next_use)
				{	match=local;
					min_next_use=new_next_use;
				}
			}
			local=next_unlocked_local(local);
		}
		return match;
	}
	
	internal_error("find_real_local_float: could not find free float local");
	return NULL;
}

/************************************************************************************
 *																					*
 *		local allocation, basic routines											*
 *																					*
 ************************************************************************************/

static VOID allocate_local(VIRTUAL_LOCAL *local, INSTRUCTION *insertion_point);

static VOID reserve_real_local (register REAL_LOCAL *local, INSTRUCTION *insertion_point)
{
	/*	This routine should be called just before storing some value in the
		real local. It reserves the local by locking it (it may be locked already
		by the way) and saves any value that is in it. Note that locking must
		preceed saving (which may introduce still more saving), to avoid saving
		some other value in THIS local (again).
		
		If a value is in the local, the use flags of the coresponding virtual local
		will indicate the use in/after this group, which is used to determine
		whether the value must be saved or not.
	*/
	
	register VIRTUAL_LOCAL *virtual_local;
	
	lock_local(local);

	if (local->virtual_local == NO_LOCAL)
		return;									/* local already empty */
		
	/*	real local contains some value, if necessary, evacuate it to virtual
		local or to another real local */
	
	virtual_local=&virtual_memory[local->virtual_local];

	if (virtual_local->group_flags == 0)
	{
		/* local is not used in this block, write to memory */
		
		if (virtual_local->used)
		{
			/* value is used lateron */
			
			if (virtual_local->offset==NOT_STORED)
			{
				/* local has never been stored, so store it */

				virtual_local->offset=virtual_memory_use;				
				++virtual_memory_use;
				
				insert_store(local->number,virtual_local->offset,insertion_point);
				
			} else if (local->altered)
				insert_store(local->number,virtual_local->offset,insertion_point);
		}
		
		/*	indicate the virtual local is not stored in some real local */
		
		virtual_local->real_local	=NO_LOCAL;
		
		/* indicate the local is free */
	
		local->virtual_local		=NO_LOCAL;
		local->next_use				=UNUSED;

	 } else
	 {
	 	/*	the local is used in this block, but apparantly it must be moved
	 		to some other local. If the local is not fixed, this means probably
	 		that some fixed local is about to overwrite this real local. If it is
	 		fixed, it is not at its 'own' position. Other situations should
	 		not occur. It is sufficient to check these conditions (not really
	 		needed) and to call allocate_local for the virtual local.
	 	*/
	 	
	 	if (virtual_local->group_flags & FIXED)
	 	{
	 		if (virtual_local->number != virtual_local->real_local)
	 			allocate_local(virtual_local, insertion_point);
	 		else
	 			internal_error("reserve_real_local: overwriting fixed local at correct position");
	 	} else
	 	{
	 		if (virtual_memory[local->number].group_flags & FIXED)
	 			allocate_local(virtual_local, insertion_point);
	 		else
	 			internal_error("reserve_real_local: overwriting fixed local by non-fixed one");
	 	}
	 	
	 	/* check if real local is free now indeed */

	 	if (local->virtual_local != NO_LOCAL)
	 		internal_error("reserve_real_local: reservation unsuccessfull");
	}
}

static VOID load_virtual_local (REAL_LOCAL *real_local, VIRTUAL_LOCAL *virtual_local, INSTRUCTION *insertion_point)
{
	/*	get the virtual local from memory in the real local, which must be
		reserved first. the virtual local must be needed in this group */
	
	reserve_real_local(real_local,insertion_point);
	
	virtual_local->real_local	=real_local->number;

	real_local->virtual_local	=virtual_local->number;
	real_local->altered			=false;
	
	/*	new real local uses are computed later, insert instruction to load value
		from memory if value is used */
	
	if (virtual_local->group_flags & USE)
	{
		if (virtual_local->offset != NOT_STORED)
			insert_load(real_local->number,virtual_local->offset,insertion_point);
		else
			internal_error("load_virtual_local\n");
	}
	
	/* if local only defined, no loading necessary */
}

static VOID move_virtual_local (REAL_LOCAL *real_local, VIRTUAL_LOCAL *virtual_local, INSTRUCTION *insertion_point)
{
	REAL_LOCAL *old_real_local;
	
	/*	needed virtual local is supposed to be in the wrong real local, move it to
		the given one. If the old local is not needed by a fixed one it may be
		unlocked (this routine will certainly not be called to move the a-stack
		pointer so we don't have to worry about unlocking that local).
	*/
	
	if (virtual_local->real_local == NO_LOCAL)
		internal_error("move_virtual_local: the virtual local is not in a real one");
	
	old_real_local=&real_memory[virtual_local->real_local];
	
	reserve_real_local(real_local,insertion_point);
	
	virtual_local->real_local	=real_local->number;

	real_local->virtual_local	=virtual_local->number;
	real_local->altered			=old_real_local->altered;
		
	/* free old real register without evacuating its value */
	
	old_real_local->virtual_local	=NO_LOCAL;
	old_real_local->next_use		=UNUSED;

	if (!(virtual_memory[old_real_local->number].group_flags & FIXED))
		unlock_local(old_real_local);

	if (virtual_local->group_flags & USE)
		insert_move(real_local->number,old_real_local->number,insertion_point);
		
	/*	if local only defined by this group, no moving is necessary, altered
		flag will ensure it is saved lateron if needed */
}


/********************************************************************************
 *																				*
 *		local allocation: main allocation routines								*
 *																				*
 ********************************************************************************/

static VOID allocate_non_fixed_float (VIRTUAL_LOCAL *low_virtual,INSTRUCTION *insertion_point)
{
	/*	this routine should only be called for a NON-FIXED virtual low word of
		a float */
		
	VIRTUAL_LOCAL *high_virtual;
	INT16 low_number, high_number;
	REAL_LOCAL *low_real, *high_real;
	
	high_virtual=low_virtual+1;
		
	/* see where both words are */
	
	low_number	=low_virtual->real_local;
	high_number	=high_virtual->real_local;
	
	if (low_number == NO_LOCAL)
	{
		if (high_number == NO_LOCAL)
		{
			/* both locals not allocated, find 2 subsequent locals */
		
			low_real	=find_real_float_local();
			high_real	=low_real+1;
			load_virtual_local(low_real,low_virtual,insertion_point);
			load_virtual_local(high_real,high_virtual,insertion_point);
			
		} else
		{
			/* high is allocated, low not */
			
			if (!(virtual_memory[high_number].group_flags & FIXED) && is_unlocked_local(high_number+1))
			{
				/*	high local does not have to be moved and local before it is free,
					so only the low local has to be loaded */
				
				low_real=&real_memory[high_number-1];
				load_virtual_local(low_real,low_virtual,insertion_point);
							
			} else
			{
				/*	high local must be moved and low local must be loaded in newly
					found float local */
				
				low_real	=find_real_float_local();
				high_real	=low_real+1;
				load_virtual_local(low_real,low_virtual,insertion_point);
				move_virtual_local(high_real,high_virtual,insertion_point);
							
			}
		}
	} else
	{
		if (high_number == NO_LOCAL)
		{
			/* low allocated, high not */
			
			if (!(virtual_memory[low_number].group_flags & FIXED) && is_unlocked_local(low_number+1))
			{
				/*	low local does not have to be moved and local after it is free,
					so only the high local has to be loaded */
				
				high_real=&real_memory[low_number+1];
				load_virtual_local(high_real,high_virtual,insertion_point);
				
			} else
			{
				/*	low local must be moved and high local must be loaded in newly
					found float local */

				low_real	=find_real_float_local();
				high_real	=low_real+1;
				move_virtual_local(low_real,low_virtual,insertion_point);
				load_virtual_local(high_real,high_virtual,insertion_point);
			}
		} else
		{
			/* both locals are allocated */
			
			if (low_number+1 == high_number)
			{
				/*	locals are at subsequent positions already, move if necessary,
					else leave values where they are */
				
				if (virtual_memory[low_number].group_flags & FIXED ||
					virtual_memory[high_number].group_flags & FIXED)
				{
					/* the float must be moved */
					
					low_real=find_real_float_local();
					high_real=low_real+1;
					
					move_virtual_local(low_real,low_virtual,insertion_point);
					move_virtual_local(high_real,high_virtual,insertion_point);
				}
				
			} else
			{
				/* both allocated, but not at subsequent positions */
				
				if (virtual_memory[low_number].group_flags & FIXED)
				{
					if (!(virtual_memory[high_number].group_flags & FIXED) && is_unlocked_local(high_number-1))
					{
						/*	low local must be moved, but high word not and local before
							it is free */
				
						low_real=&real_memory[high_number-1];
						move_virtual_local(low_real,low_virtual,insertion_point);
						
					} else
					{
						/* both locals must be moved */
						
						low_real	=find_real_float_local();
						high_real	=low_real+1;
						
						move_virtual_local(low_real,low_virtual,insertion_point);
						move_virtual_local(high_real,high_virtual,insertion_point);
					
					}
				} else
				{
					/*	low local may stay where it is, first try local after it,
						regardless of status of high local */
					
					if (is_unlocked_local(low_number+1))
					{
						/* move high word after low word */
						
						high_real=&real_memory[low_number+1];
						move_virtual_local(high_real,high_virtual,insertion_point);
						
					} else
					{
						/*	local after low word not available */
						
						if (!(virtual_memory[high_number].group_flags & FIXED) && is_unlocked_local(high_number-1))
						{
							/* local before high word available, move low word to it */
					
							low_real=&real_memory[high_number-1];
							move_virtual_local(low_real,low_virtual,insertion_point);
						
						} else
						{
							/* both words must be moved to new position */
							
							low_real	=find_real_float_local();
							high_real	=low_real+1;
						
							move_virtual_local(low_real,low_virtual,insertion_point);
							move_virtual_local(high_real,high_virtual,insertion_point);
						}
					}
				}
			}
		}
	}
}

static VOID allocate_local (VIRTUAL_LOCAL *virtual_local, INSTRUCTION *insertion_point)
{
	/* allocate some virtual local */
	
	if (virtual_local->group_flags == 0)
		return;											/* not needed, do nothing */

	if (virtual_local->group_flags & FIXED)
	{
		INT16 real_number, new_real_number;
		REAL_LOCAL *new_real_local;

		/*	this is a needed fixed local,
			allocate local at same number as virtual local
			fixed floating point locals are allocated at consecutive places
			automatically (as they reside at consecutive places in virtual memory).
		*/
		
		real_number		=virtual_local->real_local;
		new_real_number	=virtual_local->number;
		
		/* if local already at correct place, return */
		
		if (real_number == new_real_number)
			return;
	
		/*	local not at correct place, so move it or load it. */
		
		new_real_local=&real_memory[new_real_number];
		
		if (real_number == NO_LOCAL)
			load_virtual_local(new_real_local,virtual_local,insertion_point);
		else
			move_virtual_local(new_real_local,virtual_local,insertion_point);
		
	} else
	{
		INT16 real_number;
	
		/*	First see if it is part of a float. If so, call a special routine that
			ensures both words of the float are allocated in subsequent
			real locals
		*/

		if (virtual_local->group_flags & LOW)
		{	allocate_non_fixed_float(virtual_local,insertion_point);
			return;
		}
		
		if (virtual_local->group_flags & HIGH)
		{	allocate_non_fixed_float(virtual_local-1,insertion_point);
			return;
		}

	 	/*	This is a single needed non-fixed local. If is not yet allocated,
	 		find a local and load it. If it is allocated at the place of a
	 		needed fixed local it must be moved. */
	 	
		real_number=virtual_local->real_local;
	
		if (real_number == NO_LOCAL)
			load_virtual_local(find_real_local(), virtual_local,insertion_point);
		else if (virtual_memory[real_number].group_flags & FIXED)
			move_virtual_local(find_real_local(), virtual_local, insertion_point);
	}
}

static VOID allocate_locals (INSTRUCTION_BLOCK *block,INT16 n_virtual_locals)
{
	/*	Allocate locals per group. Walk all instructions from front to back.
		Each group should start with a group instruction. The last group
		instruction also denotes the end of all groups for which scheduling
		is necessary. */
	
	register INT16 n;
	INSTRUCTION_GROUP *group, *next_group;
	INSTRUCTION *first_instruction, *last_instruction;
	VIRTUAL_LOCAL *local;

	group=(INSTRUCTION_GROUP *)block->instructions;
	
	if (group == NULL)
		internal_error("allocate_locals: block does not start with group");
	
	next_group=group->next_group;
	
	while(next_group != NULL)
	{
		if (group->free < 3)
			internal_error("allocate_locals: cannot handle group with < 3 free regs");
			
		first_instruction	=group->next;
		last_instruction	=next_group->prev;
		
		if (first_instruction == NULL || last_instruction == NULL)
			internal_error("allocate_locals: group has no instructions");

		compute_group_virtual_uses(last_instruction,n_virtual_locals);		
		compute_group_locked_locals();
		
		for (n=0, local=virtual_memory; n<n_virtual_locals; ++n,++local)
			allocate_local(local,first_instruction);

		adjust_group_instructions(first_instruction);
		compute_group_real_next_uses();
		
		group=next_group;
		next_group=next_group->next_group;
	}
	
	/*	Last group reached, insert instructions to get parameters at the right
		position. Note that parameters may reside in:
		-	their correct local
		-	in memory
		
		They cannot reside in an incorrect local: parameters are always fixed
		(floating point parameters as well) and this means that if they have
		been stored in locals earlier, they will be at their correct positions
		already.
		
		Note that the last group should either have no instructions, or some
		special instructions */
	
	
	if (group->free < 2)
		internal_error("allocate_locals: cannot handle last group with < 2 free regs");
	
	first_instruction=group->next;
	
	for (n=0,local=virtual_memory; n<N_LOCALS; ++n,++local)
	{
		if (end_locals is_bit_set bit(n))
		{
			INT16 place;
			
			place=local->real_local;
			
			if (place== NO_LOCAL)
			{
				if (first_instruction == NULL)
					append_load(n,local->offset,block);			
				else
					insert_load(n,local->offset,first_instruction);
			}
			else if (place != n)
				internal_error("parameter local at incorrect end position");
		}
	}	
}

/************************************************************************************
 *																					*
 *		exported routines															*
 *																					*
 ************************************************************************************/

VOID reset_end_locals (VOID)
{
	end_locals=
			(bit (RETURN_ADDRESS))
		|	(bit (A_STACK_POINTER))
		|	(bit (B_STACK_POINTER))
		|	(bit (HEAP_START))
		|	(bit (HEAP_END));
}

VOID set_end_local (INT16 n)
{
	end_locals set_bit (bit(n));
}

VOID schedule_locals(INSTRUCTION_BLOCK *block,INT16 n_virtual_locals)
{
	/*	this routine should only be called when scheduling is necessary, that
		is, when the number of virtual locals used is higher than the number
		of real locals (otherwise the real registers can be used). Note that
		the increase of the a-stack in this block should be known and that
		combining of stack checks must happen afterwards.
	*/

	virtual_memory_use=0;
	
	if (block->a_increase < 0)
		virtual_memory_offset=-block->a_increase;
	else
		virtual_memory_offset=block->a_increase;

	initialise_locals(block,n_virtual_locals);
	allocate_locals(block,n_virtual_locals);
	
	/*	increase check size of block */
	
	block->check_size+=virtual_memory_use;	
}