#include <string.h>

#include "ifinterface.h"
#include "cgmemory.h"

/********************************************************************************
 *																				*
 *		memory management														*
 *		for the Mac memory is aligned on 4-byte boundaries						*
 *		for the Sun memory is aligned on 8-byte boundaries						*
 *																				*
 ********************************************************************************/

#define MEM_BLOCK_SIZE 16380

typedef struct MEM_BLOCK
{
	struct MEM_BLOCK	*next;
	BYTE				block [MEM_BLOCK_SIZE];

} MEM_BLOCK;


static MEM_BLOCK	*heap;
static INT16		heap_size;

static MEM_BLOCK	*graph_heap;
static INT16		graph_heap_size;

static MEM_BLOCK	*block_heap;
static INT16		block_heap_size;

static MEM_BLOCK	*macro_heap;
static INT16		macro_heap_size;

static VOID *alloc_mem (U_INT16 size)
{
	POINTER p;

#ifdef MACINTOSH
	if ((p=(POINTER)NewPtr(size))!=NULL)
#else		
	if ((p=malloc(size))!=NULL)
#endif
		return p;
	
	fatal_error("out of memory");
	
	return NULL;
}

static VOID free_mem (POINTER p)
{
#ifdef MACINTOSH
	DisposPtr(p);
#else
	free(p);
#endif
}

VOID *alloc_heap_mem(U_INT16 size)
{
	/* allocates persistent heap */

	word_align(size);

	if (heap_size+size<=MEM_BLOCK_SIZE)
	{	POINTER p;
		
		p=&heap->block[heap_size];
		heap_size+=size;
		
		return p;

	} else
	{
		register MEM_BLOCK *new_block;
		
		new_block=alloc_mem(sizeof (MEM_BLOCK));
		new_block->next=heap;
		heap=new_block;
		heap_size=size;
		
		return new_block->block;
	}
}

POINTER alloc_graph_mem(U_INT16 size)
{
	/* allocates heap related to graphs (freed after a sequence of basic blocks has
	   been linearised) */

	word_align(size);
	
	if (graph_heap_size+size<=MEM_BLOCK_SIZE)
	{	POINTER p;
		
		p=&graph_heap->block[graph_heap_size];
		graph_heap_size+=size;
		return p;

	} else
	{
		register MEM_BLOCK *new_block;
		
		new_block=alloc_mem(sizeof (MEM_BLOCK));
		new_block->next=graph_heap;
		graph_heap=new_block;
		graph_heap_size=size;
		return new_block->block;
	}
}

POINTER alloc_block_mem(U_INT16 size)
{
	/* allocates heap that is freed after each block has been linearised */
	
	word_align(size);
	
	if (block_heap_size+size<=MEM_BLOCK_SIZE)
	{	POINTER p;
		
		p=&block_heap->block[block_heap_size];
		block_heap_size+=size;
		return p;

	} else
	{
		register MEM_BLOCK *new_block;
		
		new_block=alloc_mem(sizeof (MEM_BLOCK));
		new_block->next=block_heap;
		block_heap=new_block;
		block_heap_size=size;
		return new_block->block;
	}
}

POINTER alloc_macro_mem(U_INT16 size)
{
	/* allocates heap that is freed after each completed macro definition */
	
	word_align(size);
	
	if (macro_heap_size+size<=MEM_BLOCK_SIZE)
	{	POINTER p;
		
		p=&macro_heap->block[macro_heap_size];
		macro_heap_size+=size;
		return p;

	} else
	{
		register MEM_BLOCK *new_block;
		
		new_block=alloc_mem(sizeof (MEM_BLOCK));
		new_block->next=macro_heap;
		macro_heap=new_block;
		macro_heap_size=size;
		return new_block->block;
	}
}

VOID free_block_mem(VOID)
{	
	while(block_heap->next!=NULL)
	{
		MEM_BLOCK *next_block;
				
		next_block=block_heap->next;
		free_mem(block_heap);
		block_heap=next_block;
	}
		
	block_heap_size=0;
}

VOID free_macro_mem(VOID)
{		
	while(macro_heap->next!=NULL)
	{
		MEM_BLOCK *next_block;
				
		next_block=macro_heap->next;
		free_mem(macro_heap);
		macro_heap=next_block;
	}
		
	macro_heap_size=0;
}

VOID free_graph_mem(VOID)
{
	while(graph_heap->next!=NULL)
	{
		MEM_BLOCK *next_block;
				
		next_block=graph_heap->next;
		free_mem(graph_heap);
		graph_heap=next_block;
	}
		
	graph_heap_size=0;
}

static VOID free_heap (VOID)
{
	free_graph_mem();
	
	while(heap->next!=NULL)
	{
		MEM_BLOCK *next_block;
				
		next_block=heap->next;
		free_mem(heap);
		heap=next_block;
	}
		
	heap_size=0;
}

CHAR *new_chars_from_chars (CHAR *s)
{
	CHAR *p;
	
	p=alloc_heap_mem(strlen(s)+1);
	return strcpy(p,s);
}

CHAR *new_chars_from_string (STRING *s)
{
	CHAR *p;
	
	p=alloc_heap_mem(s->length+1);
	memcpy(p,s->string,s->length+1);
	
	return p;
}

STRING *new_string_from_string (STRING *s)
{
	STRING *p;
	
	p=alloc_heap_mem (sizeof(STRING));
	
	p->string=alloc_heap_mem(s->length+1);
	p->length=s->length;
	
	memcpy(p->string,s->string,s->length+1);
	
	return p;
}

VOID put_new_string_from_string (STRING *source, STRING *dest)
{
	dest->string=alloc_heap_mem(source->length+1);
	dest->length=source->length;
	
	memcpy(dest->string,source->string,source->length+1);
}

CHAR *alloc_macro_string (STRING *s)
{
	CHAR *p;
	
	p=alloc_macro_mem(s->length+1);
	memcpy(p,s->string,s->length+1);
	
	return p;
}

VOID alloc_code_memory(VOID)
{
	nr_errors = 0;
	nr_warnings = 0;

	heap_size=0;
	heap=alloc_mem(sizeof (MEM_BLOCK));
	heap->next=NULL;
	
	graph_heap_size=0;
	graph_heap=alloc_mem(sizeof (MEM_BLOCK));
	graph_heap->next=NULL;
	
	block_heap_size=0;
	block_heap=alloc_mem(sizeof (MEM_BLOCK));
	block_heap->next=NULL;
	
	macro_heap_size=0;
	macro_heap=alloc_mem(sizeof (MEM_BLOCK));
	macro_heap->next=NULL;
}

VOID free_code_memory()
{
	free_heap();
	free_macro_mem();
	
	free_mem(heap);
	free_mem(graph_heap);
	free_mem(block_heap);
	free_mem(macro_heap);
}