#include "ifinterface.h"

#include "cgopt.h"
#include "cgmemory.h"
#include "cglabels.h"

#include "ccintermediate.h"

/****************************************************************
 *																*
 *	inserting / deleting instructions							*
 *																*
 ****************************************************************/
 
static INSTRUCTION *insert_before(	INSTRUCTION_BLOCK *current_block,
									INSTRUCTION *next_instruction,
									INT16 instruction_code,INT16 arg_size)

/* next_instruction should point to a valid instruction */

{
	INSTRUCTION *instruction,*prev_instruction;
	
	instruction=alloc_heap_mem(sizeof (INSTRUCTION)+arg_size);
	instruction->code=instruction_code;

	prev_instruction = next_instruction->prev;
	if (prev_instruction==NULL)
		current_block->instructions=instruction;
	else
		prev_instruction->next=instruction;

	next_instruction->prev=instruction;
	
	instruction->next=next_instruction;
	instruction->prev=prev_instruction;
	
	return instruction;
}

static VOID delete_before(	INSTRUCTION_BLOCK *current_block,
							INSTRUCTION *next_instruction)

/* next_instruction should point to a valid instruction */

{
	INSTRUCTION *instruction,*prev_instruction;
	
	instruction = next_instruction->prev;
	if (instruction==NULL)
		internal_error("delete_before: no instruction before current one");
	else
	{	prev_instruction = instruction->prev;
		if (prev_instruction == NULL)
		{	current_block->instructions = next_instruction;
			next_instruction->prev = NULL;
		} else
		{	prev_instruction->next = next_instruction;
			next_instruction->prev = prev_instruction;
		}
	}
}

#if _COMPILE_UNUSED_

static INSTRUCTION *insert_after (	INSTRUCTION_BLOCK *current_block,
									INSTRUCTION *prev_instruction,
									INT16 instruction_code,INT16 arg_size)
{
	INSTRUCTION *instruction,*next_instruction;
	
	instruction=alloc_heap_mem(sizeof (INSTRUCTION)+arg_size);
	instruction->code=instruction_code;

	if (prev_instruction==NULL)
	{	next_instruction=current_block->instructions;
		current_block->instructions=instruction;
	} else 
	{	next_instruction=prev_instruction->next;
		prev_instruction->next=instruction;
	}
	if (next_instruction!=NULL)
		next_instruction->prev=instruction;
		
	instruction->next=next_instruction;
	instruction->prev=prev_instruction;

	return instruction;
}

#endif

static VOID instruction_integer_before(	INSTRUCTION_BLOCK *block,
										INSTRUCTION *next_instr,
										INT16 instruction_code,
										INT32 constant)
{
	INSTRUCTION *instruction;
	
	instruction = insert_before(block,next_instr,instruction_code,sizeof (PARAMETER));
	
	instruction->parameters[0].int32 = constant;
}

#if _COMPILE_UNUSED_

static VOID instruction_integer_after(	INSTRUCTION_BLOCK *block,
										INSTRUCTION *prev_instr,
										INT16 instruction_code,
										INT32 constant)
{
	INSTRUCTION *instruction;
	
	instruction = insert_after(block, prev_instr,instruction_code,sizeof (PARAMETER));
	
	instruction->parameters[0].int32 = constant;
}

#endif

/****************************************************************
 *																*
 *	optimisation of (conditional) jumps							*
 *																*
 ****************************************************************/
 
static VOID invert_conditional_jump (	INSTRUCTION *cond_jump,
										INSTRUCTION_BLOCK *cond_jump_block,
										INSTRUCTION *jump,
										INSTRUCTION_BLOCK *jump_block)
{
	INSTRUCTION *prev_instruction;
	
	prev_instruction = cond_jump->prev;
	
	if (prev_instruction != NULL &&
		prev_instruction->code == EQC &&
		prev_instruction->parameters[0].int32 == 0)
		delete_before(cond_jump_block, cond_jump);
	else
		instruction_integer_before(cond_jump_block, cond_jump, EQC, (INT32)0);
	
	cond_jump->parameters[0].label=jump->parameters[0].label;
	
	jump_block->instructions=NULL;
}

INSTRUCTION_BLOCK *do_jump_optimisation(INSTRUCTION_BLOCK *first_block)
{
	INSTRUCTION_BLOCK *block;
	
	for (block=first_block; block!=NULL; block=block->next)
	{	INSTRUCTION *last_instruction;
		
		handle_compute_event();
				
		if ((last_instruction=block->last_instruction)!=NULL)
		{	switch (last_instruction->code)
			{	case CJUMP:
				{	
					/* last instruction is conditional jump */
				
					INSTRUCTION_BLOCK *next_block, *next_next_block;
					INSTRUCTION *prev_instruction, *prev_prev_instruction, *jump;
					
					if ((next_block=block->next)			!=NULL &&
						next_block->start_label				==NULL &&
						next_block->a_increase 				== 0 &&
						next_block->b_increase 				== 0 &&
						next_block->data 					== NULL &&
						(jump=next_block->instructions)		!=NULL &&
						jump->code							==JUMP &&
						jump->next							==NULL &&
						(next_next_block=next_block->next)	!=NULL)
					{
						LABEL *cond_jump_label;
						
						/*	the next block merely contains a jump somewhere. If
							the conditional jump jumps to the block directly
							following the jump, the test will be inverted */
						
						cond_jump_label=last_instruction->parameters[0].label;
						
						if (cond_jump_label->flags & DEF_LABEL &&
							next_next_block->start_label==cond_jump_label)
						{
							invert_conditional_jump (last_instruction,block,jump,next_block);										
							cond_jump_label->flags|=REACHABLE;
						}
					}
					
					/*	last instruction is conditional jump, delete double
						occurences of eqc 0 */
					
					if ((prev_instruction=last_instruction->prev)!=NULL)
					{	if((prev_prev_instruction=prev_instruction->prev)!=NULL)
						{	if (prev_instruction->code==EQC &&
								prev_prev_instruction->code==EQC &&
								prev_instruction->parameters[0].int32==0 &&
								prev_prev_instruction->parameters[0].int32==0)
							{	delete_before(block, last_instruction);
								delete_before(block, last_instruction);
							}
						}
					}
					break;
				}
				case JUMP:
				{
					INSTRUCTION_BLOCK *next_block;
					
					if (block->data==NULL &&
						(next_block=block->next)!=NULL)
					{
						/*	no data has to be inserted after the jump, and there
							exists a next block */
						
						if(next_block->start_label==last_instruction->parameters[0].label)
						{
							/* remove jump (label becomes directly reachable) */
							
							next_block->start_label->flags|=REACHABLE;
							
							block->last_instruction=last_instruction->prev;
							if (block->last_instruction!=NULL)
								block->last_instruction->next=NULL;
							else
								block->instructions=NULL;
						}
					}
					break;
				}
			}
		}
	}
	return first_block;
}

