#include "ifinterface.h"

#include "cgoutputs.h"
#include "cgwrites.h"
#include "cgmemory.h"
#include "cglabels.h"

#include "ccintermediate.h"

INT32 inline_instruction_number;

/************************************************************************************
 *																					*
 *		creation of temporary nodes for evaluation of macros						*
 *																					*
 ************************************************************************************/

static TINSTRUCTION_NODE *new_integer(INT16 kind, INT32 number)
{
	TINSTRUCTION_NODE *new_node;
	
	new_node=alloc_macro_mem(sizeof(TINSTRUCTION_NODE)+sizeof(TINSTRUCTION_PARAMETER));
	new_node->kind=kind;
	new_node->parameter[0].number=number;
	
	return new_node;
}

static TINSTRUCTION_NODE *new_monadic(INT16 kind, TINSTRUCTION_NODE *arg)
{
	TINSTRUCTION_NODE *new_node;
	
	new_node=alloc_macro_mem(sizeof(TINSTRUCTION_NODE)+sizeof(TINSTRUCTION_PARAMETER));
	new_node->kind=kind;
	new_node->parameter[0].pointer=arg;
	
	return new_node;
}

static TINSTRUCTION_NODE *new_dyadic(INT16 kind, TINSTRUCTION_NODE *arg1, TINSTRUCTION_NODE *arg2)
{
	TINSTRUCTION_NODE *new_node;
	
	new_node=alloc_macro_mem(sizeof(TINSTRUCTION_NODE)+2*sizeof(TINSTRUCTION_PARAMETER));
	new_node->kind=kind;
	new_node->parameter[0].pointer=arg1;
	new_node->parameter[1].pointer=arg2;
	
	return new_node;
}

/************************************************************************************
 *																					*
 *		simple evaluation															*
 *																					*
 ************************************************************************************/

static INT32 compute_integer_operation(INT16 op, INT32 arg1, INT32 arg2)
{
	switch (op)
	{
		case AADD:
			return arg1+arg2;
		case ASUB:
			return arg1-arg2;
		case AMUL:
			return arg1*arg2;
		case ADIV:
			return arg1/arg2;
		case AMOD:
			return arg1%arg2;
		
		default:
			internal_error("compute_integer_operation");
			return 0;
	}
}

/********************************************************************************
 *																				*
 *		simplification of expressions											*
 *																				*
 ********************************************************************************/

static TINSTRUCTION_NODE *macro_combine_comm_any (TINSTRUCTION_NODE *a_op_b,INT16 op2,TINSTRUCTION_NODE *c)
{
	/*	a_op_b = 'a op b', the result of this function should be 'a op b op2 c' in
		a simplified way. c is a constant, and a and b are not constant both, op and
		op2 have the same priority, and 'op' should be commutative:
		a op b op2 c -> a op (b op2 c),
		a op b op2 c -> b op a op2 c -> b op (a op2 c)
	*/

	INT16 op;
	TINSTRUCTION_NODE *a, *b;
	
	op	=a_op_b->kind;
	a	=a_op_b->parameter[0].pointer;
	b	=a_op_b->parameter[1].pointer;	
	
	if (a->kind == AINTEGER)
		return new_dyadic(op,b,new_integer(AINTEGER,compute_integer_operation(op2,a->parameter[0].number,c->parameter[0].number)));
	
	if (b->kind == AINTEGER)
		return new_dyadic(op,a,new_integer(AINTEGER,compute_integer_operation(op2,b->parameter[0].number,c->parameter[0].number)));
	
	/* no useful combination, just return a_op_b op2 c */
	
	return new_dyadic(op2,a_op_b,c);
}

/************************************************************************************
 *																					*
 *		evaluation of expressions													*
 *																					*
 ************************************************************************************/

static TINSTRUCTION_NODE *evaluate_integer_expression(TINSTRUCTION_NODE *e)
{
	/* this is needed when expressions have been defined after their application.
	   Note that this routine cannot be used for macros as it would alter the
	   macro definition.
	*/
	
	register TINSTRUCTION_NODE *e1,*e2;
	
	switch(e->kind)
	{	
		case ABRACKET:
		{
			e1=evaluate_integer_expression(e->parameter[0].pointer);
			
			switch (e1->kind)
			{
				case AINTEGER:			/* integer */
				case ALOCAL:			/* local macro label */
				case ALABEL:			/* global label */
				case AJLABEL:
				case AGLABEL:
				case ADLABEL:
				case AUNDEFINED:
					return e1;
								
				default:
					e->parameter[0].pointer=e1;
					return e;
			}
		}
		case AMONADD:
		{
			e1=evaluate_integer_expression(e->parameter[0].pointer);
			
			if (e1->kind == AINTEGER || e1->kind == AUNDEFINED)
				return e1;
				
			e->parameter[0].pointer=e1;
			return e;
		}
		case AMONSUB:
		{
			e1=evaluate_integer_expression(e->parameter[0].pointer);
			
			if (e1->kind == AUNDEFINED)
				return e1;
			
			if (e1->kind == AINTEGER)
			{	
				e->kind=AINTEGER;
				e->parameter[0].number = -e1->parameter[0].number;
				return e;
			}
			e->parameter[0].pointer=e1;
			return e;
		}
		case AMUL:
		{
			e1=evaluate_integer_expression(e->parameter[0].pointer);
			if (e1->kind == AUNDEFINED)
				return e1;
			e2=evaluate_integer_expression(e->parameter[1].pointer);
			if (e2->kind == AUNDEFINED)
				return e2;
				
			if (e1->kind == AINTEGER && e2->kind == AINTEGER)
			{
				e->kind=AINTEGER;
				e->parameter[0].number=e1->parameter[0].number*e2->parameter[0].number;
				return e;
			}
			e->parameter[0].pointer=e1;
			e->parameter[1].pointer=e2;
			return e;
		}
		case ADIV:
		{
			e1=evaluate_integer_expression(e->parameter[0].pointer);
			if (e1->kind == AUNDEFINED)
				return e1;
			e2=evaluate_integer_expression(e->parameter[1].pointer);
			if (e2->kind == AUNDEFINED)
				return e2;
			
			if (e1->kind == AINTEGER && e2->kind == AINTEGER)
			{
				e->kind=AINTEGER;
				e->parameter[0].number=e1->parameter[0].number/e2->parameter[0].number;
				return e;
			}
			e->parameter[0].pointer=e1;
			e->parameter[1].pointer=e2;
			return e;
		}
		case AMOD:
		{
			e1=evaluate_integer_expression(e->parameter[0].pointer);
			if (e1->kind == AUNDEFINED)
				return e1;
			e2=evaluate_integer_expression(e->parameter[1].pointer);
			if (e2->kind == AUNDEFINED)
				return e2;
			
			if (e1->kind == AINTEGER && e2->kind == AINTEGER)
			{
				e->kind=AINTEGER;
				e->parameter[0].number=e1->parameter[0].number%e2->parameter[0].number;
				return e;
			}
			e->parameter[0].pointer=e1;
			e->parameter[1].pointer=e2;
			return e;
		}
		case AADD:
		{
			e1=evaluate_integer_expression(e->parameter[0].pointer);
			if (e1->kind == AUNDEFINED)
				return e1;
			e2=evaluate_integer_expression(e->parameter[1].pointer);
			if (e2->kind == AUNDEFINED)
				return e2;
			
			if (e1->kind == AINTEGER && e2->kind == AINTEGER)
			{
				e->kind=AINTEGER;
				e->parameter[0].number=e1->parameter[0].number+e2->parameter[0].number;
				return e;
			}
				
			e->parameter[0].pointer=e1;
			e->parameter[1].pointer=e2;
			return e;
		}
		case ASUB:
		{
			e1=evaluate_integer_expression(e->parameter[0].pointer);
			if (e1->kind == AUNDEFINED)
				return e1;
			e2=evaluate_integer_expression(e->parameter[1].pointer);
			if (e2->kind == AUNDEFINED)
				return e2;
			
			if (e1->kind == AINTEGER && e2->kind == AINTEGER)
			{
				e->kind=AINTEGER;
				e->parameter[0].number=e1->parameter[0].number-e2->parameter[0].number;
				return e;
			}
			e->parameter[0].pointer=e1;
			e->parameter[1].pointer=e2;
			return e;
		}
		case AINTEGER:
			return e;
		
		case ALABEL:
		case AJLABEL:
		case AGLABEL:
		case ADLABEL:
		{
			register LABEL *label;
			
			label=e->parameter[0].label;
			
			if (label->def_type == D_EXPRESSION)
			{
				TINSTRUCTION_NODE *def;
				
				/* return definition of label, but watch out for circular dependencies */
				
				if (label->def.expression == NULL)
				{
					error("evaluate expression: circular dependency in definition of \'%s\'\n",label->name);
					
					e->kind=AUNDEFINED;
					return e;
				}
				
				def=label->def.expression;
				label->def.expression=NULL;
				
				label->def.expression=evaluate_integer_expression(def);
				return label->def.expression;
			}
			return e;
		}
		
		case AUNDEFINED:
			error("evaluate expression: undefined value encountered\n");
			return e;
			
		default:
			internal_error("(evaluate_integer_expression) invalid expression: %d",e->kind);
			return e;
	}
}

static TINSTRUCTION_NODE *evaluate_macro_expression(TINSTRUCTION *macro, TINSTRUCTION_NODE *e)
{
	/* this instruction evaluates macro expressions. Note that the orginal macro
	   definition may not be altered, so new nodes are created when necessary... */

	register TINSTRUCTION_NODE *e1,*e2;
	
	switch(e->kind)
	{
		case ABRACKET:
		{
			e1=evaluate_macro_expression(macro,e->parameter[0].pointer);
			
			switch (e1->kind)
			{
				case AINTEGER:			/* integer */
				case ALOCAL:			/* local macro label */
				case ALABEL:			/* global label */
				case AJLABEL:
				case AGLABEL:
				case ADLABEL:
				case AUNDEFINED:
					return e1;
				
				default:
					return new_monadic(ABRACKET,e1);
			}
		}
		case AMONADD:
		{
			e1=evaluate_macro_expression(macro,e->parameter[0].pointer);
			
			if (e1->kind == AINTEGER || e1->kind == AUNDEFINED)
				return e1;
			
			return new_monadic(AMONADD,e1);
		}
		case AMONSUB:
		{
			e1=evaluate_macro_expression(macro,e->parameter[0].pointer);
			
			if (e1->kind == AUNDEFINED)
				return e1;
			
			if (e1->kind == AINTEGER)	
				return new_integer(AINTEGER,-e1->parameter[0].number);
			
			return new_monadic(AMONSUB,e1);
		}
		case AMUL:
		{
			e1=evaluate_macro_expression(macro,e->parameter[0].pointer);
			if (e1->kind == AUNDEFINED)
				return e1;
			e2=evaluate_macro_expression(macro,e->parameter[1].pointer);
			if (e2->kind == AUNDEFINED)
				return e2;
			
			if (e1->kind == AINTEGER && e2->kind == AINTEGER)
				return new_integer(AINTEGER,e1->parameter[0].number*e2->parameter[0].number);
			
			return new_dyadic(AMUL,e1,e2);
		}
		case ADIV:
		{
			e1=evaluate_macro_expression(macro,e->parameter[0].pointer);
			if (e1->kind == AUNDEFINED)
				return e1;
			e2=evaluate_macro_expression(macro,e->parameter[1].pointer);
			if (e2->kind == AUNDEFINED)
				return e2;
			
			if (e1->kind == AINTEGER && e2->kind == AINTEGER)
				return new_integer(AINTEGER,e1->parameter[0].number/e2->parameter[0].number);
			
			return new_dyadic(ADIV,e1,e2);
		}
		case AMOD:
		{
			e1=evaluate_macro_expression(macro,e->parameter[0].pointer);
			if (e1->kind == AUNDEFINED)
				return e1;
			e2=evaluate_macro_expression(macro,e->parameter[1].pointer);
			if (e2->kind == AUNDEFINED)
				return e2;
			
			if (e1->kind == AINTEGER && e2->kind == AINTEGER)
				return new_integer(AINTEGER,e1->parameter[0].number%e2->parameter[0].number);
			
			return new_dyadic(AMOD,e1,e2);
		}
		case AADD:
		{
			e1=evaluate_macro_expression(macro,e->parameter[0].pointer);
			if (e1->kind == AUNDEFINED)
				return e1;
			e2=evaluate_macro_expression(macro,e->parameter[1].pointer);
			if (e2->kind == AUNDEFINED)
				return e2;
			
			if (e1->kind == AINTEGER)
			{
				if (e2->kind == AINTEGER)
					return new_integer(AINTEGER,e1->parameter[0].number+e2->parameter[0].number);
			
				if (e2->kind == AADD)
					return macro_combine_comm_any(e2,AADD,e1);					
			}
			
			if (e2->kind == AINTEGER && e1->kind == AADD)
				return macro_combine_comm_any(e1,AADD,e2);
			
			return new_dyadic(AADD,e1,e2);
		}
		case ASUB:
		{
			e1=evaluate_macro_expression(macro,e->parameter[0].pointer);
			if (e1->kind == AUNDEFINED)
				return e1;
			e2=evaluate_macro_expression(macro,e->parameter[1].pointer);
			if (e2->kind == AUNDEFINED)
				return e2;

			if (e1->kind == AINTEGER && e2->kind == AINTEGER)
				return new_integer(AINTEGER,e1->parameter[0].number-e2->parameter[0].number);
			
			if (e2->kind == AINTEGER && e1->kind == AADD)
				return macro_combine_comm_any(e1,ASUB,e2);
			
			return new_dyadic(ASUB,e1,e2);
		}
		case AINTEGER:
			return e;
		
		case ALABEL:
		case AJLABEL:
		case AGLABEL:
		case ADLABEL:
		{
			register LABEL *label;
			
			label=e->parameter[0].label;
			
			if (label->def_type == D_EXPRESSION)
			{
				TINSTRUCTION_NODE *def;
				
				/* return definition of label, but watch out for circular dependencies */
				
				if (label->def.expression == NULL)
				{
					error("evaluate expression: circular dependency in definition of \'%s\'\n",label->name);
					
					e->kind=AUNDEFINED;
					return e;
				}
				
				def=label->def.expression;
				label->def.expression=NULL;
				
				label->def.expression=evaluate_integer_expression(def);
				return label->def.expression;
			}
			return e;
		}
		case APARAMETER:
		{
			/* return macro argument */
			
			return macro->operand[e->parameter[0].number];
		}
		case ALOCAL:
			return e;

		case AUNDEFINED:
			error("evaluate macro expression: undefined value encountered\n");
			return e;
			
		default:
			error("(evaluate macro expression) invalid expression: %d",e->kind);
			return e;
	}
}

/************************************************************************************
 *																					*
 *		output of macros															*
 *																					*
 ************************************************************************************/

static VOID out_integer_macro_expression (TINSTRUCTION *macro,TINSTRUCTION_NODE *e,INT16 *label_number)
{
	switch(e->kind)
	{
		case ABRACKET:
		{
			o_char('(');
			out_integer_macro_expression(macro,e->parameter[0].pointer,label_number);
			o_char(')');
			return;
		}
		case AMONADD:
		{
			o_char('+');
			out_integer_macro_expression(macro,e->parameter[0].pointer,label_number);
			return;
		}
		case AMONSUB:
		{
			o_char('-');
			out_integer_macro_expression(macro,e->parameter[0].pointer,label_number);
			return;
		}
		case AMUL:
		{
			out_integer_macro_expression(macro,e->parameter[0].pointer,label_number);
			o_char('*');
			out_integer_macro_expression(macro,e->parameter[1].pointer,label_number);
			return;
		}
		case ADIV:
		{
			out_integer_macro_expression(macro,e->parameter[0].pointer,label_number);
			o_char('/');
			out_integer_macro_expression(macro,e->parameter[1].pointer,label_number);
			return;
		}
		case AMOD:
		{
			out_integer_macro_expression(macro,e->parameter[0].pointer,label_number);
			o_char('%');
			out_integer_macro_expression(macro,e->parameter[1].pointer,label_number);
			return;
		}
		case AADD:
		{
			out_integer_macro_expression(macro,e->parameter[0].pointer,label_number);
			o_char('+');
			out_integer_macro_expression(macro,e->parameter[1].pointer,label_number);
			return;
		}
		case ASUB:
		{
			out_integer_macro_expression(macro,e->parameter[0].pointer,label_number);
			o_char('-');
			out_integer_macro_expression(macro,e->parameter[1].pointer,label_number);
			return;
		}					
		case AINTEGER:
			fprintf(codefile,"%ld",(long)e->parameter[0].number);
			return;
		
		case ALABEL:
			o_inline_label(e->parameter[0].label);
			return;
		case AJLABEL:
			o_jlabel(e->parameter[0].label);
			return;
		case AGLABEL:
			o_glabel(e->parameter[0].label);
			return;
		case ADLABEL:
			o_dlabel(e->parameter[0].label);
			return;
		
		case APARAMETER:
			out_integer_macro_expression(macro,macro->operand[e->parameter[0].number],label_number);
			return;
		
		case ALOCAL:
			o_luse_label(label_number[e->parameter[0].number]);
			return;

		case AUNDEFINED:
			fputs("UNDEFINED",codefile);
			return;
			
		default:
			error("invalid macro expression");
	}
}

static VOID out_bytes_macro_expression (TINSTRUCTION *macro,TINSTRUCTION_NODE *e,INT16 *label_number)
{
	switch(e->kind)
	{	
		case ACOMMA:
			out_bytes_macro_expression(macro,e->parameter[0].pointer,label_number);
			o_char(',');
			out_bytes_macro_expression(macro,e->parameter[1].pointer,label_number);
			return;

		case ASTRING:
			o_string(e->parameter[0].string->string);
			return;
		
		case APARAMETER:
		{
			e=macro->operand[e->parameter[0].number];
			
			if (e->kind == ASTRING)
				o_string(e->parameter[0].string->string);
			else
				out_integer_macro_expression(macro,e,label_number);
			return;
		}
		default:
			e=evaluate_macro_expression(macro,e);
			out_integer_macro_expression(macro,e,label_number);
			return;
	}
}

static VOID out_words_macro_expression (TINSTRUCTION *macro,TINSTRUCTION_NODE *e,INT16 *label_number)
{
	switch(e->kind)
	{	
		case ACOMMA:
			out_words_macro_expression(macro,e->parameter[0].pointer,label_number);
			o_char(',');
			out_words_macro_expression(macro,e->parameter[1].pointer,label_number);
			return;
		
		default:
			e=evaluate_macro_expression(macro,e);
			out_integer_macro_expression(macro,e,label_number);
			return;
	}
}

static VOID out_macro (TINSTRUCTION *macro)
{
	register TINSTRUCTION	*i;
	INT16					n, n_labels, arity;
	INT16					*label_number;
	
	n_labels	=macro->operator->n_labels;
	arity		=macro->operator->arity;

	/* get new label ids for all local labels */
	
	label_number=alloc_macro_mem(n_labels * sizeof(INT16));
	for(n=0; n < n_labels; n++)
		label_number[n]=new_label_id();
	
	/* evaluate all parameter expressions */
	
	for (n=0; n<arity; n++)
	{
		if (macro->operand[n]->kind != ASTRING)
			macro->operand[n]=evaluate_integer_expression(macro->operand[n]);
	}
	
	i=macro->operator->instructions;
	
	while (i != NULL)
	{
		register TINSTRUCTION_DEF *instr;
		
		instr=i->operator;
		switch (instr->kind)
		{

			/* instructions */
	
			case TINSTR:
			{							
				switch (instr->arity)
				{
					case 0:
					
						fprintf(codefile,"\t%s\n",instr->name);
						break;
						
					case 1:
					
						fprintf(codefile,"\t%s\t",instr->name);
						out_integer_macro_expression(macro,evaluate_macro_expression(macro,i->operand[0]),label_number);
						o_newline();
						break;
						
					default:
						internal_error("out_macro: invalid arity");
				}
				break;
			}
			case TMACRO:
				internal_error("out_macro: no macro within macro");
				break;
			
			/* definitions */
	
			case TWORD_DEF:
			{
				TINSTRUCTION_NODE *args;
				
				args=i->operand[0];

				o_iwords16(0);
				out_words_macro_expression(macro,args,label_number);
				o_newline();
				break;

				break;
			}
			case TBYTE_DEF:
			{
				TINSTRUCTION_NODE *args;

				args=i->operand[0];
				o_ibytes16(0);
				out_bytes_macro_expression(macro,args,label_number);
				o_newline();
				break;
			}	
			case TSTRING_DEF:
			{
				TINSTRUCTION_NODE	*arg;
				STRING	*string;

				arg=i->operand[0];
				
				switch (arg->kind)
				{
					case ASTRING:
					{
						string=arg->parameter[0].string;
				
						o_iwords16(1,string->length);
						o_ibytes16(0);				
						o_string(string->string);
						o_newline();
						
						break;
					}
					case APARAMETER:
					{
						INT32 argn;
						
						argn=arg->parameter[0].number;
						
						arg=macro->operand[argn];
						
						if (arg->kind != ASTRING)
							error("macro argument %d should be a string\n",argn);
						else
						{
							string=arg->parameter[0].string;
				
							o_iwords16(1,string->length);
							o_ibytes16(0);				
							o_string(string->string);
							o_newline();
						}
						break;
					}
					default:
						internal_error("out_macro: invalid string definition");
				}
				break;
			}
			case TFLOAT_DEF:
			{
				TINSTRUCTION_NODE	*arg;

				arg=i->operand[0];
				
				switch (arg->kind)
				{
					case AFLOAT:
					{
						o_float((FLOAT *)&arg->parameter[0].number);						
						break;
					}
					case APARAMETER:
					{
						INT32 argn;
						
						argn=arg->parameter[0].number;
						
						arg=macro->operand[argn];
						
						if (arg->kind != AFLOAT)
							error("macro argument %d should be a float\n",argn);
						else
						{
							o_float((FLOAT *)&arg->parameter[0].number);						
						}
						break;
					}
					default:
						internal_error("out_macro: invalid float definition");
				}
				break;
			}
			case TLABEL_DEF:
			{
				TINSTRUCTION_NODE *label;
				
				label=i->operand[0];
				
				if (label->kind != ALOCAL)
					internal_error("out_macro: invalid label definition");
				
				o_ldef_label(label_number[label->parameter[0].number]);
				break;
			}
			
			/* directives */
	
			case TALIGN:
			{
				o_align();
				break;
			}			
			case TCSEC:
			{	
				o_csec();
				break;
			}
			case TDSEC:
			{
				o_dsec();
				break;
			}
			
			default:
				internal_error("out_macro: invalid instruction");
		}
		i=i->next;
	}
	free_macro_mem();
}

/********************************************************************************
 *																				*
 *		output expressions														*
 *																				*
 ********************************************************************************/

static VOID out_integer_expression (register TINSTRUCTION_NODE *e)
{
	switch(e->kind)
	{	
		case ABRACKET:
		{
			o_char('(');
			out_integer_expression(e->parameter[0].pointer);
			o_char(')');
			return;
		}
		case AMONADD:
		{
			o_char('+');
			out_integer_expression(e->parameter[0].pointer);
			return;
		}
		case AMONSUB:
		{
			o_char('-');
			out_integer_expression(e->parameter[0].pointer);
			return;
		}
		case AMUL:
		{
			out_integer_expression(e->parameter[0].pointer);
			o_char('*');
			out_integer_expression(e->parameter[1].pointer);
			return;
		}
		case ADIV:
		{
			out_integer_expression(e->parameter[0].pointer);
			o_char('/');
			out_integer_expression(e->parameter[1].pointer);
			return;
		}
		case AMOD:
		{
			out_integer_expression(e->parameter[0].pointer);
			o_char('%');
			out_integer_expression(e->parameter[1].pointer);
			return;
		}
		case AADD:
		{
			out_integer_expression(e->parameter[0].pointer);
			o_char('+');
			out_integer_expression(e->parameter[1].pointer);
			return;
		}
		case ASUB:
		{
			out_integer_expression(e->parameter[0].pointer);
			o_char('-');
			out_integer_expression(e->parameter[1].pointer);
			return;
		}					
		case AINTEGER:
			fprintf(codefile,"%ld",(long)e->parameter[0].number);
			return;
		
		case ALABEL:
			o_inline_label(e->parameter[0].label);
			return;
		case AJLABEL:
			o_jlabel(e->parameter[0].label);
			return;
		case AGLABEL:
			o_glabel(e->parameter[0].label);
			return;
		case ADLABEL:
			o_dlabel(e->parameter[0].label);
			return;
		
		case AUNDEFINED:
			fputs("UNDEFINED",codefile);
			return;
					
		default:
			internal_error("out_integer_expression: invalid node encountered");
	}
}

static VOID out_bytes_expression (register TINSTRUCTION_NODE *e)
{
	switch(e->kind)
	{	
		case ACOMMA:
			out_bytes_expression(e->parameter[0].pointer);
			o_char(',');
			out_bytes_expression(e->parameter[1].pointer);
			return;

		case ASTRING:
			o_string(e->parameter[0].string->string);
			return;
		
		default:
			e=evaluate_integer_expression(e);
			out_integer_expression(e);
			return;
	}
}

static VOID out_words_expression (register TINSTRUCTION_NODE *e)
{
	switch(e->kind)
	{
		case ACOMMA:
			out_words_expression(e->parameter[0].pointer);
			o_char(',');
			out_words_expression(e->parameter[1].pointer);
			return;
		
		default:
			e=evaluate_integer_expression(e);
			out_integer_expression(e);
			return;
	}
}

/************************************************************************************
 *																					*
 *		output normal instructions													*
 *																					*
 ************************************************************************************/

VOID out_inline_instructions(register TINSTRUCTION *i)
{	
	while (i != NULL)
	{
		register TINSTRUCTION_DEF *instr;
		
		instr=i->operator;
		switch (instr->kind)
		{

			/* instructions */
	
			case TINSTR:
			{							
				switch (instr->arity)
				{
					case 0:
					
						fprintf(codefile,"\t%s\n",instr->name);
						break;
						
					case 1:
					
						fprintf(codefile,"\t%s\t",instr->name);
						out_integer_expression(evaluate_integer_expression(i->operand[0]));
						o_newline();
						break;
						
					default:
						internal_error("out_inline_instructions: invalid arity");
				}
				break;
			}
			case TMACRO:
				out_macro(i);
				break;
				
			/* definitions */
	
			case TWORD_DEF:
			{
				TINSTRUCTION_NODE *args;
				
				args=i->operand[0];

				o_iwords16(0);
				out_words_expression(args);
				o_newline();
				break;

				break;
			}
			case TBYTE_DEF:
			{
				TINSTRUCTION_NODE *args;

				args=i->operand[0];
				
				o_ibytes16(0);
				out_bytes_expression(args);
				o_newline();
				break;
			}	
			case TSTRING_DEF:
			{
				TINSTRUCTION_NODE	*arg;
				STRING				*string;
				
				arg=i->operand[0];
				
				if (arg->kind != ASTRING)
					internal_error("out_inline_instructions: invalid string definition");

				string=arg->parameter[0].string;
				
				o_iwords16(1,string->length);
				o_ibytes16(0);				
				o_string(string->string);
				o_newline();
				break;

				break;
			}
			case TFLOAT_DEF:
			{
				TINSTRUCTION_NODE	*arg;
				
				arg=i->operand[0];

				if (arg->kind != AFLOAT)
					internal_error("out_inline_instructions: invalid float definition");
				
				o_float((FLOAT *)&arg->parameter[0].number);
				break;
			}
			case TLABEL_DEF:
			{
				TINSTRUCTION_NODE *label;
				
				label=i->operand[0];
				
				switch (label->kind)
				{
					case ALABEL:
						o_inline_label(label->parameter[0].label);
						o_newline();
						break;

					case AJLABEL:
						o_jlabel(label->parameter[0].label);
						o_newline();
						break;

					case AGLABEL:
						o_glabel(label->parameter[0].label);
						o_newline();
						break;

					case ADLABEL:
						o_dlabel(label->parameter[0].label);
						o_newline();
						break;

					default:
						internal_error("out_inline_instructions: invalid label definition");
				}
				break;
			}
			
			/* directives */
	
			case TALIGN:
			{
				o_align();
				break;
			}
			case TATTRIB:
			{
				TINSTRUCTION_NODE *param;
				
				o_attrib();
				
				param=i->operand[0];
				
				if (param->kind != ALITERAL)
					internal_error("out_inline_instructions: attrib has no literal param");
				
				fputs(param->parameter[0].literal,codefile);
				
				o_newline();
				
				break;
			}
			case TCSEC:
			{
				o_csec();
				break;
			}
			case TDSEC:
			{
				o_dsec();
				break;
			}
			
			default:
				internal_error("out_inline_instructions: invalid instruction");
		}
		i=i->next;
		++inline_instruction_number;
		
		handle_write_code();
	}
}
