/*******************************************************************************
 *                                                                             *
 *     Concurrent Clean scanner.c                                              *
 *     ==========================                                              *
 *                                                                             *
 *                                                                             *
 *                                                                             *
 *     Author: Eric Nocker, John van Groningen, Marco Kesseler                 *
 *     At: Department of Computer Science                                      *
 *         University of Nijmegen                                              *
 *                                                                             *
 *     Version: 0.6                                                            *
 *     Date:    July 19, 1990                                                  *
 *                                                                             *
 ******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "ifinterface.h"

#include "cgscanner.h"
#include "cgmemory.h"

#include "ccintermediate.h"
#include "ccmacros.h"

/********************************************************************************
 *																				*
 *		global variables														*
 *																				*
 ********************************************************************************/

typedef struct FILE_NAME
{	
	struct FILE_NAME *left;
	struct FILE_NAME *right;
	CHAR *file_name;

} FILE_NAME;

static FILE_NAME *file_names;					/* to keep track of files that have been opened */

static FILE_DESCRIPTOR	*file_stack;

static FILE			*curr_in_file;	/* the current file	pointer	for input files		*/
static FILE 		*curr_out_file;	/* the current file pointer for output files	*/
static CHAR			*filenm;		/* the file name of the current file			*/
static U_INT16		line_nr;		/* the line number within the current file		*/
static INT16		curchar;		/* the current CHAR								*/

static long			source_file_size;

static BOOL alpha_num_table[257];			/* table with True for ident chars, false otherwise */
static BOOL a_alpha_num_table[257];			/* table with True for assembly ident chars, false otherwise */

#ifdef MACINTOSH
	static long ftype		='TEXT';
	static long fcreator	='MKcg';
#endif

/********************************************************************************
 *																				*
 *		File Administration														*
 *																				*
 ********************************************************************************/

static BOOL scanned_file (register STRING *file_name)
{
	register FILE_NAME **file_p,*new_file;

	file_p=&file_names;
	while (*file_p!=NULL)
	{	FILE_NAME *file=*file_p;
		INT16 r=strcmp (file_name->string,file->file_name);
		if (r==0)
			return true;
		if (r<0)
			file_p=&file->left;
		else
			file_p=&file->right;
	}
	
	new_file=alloc_heap_mem(sizeof (FILE_NAME));
	new_file->left=NULL;
	new_file->right=NULL;
	new_file->file_name=new_chars_from_string(file_name);
	
	*file_p=new_file;
	return false;
}

/********************************************************************************
 *																				*
 *			errors																*
 *																				*
 ********************************************************************************/

VOID scan_error (CHAR *str, CHAR *mess, ...)
{
	va_list argp;
	
	error("SYNTAX ERROR at line %u of %s :\n    ", line_nr, filenm);

	if (str != (CHAR *) NULL)
		error("%s\n", str);

	if (mess != (CHAR *) NULL)
	{
		va_start(argp, mess);
		verror(mess,argp);
		va_end(argp);
		error("\n");
    }
        
	nr_errors++;
}

VOID scan_warning (CHAR *str, CHAR *mess, ...)
{
	if (options.verbose_warnings)
	{
		va_list argp;

		error("SYNTAX WARNING: line %u of %s :\n    ", line_nr, filenm);

		if (str != (CHAR *)NULL)
			error("%s\n", str);

		if (mess != (CHAR *)NULL)
		{
			va_start(argp, mess);
			verror(mess,argp);
			va_end(argp);
			error("\n");
		}
    }
    nr_warnings++;
}

/********************************************************************************
 *																				*
 *		The Scanner																*
 *																				*
 *		Scanning is quite straightforward. No look ahead etc is needed. For		*
 *		each kind of item a scanning function is defined. Also some				*
 *		supporting procedures are present.										*
 *																				*
 ********************************************************************************/

/********************************************************************************
 *																				*
 *		low level routines														*
 *																				*
 ********************************************************************************/

#define next_char() getc (curr_in_file)

static VOID initialise_alpha_num_table(VOID)
{
	INT16 i;
	
	U_CHAR extra_alpha_nums[]="@#$%^&*-+/=<>\\_~.?\'\"";
	U_CHAR extra_a_alpha_nums[]="_.$";
	
	for (i=0; i<257; ++i)
		alpha_num_table[i]=false;	
	for (i='a'; i<='z'; ++i)
		alpha_num_table[i+1]=true;
	for (i='A'; i<='Z'; ++i)
		alpha_num_table[1+i]=true;
	for (i='0'; i<='9'; ++i)
		alpha_num_table[1+i]=true;
	for (i=0; i<sizeof (extra_alpha_nums); ++i)
		alpha_num_table[1+extra_alpha_nums[i]]=true;

	for (i=0; i<257; ++i)
		a_alpha_num_table[i]=false;
	for (i='a'; i<='z'; ++i)
		a_alpha_num_table[i+1]=true;
	for (i='A'; i<='Z'; ++i)
		a_alpha_num_table[1+i]=true;
	for (i='0'; i<='9'; ++i)
		a_alpha_num_table[1+i]=true;
	for (i=0; i<sizeof (extra_a_alpha_nums); ++i)
		a_alpha_num_table[1+extra_a_alpha_nums[i]]=true;
}

#define is_ident_char(c) (alpha_num_table[(c)+1])
#define is_assembler_ident_char(c) (a_alpha_num_table[(c)+1])

/************************************************************************************
 *																					*
 *		skipping																	*
 *																					*
 ************************************************************************************/

VOID skip_whites (VOID)
{
	while (curchar == ' ' || curchar == '\t')
		curchar = next_char ();
}

VOID SkipToChar (register CHAR c)
{
	for (; (c != curchar) && (curchar != EOF); curchar = next_char ())
		;
}

INT16 SkipRest(VOID)
{
	SkipToChar ('\n');
	if (curchar == '\n')
	{	curchar = next_char ();
		++line_nr;
		return PROCEED;
	}
	return END_OF_FILE;
}


INT16 SkipComment(VOID)
{
	skip_whites ();
		
	if (curchar == '|')
		SkipToChar ('\n');

	if (curchar == '\n')
	{	curchar = next_char ();
		++line_nr;
		return PROCEED;
	}
	
	if ((INT16) curchar == EOF)
		return END_OF_FILE;

	scan_error ("comment expected", NULL);
	return SkipRest();
}

INT16 skip_assembler_comment(VOID)
{
	skip_whites ();
	
	if (curchar == ']')
	{
		/*	end of inline instructions, do not skip rest of line, info may
			follow */

		curchar = next_char();
		return END_OF_INLINE;
	}
	
	if (curchar == ';')
		SkipToChar ('\n');

	if (curchar == '\n')
	{	curchar = next_char ();
		++line_nr;
		return PROCEED;
	}
	
	if ((INT16) curchar == EOF)
		return END_OF_FILE;

	scan_error ("comment expected", NULL);
	return SkipRest();
}

INT16 skip_assembler(VOID)
{
	for (;;)
	{
		SkipToChar (']');
		
		if (curchar == ']')
		{
			/* end of inline instructions, skip rest of line */
	
			SkipToChar ('\n');
			return END_OF_INLINE;
		}
	
		if (curchar == '\n')
		{	curchar = next_char ();
			++line_nr;
			continue;
		}

		if ((INT16) curchar == EOF)
			return END_OF_FILE;
		
		internal_error("skip_assembler");
	}
}

/************************************************************************************
 *																					*
 *		test routines																*
 *																					*
 ************************************************************************************/

static BOOL is_hexa_char (CHAR c)
{
	if (is_digit(c))
		return true;
	
	return ((c >= 'a' &&  c <= 'f') || (c >= 'A' && c <= 'F'));	
}

BOOL is_directive(VOID)
{
	return (curchar=='.');
}

BOOL is_char(VOID)
{
	return (curchar=='\'');
}

BOOL is_string(VOID)
{
	return (curchar=='\"');
}

BOOL is_number(VOID)
{
	return is_digit(curchar);
}

/************************************************************************************
 *																					*
 *		identifiers																	*
 *																					*
 ************************************************************************************/

BOOL scan_ident (register STRING *s)
{
	register U_INT16 curpos = 0;
	
	skip_whites ();

	while ( is_ident_char (curchar) )
	{	s->string[curpos] = curchar;
		curchar = next_char ();
		if (++curpos > STRINGSIZE)
		{	s->string[--curpos] = '\0';
			s->length=curpos;
			scan_error ("identifier too long", s->string);
			return true;
		}
	}
	s->string[curpos] = '\0';
	s->length = curpos;
	
	return (curpos);
}

BOOL scan_assembler_ident (register STRING *s)
{
	register U_INT16 curpos = 0;
	
	skip_whites();
	
	if (is_digit(curchar))
		return false;

	while ( is_assembler_ident_char (curchar))
	{	s->string[curpos] = curchar;
		curchar = next_char ();
		if (++curpos > STRINGSIZE)
		{	s->string[--curpos] = '\0';
			s->length=curpos;
			scan_error ("identifier too long", s->string);
			return true;
		}
	}
	s->string[curpos] = '\0';
	s->length=curpos;
	
	return (curpos);
}

/************************************************************************************
 *																					*
 *		integer values																*
 *																					*
 ************************************************************************************/

static BOOL scan_u32_hexadecimal (U_INT32 *value)
{
	register U_INT32 n;
	
	skip_whites();
	
	if (! is_hexa_char (curchar))
		return false;
	
	n=0;
	for (;;)
	{
		if (is_digit(curchar))
		{
			n=(n<<4) | curchar-'0';
			curchar = next_char();
			
		} else if (curchar >= 'A' && curchar <= 'F')
		{
			n=(n<<4) | curchar-'A'+10;
			curchar = next_char();
			
		} else if (curchar >= 'a' && curchar <= 'f')
		{
			n=(n<<4) | curchar-'a'+10;
			curchar = next_char();
			
		} else
			break;
	}
	
	*value=n;
	
	return true;
}

static BOOL scan_u16_hexadecimal (U_INT16 *value)
{
	register U_INT16 n;
	
	skip_whites();
	
	if (! is_hexa_char (curchar))
		return false;
	
	n=0;
	for (;;)
	{
		if (is_digit(curchar))
		{
			n=(n<<4) | curchar-'0';
			curchar = next_char();
			
		} else if (curchar >= 'A' && curchar <= 'F')
		{
			n=(n<<4) | curchar-'A'+10;
			curchar = next_char();
			
		} else if (curchar >= 'a' && curchar <= 'f')
		{
			n=(n<<4) | curchar-'a'+10;
			curchar = next_char();
			
		} else
			break;
	}

	*value=n;
	
	return true;
}

static BOOL scan_u32_decimal (U_INT32 *value)
{
	register U_INT32 n;

	skip_whites();
	
	if (! is_digit (curchar))
		return false;
	
	n = curchar - '0';
	curchar = next_char();
	while (is_digit (curchar))
	{	n = (n * 10) + (curchar - '0');
		curchar = next_char ();
	}

	*value=n;
	
	return true;
}

static BOOL scan_u16_decimal (U_INT16 *value)
{
	register U_INT16 n;
	
	skip_whites();
	
	if (! is_digit (curchar))
		return false;
	
	n = curchar - '0';
	curchar = next_char();
	while (is_digit (curchar))
	{	n = (n * 10) + (curchar - '0');	
		curchar = next_char ();
	}

	*value=n;
	
	return true;
}

BOOL scan_unumber32 (U_INT32 *value)
{
	/* scans a decimal or hexadecimal value of 32 bits */
	
	skip_whites();

	if (curchar == '0')
	{	
		curchar=next_char();
		
		if (curchar == 'x')
		{
			curchar = next_char();
			return scan_u32_hexadecimal(value);
		
		} else if (!is_digit(curchar))
		{
			/* there is no digit after the 0 */
			
			*value=0;
			return true;
		}
	}
	
	return scan_u32_decimal(value);
}

BOOL scan_number32 (INT32 *value)
{
	/* scans a decimal or hexadecimal value of 32 bits */

	BOOL	pos=true;
	U_INT32	n;
	
	skip_whites();

	if (curchar == '-')
	{	pos = false;
		curchar = next_char ();
	} else if (curchar == '+')
		curchar = next_char ();

	if (!scan_unumber32(&n))
		return false;
	
	if (pos)
		*value = (INT32)n;
	else
		*value = 0-(INT32)n;
	
	return true;
}

BOOL scan_unumber16 (U_INT16 *value)
{
	/* scans a decimal or hexadecimal value */
	
	skip_whites();

	if (curchar == '0')
	{
		curchar=next_char();

		if (curchar == 'x')
		{
			curchar = next_char();			
			return scan_u16_hexadecimal(value);
		
		} else if (!is_digit(curchar))
		{
			/* there is no digit after the 0 */
			
			*value=0;
			return true;
		}
	}
	
	return scan_u16_decimal(value);
}

BOOL scan_number16 (INT16 *value)
{
	/* scans a decimal or hexadecimal value */

	BOOL	pos=true;
	U_INT16 n;
		
	skip_whites();

	if (curchar == '-')
	{	pos = false;
		curchar = next_char ();
	} else if (curchar == '+')
		curchar = next_char ();

	if (!scan_unumber16(&n))
		return false;
		
	if (pos)
		*value = (INT16)n;
	else
		*value = 0-(INT16)n;
	
	return true;
}

#if _COMPILE_UNUSED_

BOOL ScanSize (register INT32 *l)
{	INT32 digs;

	if (! scan_unsigned_integer (&digs))
		return false;
	
	switch (curchar)
	{	case 'k':
		case 'K':
					*l = digs * 1024;
					curchar = next_char ();
					break;
		case 'm':
		case 'M':
					*l = digs * 1024 * 1024;
					curchar = next_char ();
					break;
		default:
					*l = digs;
					break;
	}
	
	return true;
}

#endif

/************************************************************************************
 *																					*
 *		reals																		*
 *																					*
 ************************************************************************************/

static VOID next_real_char(register CHAR *string,register INT16 *length)
{
	if (*length<STRINGSIZE)
		string[(*length)++]=curchar;
	curchar=next_char();
}

static BOOL scan_real_digits(CHAR *string,INT16 *length)
{
	if (!is_digit(curchar))
		return false;
	
	next_real_char(string,length);
	
	while(is_digit(curchar))
		next_real_char(string,length);
	
	return true;
}

BOOL scan_real (FLOAT *value)
{	
	SSTRING string;
	INT16 length;

	skip_whites();
	
	length=0;

	if (curchar == '-')
		next_real_char(string,&length);
	else if (curchar == '+')
		curchar=next_char();
	
	if (!scan_real_digits(string,&length))
		return false;
	
	if (curchar!='.')
		return false;

	next_real_char(string,&length);
	
	if (!scan_real_digits(string,&length))
		return false;
	
	if (curchar=='e' || curchar=='E')
	{	next_real_char(string,&length);
		
		if(curchar == '-')
			next_real_char(string,&length);
		else if(curchar == '+')
			curchar=next_char();
		
		if (!scan_real_digits(string,&length))
			return false;
	}
	
	if (length>=STRINGSIZE)
	{	string[STRINGSIZE]='\0';
		scan_error ("real too long", string);
		return false;
	}

	string[length]='\0';
	value->f=(DOUBLE) atof (string);
	
	return true;
}

/************************************************************************************
 *																					*
 *		strings and chars															*
 *																					*
 ************************************************************************************/

static INT16 scan_BS_char (VOID)
{
	/* supported are:
	   - octal excape sequence
	   - \b, \f, \n, \r and \t
	   - \'printable char', e.g. \"
	*/

	curchar=next_char();				/* skip backslash */
	
	if (curchar>='0' && curchar<='7')
	{	
		/* octal number */
		
		register U_INT16 c;
	
		c=curchar-'0';
		curchar=next_char();
		if (curchar>='0' && curchar<='7')
		{	c=(c<<3)+(curchar-'0');
			curchar=next_char();
			if (curchar>='0' && curchar<='7')
			{	c=(c<<3)+(curchar-'0');
				curchar=next_char();
			}
		}
		return c;
	
	} else
	{	switch (curchar)
		{
			case 'b':	curchar=next_char(); return '\b';
			case 'f':	curchar=next_char(); return '\f';
			case 'n':	curchar=next_char(); return '\n';
			case 'r':	curchar=next_char(); return '\r';
			case 't':	curchar=next_char(); return '\t';
			case EOF:	return EOF;
			default:
			{
				INT16 c;
				
				c=curchar;
				curchar=next_char();
				return c;
			}
		}
	}
}


BOOL scan_string (register STRING *string)
{
	register INT16 length = 0;

	length=0;
	
	skip_whites ();
	
	if (curchar != '\"')
		return false;
	
	curchar = next_char ();
	while ( curchar != '\"' )
	{	if (curchar == '\\')
			string->string[length]=scan_BS_char();
		else
		{	string->string[length]=curchar;
			curchar = next_char ();
		}
		if (++length > STRINGSIZE)
		{	string->string[STRINGSIZE] = '\0';
			string->length=STRINGSIZE;
			scan_error ("string too long", string->string);
			return false;
		}
	}
	curchar = next_char ();
	
	string->string[length] = '\0';
	string->length=length;
	
	return true;
}

#if _COMPILE_UNUSED_

BOOL ScanFilename (register CHAR *s)
{	register INT16 curpos = 0;

	skip_whites ();
	
	if (curchar != '\"')
		return false;
	
	curchar = next_char ();
	while ( curchar != '\"' )
	{	s [curpos] = curchar;
		curchar = next_char ();
		if (++curpos > STRINGSIZE)
		{	s [--curpos] = '\0';
			scan_error ("string too long", s);
			return true;
		}
	}
	
	s [curpos] = '\0';
	curchar = next_char ();
	return (curpos > 0);
}

#endif
	

BOOL scan_char (INT32 *c)
{
	skip_whites ();
	
	if (curchar != '\'')
		return false;
	
	curchar = next_char ();
	if (curchar == '\\')
		*c=scan_BS_char();
	else
	{
		*c=curchar;
		curchar=next_char();
	}

	if ( curchar != '\'' )
		return false;
	curchar = next_char ();
	return true;
}

/************************************************************************************
 *																					*
 *		bools																		*
 *																					*
 ************************************************************************************/

BOOL scan_bool (INT32 *b)
{	
	skip_whites ();
	
	if (curchar == 'T')
	{	if ((curchar=next_char ())=='R' &&
			(curchar=next_char ())=='U' &&
			(curchar=next_char ())=='E')
		{	*b = 1;
			curchar=next_char();
			return true;
		}
	}	
	else if (curchar == 'F')
	{	if ((curchar=next_char ())=='A' &&
			(curchar=next_char ())=='L' &&
			(curchar=next_char ())=='S' &&
			(curchar=next_char ())=='E')
		{	*b = 0;
			curchar=next_char();
			return true;
		}
	}
	return false;
}

/************************************************************************************
 *																					*
 *		vectors																		*
 *																					*
 ************************************************************************************/

static VOID SkipBVector (VOID)
{
	skip_whites();
	while  (curchar == 'i' || curchar == 'I' ||
			curchar == 'b' || curchar == 'B' ||
			curchar == 'c' || curchar == 'C' ||
			curchar == 'f' || curchar == 'F' ||
			curchar == 'r' || curchar == 'R' ||
			curchar == 'p' || curchar == 'P')
	{	curchar = next_char();
		skip_whites();
	}
}

BOOL ScanBVector (INT16 *size, register U_INT32 *vector)
{
	register INT16 i;

	if (*size>MAX_VECTOR_SIZE*VECTOR_ELEMENT_SIZE)
	{
		scan_error ("vector for b-stack too big (empty vector will be used instead)", (CHAR *) NULL);
		SkipBVector ();
		*size = 0;
		return true;
	}
	
	skip_whites();
	
	i=0;
	while (i!=*size)
	{	switch (curchar)
		{
			case 'i':	case 'I':
			case 'b':	case 'B':
			case 'c':	case 'C':
			case 'p':	case 'P':
				clear_vector_bit(vector,i);
				i+=1;
				break;
			case 'f':	case 'F':
				clear_vector_bit(vector,i);
				i+=1;
				clear_vector_bit(vector,i);
				i+=1;
				break;
			case 'r':	case 'R':
				set_vector_bit(vector,i);
				i+=1;
				set_vector_bit(vector,i);
				i+=1;	
				break;
			default:				
				return false;
		}
		curchar=next_char();
		skip_whites();
	}
	return true;
}


BOOL scan_parameters (INT16 *n_a_params, INT16 *n_b_params, U_INT32 *vector, STRING *params)
{
	register	INT16 a_param, b_param, curpos;
	register	CHAR *p;
	INT16		offset;
	
	a_param	=0;
	b_param	=0;
	curpos	=0;
	
	p=params->string;
	
	for(;;)
	{
		skip_whites();

		switch(curchar)
		{
			case 'a':
			case 'A':
			
				/* check space in string */
				
				curpos+=6;
				if (curpos > STRINGSIZE)
				{
					params->length=0;
					return false;
				}

				*p++='A';

				curchar=next_char();
				
				/* read offset */
				
				if (!scan_unumber16((U_INT16 *)&offset))
					return false;
				
				put_int16(p,offset);
				
				/* skip colon */
				
				if (curchar != ':')
					return false;
				curchar=next_char();
				
				/* read one-letter place code */
				
				*p++=curchar;
				curchar=next_char();
				
				/* read place offset */
				
				if (!scan_unumber16((U_INT16 *)&offset))
					return false;
				put_int16(p,offset);
				
				++a_param;
				break;
			
			case 'b':
			case 'B':
			
				/* check space in string */
				
				curpos+=(6);
				if (curpos > STRINGSIZE)
				{
					params->length=0;
					return false;
				}

				*p++='B';
				curchar=next_char();

				/* read offset */
				
				if (!scan_unumber16((U_INT16 *)&offset))
					return false;
				put_int16(p,offset);
				
				/* skip colon */
				
				if (curchar != ':')
					return false;
				curchar=next_char();
				
				/* read place code, this may be prefixed by an 'R', which indicates
				   that a real is placed at the location that follows*/
				
				if (curchar=='R' || curchar=='r')
				{
					/*	this is a real, indicate this in the vector returned, and
						the amount of parameters */
					
					curchar=next_char();
					
					set_vector_bit(vector,b_param);
					++b_param;
					set_vector_bit(vector,b_param);
					++b_param;
					
				} else
				{
					clear_vector_bit(vector,b_param);
					++b_param;
				}
				
				*p++=curchar;
				curchar=next_char();
				
				/* read place offset */
				
				if (!scan_unumber16((U_INT16 *)&offset))
					return false;
				put_int16(p,offset);

				break;
			
			default:
				params->length=curpos;
				*n_a_params=a_param;
				*n_b_params=b_param;
				return true;
		}
	}
}

/************************************************************************************
 *																					*
 *		constructors																*
 *																					*
 ************************************************************************************/

BOOL scan_inline_start (VOID)
{
	if (curchar == '[')
	{
		curchar=next_char();
		return true;
	}
	return false;
}

BOOL scan_colon(CHAR *colon)
{
	skip_whites();
	
	if (curchar==':')
	{	
		*colon=curchar;
		curchar=next_char();
		return true;
	}
	return false;
}

BOOL scan_open(CHAR *open)
{
	skip_whites();
	
	if (curchar=='(')
	{	
		*open=curchar;
		curchar=next_char();
		return true;
	}
	return false;
}

BOOL scan_close(CHAR *close)
{
	skip_whites();
	
	if (curchar==')')
	{	
		*close=curchar;
		curchar=next_char();
		return true;
	}
	return false;
}

BOOL scan_begin_macro(CHAR *open)
{
	skip_whites();
	
	if (curchar=='{')
	{	
		*open=curchar;
		curchar=next_char();
		return true;
	}
	return false;
}

BOOL scan_end_macro(CHAR *close)
{
	skip_whites();
	
	if (curchar=='}')
	{	
		*close=curchar;
		curchar=next_char();
		return true;
	}
	return false;
}

BOOL scan_assembler_separator(CHAR *separator)
{
	skip_whites();
	
	if (curchar==',')
	{
		*separator=curchar;
		curchar=next_char();
		return true;
	}
	return false;
}

/************************************************************************************
 *																					*
 *		operators																	*
 *																					*
 ************************************************************************************/

BOOL scan_monadic_assembler_operator(INT16 *operator)
{
	skip_whites();
	
	switch(curchar)
	{
		case '-':
			*operator=AMONSUB;
			curchar=next_char();
			return true;
		
		case '+':
			*operator=AMONADD;
			curchar=next_char();
			return true;
		
		default:
			return false;
	}
}

BOOL scan_prio2_assembler_operator(INT16 *operator)
{
	skip_whites();
	
	switch (curchar)
	{
		case '*':
			*operator=AMUL;
			curchar=next_char();
			return true;

		case '/':
			*operator=ADIV;
			curchar=next_char();
			return true;

		case '%':
			*operator=AMOD;
			curchar=next_char();
			return true;

		default:
			return false;
	}
}

BOOL scan_prio3_assembler_operator(INT16 *operator)
{
	skip_whites();
	
	switch (curchar)
	{
		case '+':
			*operator=AADD;
			curchar=next_char();
			return true;

		case '-':
			*operator=ASUB;
			curchar=next_char();
			return true;

		default:
			return false;
	}
}

BOOL scan_prio4_assembler_operator(INT16 *operator)
{
	skip_whites();
	
	switch (curchar)
	{
		case '|':
			*operator=AOR;
			curchar=next_char();
			return true;

		case '&':
			*operator=AAND;
			curchar=next_char();
			return true;

		default:
			return false;
	}
}

BOOL scan_equal(CHAR *equal)
{
	skip_whites();
	
	if (curchar=='=')
	{	
		*equal=curchar;
		curchar=next_char();
		return true;
	}
	return false;
}

/********************************************************************************
 *																				*
 *			files																*
 *																				*
 ********************************************************************************/

static CHAR *get_file_extension (CHAR *name, CHAR *extension)
{
	CHAR	*p;
		
	p=name+strlen(name)-strlen(extension);
	
	if (p >= name && strcmp(p,extension)==0)
		return p;								/* correct extension found */
	
	return NULL;
}

static CHAR *append_extension (CHAR *name, CHAR *extension)
{
	CHAR *p;
	
	if ((p=get_file_extension(name,extension))!=NULL)
	{
		return name;						/* correct extension present, return name */
	
	} else
	{
		/* no extension, append correct extension */
		
		CHAR *new_name;
		
		new_name=alloc_heap_mem(strlen (name) + strlen (extension) + 1);
		strcpy (new_name, name);
		strcat (new_name, extension);
		
		return new_name;
	}
}

static CHAR *strip_extension (CHAR *name, CHAR *extension)
{
	CHAR *p;
	
	if (!(p=get_file_extension(name,extension)))
		return name;								/* no extension present, return name */
	else
	{
		CHAR *new_name;
		INT16 length;
		
		length=strlen(name)-strlen(extension);
		
		new_name=alloc_heap_mem(length+1);
		memcpy(new_name, name, length);
		new_name[length]='\0';
		
		return new_name;
	}
}

static FILE *open_file(BOOL create, CHAR *mode, CHAR *name, CHAR *extension, U_INT16 number, ...)
{
	FILE *f;
	CHAR *filename;
	va_list argp;
	
	filename=append_extension(name, extension);

	va_start(argp, number);
	while (number--)
	{
		CHAR *path;
		CHAR *full_name;
		
		path=next_string(argp);
		
		full_name=alloc_heap_mem(strlen(path) + strlen(filename) + 1);
		strcpy (full_name, path);
		strcat (full_name, filename);

		if ((f = fopen (full_name, mode))!=NULL)
		{
		
#ifdef MACINTOSH

			if (create)
			{
				FileParam	pb;
				Str255		full_name2;
	
				set_str255(full_name2,full_name);

				pb.ioNamePtr	=full_name2;
				pb.ioVRefNum	=0;
				pb.ioFVersNum	=0;
				pb.ioFDirIndex	=0;
			
				if (PBGetFInfoSync((ParmBlkPtr)&pb)==noErr)
				{
					pb.ioFlFndrInfo.fdType		=ftype;
					pb.ioFlFndrInfo.fdCreator	=fcreator;
			
					PBSetFInfoSync((ParmBlkPtr)&pb);
				}
			}
#endif
	
			va_end(argp);
			return f;
		}
	}
	va_end(argp);
	return NULL;
}

static VOID push_file (VOID)
{
	FILE_DESCRIPTOR *file_desc;
	
	if (curr_in_file==NULL)
		internal_error("push_file: no file to push");
	
	file_desc=alloc_heap_mem(sizeof(FILE_DESCRIPTOR));
	
	/* push current status */
	
	file_desc->file		=curr_in_file;
	file_desc->filename	=filenm;
	file_desc->line_nr	=line_nr;
	file_desc->curchar	=curchar;

	file_desc->prev		=file_stack;
	file_stack			=file_desc;
}

static VOID pop_file (VOID)
{
	curr_in_file = file_stack->file;
	filenm  = file_stack->filename;
	line_nr = file_stack->line_nr;
	curchar = file_stack->curchar;

	file_stack = file_stack->prev;
}

FILE *open_source_file(CHAR *filename)
{	
	if ((curr_in_file=open_file(false,"r",filename,options.source_extension,3,"",options.source_directory,options.depend_directory))!=NULL)
	{
		setvbuf (curr_in_file,NULL,_IOFBF,IO_BUF_SIZE);

		/*	get file size */
		
		if (fseek (curr_in_file,0L,SEEK_END) != 0)
		{	fclose(curr_in_file);
			return NULL;
		}
		
		source_file_size=ftell(curr_in_file);
		
		if (fseek (curr_in_file,0L,SEEK_SET) != 0)
		{	fclose(curr_in_file);
			return NULL;
		}
				
		filenm = filename;	
		curchar = next_char ();
		line_nr = 1;
	
		return curr_in_file;				
	}
	return NULL;
}
	
FILE *open_depend_file(STRING *filename)
{
	/* save current state */
	
	push_file();
		
	if ((curr_in_file=open_file(false,"r",filename->string,options.source_extension,3,"",options.source_directory,options.depend_directory))!=NULL)
	{	
		setvbuf (curr_in_file,NULL,_IOFBF,IO_BUF_SIZE);
	
		if (scanned_file(filename))
		{
			error("warning: file \'%s\' scanned twice (skipped)\n",filename->string);

			/* skip to end of file */
			
			if (fseek (curr_in_file,0l,SEEK_END)!=0)
			{
				error("open_depend_file: cannot skip to end of file\n");
				return NULL;
			}
		}
		
		filenm = filename->string;
		curchar = next_char ();
		line_nr = 1;
	
		/* return new state */
		
		return curr_in_file;	
    }
    
    /* opening failed, restore previous state */
    
    pop_file();
    return NULL;
}

FILE *open_assembly_file(char *name)
{
	name=strip_extension(name,options.source_extension);
	
	if ((curr_out_file=open_file(true,"w",name,options.assembly_extension,1,options.assembly_directory))!=NULL)
	{		
		setvbuf (curr_out_file,NULL,_IOFBF,2*IO_BUF_SIZE);		
		return curr_out_file;
	}
	return NULL;
}

FILE *open_intermediate_file(char *name, char *extension)
{
	name=strip_extension(name,options.source_extension);
	
	if ((curr_out_file=open_file(true,"w",name,extension,1,options.assembly_directory))!=NULL)
	{
		setvbuf (curr_out_file,NULL,_IOFBF,2*IO_BUF_SIZE);
		return curr_out_file;
	}
	return NULL;
}

VOID close_input_file(VOID)
{
	/*	only to be called when opening of file succeeded, else it will close
		the 'previous' file */
	
	if (curr_in_file == NULL)
		internal_error("close_file: no file to close");
	
	fclose(curr_in_file);

	/* restore previous state if needed */
	
	if (file_stack == NULL)
		curr_in_file = NULL;
	else
		pop_file();
}

VOID close_output_file(VOID)
{
	if (curr_out_file == NULL)
		internal_error("close_output_file: no file to close");
	
	fclose(curr_out_file);
}

VOID close_all_files(VOID)
{
	if (curr_out_file != NULL)
		fclose(curr_out_file);
		
	if (curr_in_file != NULL)
		fclose(curr_in_file);
	
	while(file_stack != NULL)
	{
		pop_file();
		fclose(curr_in_file);
	}
}

void get_source_position(long *position, long *max)
{
	if (file_stack == NULL && curr_in_file != NULL)
	{	*position	=ftell(curr_in_file);
		*max		=source_file_size;
	} else
	{	*position	=-1L;
		*max		=-1L;
	}
}

char *get_curr_file_name(short *main)
{
	/* main indicates if this is the main module, the filename of the
		current input file is returned */
	
	*main=(file_stack==NULL);
	
	if (curr_in_file != NULL)
		return filenm;
	
	return NULL;
}

/********************************************************************************
 *																				*
 *			initialisation														*
 *																				*
 ********************************************************************************/

VOID initialise_scanner(VOID)
{
	initialise_alpha_num_table();
	
	curr_in_file	=NULL;
	curr_out_file	=NULL;
	file_names		=NULL;
	file_stack		=NULL;
}
