#include "TCP.h"

#include "ifinterface.h"
#include "ifmacinterface.h"
#include "ifplot.h"
#include "ifserver.h"

/************************************************************************************
 *																					*
 *		local typedefs																*
 *																					*
 ************************************************************************************/

typedef struct CommandStream
{
	TCPiopb pb;							/* parameter block for TCP commands */
	StreamPtr stream;					/* stream pointer */
	char *tcp_buf;						/* buffer for TCP routines */
	Boolean open;						/* true if stream is open */
	Boolean	created;					/* true if stream has been created */
	Boolean active;						/* this boolean has only meaning for an
										   open stream. normally it is true (set
										   by the open routine). Clearing it will
										   result in closing down the connection */
} CommandStream;

/************************************************************************************
 *																					*
 *		global variables															*
 *																					*
 ************************************************************************************/

static struct CommandStream		graph_stream;

static Boolean	tcp_avail;
static Boolean	handling_remote_commands;
static Boolean	remote_enabled;
static Boolean	little_endian_client;					/* true if little endian client */

/************************************************************************************
 *																					*
 *		little-endian / big-endian conversions										*
 *																					*
 ************************************************************************************/

static unsigned short swap_short(unsigned short v)
{
	/* works for big endian machines */

	register unsigned char *p;
	
	p=(unsigned char *)&v;
	return (*p | (*(p+1) << 8));
}

static unsigned long swap_long(unsigned long v)
{
#if 1

	register unsigned char *p;

	p=(unsigned char *)&v;
	return (((((((long)*(p+3) << 8) | (long)*(p+2)) << 8) | (long)*(p+1)) << 8) | (long)*p);
#else
	return v;
#endif
}

/************************************************************************************
 *																					*
 *		access to TCP/IP															*
 *																					*
 ************************************************************************************/

static pascal void remote_notify(StreamPtr tcpStream,unsigned short eventCode,Ptr userDataPtr,unsigned short terminReason,struct ICMPReport *icmpMsg)
{
	/* this routine handles events on a stream, ... it basically does nothing */
	
	struct CommandStream *s;
	
	s=(struct CommandStream *)userDataPtr;
	
	switch (eventCode)
	{
		case TCPClosing:
		case TCPULPTimeout:
			break;
			
		case TCPTerminate:
		{
			/* connection has terminated, handle it */
						
			switch (terminReason)
			{
				case TCPULPClose:
				case TCPULPAbort:
							
					/* stream has been closed, notify others */
					
					s->open=false;
					break;					

				case TCPRemoteAbort:
				case TCPNetworkFailure:
				case TCPSecPrecMismatch:
				case TCPULPTimeoutTerminate:
				case TCPServiceError:
				default:
					break;
			}
			break;
		}
		case TCPDataArrival:
		case TCPUrgent:
		case TCPICMPReceived:
		default:
			break;
	}
}

static void init_stream(struct CommandStream *s)
{
	s->open			=false;
	s->created		=false;
}

static void remote_create(struct CommandStream *s)
{
	OSErr	err;

	if (!s->created)
	{
		/* synchronous creation of stream */
		
		/* allocate buffer space */
		
		s->tcp_buf=NewPtr(4096);
		
		if (s->tcp_buf == NULL)
		{
			internal_error("remote_create: could not allocate buffer");
			return;
		}
		
		err=_TCPCreate(&s->pb,&s->stream,s->tcp_buf,4096,remote_notify,(Ptr) s,NULL,false);
		
		if (err!=noErr)
		{	internal_error ("creating stream failed");
			return;
		}
		
		s->created	=true;
	}
}

static void remote_release(struct CommandStream *s)
{
	OSErr	err;
	long	stream_buf_len;
	char	*stream_buf_ptr;
	
	/* synchronous release of stream, if it has been created */
		
	if (s->created)
	{
		s->created=false;

		if (s->tcp_buf != NULL)
		{
			DisposPtr(s->tcp_buf);
			s->tcp_buf=NULL;
		}

		err=_TCPRelease (&s->pb,s->stream,&stream_buf_ptr,&stream_buf_len,NULL,false);
	
		if (err!=noErr)
			internal_error ("releasing stream failed");		
	}
}


static Boolean remote_open(struct CommandStream *s, void(*idle)(void), Boolean *able)
{
	/* Routine to asynchronously open a stream. */
	
	if (!s->open)
	{
		OSErr	err;
	
		ip_addr localIP,remoteIP;
		ip_port local_port,remote_port;
	
		remoteIP	=cAnyIP;
		localIP		=cAnyIP;
	
		local_port	=5678;
		remote_port	=cAnyPort;
		
		/* try open stream asynchronously */
		
		for (;;)
		{
			err=_TCPPassiveOpen (&s->pb,s->stream,&remoteIP,&remote_port,&localIP,&local_port,NULL,NULL,true);
			
			if (err != noErr)
			{
				error ("open request could not be issued\n");
				return false;
			}
			
			/* call was asynchronous, poll until ready... */
	
			while (s->pb.ioResult == 1)
			{
				(*idle)();
				
				/* check if stream has been disabled meanwhile */
				
				if (!(*able))
					return false;
			}
			
			switch (s->pb.ioResult)
			{
				case noErr:

					/* stream is open */
					
					s->open=true;
					s->active=true;		/* it is active untill told otherwise */
					
					return true;
				
				case commandTimeout:
				
					/* a time-out has occured, try again... */
					
					break;
				
				default:
				
					error("connection could not be opened (id = %d)\n",s->pb.ioResult);
					return false;
			}
		}
	}
	else
	{
		internal_error("remote_open: stream already open");
		return false;
	}
}


static void remote_close(struct CommandStream *s, void (* idle)(void))
{
	OSErr err;
	
	/* synchronous close, if stream open */
	
	if (s->open)
	{
		err=_TCPClose (&s->pb,s->stream,NULL,NULL,false);
			
		if (err!=noErr)
			error ("closing stream failed\n");
		
		while (s->open)
			(*idle)();
	}
}

static void remote_abort(struct CommandStream *s, void (* idle)(void))
{
	OSErr err;
	
	/* synchronous abort, but only if stream is open */
	
	if (s->open)
	{
		err=_TCPAbort (&s->pb,s->stream,NULL,NULL,false);
		
		if (err!=noErr)
			error ("aborting stream failed\n");

		while (s->open)
			(*idle)();
	}
}

static int remote_receive(unsigned char *buffer, int size, struct CommandStream *s, void(*idle)(void))
{
	/* asynchronous receive */
	
	unsigned char	*ptr;
	unsigned char	*end;

	ptr=buffer;
	end=&buffer[size];

	do
	{
		OSErr err;
		unsigned short	length;

		/* length is amount free in buffer */
	
		length=end-ptr;

		err=_TCPRcv (&s->pb,s->stream,(Ptr)ptr,&length,NULL,NULL,true);
		
		if (err != noErr)
		{
			error ("read request could not be issued\n");
			return false;
		}
		
		/* call was asynchronous, poll until ready, it should not be
		   possible to disable the stream until explicit disconnect, so this
		   does not have to be checked after the idle function. A disconnection
		   could have been forced however... */

		while (s->pb.ioResult == 1)
		{
			(*idle)();
			
			if (s->active == false)
				return false;
		}
		
		switch (s->pb.ioResult)
		{
			case noErr:
			
				/* no error, read succeeded */
		
				ptr+=s->pb.csParam.receive.rcvBuffLen;
				break;
		
			case commandTimeout:
			
				/* a time-out has occured, try again... */
				
				break;
			
			default:
			
				error("receiving failed (id = %d)\n",s->pb.ioResult);
				return false;
		}
		
	} while (ptr<end);

	if (ptr!=end)
	{	error ("not all bytes received\n");
		return false;
	}
	
	return true;
}

/************************************************************************************
 *																					*
 *		auxiliary functions															*
 *																					*
 ************************************************************************************/

static void set_remote_able(void)
{
	/* reverse of current state */
	
	SetItem(menus.remote_menu,menu_item_remote_enable,remote_enabled ? *strings.disable_string : *strings.enable_string);
}

/************************************************************************************
 *																					*
 *		remote graphics server														*
 *																					*
 ************************************************************************************/

static Boolean handle_graph_command (long command, struct CommandStream *s, struct GraphWindow *window, unsigned char *buffer, long length)
{		
	switch (command)
	{
		case PLOT_PIXEL:
		{
			struct plot_pixel_record *pixel;
			
			pixel=(struct plot_pixel_record *)buffer;
			PlotPixel(swap_long(pixel->x),swap_long(pixel->y),swap_long(pixel->value),window);
			
			break;
		}
		case PLOT_MAP:
		{
			struct plot_map_record *map;
						
			map=(struct plot_map_record *)buffer;
			map->dx=swap_long(map->dx);
			map->dy=swap_long(map->dy);
						
			PlotPixelMap(swap_long(map->x),swap_long(map->y),map->dx,map->dy,swap_long(map->pixSize),swap_long(map->offset),length-6*sizeof(long),(Ptr)&map->data,window);

			break;
		}
		case PLOT_SIZE:
			break;
					
		case REMOTE_END:

			/* stop reading commands */
					
			return false;

		case PLOT_CLEAR_SCREEN:
				
			/* erase desktop window */
	
			SetPort(window->window);
			EraseRect(&window->window->portRect);

			if(window->off_screen != NULL)
			{
				/* erase off-screen port as well */
				
				SetPort((GrafPtr)window->off_screen);
				EraseRect(&window->off_screen->portRect);
			}
			
			return true;

		default:
		{
			error("unknown command received, ignored\n");
			return true;
		}
	}
	
	return true;
}

static void handle_graph_commands(struct CommandStream *s, struct GraphWindow *window)
{
	long message[2];
	unsigned char *buffer;
	
	/* allocate a large message buffer */
	
	buffer=(unsigned char *)NewPtr(10*1024);
	
	if (buffer == NULL)
	{
		error("could not allocate message buffer\n");
		return;
	}
	
	/*
		each message is preceded by its length and its kind, events are handled
		during each reception, so it is not necessary to handle these explicitly
		here
	*/
	
	while (remote_receive((unsigned char *)message,2*sizeof(long),s,handle_compute_event))
	{
		long length;
		long command;
		
		length=swap_long(message[0]);
		
		if (length < 0 || length > 10*1024)
		{
			int c;
		
			/*
						NOTE NOTE NOTE NOTE NOTE NOTE NOTE NOTE
				
				Helios (grrrmblr...) sometimes overwrites the start of the
				buffer with the string ".../IO/window/console". The exact
				format depends on the resource map that Helios uses. It may
				be something like "/sun2/IO/..." or whatever. To work around this
				problem the clean server on the Helios processor inserts a dummy
				of 8 (!) 4-byte words at the start of each message, so that no
				data is lost. The first word of this dummy contains "/...". When
				something strange happens this server checks if the message starts
				with '/' and skips the remaining 6 4-byte words if so...
			*/
			
			c=(int)(*((unsigned char *)&length+3));
			
			if (c == '/')
			{
				/* skip a number of bytes*/

				unsigned long skip[6];
			
				if (remote_receive((unsigned char *)skip,6*sizeof(long),s,handle_compute_event))
					continue;
					
			} else
				error("handle_command: message has invalid length (remote %ld) (local %ld)\n",length,swap_long(length));
			
			break;
		}

		/* length valid, read the rest of the message */
		
		if (length > 0)
		{
			if (!remote_receive(buffer,length,s,handle_compute_event))
				break;
		}

		command=swap_long(message[1]);
		
		if (!handle_graph_command(command,s,window,buffer,length))
			break;
		
		/* handle events in case something goes on... */
		
		handle_compute_event();
		
		if (s->active == false)
			break;
	}
	
	DisposPtr((Ptr)buffer);
}

static void HandleRemoteEvents(struct CommandStream *s)
{
	/* open a graphics window and open a command stream to which clients
	   can connect to send graphics to...
	*/

	handling_remote_commands=true;

	AllocWindowOffScreen(&windows.graphics_window,8);
	
	if (windows.graphics_window.window != FrontWindow())
		SelectWindow(windows.graphics_window.window);
	
	ShowWindow(windows.graphics_window.window);

	/* take care not to destroy anything vital while drawing, i.e. disallow
	   disabling of remote commands and disable quitting. Note that this
	   would be somewhat dangerous if the other side crashes or does not
	   release the connection properly. For this reason a 'force disconnect'
	   command is present. This clears the 'active' flag of the stream, which
	   forces 'handle_graph_commands' to return. */
	
	DisableItem(menus.remote_menu,menu_item_remote_enable);	
	EnableItem(menus.remote_menu,menu_item_force_disconnect);
	
	handle_graph_commands(s,&windows.graphics_window);
	
	EnableItem(menus.remote_menu,menu_item_remote_enable);
	DisableItem(menus.remote_menu,menu_item_force_disconnect);
	
	handling_remote_commands=false;
}

/********************************************************************************
 *																				*
 *		exported routines														*
 *																				*
 ********************************************************************************/

void InitTCP(void)
{
	tcp_avail=(_TCPInit() == noErr);
	
	handling_remote_commands	=false;
	remote_enabled				=false;
	little_endian_client		=true;	/* transputer is little endian */
}

void InitRemote(void)
{
	/* set menu string according to options */
	
	set_remote_able();
	
	if (tcp_avail)
		init_stream(&graph_stream);
	else
	{
		DisableItem(menus.remote_menu,0);
		DisableItem(menus.window_menu,menu_item_graphics_window);
	}
}

void DisconnectRemote(void)
{
	graph_stream.active=false;
}
 
Boolean DisposeRemote (void)
{
	if (handling_remote_commands)
	{
		error("remote stream is active\n");
		return false;
	}

	if (tcp_avail)
		remote_release(&graph_stream);

	return true;
}

void FlipRemoteAble(void)
{
	remote_enabled=!remote_enabled;
	set_remote_able();
}

/********************************************************************************
 *																				*
 *		main event loop															*
 *																				*
 ********************************************************************************/

void EventHandler(void)
{
	for (;;)
	{
		if (tcp_avail && remote_enabled)
		{
			/* create stream, open it, receive commands on it, etc. */
			
			remote_create(&graph_stream);
			
			while (remote_open(&graph_stream, handle_event, &remote_enabled))
			{
				/* stream could be opened, and is set to active */
							
				HandleRemoteEvents(&graph_stream);
				
				/* if stream has been deactivated, abort the connection,
				   else close it neatly */
				
				if (graph_stream.active == false)
					remote_abort(&graph_stream, handle_event);
				else
					remote_close(&graph_stream, handle_event);
			}
			
			remote_release(&graph_stream);
		
		} else
			handle_event();
	}
}
