#include "ifwindows.h"
#include "ifkeys.h"
#include "ifaux.h"
#include "ifinterface.h"
#include "ifmacinterface.h"

/************************************************************************************
 *																					*
 *		allocation/deallocation of off-screen ports									*
 *																					*
 ************************************************************************************/
 
void AllocWindowOffScreen(struct GraphWindow *window,int pixelsize)
{
	Ptr				pixels;
	PixMapPtr		pix_map_ptr;
	unsigned long	hpixels, vpixels, rowbytes, nbytes;
	WindowPtr		w;
	CGrafPtr		p;
	
	w=window->window;
	
	if (window->off_screen != NULL)
	{
		/*
			an off-screen bitmap already exists for this window, if the pixel size
			of the existing pixel map is the same as the one requested, nothing
			has to be done, just keep this pixel map and don't erase the screen
			either
		*/
		
		p=window->off_screen;
		
		if ((*p->portPixMap)->pixelSize==pixelsize)
			return;

		/* a pixel map of different format has to be allocated */
		
		DisposPtr((*p->portPixMap)->baseAddr);		/* get rid of actual pixels */
		
	} else
	{
		/* allocate new port */
		
		p=(CGrafPtr)NewPtr(sizeof(struct CGrafPort));
		
		if (p==NULL)
		{
			error("could not allocate off-screen port\n");
			return;
		}
		
		window->off_screen=p;
		
		OpenCPort(p);
		
		/*	the pixel map is a copy of the current devices pixel map. The colour
	   		table is a handle to the current devices colour table. The graph port
			has been set to this new port.
		*/
	}

	/*
		determine amount of off-screen pixels. The amount of pixels allocated depends
		on the size of the original window (according to its portRect), NOT on
		the size of the actual bit image. This means one has to be carefull
		when the window is resized.
	*/

	hpixels=w->portRect.right - w->portRect.left;		/* horizontal pixels */
	vpixels=w->portRect.bottom - w->portRect.top;		/* vertical pixels */
	
	/* determine amount of bytes in a single row */
	
	switch (pixelsize)
	{
		case 1: rowbytes=(hpixels+7)>>3; break;		/* divide by 8 */
		case 2: rowbytes=(hpixels+3)>>2; break;		/* divide by 4 */
		case 4: rowbytes=(hpixels+1)>>1; break;		/* divide by 2 */
		case 8: rowbytes=hpixels; break;
		
		default:
			internal_error("AllowWindowPixMap: pixel size %d not supported",pixelsize);
			return;
	}
	
	/* rowbytes must be even, but for optimal performance a multiple of 4 */
	
	rowbytes=(rowbytes+3)&~3;
	nbytes=vpixels*rowbytes;

	/* allocate appropriate amount of bytes */

	pixels=NewPtr(nbytes);
	
	if (pixels==NULL)
	{
		/* could not allocate enough storage, throw away port as well */
		
		CloseCPort(p);
		DisposPtr((Ptr)p);
		window->off_screen=NULL;
		
		error("could not allocate off-chip pixels\n");

		return;
	}
	
	/* initialise pixel map */
	
	pix_map_ptr=*p->portPixMap;
	
	pix_map_ptr->baseAddr=pixels;
	
	pix_map_ptr->rowBytes=rowbytes;
	if (pixelsize != 1)
		pix_map_ptr->rowBytes |= 0x8000;		/* pixel map flag */
	
	/* rectangle in pixels */
	
	pix_map_ptr->bounds.top		=0;
	pix_map_ptr->bounds.bottom	=vpixels;
	pix_map_ptr->bounds.left	=0;
	pix_map_ptr->bounds.right	=hpixels;
	
	pix_map_ptr->pmVersion		=0;				/* for future compatibility */
	
	pix_map_ptr->packType		=0;				/* no packing */
	pix_map_ptr->packSize		=0;
	
	pix_map_ptr->pixelType		=0;				/* chunky */
	pix_map_ptr->pixelSize		=pixelsize;		/* bits per pixel */
	pix_map_ptr->cmpCount		=1;				/* 1 for chunky pixels */
	pix_map_ptr->cmpSize		=pixelsize;		/* pixelSize for chunky pixels */
	pix_map_ptr->planeBytes		=0;				/* 0 for chunky pixels */
	
	/* initialise port rect (same as pixel map) */
	
	p->portRect=pix_map_ptr->bounds;
	
	/* initialise visible region */
		
	RectRgn(p->visRgn,&p->portRect);
	
	/* initialise clip region */
	
	RectRgn(p->clipRgn,&p->portRect);
	
	/* erase the (off-screen) screen as it's data can be anything */
	
	SetPort((GrafPtr)p);
	EraseRect(&p->portRect);
	SetPort(w);
	EraseRect(&w->portRect);
}

void DisposWindowOffScreen(struct GraphWindow *window)
{
	CGrafPtr		p;

	if (window->off_screen==NULL)
		return;
		
	/* remove off-screen pixels */
	
	p=window->off_screen;
	DisposPtr((*p->portPixMap)->baseAddr);
	
	/* remove port */
	
	CloseCPort(p);
	DisposPtr((Ptr)p);
	window->off_screen=NULL;
}

/************************************************************************************
 *																					*
 *		initialisation of a window													*
 *																					*
 ************************************************************************************/

void InitGraphWindow(struct GraphWindow *window,int id, WindowPosition *pos, Boolean hscroll, Boolean vscroll, Boolean text, Boolean colour)
{
	Rect text_rect;
	Rect scroll_rect;
	WindowPtr w;

	if (colour)
		window->window=GetNewCWindow(id,NULL,(WindowPtr)-1);
	else
		window->window=GetNewWindow(id,NULL,(WindowPtr)-1);
	
	if ((w=window->window)==NULL)
		return;

	SetPort(w);
	
	if (pos->hsize > 0 && pos->vsize > 0)
	{
		MoveWindow(w,pos->hpos,pos->vpos,false);
		SizeWindow(w,pos->hsize,pos->vsize,true);
	}

	if (text)
	{
		/* default settings for text */
		
		TextFont(monaco);
		TextSize(9);
		
		text_rect.left		=4;
		text_rect.top		=4;
		text_rect.right		=qd.screenBits.bounds.right;
		text_rect.bottom	=qd.screenBits.bounds.bottom;
		
		window->text=TENew(&text_rect,&text_rect);
		
	} else
		window->text=NULL;
	
	if (hscroll)
	{
		scroll_rect.left	=w->portRect.left-1;
		scroll_rect.bottom	=w->portRect.bottom+1;
		scroll_rect.top		=w->portRect.bottom+1-16;
		scroll_rect.right	=w->portRect.right+2-16;
	
		window->hbar=NewControl(w,&scroll_rect,NULL,true,0,0,scroll_resolution,scrollBarProc,(long)window);
	} else
		window->hbar=NULL;
	
	if (vscroll)
	{
		scroll_rect.top		=w->portRect.top-1;
		scroll_rect.right	=w->portRect.right+1;
		scroll_rect.left	=w->portRect.right+1-16;
		scroll_rect.bottom	=w->portRect.bottom+2-16;
		
		window->vbar=NewControl(w,&scroll_rect,NULL,true,0,0,0,scrollBarProc,(long)window);
	} else
		window->vbar=NULL;
	
	if (text)
	{
		window->text_rect.left		=w->portRect.left;
		window->text_rect.right		=w->portRect.right-15;
		window->text_rect.top		=w->portRect.top;
		window->text_rect.bottom	=w->portRect.bottom-15;
	
		ClipRect(&window->text_rect);
	
		window->first_vis_line=0;
	}
	
	window->off_screen=NULL;
}

/********************************************************************************
 *																				*
 *		cleaning up a window													*
 *																				*
 ********************************************************************************/

void DisposeGraphWindow		(struct GraphWindow *window)
{
	DisposeWindow(window->window);
}

/********************************************************************************
 *																				*
 *		mouse-clicks in a window												*
 *																				*
 ********************************************************************************/

#define get_vis_lines(window) ((window->text_rect.bottom-window->text_rect.top)/(*window->text)->lineHeight)

static void get_grow_box(WindowPtr window,Rect *rect)
{
	SetRect(rect,
		window->portRect.right-15,
		window->portRect.bottom-15,
		window->portRect.right,
		window->portRect.bottom);
}

static void get_hbar_box(WindowPtr window,Rect *rect)
{
	SetRect(rect,
		window->portRect.left,
		window->portRect.bottom-15,
		window->portRect.right-15,
		window->portRect.bottom);
}

static void get_vbar_box(WindowPtr window,Rect *rect)
{
	SetRect(rect,
		window->portRect.right-15,
		window->portRect.top,
		window->portRect.right,
		window->portRect.bottom-15);
}

static int get_last_line(struct GraphWindow *window)
{
	int length;
	int n_lines;
	char *p;
	
	n_lines=(*window->text)->nLines;
	length=(*window->text)->teLength;
	
	if (n_lines==0)
		return 1;
	
	p=(char *)*(TEGetText(window->text))+length-1;
		
	if (*p==returnKey)
		return n_lines+1;
	return n_lines;
}

static int get_vbar_position(struct GraphWindow *window)
{
	long curr_line;
	long last_line;
	long vis_lines;
	
	curr_line=window->first_vis_line;
	vis_lines=get_vis_lines(window);
	last_line=get_last_line(window)-vis_lines+1;

	if (last_line<=0)
	{	if (curr_line==0)
			return 0;
		return scroll_resolution;
	}	
	return (int)((curr_line*scroll_resolution)/last_line);
}

static void set_vbar(struct GraphWindow *window)
{
	SetCtlValue(window->vbar,get_vbar_position(window));

	if (get_last_line(window)>=get_vis_lines(window))
		SetCtlMax(window->vbar,scroll_resolution);
	else
	{
		if (window->first_vis_line==0)
			SetCtlMax(window->vbar,0);
		else
			SetCtlMax(window->vbar,scroll_resolution);
	}
}

static pascal void ScrollVText(ControlHandle bar, int part_code)
{
	struct GraphWindow *window;

	window=(struct GraphWindow *)GetCRefCon(bar);
		
	switch(part_code)
	{
		case inUpButton:
			
			if (window->first_vis_line>0)
			{	--window->first_vis_line;
				set_vbar(window);

				ClipRect(&window->text_rect);
				TEScroll(0,(*window->text)->lineHeight,window->text);
				ClipRect(&window->window->portRect);
			}
			
			break;
		
		case inDownButton:
			
			if(window->first_vis_line+get_vis_lines(window)<=get_last_line(window))
			{	++window->first_vis_line;
				set_vbar(window);

				ClipRect(&window->text_rect);
				TEScroll(0,-((*window->text)->lineHeight),window->text);
				ClipRect(&window->window->portRect);
			}

			break;
			
		case inPageUp:

			if (window->first_vis_line>0)
			{
				int scroll_lines;
				
				scroll_lines=get_vis_lines(window)-1;
				if (scroll_lines>window->first_vis_line)
					scroll_lines=window->first_vis_line;
			
				window->first_vis_line-=scroll_lines;
				set_vbar(window);

				ClipRect(&window->text_rect);
				TEScroll(0,(*window->text)->lineHeight*scroll_lines,window->text);
				ClipRect(&window->window->portRect);
			}

			break;
		
		case inPageDown:
		{
			int scroll_lines;
			
			if(window->first_vis_line+get_vis_lines(window)<=get_last_line(window))
			{	
				int last_line;
				
				last_line=get_last_line(window);
				scroll_lines=get_vis_lines(window)-1;

				if (scroll_lines>(last_line-window->first_vis_line-scroll_lines))
					scroll_lines=last_line-window->first_vis_line-scroll_lines;
				
				window->first_vis_line+=scroll_lines;
				set_vbar(window);

				ClipRect(&window->text_rect);
				TEScroll(0,-((*window->text)->lineHeight)*scroll_lines,window->text);
				ClipRect(&window->window->portRect);
			}
			break;
		}
		default:
			break;
	}
	ClipRect(&window->window->portRect);
}

void GraphWindowClick(struct GraphWindow *window,EventRecord *event)
{
	static ControlActionUPP scroll_vertical=NULL;

	ControlHandle bar;
	int part_code;
	
	if (scroll_vertical == NULL)
		scroll_vertical = NewControlActionProc(ScrollVText);

	SetPort(window->window);
	GlobalToLocal(&event->where);
	
	part_code=FindControl(event->where,window->window,&bar);
	
	if (part_code)
	{
		ClipRect(&window->window->portRect);

		if (bar==window->vbar)
		{			
			if (part_code!=inThumb)
			{
				part_code=TrackControl(bar,event->where,scroll_vertical);
				
			} else
				part_code=TrackControl(bar,event->where,NULL);
		}
		
		ClipRect(&window->text_rect);

	} else
		TEClick(event->where,(event->modifiers & shiftKey)!=0,window->text);
}

void GraphWindowGrow(struct GraphWindow *window,EventRecord *event)
{
	long result_size;
	int new_height;
	int new_width;
	
	Rect rect;
	
	WindowPtr w;
	
	w=window->window;
	
	SetRect(&rect,150,80,qd.screenBits.bounds.right-8,qd.screenBits.bounds.bottom-28);
	
	result_size=GrowWindow(w,event->where,&rect);
	new_width	=LoWord(result_size);
	new_height	=HiWord(result_size);

	SetPort(w);

	/* first erase old bars */
		
	ClipRect(&w->portRect);

	get_grow_box(w,&rect);
	EraseRect(&rect);
	InvalRect(&rect);
	HideControl(window->hbar);
	HideControl(window->vbar);
	
	/* compute new sizes */
	
	SizeWindow(w,new_width,new_height,true);
	
	MoveControl(window->hbar,w->portRect.left-1,w->portRect.bottom+1-16);
	MoveControl(window->vbar,w->portRect.right+1-16,w->portRect.top-1);
	SizeControl(window->hbar,w->portRect.right+3-16,16);
	SizeControl(window->vbar,16,w->portRect.bottom+3-16);

	window->text_rect.left		=w->portRect.left;
	window->text_rect.right		=w->portRect.right-15;
	window->text_rect.top		=w->portRect.top;
	window->text_rect.bottom	=w->portRect.bottom-15;

	set_vbar(window);

	/* and draw new bars */
	
	ClipRect(&w->portRect);
	
	DrawGrowIcon(w);
	ShowControl(window->hbar);
	ShowControl(window->vbar);
	
	get_grow_box(w,&rect);
	ValidRect(&rect);
	get_hbar_box(w,&rect);
	ValidRect(&rect);
	get_vbar_box(w,&rect);
	ValidRect(&rect);
	
	ClipRect(&window->text_rect);
}

void GraphWindowActivate(struct GraphWindow *window,EventRecord *event)
{
	WindowPtr w;
	Rect bad_rect;

	w=window->window;	
	
	SetPort(w);
			
	if (event->modifiers & activeFlag)
	{
		if (window->hbar!=NULL)
			HiliteControl(window->hbar,0);
		if (window->vbar!=NULL)
			HiliteControl(window->vbar,0);
		
		DisableItem(menus.edit_menu,menu_item_undo);
		DisableItem(menus.edit_menu,menu_item_cut);
		DisableItem(menus.edit_menu,menu_item_copy);
		DisableItem(menus.edit_menu,menu_item_paste);
		DisableItem(menus.edit_menu,menu_item_clear);

		if (window->text!=NULL)
			TEActivate(window->text);
	} else
	{
		if (window->hbar!=NULL)
			HiliteControl(window->hbar,255);
		if (window->vbar!=NULL)
			HiliteControl(window->vbar,255);

		EnableItem(menus.edit_menu,menu_item_undo);
		EnableItem(menus.edit_menu,menu_item_cut);
		EnableItem(menus.edit_menu,menu_item_copy);
		EnableItem(menus.edit_menu,menu_item_paste);
		EnableItem(menus.edit_menu,menu_item_clear);

		if (window->text!=NULL)
			TEDeactivate(window->text);
	}
	
	if (window->hbar!=NULL || window->vbar!=NULL)
	{
		get_grow_box(w,&bad_rect);
		InvalRect(&bad_rect);
	}
	if (window->hbar!=NULL)
	{
		get_hbar_box(w,&bad_rect);
		InvalRect(&bad_rect);
	}
	if (window->vbar!=NULL)
	{
		get_vbar_box(w,&bad_rect);
		InvalRect(&bad_rect);
	}
}

void GraphWindowUpdate(struct GraphWindow *window,EventRecord *event)
{
	WindowPtr w;
	
	w=window->window;
	
	SetPort(w);

	ClipRect(&w->portRect);

	if (window->hbar!=NULL || window->vbar!=NULL)
		DrawGrowIcon(w);
	
	DrawControls(w);

	if (window->text!=NULL)
	{
		ClipRect(&window->text_rect);

		EraseRect(&window->text_rect);
		TEUpdate(&window->text_rect,window->text);
	}
	
	if (window->off_screen != NULL)
	{
		/* copy off-screen pixel map to window */
		
		CopyBits ((BitMap *)*(window->off_screen->portPixMap),&(window->window)->portBits,&(window->window)->portRect,&(window->window)->portRect,srcCopy,NULL);
	}
}

/********************************************************************************
 *																				*
 *		keyboard handling for a window											*
 *																				*
 ********************************************************************************/

void GraphWindowKey(struct GraphWindow *window, char c)
{
	/* this routine is only valid for windows containing text */
	
	Point position;
	
	SetPort(window->window);
	GetPen(&position);
	
	/*
	
	under construction
	
	if (position.v <= window->window->portRect.top+4)
		TEScroll(0,(window->window->portRect.top+4)-position.v,window->text);
	if (position.v > window->window->portRect.bottom-20)
		TEScroll(0,(window->window->portRect.bottom-20)-position.v,window->text);
	*/
	
	TEKey(c,window->text);
	
	ClipRect(&window->window->portRect);
	set_vbar(window);
	ClipRect(&window->text_rect);
}

/********************************************************************************
 *																				*
 *		maintaining window positions											*
 *																				*
 ********************************************************************************/

void GetWindowPosition (struct GraphWindow *window, WindowPosition *pos)
{
	WindowPtr	w;
	Point		origin={0,0};

	/* get window position */

	w=window->window;

	SetPort(w);
	LocalToGlobal(&origin);

	pos->hpos		=origin.h;
	pos->vpos		=origin.v;

	/* get window width and heigth */

	pos->hsize		=w->portRect.right - w->portRect.left;
	pos->vsize		=w->portRect.bottom - w->portRect.top;
}